/*
 * AlgoState.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science,
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
 
package kea;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.GregorianCalendar;

import kea.algorithms.AbstractAlgorithm;

/**
 * This class is describing an optimization run.
 * Methods for saving the data of a run as txt-file on disc are also found here.
 *
 * @author Stephan Sigg
 * @version $Revision: 723 $
 *
 */
public class AlgoState extends kea.Datacontainer implements Serializable {
  /**
   * This attribute contains the whole algorithm object of the current run.
   * This is needed to continue a halted run. By serialising the AlgoState
   * object, the status of the alorithm at the halt is stored and can be
   * continued.
   */
  private AbstractAlgorithm algorithm;
  /**
   * Counts the total number of evaluations (this is needed to continue a interrupted run)
   */
  private int evalCount;
  /**
   * this Variable defines, after how many iterations the Algostate is to be reported in a txt-file.
   * (you can archieve this with the command this.export()
   */
  private int exportStepSize;
  // counts the total number of iterations.
  private int iterationCount;
  // the last reported state of the Algo is specified here
  private Data[] results;
  // the runtime of the algo.
  private long runtime;
  /**
   * this integer-value specifies the time window which difines the exact time
   * between two (binary) Save-steps with this.save().
   */
  private int saveStepDuration;
  // This String contains the Filename of the txt-file (without the ending .ast for AlgoState)
  private String exportFileName;
  // this String contains the Filename of the binary file ( without the ending .sav)
  private String saveFileName;
  // this String contains the Filename of the AlgoState
  private String loadFileName;
  // boolean Variable to indicate weather the file contains a header or not.
  private boolean printHeader;
  // boolean Variable to indicate, if the .ast and .sav filenames are default filenames or not (comp. setFilename(String))
  private boolean savOldFileName;
  private boolean exOldFileName;
  private String exPath;
  private String svPath;

  public void setExportPath(String newPath) {
    exPath = newPath;
  }

  public void setSavePath(String newPath) {
    svPath = newPath;
  }
  /**
   * constructor - all Variables are initialized here with default values.
   */
  public AlgoState() {
    this.evalCount=0;
    this.exportStepSize=5;
    this.iterationCount=0;
    this.results=null;
    this.runtime=0;
    this.saveStepDuration=5;
    this.exportFileName = "";
    this.saveFileName = "";
    this.printHeader = true;
    this.savOldFileName = true;
    this.exOldFileName = true;
  }

  public AbstractAlgorithm getAlgorithm(){
    return algorithm;
  }
  public void setAlgorithm(AbstractAlgorithm algo){
    if (algo != null){
      this.algorithm = algo;
    }
    else{
      Kea.logger.logError("Sorry, but the algorithm may not be initialized. Please set it properly",this);
    }
  }
  public int getEvalCount(){
    return evalCount;
  }
  public void setEvalCount(int evalCount){
    this.evalCount = evalCount;
  }
  public int getExportStepSize(){
    return exportStepSize;
  }
  public void setExportStepSize(int exportStepSize){
    this.exportStepSize = exportStepSize;
  }
  public int getIterationCount(){
    return iterationCount;
  }
  public void setIterationCount(int iterationCount){
    this.iterationCount = iterationCount;
  }
  public Data[] getResults(){
    return results;
  }
  public void setResults(Data[] results){
    if (results != null){
      this.results = results;
    }
	else{
	  Kea.logger.logError("Sorry, but the results may not be initialized.Please set them properly ",this);
	}
  }
  public long getRuntime(){
    return runtime;
  }
  public void setRuntime(long runtime){
    Kea.logger.logInfo("New runtime set in AlgoState",this);
    this.runtime = this.runtime + runtime;
  }
  public int getSaveStepDuration(){
    return saveStepDuration;
  }
  public void setSaveStepDuration(int saveStepDuration){
    Kea.logger.logInfo("saveStepDuration set in AlgoState",this);
    this.saveStepDuration = saveStepDuration;
  }

  /**
   *
   */
  private String createDefaultFileName(){
    String retVal = "";
    GregorianCalendar time = (GregorianCalendar)GregorianCalendar.getInstance();
    String algoName = algorithm.getParameter().getMembership();
    String probName = algorithm.getProblem().getParameter().getMembership();
    SimpleDateFormat format = new SimpleDateFormat("_yyMMdd_HHmmss");

    // build name: AlgoName!ProblemName
    retVal = algoName.substring(algoName.lastIndexOf(".")+1,algoName.length()) + "!"
             + probName.substring(probName.lastIndexOf(".")+1,probName.length());

    // generate unic Timestamp for filename
    retVal = retVal + format.format(new Date());

    // to avoid problems with your Microsoft Windows system:
    retVal = retVal.replace(':','_');
    retVal = retVal.replace('/','-');
    retVal = retVal.replace(' ','%');
    return retVal;
  }

  /**
   * Creates a txt-file which contains the Parameters and Values of the
   * corresponding Algo iff the file not already exists.
   * In every case <code>export</code> calls <code>appendResults()</code>
   * to append the newest Results.
   * The created file can  be specified as commandline-parameter of Kea
   * using <code>-f \<filename\></code>
   */
  public void export() throws KeaCoreException{
    //check if user gave filename, if not generate a default name
    if (this.exportFileName == "") {
      this.exportFileName = (createDefaultFileName() + ".ast");
    }
    
    File temp = new File(exPath, this.exportFileName);
   
    //check if it is the first call of export, then generate new File and
    //write the header
    if (this.printHeader || !temp.exists()){
      Kea.logger.logInfo("Writing File Header", this);
      this.printHeader = false;
      
      try
	{
	  if (temp.exists()) {
	    throw new KeaCoreException("Cannot write AlgoState File! " + temp.getAbsoluteFile() + " already exists!");
	  }
	  
	  /*
	   *   Output:  setOfParameters as binary in a "GZIP-Header"
	   *   If you want it replace null with the outcommented text.
	   */
	  
	  Datacontainer.writeSetOfParametersGZipTo
	    (null,
	     // (new Parameters[]{algorithm.getParameter(),
	     //	       	          algorithm.getProblem().getParameter()},
	     temp.getAbsolutePath());
	  
	  
	  /*
	   *   outputs parameters in a human readable form
	   */
	  
	  PrintWriter writer = 
	    new PrintWriter
	    ( new BufferedWriter
	      ( new FileWriter(temp.getAbsolutePath(),true)
		)
	      );
	  
	  // first write the algorithm-parameter
	  writer.println(this.algorithm.getParameter().toString());
	  writer.println();
	  
	  //second write the problem-parameter
	  writer.println(this.algorithm.getProblem().getParameter().toString());
	  writer.println();
	  
	  writer.println("DimDecSpace = " + algorithm.getResults()[0].getDecVector().length);
	  writer.println("DimObjSpace = " + algorithm.getResults()[0].getObjVector().length);
	  writer.println();
	  
	  // All additional variables of AlgoState plus the corresponding Values.
	  writer.println("exportStepSize = " + exportStepSize);
	  writer.println("saveStepDuration = " + saveStepDuration);
	  
	  //set mark to identify the break between Parameter and Data
	  writer.println(Datacontainer.HEADER_END_IDENTIFIER);
	  writer.println();
	  writer.close();
	  if(Kea.logger.infoOn())
	    Kea.logger.logInfo("New file " + temp.getAbsolutePath() +" created.", this);
	  
	}// end of try
      catch (IOException e) {
	Kea.logger.logError("Fehler",e,this);
      }
    }// end of if
      appendResults();
  }// end of export
  
  
  
  /**
   * The results of one run are written (i.e. appended) into the already created textfile.
   */
  private void appendResults(){
    try{

      File temp = new File(exPath, this.exportFileName);
      PrintWriter writer = new PrintWriter(new BufferedWriter(new FileWriter(temp.getAbsolutePath(),true)));
         Data[] data = this.algorithm.getResults();
      for (int k = 0;k < data.length; k++){
	// decisionVector
        Value[] decVector = data[k].getDecVector();
	for(int j = 0; j < decVector.length; j++){
	  writer.write(decVector[j].getData().toString() + " ");
	}// end of for j

	writer.write("/");

        //objectiveVector
        Value[] objVector = data[k].getObjVector();
	for (int j = 0; j < objVector.length; j++) {
	  writer.write(objVector[j].getData().toString() + " ");
	}// end of for j

	writer.println();
      }// end of for k

      writer.println("evaluationCount = " + evalCount);
      writer.println("iterationCount = " + iterationCount);
      writer.println("runtime = " + runtime);
      writer.println();
      writer.close();
    }// end of try
    catch(Exception e){
      Kea.logger.logError("Unable to append results : "+ e.toString() +" caught",this);
      e.printStackTrace();
    }
  }
  
  
   /**
   *
   */
  public String getExportFileName(){
    return this.exportFileName;
  }

  /**
   *
   */
  public void setExportFileName(String newName){
    if (!newName.endsWith(".ast")) {
      exportFileName = newName + ".ast";
    }
    else {
      //take name as given
      exportFileName = newName;
    }
    printHeader = true;
  }


  public String getSaveFileName(){
    if (saveFileName == "") {
      saveFileName = createDefaultFileName() + ".sav" ;
    }
    //return (svPath + saveFileName);
    return (saveFileName);
  }

  /**
   *
   */
  public void setSaveFileName(String newName){
    if (!newName.endsWith(".sav")) {
      saveFileName = newName + ".sav";
    }
    else {
      //take name as given
      saveFileName = newName;
    }
    printHeader = true;
  }// end of setSaveFileName


}
 
  
  
  
  
