/*
 * CompResult.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.GregorianCalendar;

/**
 * This subclass of Datacontainer collects all data belonging to a
 * special comparison.  You may save it.
 *
 * Attributs:
 * <ol>
 * <li> setOfParameters: These are the parameters used for comparison,
 * algorithms and/or evaluation;
 * <li> compResult: It's a String-array which contains the results of the
 * respective comparison.
 * </ol>
 * 
 * @author (dk)
 * @version 0.1.0
 * @see kea.Datacontainer
 */
public class CompResult extends Datacontainer implements Serializable {
  /**
   * Creates a new <code>CompResult</code> instance with a standard
   * save- and export-filename and the other member variables set to
   * <code>null</code>.
   *
   */
  public CompResult(){
    String fileName;
    GregorianCalendar time =(GregorianCalendar)GregorianCalendar.getInstance();
    SimpleDateFormat format = new SimpleDateFormat();
    format.setCalendar(time);
    fileName =
      format.format(time.getTime())
      + "_" + time.get(time.SECOND);
    // to avoid problems with your Microsoft Windows system:
    fileName = fileName.replace(':','_');
    fileName = fileName.replace('/','-');
    fileName = fileName.replace(' ','%');
    this.exportFileName = fileName + ".crs";
    this.saveFileName = fileName + "_crs.sav";

    this.setCompResult(new String[]{});
    this.setSetOfParameters(new Parameters[]{});
  }


  /**
   * <code>getCompResult</code> returns the membervariable
   * compResult set in CompResult.
   *
   * @return <b>null</b> if not initialized different with
   * setCompMethod().  So don't try to find out anything about the
   * compMethod, if you didn't initialize it properly. Otherwise
   * you've to handle NullPointerExceptions.
   */
  public String[] getCompResult(){
    if(this.compResult == null) Kea.logger.logWarn("CompResult == null", this);
    else if (Kea.logger.infoOn()) Kea.logger.logInfo("Return compResult != null", this);
    return this.compResult;
  }

  /**
   * <code>setCompResult</code> sets the result of a comparison in
   * CompResult.
   *
   * @param compResult a <code>String-array</code>, which contains
   * the result of a special comparison. Normally this should be
   * strings.
   */
  public void setCompResult(String[] compResult){
    this.compResult = compResult;
  }

  /**
   * <code>getSetOfParameters</code> returns the set of Parameters
   * belonging to the comparison either since they're belonging to
   * the comparison-method itself or to algorithms which results
   * were compared.
   *
   * @return <b>null</b> if not initialized different with
   * setCompMethod().  So don't try to find out anything about the
   * compMethod, if you didn't initialize it properly. Otherwise
   * you've to handle NullPointerExceptions.
   */
  public Parameters[] getSetOfParameters(){
    return this.setOfParameters;
  }

  /**
   * <code>setSetOfParameters</code> sets the set of parameters inside
   * CompResult.
   *
   * @param parameters <code>Parameters[]</code> should contain all
   * parameters belonging to the results, which were compared, i. e.
   * from the respective comparison-method, AlgoStates or EvalResults.
   * (Since they're also set inside the two last objects, it could be
   * enough to receive a file-reference in a later version.)
   */
  public void setSetOfParameters(Parameters[] parameters){
    this.setOfParameters = parameters;
  }

  /* ******************************************************************
   *
   * Methods for the classes inherited from   D a t a C o n t a i n e r
   *
   * *****************************************************************/

  /**
   * <code>export</code> exports all relevant data of CompResult in a
   * <code>ExportFileName</code>.crs-file.
   * This file will contain the respective date, parameters and result
   * of the comparison.
   *
   */

  public void export(){
    try {
      /*
       *   Output:  setOfParameters as binary in a "GZIP-Header"
       */

      Datacontainer.writeSetOfParametersGZipTo
	( null, getExportFileName());


      /*
       *   Output:  Parameters and results in a more readable version
       */

      PrintWriter outPrint =
	new PrintWriter
	( new BufferedWriter
	  ( new FileWriter(getExportFileName(),true)));

      try{
	for(int i=0; i<this.setOfParameters.length; i++){
	  outPrint.println(this.setOfParameters[i].toString());
	  if (i<this.setOfParameters.length-1) outPrint.println();
	}
      }catch(NullPointerException e){
	outPrint.println("#   <<Did you initialize your objects properly?>>");
	outPrint.println("#   Otherwise you can't expect to find any parameter-infos here.");
	Kea.logger.logError(" setOfParameters isn't initialized properly", this);
      }

      if(Kea.logger.infoOn()) Kea.logger.logInfo("Exporting compResults", this);

      try{
	outPrint.println("### ************   R E S U L T S   ************");
	outPrint.println();
	for(int i = 0; i < compResult.length; i++){
	  outPrint.println(compResult[i]);
	}
      }catch(NullPointerException e){
	outPrint.println("#   No comparison results found.");
	Kea.logger.logWarn("No compResults to export", e, this);
      }
      catch(ClassCastException e){
	Kea.logger.logError("Can't cast CompResults to string.", e, this);
      }

      outPrint.flush();
      outPrint.close();

    } catch(IOException e) {
      Kea.logger.logError
	( "IOError occured while working with " + getExportFileName(), e, this);
    }

    if(Kea.logger.infoOn()){
      Kea.logger.logInfo
	("Creating file " + getExportFileName() + "............OK", this);
      Kea.logger.logInfo
	("Export to file " + getExportFileName() + "............OK", this);
    }

    /* ensures that the exported file can't be written by mistake */
    // File momentFile = new File(getExportFileName());
    // momentFile.setReadOnly();

    /* // otherwise, it's still possible to delete the file using:
     *
     *      momentFile.delete();
     */

  }

  /* idea for GZIP-Output to reduce space needed for saving
   *
   * public void exportZIP(){
   try {
   PrintStream outPrint =
   new PrintStream
   ( new BufferedOutputStream
   ( new GZIPOutputStream
   ( new FileOutputStream(getExportFileName(),false))));
   exportWriteData(outPrint);
   } catch(IOException e) {
   Kea.logger.logError
   ( "IOError occured while working with " + exportFileNameMod, e, this);
   }
   }
  */

  /**
   * <code>getExportFileName</code> returns the name of the file, which
   * will contain the data of <code>CompResult</code> in ASCII-format.
   *
   * @return a <code>String</code> value containing the momentarily set
   * filename for the export.
   */
  public String getExportFileName(){
    /* // allows a ascending numbering of exported files;
    // useful if they would get the same name otherwise.
    ++this.exportFileNumber;
    this.exportFileNameMod =  this.exportFileName.concat
    ( new Integer(this.exportFileNumber).toString().substring(1) + ".crs");;
    */
    this.exportFileNameMod = this.exportFileName; // only, if not used above FileNumbering
    return this.exportFileNameMod;
  }

  /**
   * <code>setExportFileName</code> defines the name of the ASCII-file,
   * in which the data of <code>CompResult</code> will be exported.
   * The parameter exportFileName will be extended by .crs.
   *
   *
   * @param exportFileName is the name of the ASCII-file which shall be
   * created by <code>export</code>
   */
  public void setExportFileName(String exportFileName){
    /*
      if (exportFileName.endsWith(".crs")){
      int lIO = exportFileName.lastIndexOf(".");
      this.exportFileName = exportFileName.substring(0,lIO);
      }
      else
      this.exportFileName = exportFileName;
      this.exportFileNumber = 1000;
    */
    if (!exportFileName.endsWith(".crs"))
      this.exportFileName = exportFileName + ".crs"; //only, if not used above FilePreparations
    else
      this.exportFileName = exportFileName;
  }

  /**
   * <code>getSaveFileName</code> returns the name of the object-file,
   * which will be created, if <code>save</code> is used.
   *
   * @return a <code>String</code> value of the name of the saved
   * <code>CompResult</code>-object.
   */
  public String getSaveFileName(){
    return this.saveFileName;
  }

  /**
   * <code>setSaveFileName</code> defines the name of the object-file,
   * which will contain <code>CompResult</code> as binary.
   *
   * @param saveFileName is the name of the object-file of
   *  <code>CompResult</code> if it's saved.
   */
  public void setSaveFileName(String saveFileName){
    if(!saveFileName.endsWith("_crs.sav"))
      this.saveFileName = saveFileName + "_" + getSaveFileName();
    else this.saveFileName = saveFileName.substring(0,saveFileName.length()-8) + "_" + getSaveFileName();
  }

  private String[] compResult;
  //f\uFFFDr detaillierte export()
  private Parameters[] setOfParameters;

  // Variablen f\uFFFDr Datacontainer-Methoden
  private String exportFileName;
  private String saveFileName;

  // needed for export()
  private int exportFileNumber;
  private String exportFileNameMod;

  /*
   *    M A I N   - Method for testing CompResult
   *
   */

  /*  public static void main (String[] argv){
      CompResult crTest = new CompResult();
      
      Parameter parObject = new Parameter();
      parObject.setName("Name1");
      parObject.setToolTip("jff");
      parObject.setType(new Type().STRING);
      System.out.println("1st setData.  Success?: "+parObject.setData("just data"));
      Parameter parObjecta = new Parameter();
      parObjecta.setName("Name2");
      parObjecta.setToolTip(null);
      parObjecta.setType(new Type().STRING);
      System.out.println("2nd setData.  Success?: "+parObjecta.setData(""));
      Parameters params = new Parameters();
      params.setParameterArray(new Parameter[]{parObject,parObjecta});
      params.setMembership("Test.test");
      Parameters[] paramsSet = new Parameters[]{params, params, params};
      
      crTest.setSetOfParameters(paramsSet);
      
      Parameters[] testParams= {new Parameters(), new Parameters()};
      String[] testCompRes = {"Niete","Noch 'ne Niete"};
      
      CompResult newcrTest = crTest;
      System.out.println("testCompRes.length = "+testCompRes.length);
      crTest.setCompResult(testCompRes);
      crTest.setSaveFileName("crTest");
      crTest.setExportFileName("crTest");
      crTest.save();
      crTest.export();
      CompResult mycrTest = (CompResult) CompResult.load("crTest.sav");
      mycrTest.getSetOfParameters()[1].getParameterArray()[1].setData("changed");
      mycrTest.setSaveFileName("mycrTest");
      mycrTest.setExportFileName("mycrTest");
      mycrTest.save();
      mycrTest.export();
      newcrTest.setExportFileName("newcrTest");
      newcrTest.setSaveFileName("newcrTest");
      crTest.export(); // sollte ursprnglich append testen
      //mycrTest.export(); // in jetziger CR-Version ausgestellt
      }
  */
}
