/*
 * Data.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.Serializable;

/**
 * @author Stefan Kusper
 * @version 0.1.0
 */
public class Data implements Serializable, Cloneable {
    public Value[] decVector;
    public Value[] objVector;

    /**
     * Standard constructor. Does not set anything.
     */
	public Data() {
        decVector=null;
        objVector=null;
    }

    /**
     * Constructor for initializing the decision vector with a known vector and
     * the objective vector with null.
     * @param dv a decision vector to be stored
     */
    public Data(Value[] dv) {
        decVector=dv;
        objVector=null;
    }

    /**
     * Constructor for reconstruction. Use if you want to set both
     * decision vector and objective vector.
     * @param dv a decision vector
     * @param ov a objective vector to be stored
     */
    public Data(Value[] dv, Value[] ov) {
        this.decVector=dv;
        this.objVector=ov;
    }

    /**
     * Constructor for reconstruction. Use, if you want to set the
     * types of the decision and objective vector.
     * @param decTypes a Type-vector of the decision space
     * @param objTypes a Type-vector of the objective space
     */
    public Data(Type[] decTypes, Type[] objTypes, Range[] decRanges) {
      // set decision vector
      if (decTypes == null) {
        this.decVector = null;
      }
      else {
        this.decVector = new Value[decTypes.length];
        for (int i = 0; i < decTypes.length; ++i) {
          this.decVector[i] = new Value(decTypes[i].getTypeInfo(), decRanges[i]);
        }
      }
      // objective vector
      if (objTypes == null) {
        this.objVector = null;
      }
      else {
        this.objVector = new Value[objTypes.length];
        for (int i = 0; i < objTypes.length; ++i) {
          this.objVector[i] = new Value(objTypes[i]);
        }
      }
    }

    /**
     * Clears both vectors in this object.
     */
	public void clear(){
        this.decVector = null;
        this.objVector = null;
    }

    /**
     * Returns the decision vector.
     * @return the decision vector
     */
    public Value[] getDecVector(){
        if (this.decVector!=null)
        	return this.decVector;
        else return null;
        //else throw new AlgoException("Data.getDecVector: reference is null");
    };

    /**
     * Returns the objective vector.
     * @return the objective vector
     */
    public Value[] getObjVector() {
        if (this.objVector!=null)
            return this.objVector;
        else return null;
        //else throw new AlgoException("Data.getObjVector: reference is null");
    }

    /**
     * Sets the decision vector to the given Value array.
     * @param dv a decision vector to be stored
     */
	public void setDecVector(Value[] dv) {
        if (dv!=null)
	        this.decVector = dv;
        else this.decVector = null;
        //else throw new AlgoException("Data.setDecVector: parameter Value[] is null");
    }

    /**
     * Sets the objective vector to the given Value array.
     * Usefull for loading and continuing.
     * @param ov a objective vector to be stored
     */
    public void setObjVector(Value[] ov) {
        if (ov!=null)
	        this.objVector = ov;
        else this.objVector = null;
        //else throw new AlgoException("Data.setObjVector: parameter Value[] is null");
    }

    /**
     * @return the string representation of this object.
     */
//	public String toString() {
//        String ds=new String();
//        String os=new String();
//        if(this.objVector!=null)
//	        for (int i=0; i<this.objVector.length; ++i){
//				os += this.objVector[i].toString();
//        	    os += " ";
//     	   }
//        else os = "null";
//        if(this.decVector!=null)
//	        for (int i=0; i<this.decVector.length; ++i){
//				ds += this.decVector[i].toString();
//        	    ds += " ";
//      		}
//        else ds = "null";
//		return "Decision Vector: " + ds + "\n" + "Objective Vector: " + os + "\n";
//    }

    public Object clone() {
        Data retval = null;
        try {
            retval = (Data)super.clone();
            Value[] n_dec = new Value[this.decVector.length];
            Value[] n_obj = new Value[this.objVector.length];
            for (int i = 0; i < n_dec.length; i++) {
                n_dec[i] = (Value)(this.decVector[i].clone());
            }
            for (int i = 0; i < n_obj.length; i++) {
                n_obj[i] = (Value)(this.objVector[i].clone());
            }
            retval.setDecVector(n_dec);
            retval.setObjVector(n_obj);
        }
        catch (CloneNotSupportedException e) {
            System.out.println(e);
        }
        return retval;
    }

    /**
     * main method for testing
     */
	public static void main(String[] argv) {
        System.out.println("Beginne Test OptimisationData:");
        Data d = new Data();
        System.out.println("leeres Data Objekt angelegt");
        Value[] dec, obj;
        dec = new Value[10];
        obj = new Value[10];
        for(int i=0; i<dec.length; ++i) {
            dec[i]=new Value(Math.random() );
            obj[i]=new Value(Math.random() );
        } // end of for
		d = new Data(dec);
		System.out.println("nicht leeres Data Objekt angelegt: decVector");
        d = new Data(dec, obj);
        System.out.println("nicht leeres Data Objekt angelegt: decVector, objVector");
        d.setDecVector(obj);
        d.setObjVector(dec);
        System.out.println("Versuche zu klonen:\n");
        Data d1 = (Data)d.clone();
		System.out.println(d);
        d.clear();
        System.out.println("Objektinhalt gelscht");
        System.out.println(d);
        System.out.println("clone:\n" + d1);
    } // end of main
} // end of class
