/*
 * Datacontainer.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.PrintWriter;
import java.util.Date;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

/**
 * Interface which ensures that the implementing classes can be saved and exported.
 *
 * <ol>
 * <li>exportFileName: name for the ASCII-file created using the <code>export</code>-method
 * </ol>
 * 
 * @author PG 419, University of Dortmund (dk)
 * @version 0.1.0
 */
abstract public class Datacontainer {

  /**
   * Constant <code>GZIP_END_IDENTIFIER</code> shall mark the end
   * of the GZipHeader which contains the <code>parameters[]</code>.  It
   * is produced during <code>export()</code> of the subclasses using
   * <code>writeSetOfParametersGZipTo()</code>.
   *
   */
  public static final String GZIP_END_IDENTIFIER = "# A A A";
  /**
   * Constant <code>KEA_CLASS_IDENTIFIER</code> shall mark the lines
   * which contain the output of Parameters.getMembership().  Those
   * are produced during export() of the subclasses.
   *
   */
  public static final String KEA_CLASS_IDENTIFIER = ">>>" ;
  /**
   * Constant <code>HEADER_END_IDENTIFIER</code> shall mark the end
   * of the header of each exported file.  After this block normally
   * some results follow.
   *
   */
  public static final String HEADER_END_IDENTIFIER = "###";

  /**
   * <code>save</code> saves the calling <code>Datacontainer</code>object
   * as binary.  In the case of <code>AlgoState</code>, this should allow
   * to continue the execution of a run of an alogrithm using the saved data.
   */
  public void save(){
    try {
      ObjectOutputStream outObject =
	new ObjectOutputStream
	( new BufferedOutputStream
	  ( new GZIPOutputStream
	    ( new FileOutputStream(getSaveFileName())
	      )
	    )
	  );
	  Kea.logger.logDebug("befor saving this file: "+getSaveFileName()+".",  this.toString());

      outObject.writeObject(this);
	  Kea.logger.logDebug("after saving this file: "+getSaveFileName()+".",  this.toString());

      outObject.flush();
      outObject.close();
    } catch (IOException e) {
      Kea.logger.logError("Unable to write file"+getSaveFileName()+".", e, this);
    }
  }

  /**
   * <code>load</code> loads the binary object belonging to the respective loadFileName.
   * All values inside the object should be restored the way they were saved.
   *
   * This method is static to allow loading an object without creating an <code>
   * Datacontainer</code>-object before.
   * @param loadFileName <code>String</code> naming the .sav-file to load
   * @return <code>Datacontainer</code> needs to be casted to AlgoState, EvalResult or
   *         CompResult, respective.
   */
  public static Datacontainer load(String loadFileName){
    Datacontainer tempDC = null;

    try{
      ObjectInputStream inObject =
	new ObjectInputStream
	( new BufferedInputStream
	  ( new GZIPInputStream
	    ( new FileInputStream(loadFileName)
	      )
	    )
	  );

      tempDC = (Datacontainer) inObject.readObject();

      inObject.close();
    } catch (Exception e) {
      Kea.logger.logError("(Datacontainer or inheriting class) Unable to read file "
			  + loadFileName + ".");
      e.printStackTrace();
    }

    return tempDC;
  }

  /**
   * <code>readSetOfParametersGZipFrom</code> reads the gZiped Parameters
   * @param gzImport Name of the file from which the files shall be read
   * @return <code>Parameters[]</code> if it was found else <code>null</code>
   */

  public static Parameters[] readSetOfParametersGZipFrom(String gzImport){
    Parameters[] tempParameters = null;
    try{
      ObjectInputStream inObject =
	new ObjectInputStream
	( new BufferedInputStream
	  ( new GZIPInputStream
	    ( new FileInputStream(gzImport) )
	    )
	  );

      tempParameters = (Parameters[])inObject.readObject();
      inObject.close();
    }
    catch(IOException e){
      Kea.logger.logWarn("\n[kea.Datacontainer]\nFailure in GZIP-Header of "
			 + gzImport +".  It may be corrupted or missing.\n");
    }
    catch(ClassNotFoundException e){
      Kea.logger.logError("Couldn't find class, corresponding to GZIP-Header\n");
      e.printStackTrace();
    }
    return tempParameters;
  }

  /**
   * <code>writeSetOfParametersGZipTo</code> writes the given set of parameters
   * to the specified file.  <em>Attention:</em>  If it already exists it will
   * be replaced.  Make sure that subsequent writes to this file are appends,
   * otherwise you will lose the GZip-Header again.
   *
   * @param gzParametersArr Set of parameters to write to file.  If you pass
   * <code>null</code> only the standard intro-lines will be written.
   * @param gzExport Name of the file the header shall be written to.
   */

  public static void writeSetOfParametersGZipTo(Parameters[] gzParametersArr, String gzExport){
    PrintWriter writer;
    try{
      if(gzParametersArr != null)
	{
	  ObjectOutputStream outObject =
	    new ObjectOutputStream
	    ( new BufferedOutputStream
	      ( new GZIPOutputStream
		( new FileOutputStream(gzExport)
		  )
		)
	      );
	  outObject.writeObject(gzParametersArr);
	  outObject.flush();
	  outObject.close();

	  writer =
	    new PrintWriter
	    ( new BufferedWriter
	      ( new FileWriter(gzExport,true)
		)
	      );
	}
      else
	{
	  writer =
	    new PrintWriter
	    ( new BufferedWriter
	      ( new FileWriter(gzExport)
		)
	      );
	}
      if(gzParametersArr != null)
	{
	  writer.println();
	  writer.println();
	  writer.println(GZIP_END_IDENTIFIER + "   The following Parameters "
			 + "as GZiped binary - Do not edit   A A A ");
	}
      writer.println("#");
      writer.println("#   " + new Date().toString());
      writer.println("#");
      writer.println("#   The following lines point out which algorithms, evaluations");
      writer.println("#   and/or comparison method were used and show their corresponding");
      writer.println("#   parameters or parameter-file-names:");
      writer.println("#");
      writer.close();
    }
    catch(Exception e){
      e.printStackTrace();
    }
  }

    /**
     * Creates a ASCII-file containing all values set in the calling, inheriting
     * class.  Files created with this method can be adressed via the Kea-command
     * using the -f <fileName> flag.
     */
    abstract void export() throws KeaCoreException;

    abstract String getExportFileName();

    abstract void setExportFileName(String exportFileName);

    abstract String getSaveFileName();

    abstract void setSaveFileName(String saveFileName);


  public static String replaceAll(String in, String regex, String replacement){
    int idx = -1;
    while ((idx = in.indexOf(regex,idx))!=-1){
      in = in.substring(0,idx)
	+ replacement
	+ in.substring((idx + regex.length()), in.length());
      idx++;
    }
    return in;
  }

  /**
   * <code>split</code> does (hopefully) the same for JDKs < 1.4
   * as java.lang.String.split() since JDK1.4
   *
   * @param toSplit <code>String</code> which shall be split
   * @param separator <code>String</code> identifies the positions
   * where \"toSplit\" shall be split
   * @return <code>String[]</code> contains the split \"toSplit\"
   */
  public static String[] split(String toSplit, String separator)
  {
    int lowIdx = -1;
    int hiIdx = toSplit.indexOf(separator,lowIdx);
    int arrayIdx = 1;
    
    /* determine length for returnString[] */
    while(hiIdx > lowIdx ){
      lowIdx = hiIdx;
      hiIdx = toSplit.indexOf(separator,lowIdx+1);
      arrayIdx++;
    }
    
    /* fill array entries with different splitted data each */
    String[] splitArr = new String[arrayIdx];
    hiIdx = toSplit.length();
    while( arrayIdx > 0)
      {
	splitArr[arrayIdx-1] = toSplit.substring(lowIdx+1, hiIdx).trim();
	//System.out.println("splitArr[" + (arrayIdx-1) +"] = "+splitArr[arrayIdx-1]);
	arrayIdx--;
	hiIdx = lowIdx;
	lowIdx = toSplit.lastIndexOf(separator,hiIdx-1);
      }
    return splitArr;
  }




}
