/*
 * EvalResult.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.BufferedWriter;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.io.Serializable;
import java.text.SimpleDateFormat;
import java.util.GregorianCalendar;
import java.util.Vector;

/**
 * This subclass of Datacontainer collects all data belonging to a particular evaluation.  It containes methodes to save the result.
 *
 * Attribute:
 * <ol>
 * <li> evalParameter: These are the parameters use for evaluation and algorithm.
 * <li> evalResult: It's a vector which contains the results of the particular comparison
 * </ol>
 * @author (dz)
 * @version 0.1.0
 * @see kea.Datacontainer
 */
public class EvalResult extends Datacontainer implements Serializable {

  private String evalMethod;
  private Parameters evalParameter;
  private Parameters[] astParameters;
  private Vector individuum;
  private Data[] evalResult;
  //  Variablen fuer Datacontainer-Methoden
  private String exportFileName;


  /**
   * Default Constructor
   * Creates a new <code>EvalResult</code> instance with a standard
   * save- and export-filename and the other member variables set to
   * <code>null</code>.
   */
  public EvalResult() {
    GregorianCalendar time =(GregorianCalendar)GregorianCalendar.getInstance();
    String fileName;
    SimpleDateFormat format = new SimpleDateFormat();
    format.setCalendar(time);
    fileName =
      format.format(time.getTime())
      + "_" + time.get(time.SECOND);
    // to avoid problems with your Microsoft Windows system:
    fileName = fileName.replace(':','_');
    fileName = fileName.replace('/','-');
    fileName = fileName.replace(' ','%');
    this.exportFileName = fileName + ".ers";

    //this.evalMethod = "unknown";
    this.evalResult = new Data[]{};
    this.evalParameter = new Parameters();
    this.astParameters = new Parameters[]{};
  }


  /** Constructor
   * Creates a new <code>EvalResult</code> instance with the provided
   * save- and export-filename and the other member variables set to
   * <code>null</code>.
   */
  public EvalResult(String filename) {
    //this.evalMethod = "unknown";
    this.evalResult = new Data[]{};
    this.evalParameter = new Parameters();
    this.astParameters = new Parameters[]{};
    this.exportFileName = filename + ".ers";
  }

  /* public String getEvalMethod(){
     return this.evalMethod;
     }

     public void setEvalMethod(String evalMethod){
     this.evalMethod = evalMethod;
     }
  */

  public Parameters getEvalParameter(){
    return this.evalParameter;
  }

  public Parameters[] getAstParameters(){
    return this.astParameters;
  }

  public void setEvalParameter(Parameters evalParameter){
    this.evalParameter = evalParameter;
  }

  public void setAstParameters(Parameters[] astParameters){
    this.astParameters = astParameters;
  }

  public void setIndividuum(Vector it){
    this.individuum = it;
  }

  public Data[] getEvalResult(){
    return this.evalResult;
  }

  public void setEvalResult(Data[] evalResult){
    this.evalResult = evalResult;
  }


  /**
   * <code>export</code> exports all relevant data of EvalResult in a
   * <code>ExportFileName</code>.ers-file.
   * This file will contain the respective date, parameters and result
   * of the evaluation.
   */
  public void export() {
    Value[] decVector, objVector;
    try {
      Parameters[] tmpParamsArr;
      if (astParameters != null) {
	tmpParamsArr = new Parameters[this.astParameters.length + 1];
	System.arraycopy(this.astParameters, 0,
			 tmpParamsArr, 0, this.astParameters.length );
	tmpParamsArr[astParameters.length] = evalParameter;
      }
      else {
	tmpParamsArr = new Parameters[1];
	tmpParamsArr[0] = evalParameter;
      }


      /* create a GZip-Header containing above tmpParamsArr */
      /* if you want this replace null with the outcommented */

      Datacontainer.writeSetOfParametersGZipTo
	(null,
	//( tmpParamsArr,
	  exportFileName );

      PrintWriter outPrint = new PrintWriter(new BufferedWriter(new FileWriter(exportFileName,true)));
      try{
	outPrint.println("#");
        outPrint.println("#   ***   Begin Parameters of evaluated file   ***");
	Parameter[] tempParamArray;
	for (int i=0; i < astParameters.length; i++){
	  outPrint.println(astParameters[i].toString());
	  if(i < astParameters.length-1) outPrint.println();
	}
      }
      catch(NullPointerException e){
        outPrint.println("#   <<No parameters available!>>");
        Kea.logger.logWarn("No parameters of evaluated file available. Evaluation continues without them.", this);
      }
      outPrint.println("#   ***   End Parameters of evaluated file   ***");

      try{
	outPrint.println("#");
	outPrint.println("#   Evaluated with: ");
	outPrint.println("#");
	outPrint.println(this.evalParameter.toString());
      }
      catch(NullPointerException e){
        outPrint.println("#   <<No parameters available!>>");
        Kea.logger.logWarn("No parameters of evaluaion available. Evaluation continues without them.", this);
      }
      try{
        outPrint.println("###    Evaluation results ...");
        for (int k = 0;k < evalResult.length; k++){
          switch (((Integer)individuum.elementAt(0)).intValue()){

            case 0: for(int j = 0; j < evalResult[k].getDecVector().length; j++){
                      decVector = evalResult[k].getDecVector();
                      outPrint.write(decVector[j].getData().toString() + " ");
                    }
                    outPrint.write("/");
                    // output the Objective Vector of the k-th individual
                    for(int j = 0; j < evalResult[k].getObjVector().length; j++){
                      objVector = evalResult[k].getObjVector();
	              outPrint.write(objVector[j].getData().toString() + " ");
                    }
                    outPrint.println();
                    break;

            case 1: for(int j = 0; j < evalResult[k].getDecVector().length; j++){
                      decVector = evalResult[k].getDecVector();
                      outPrint.write(decVector[j].getData().toString() + " ");
                    }
                    outPrint.println();
                    break;

            case 2: outPrint.println();
                    break;
          }
          individuum.remove(0);
        }
      }
      catch(NullPointerException e){
        outPrint.println("#   No evaluation results found.");
        outPrint.close();
        throw new KeaException("No evaluationdata to write found!");
      }
      catch(ClassCastException e){
        outPrint.close();
        throw new KeaException("Can't cast evaluationresults to string!");
      }
      outPrint.flush();
      outPrint.close();
      if (Kea.logger.infoOn()) {
        Kea.logger.logInfo("Creating file " + exportFileName + "....................OK", this);
        Kea.logger.logInfo("Export to file " + exportFileName + "...................OK", this);
      }
    }
    catch(IOException e) {
      throw new KeaException("IOError occured while working with " + this.exportFileName+".");
    }
  }

  /**
   * <code>getExportFileName</code> returns the name of the file, which
   * will contain the data of <code> EvalResult</code> in ASCII-format.
   *
   * @return a <code>String</code> value containing the actually set
   * filename for export.
   */
  public String getExportFileName() {
    if(this.exportFileName.length()!=0){
        return this.exportFileName;}
        else return null;
    }

   /**
     * <code>setExportFileName</code> defines the name of the ASCII-file,
     * in which the data of <code>EvalResult</code> will be exported.
     * The parameter exportFileName will be extended by .ers.
     *
     * @param exportFileName is the name of the ASCII-file which shall be
     * created by <code>export</code>
     */
    public void setExportFileName(String exportFileName) {
      if (!exportFileName.endsWith(".ers"))
	this.exportFileName = exportFileName + ".ers"; //only, if not used above FilePreparations
      else
	this.exportFileName = exportFileName;
    }
  public void save(){}
  public String getSaveFileName(){return "";}
  public void setSaveFileName(String it){}
}



