/*
 * Kea.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
 
package kea;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.GregorianCalendar;

import kea.algorithms.AbstractAlgorithm;
import kea.algorithms.common.AlgoException;
import kea.comparisons.AbstractComparison;
import kea.displays.AbstractDynamicDisplay;
import kea.displays.AbstractStaticDisplay;
import kea.displays.EvaluationDisplay;
import kea.displays.OnTheFlyDisplay;
import kea.evaluations.AbstractEvaluation;
import kea.problems.AbstractProblem;

/**
 * @author Torsten Kohlen
 * @version 1.0.0
 */
public class Kea {
  private String hlp = "." + File.separator;
  private String keaPath = hlp + "kea";
  private String parameterPath = hlp + "parameter";
  private String outputPath = hlp + "results";
  private String savePath = hlp + "bak";
  private String algoPath = "";
  private String problemPath = "";

  /**
   * Parsermodul, it parses the comandline.
   */
  private Parser parser;

  /**
   * The comparion chosen to compare priviously computed data.
   */
  private AbstractComparison comparison;

  /**
   * The kind of evaluation chosen for Data.
   */
  private AbstractEvaluation evaluation;

  /**
   * The static display to use for visualization of Data.
   */
  private AbstractDynamicDisplay dynamicDisplay;

  /**
   * The dynamic display to use for visualization of Data.
   */
  private AbstractStaticDisplay staticDisplay;

  /**
   * An AlgoState holds all data of an optimisation run.
   */
  private AlgoState currentRun;

  /**
   * Threshold given by the user, which iterationstep is to be saved.
   * The default Value is 1, which means, that each (Step-) result will
   * be save to file.
   */
  private int exportFrequency = 1;

  /**
   * Variable is used to compute the new time difference to decide whether a
   * binary AlgoState is to be saved or not.
   */
  private long lastSaveTime = 0;

  /**
   * Timeperiode the Optimizationrun will be saved (in secondes).
   * default: 60 (saving every minute)
   */
  private int saveFrequency = 60;

  /**
   * Counts the iterationsteps an algorithm makes during one optimizationrun.
   */
  private int iterationCount = 0;

  /**
   * Value needed to identify, if an run is completed. If a user stops KEA,
   * e.g. with CTRL-C, an run may not been finished. The current state will
   * then be saved.
   */
  private boolean finished = false;

  /**
   * A loggingtool for KEA. All classes of this package can user this static
   * tool to write information to a file or StandardOut.
   */
  final static public KEALogger logger = KEALogger.getInstance(Kea.class.getName());

  /**
   * KEA's Version.
   */
  final static public String VERSION = "1.0.0 (RC1)";
  private boolean userBreak = false;
/*****************************************************************************************************************/

  /**
   * Constructor for KEA. Sets the parser.
   */
  public Kea() {
    try {
      String path = new File(this.getClass().getResource(".").getFile()).getParent();
      path = Datacontainer.replaceAll(path, "%20", " ");
      File ini = new File(path,"kea.ini");
      BufferedReader reader = new BufferedReader(new FileReader(ini));
      String line = reader.readLine();

      //step over header
      while (line.startsWith("#")) {
	line = reader.readLine();
      }

      //read the values
      while (line != null) {
	String value, name;
	int index = line.indexOf("=");
	name = (line.substring(0,index)).trim();
	value = line.substring(index+1,line.length());
	if (name.startsWith("KEA")) {
	  keaPath = value;
	}
	else {
	  if (name.startsWith("PARAMETER")) {
	    parameterPath = value;
	  }
	  else {
	    if (name.startsWith("RESULTS")) {
	      outputPath = value;
	    }
	    else {
	      if (name.startsWith("BACKUP")) {
		savePath = value;
	      }
	    }
	  }
	}
	line = reader.readLine();
      }

      if (!keaPath.endsWith(File.separator)){
        keaPath = keaPath + File.separator;
      }
      if (!parameterPath.endsWith(File.separator)){
        parameterPath = parameterPath + File.separator;
      }
      if (!outputPath.endsWith(File.separator)){
        outputPath = outputPath + File.separator;
      }
      if (!savePath.endsWith(File.separator)){
        savePath = savePath + File.separator;
      }

      new File(parameterPath).mkdir();
      new File(outputPath).mkdir();
      new File(savePath).mkdir();

     if (new File(parameterPath).isFile() || new File(outputPath).isFile() || new File(savePath).isFile()) {
       Kea.logger.logError("\nAt least one of the following directories \'parameter\', \'results\' or \'bak\'\n"
	  	           + "is not in the path named by kea.ini and couldn't even be created.",this);
       System.exit(-1);
     }
   }
   catch(Exception e) {
     e.printStackTrace();
     makeIni();
     System.exit(-1);
   }

    parser = new Parser();
    parser.setParaPath(parameterPath);
  }

/*****************************************************************************************************************/
public void setUserBreak(boolean s) {
	userBreak = s;
}

  /**
   *
   */
  private void makeIni(){
    String path = new File(this.getClass().getResource(".").getFile()).getParent();
    path = Datacontainer.replaceAll(path, "%20", " ");

    File ini = new File(path,"kea.ini");
    if (!ini.exists()) {
	System.err.println("***** WARNING - There was no kea.ini-file found!\n" +
	                   "***** WARNING - You cannot run KEA without it.\n" +
	                   "***** WARNING - KEA will now generate a default kea.ini-file\n" +
			   "***** WARNING - and corresponding working directories as well.\n" +
	                   "***** WARNING - If you want to modify the kea.ini-file to your own\n" +
	                   "***** WARNING - purpose (not necessary), you have to restart KEA.\n" +
                           "***** WARNING - Otherwise you can go on directly.");

        //writing default ini-File
	try{
	  BufferedWriter writer = new BufferedWriter(new FileWriter(ini));
	  String fs = File.separator;
	  writer.write("###########################################################");
	  writer.newLine();
	  writer.write("#        KEA INI File - automatic generated by KEA        #");
	  writer.newLine();
	  writer.write("###########################################################");
	  writer.newLine();
	  writer.write("KEA=" + path + fs);
	  writer.newLine();
	  writer.write("PARAMETER="+ path + fs + "parameter" + fs);
	  writer.newLine();
	  writer.write("RESULTS="+ path + fs + "results" + fs);
	  writer.newLine();
	  writer.write("BACKUP="+ path + fs + "bak" + fs);
	  writer.close();
	}
	catch (Exception e) {
	  e.printStackTrace();
	}
      }
  }


  private void setPath(String pathName, String newPath) {
    if (pathName.equalsIgnoreCase("parameter")) {
      this.parameterPath = newPath;
    }
  }


  /**
   *
   */
  private String createDefaultFileName(String[] nameComp){
    String retVal = "";
    GregorianCalendar time = (GregorianCalendar)GregorianCalendar.getInstance();
    SimpleDateFormat format = new SimpleDateFormat("_yyMMdd_HHmmss");

    // build name: string[0]!string[1]!...!string[n]
    for (int i=0;i<nameComp.length;i++) {
      retVal = retVal + nameComp[i].substring(nameComp[i].lastIndexOf(".")+1,nameComp[i].length()) + "!";
    }

    // remove final "!"
    retVal = retVal.substring(0,retVal.length()-1);

    // generate unic Timestamp for filename
    retVal = retVal + format.format(new Date());

    // to avoid problems with your Microsoft Windows system:
    retVal = retVal.replace(':','_');
    retVal = retVal.replace('/','-');
    retVal = retVal.replace(' ','%');
    return retVal;
  }


  /**
   * Creates an Object by the Objects name given as String.
   *
   * @param name the name of the Class, as String, which will be created.
   * @return Object the new created Object.
   */
  private Object makeObject(String name){
    logger.logDebug(name,this);
    Class c;
    try {
      c = Class.forName(name);
      return c.newInstance();
    }
    catch(Exception e){
      logger.logError("Error while creating " + name, e, this);
      return null;
    }
  }


    /**
     * Starts a new Optimizationrun. Algorithm and problem must be initialized
     * before! After each iteration step it is checked if any stopConditions
     * are reached. If the thresholds for saving the data are reached, the
     * data will be saved to file.
     */

	public void startRun() {
		long start, step = 0;

		//initialisind the Display, if user wants
		if (parser.getDisplayEnabled()) {
			String displayName = parser.getDisplayName();
		if (displayName.equals("")) {
				dynamicDisplay = new OnTheFlyDisplay();
		}
		else {
		  Object o = makeObject(displayName);
			  dynamicDisplay = (AbstractDynamicDisplay) o;
		}
		dynamicDisplay.execProgram();
	}

		logger.logInfo("Optimizationrun starts:", this);
		start = System.currentTimeMillis();

		while (!currentRun.getAlgorithm().checkStopCondition() && !userBreak) {
			iterationCount++;
			logger.logInfo(iterationCount + ". Iteration ...",this);


		currentRun.getAlgorithm().iterate();
			step = System.currentTimeMillis();

			logger.logInfo("... done",this);

		// transfer of computed results from algorithm to AlgoState
		currentRun.setIterationCount(iterationCount);
		currentRun.setEvalCount(currentRun.getAlgorithm().getProblem().getEvalCount());
			currentRun.setRuntime((step-start));

			// check, if data must be saved
		if (checkExportCondition()) {
			  try {
				currentRun.export();//ASCII
				logger.logInfo("Last Iterationresult added to file",this);
			  }
			  catch (KeaCoreException ex) {
				System.out.println("KEA aborded with following Problem: " + ex.getMessage());
				logger.logFatal(ex.getMessage());
				System.exit(1);
			  }
		}
		if (checkSaveCondition()) {
				currentRun.save();//BINARY
				logger.logInfo("System saved after:" + saveFrequency + " (sec)" ,this);
			}
			if (parser.getDisplayEnabled()){
				dynamicDisplay.prepareAndShow(currentRun);
			}
	} //while        

	// if not saved so far, save the last iteration Result
		if (!checkExportCondition()) { 
		  try{
		currentRun.export();//ASCII
			logger.logInfo("Final Iterationresult added to file",this);
		  }
		  catch (KeaCoreException ex) {
			System.out.println("KEA aborded with following Problem: " + ex.getMessage());
			System.exit(1);
		  }
		}

	if (parser.getDisplayEnabled()) {
			dynamicDisplay.close();
		}
	if (!userBreak) {
			logger.logInfo("Optimizationrun completed",this);
		}
	}


//    public void startRun() {
//        long start, step = 0;
//
//        //initialisind the Display, if user wants
//        if (parser.getDisplayEnabled()) {
//            String displayName = parser.getDisplayName();
//	    if (displayName.equals("")) {
//                dynamicDisplay = new OnTheFlyDisplay();
//	    }
//	    else {
//	      Object o = makeObject(displayName);
//              dynamicDisplay = (AbstractDynamicDisplay) o;
//	    }
//	    dynamicDisplay.execProgram();
//	}
//
//        logger.logInfo("Optimizationrun starts:", this);
//        start = System.currentTimeMillis();
//
//        while (!currentRun.getAlgorithm().checkStopCondition()) {
//            iterationCount++;
//            logger.logInfo(iterationCount + ". Iteration ...",this);
//
//
//	    currentRun.getAlgorithm().iterate();
//            step = System.currentTimeMillis();
//
//            logger.logInfo("... done",this);
//
//	    // transfer of computed results from algorithm to AlgoState
//	    currentRun.setIterationCount(iterationCount);
//	    currentRun.setEvalCount(currentRun.getAlgorithm().getProblem().getEvalCount());
//            currentRun.setRuntime((step-start));
//
//            // check, if data must be saved
//	    if (checkExportCondition()) {
//              try {
//                currentRun.export();//ASCII
//                logger.logInfo("Last Iterationresult added to file",this);
//              }
//              catch (KeaCoreException ex) {
//                System.out.println("KEA aborded with following Problem: " + ex.getMessage());
//                logger.logFatal(ex.getMessage());
//              }
//	    }
//	    if (checkSaveCondition()) {
//                currentRun.save();//BINARY
//                logger.logInfo("System saved after:" + saveFrequency + " (sec)" ,this);
//            }
//            if (parser.getDisplayEnabled()){
//                dynamicDisplay.prepareAndShow(currentRun);
//            }
//	} //while
//
//	// if not saved so far, save the last iteration Result
//        if (!checkExportCondition()) {
//          try{
//	    currentRun.export();//ASCII
//            logger.logInfo("Final Iterationresult added to file",this);
//          }
//          catch (KeaCoreException ex) {
//            System.out.println("KEA aborded with following Problem: " + ex.getMessage());
//          }
//        }
//
//	if (parser.getDisplayEnabled()) {
//            dynamicDisplay.close();
//        }
//	logger.logInfo("Optimizationrun completed",this);
//    }


    /**
     * Serves for the usage of a method (= metric) for comparing results.
     * The method used is determined by the classvariable "comparison" in Kea.
     * The object "AbstractComparison" has to be fixed before this method is evoked.
     * This has to be done in "createComparison".
     *  MARION
     */
    public void startComparison() {
	CompResult compResult = comparison.compare();
        // display.prepareDisplay(compResult);//legt nur GNUPlot-lesbare Dateien an
	compResult.export();
        logger.logInfo("KEA finished comparison", this);
    }


    /**
     * Starts a new evaluation. After the evaluation is made the result will
     * be saved to a File
     */
    public void startEvaluation() {
      System.out.println("***** KEA begins to evaluate");
      System.out.println("*****");
      EvalResult[] results = evaluation.evaluate();
      for (int i=0;i<results.length;i++) {
        results[i].export();
      }
      System.out.println("***** KEA wrote evaluationdata to File");

      if (parser.getDisplayEnabled()) {
	String displayName = parser.getDisplayName();
    	if (displayName.equals("")) {
          staticDisplay = new EvaluationDisplay();
	}
	else {
	  Object o = makeObject(displayName);
          staticDisplay = (AbstractStaticDisplay) o;
        }
        staticDisplay.prepareAndExec(results);
      }
      System.out.println("*****");
      System.out.println("***** KEA finished evaluation");
    }

/******************************************************************************/
/*****                      Initialising Methods for KEA                      */
/******************************************************************************/

  /**
   * Initializes KEA for running a user specified algorithm on a user specified
   * problem. First an instance of the chosen problem will be created an sent
   * to the parser for setting the the parameter values for this problem. The
   * same is done with the algorithm.
   */
  
  
  private void initAlgoProb() {
		AbstractProblem prob;

		System.out.println("*****");
		System.out.println("***** 2. Initialising Optimizationrun:");
		System.out.println("*****");
	System.out.println("***** Problem: " + parser.clParseProblemName());
		System.out.println("***** Algorithm: " + parser.clParseAlgoName());

		//initialising AlgoState, which holds all data for an optimization run
		System.out.println("*****");
		currentRun = new AlgoState();

		try {

		  //1. set the problem of the algorithm an the problem-prameter
		  prob = (AbstractProblem)makeObject(parser.clParseProblemName());
		  prob.setParameter(parser.clParseProblem(prob.getParameter()));

		  //2. create and setup the run
		  currentRun.setAlgorithm((AbstractAlgorithm)makeObject(parser.clParseAlgoName()));
		  currentRun.getAlgorithm().setup(parser.clParseAlgo(currentRun.getAlgorithm().getParameter()), prob);
		  currentRun.setExportStepSize(exportFrequency);
	  currentRun.setSaveStepDuration(saveFrequency);
		} //try
        
		catch (AlgoException ex) {
			System.out.println("KEA aborted with following Problem: " + ex.getMessage());
			logger.logFatal(ex.getMessage());
			System.exit(1);
		}
		catch (KeaCoreException kce) {
			System.out.println("KEA aborted with following Problem: " + kce.getMessage());
			logger.logFatal(kce.getMessage());
			System.exit(1);
		}
        
		  //3. set filename for this run
		if (parser.getFileName() != "") {
		  currentRun.setExportFileName(outputPath + parser.getFileName() + ".ast");
		  currentRun.setSaveFileName(savePath + parser.getFileName() + ".sav");
		} 
		else {
		  //user has not given a name, so create a default name
		  String[] names = new String[2];
		  names[0] = currentRun.getAlgorithm().getParameter().getMembership();
		  names[1] = currentRun.getAlgorithm().getProblem().getParameter().getMembership();
		  currentRun.setExportFileName(outputPath + createDefaultFileName(names) + ".ast");
		  currentRun.setSaveFileName(savePath + createDefaultFileName(names) + ".sav");
		}
		System.out.println("***** Data-File: " + currentRun.getExportFileName());
		System.out.println("***** Backup-File: " + currentRun.getSaveFileName());
	}

  
  
  
  
  
//  private void initAlgoProb() {
//        AbstractProblem prob;
//
//        System.out.println("*****");
//        System.out.println("***** 2. Initialising Optimizationrun:");
//        System.out.println("*****");
//	System.out.println("***** Problem: " + parser.clParseProblemName());
//        System.out.println("***** Algorithm: " + parser.clParseAlgoName());
//
//        //initialising AlgoState, which holds all data for an optimization run
//        System.out.println("*****");
//        currentRun = new AlgoState();
//
//        //1. set the problem of the algorithm an the problem-prameter
//        prob = (AbstractProblem)makeObject(parser.clParseProblemName());
//
//        prob.setParameter(parser.clParseProblem(prob.getParameter()));
//
//        //2. create and setup the run
//        currentRun.setAlgorithm((AbstractAlgorithm)makeObject(parser.clParseAlgoName()));
//        currentRun.getAlgorithm().setup(parser.clParseAlgo(currentRun.getAlgorithm().getParameter()), prob);
//    	currentRun.setExportStepSize(exportFrequency);
//	currentRun.setSaveStepDuration(saveFrequency);
//
//        //3. set filename for this run
//        if (parser.getFileName() != "") {
//          currentRun.setExportFileName(outputPath + parser.getFileName() + ".ast");
//          currentRun.setSaveFileName(savePath + parser.getFileName() + ".sav");
//        }
//        else {
//          //user has not given a name, so create a default name
//          String[] names = new String[2];
//          names[0] = currentRun.getAlgorithm().getParameter().getMembership();
//          names[1] = currentRun.getAlgorithm().getProblem().getParameter().getMembership();
//          currentRun.setExportFileName(outputPath + createDefaultFileName(names) + ".ast");
//          currentRun.setSaveFileName(savePath + createDefaultFileName(names) + ".sav");
//        }
//        System.out.println("***** Data-File: " + currentRun.getExportFileName());
//        System.out.println("***** Backup-File: " + currentRun.getSaveFileName());
//    }


    /**
     * Initializes KEA for running a user specified algorithm on a user specifed problem.
     * First an instance of the chosen problem will be created an sent to the parser for
     * setting the the parameter values for this problem. The same is done with the
     * algorithm.
     */
    private void initAlgoContinue() {
	logger.logInfo("KEA will finish stopped run " + parser.getclAlgoState(), this);
	currentRun = (AlgoState)AlgoState.load(parser.getclAlgoState());

	logger.logDebug("Problem to be continued: " + currentRun.getAlgorithm().getProblem().getParameter().getMembership(), this);
        logger.logDebug("Algorithm to be continued: " + currentRun.getAlgorithm().getParameter().getMembership(), this);

	this.iterationCount = currentRun.getIterationCount();
        logger.logDebug("Iterations made before halt: " + this.iterationCount, this);
    }//initAlgoContinue


    /**
     *
     */
    private void initComparison() {
      System.out.println("*****");
      System.out.println("***** 2. Initialising Comparison:");
      System.out.println("*****");
      System.out.println("***** Used Comparison: " + parser.clParseCompName());
      System.out.println("*****");

      //1. Create choosen comparison
      comparison = (AbstractComparison)makeObject(parser.clParseCompName());

      //2. Set the parameter for the comparison
      comparison.setParameter(parser.clParseComp(comparison.getParameter()));

      //3. Set the user given filename, if exists
      if (parser.getFileName() != "") {
        comparison.setFileName(outputPath + parser.getFileName() + ".crs");
      }
      else {
        //user has not given a name, so create a default name
        String[] names = new String[1];
        names[0] = comparison.getParameter().getMembership();
        comparison.setFileName(outputPath + createDefaultFileName(names) + ".crs");
      }
      System.out.println("***** Result-File: " + comparison.getFileName());
    }


    /**
     * Initializes KEA for doing an evaluation on former done Optimisation run.
     * The evaluation is initialized and either a user given filename is set
     * or a KEA default filename is generated.
     *
     */
    private void initEvaluation() {
      System.out.println("*****");
      System.out.println("***** 2. Initialising Evaluation:");
      System.out.println("*****");
      System.out.println("***** Used Evaluation: " + parser.clParseEvalName());
      System.out.println("*****");

      //1. Create choosen evaluation
      evaluation = (AbstractEvaluation)makeObject(parser.clParseEvalName());

      //2. Set the parameter for the evaluation
      evaluation.setParameter(parser.clParseEvaluation(evaluation.getParameter()));

      //3. Set the user given filename, if exists
      if (parser.getFileName() != "") {
        evaluation.setFileName(outputPath + parser.getFileName() + ".ers");
      }
      else {
        //user has not given a name, so create a default name
        String[] names = new String[1];
        names[0] = evaluation.getParameter().getMembership();
        evaluation.setFileName(outputPath + createDefaultFileName(names) + ".ers");
      }
      System.out.println("***** Result-File: " + evaluation.getFileName());
    } //initEvaluation


   /**
    * This method creates an object of a class from a module and then returns
    * its default parameters.
    */
   private void initDefaultParameters(){
      Parameters p = null;
      String dep = null;
      String[] defNames = parser.clParseDefaultNames();

      Object obj = makeObject(defNames[0]);

      if (obj != null) {
        if (obj instanceof AbstractAlgorithm) {
          p = ((AbstractAlgorithm) obj).getParameter();
	  if (defNames.length == 2) {
            Object o = makeObject(defNames[1]);
            ((AbstractAlgorithm) obj).setup(null, (AbstractProblem) o);
            p = ((AbstractAlgorithm) obj).getParameter();
	    dep = ((AbstractProblem) o).getParameter().getMembership();
          }//else fehler!!
	}

	      if (obj instanceof AbstractProblem) {
		  		p = ((AbstractProblem) obj).getParameter();
		  		((AbstractProblem)obj).cleanUp();
	      }
	      if (obj instanceof AbstractEvaluation) {
		  p = ((AbstractEvaluation) obj).getParameter();
	      }
	      if (obj instanceof AbstractComparison) {
		  p = ((AbstractComparison) obj).getParameter();
	      }
	  }
	  if(p==null || p.getMembership() == null){
	    logger.logError("At least Parameters.getMembership()!=null is needed ");
	    System.exit(1);   // should be replaced by a suitable exception
	  }

	  if(logger.infoOn() && p != null){
	    logger.logInfo(p.getMembership(), this);
	  }
	  createDefaultParamFile(p, dep);
      //}
   }

/*****************************************************************************************************************/
/*****
/*****************************************************************************************************************/


    /**
     * Checks if the Values of the current iterationstep must be saved to file.
     * Possible (return) Values:<p>
     *
     * - TRUE: The current step matches the exportFrequency.<p>
     * - FALSE: The current step does not match the exportFrequency.<p>
     *
     * @see AbstractAlgorithm
     * @return boolean IterationCount reached exportFrequency.
     *
     */
    private boolean checkExportCondition() {
      if (this.exportFrequency != 0) {
 	if ((iterationCount % exportFrequency) == 0) {
	    return true;
        }
	else return false;
      }
      else return false;
    }

  /**
   * The main method. After parsing the commandline Kea will be initialized depending
   * on the job the user wants KEA to do.
   *
   * @see RunTypes
   * @param argv commandline options for runnig KEA as String[].
   *
   */
//  public static void main(String[] argv) {
//    Kea run;
//    run = new Kea();
//
//    System.out.println("***** Welcome to KEA (c) 2003 by PG 419, University of Dortmund, Germany *****");
//    System.out.println("***** ");
//    System.out.println("***** KEA comes with ABSOLUTELY NO WARRANTY.");
//    System.out.println("***** This is free software, and you are welcome to redistribute it");
//    System.out.println("***** under certain conditions; see License.txt for details.");
//    System.out.println("***** ");
//    System.out.println("***** You are using KEA v" + Kea.VERSION);
//    System.out.println("***** ");
//    System.out.println("***** 1. Initialising KEA");
//
//    if (run.parser.splitCL(argv)) {
//      if (run.parser.getExportStep() >= 0) {
//        run.exportFrequency = run.parser.getExportStep();
//      }
//
//      // initialising Timecounter for savingperiods
//      run.lastSaveTime = System.currentTimeMillis();
//
//      //hard Programmbreak will be caught with this thread
//      Runtime.getRuntime().addShutdownHook(new BreakThread(run));
//
//      switch (run.parser.getRunType()){
//	case RunTypes.RUN_ALGOSTATE:
//          System.out.println("***** Requested KEA-Action: Continue Optimizationrun");
//          System.out.println("***** Exportfrequency: " + run.exportFrequency);
//          System.out.println("***** Savefrequency: " + run.saveFrequency + " (sec)");
//          run.initAlgoContinue();
//          run.startRun();
//          run.currentRun.getAlgorithm().getProblem().cleanUp();
//	  break;
//	case RunTypes.RUN_ALGOANDPROB:
//          System.out.println("***** Requested KEA-Action: New Optimizationrun");
//          System.out.println("***** Exportfrequency: " + run.exportFrequency);
//          System.out.println("***** Savefrequency: " + run.saveFrequency + " (sec)");
//          run.initAlgoProb();
//          run.startRun();
//          run.currentRun.getAlgorithm().getProblem().cleanUp();
//          break;
//	case RunTypes.RUN_COMPARISON:
//          System.out.println("***** Requested KEA-Action: Comparing Optimization results");
//          run.initComparison();
//          run.startComparison();
//	  break;
//	case RunTypes.RUN_EVALUATION:
//          System.out.println("***** Requested KEA-Action: Evaluation on Optimization results");
//          run.initEvaluation();
//          run.startEvaluation();
//	  break;
//        case RunTypes.RUN_MAKEDEFAULT:
//          System.out.println("***** Requested KEA-Action: Create default Parameter File");
//          run.initDefaultParameters();
//          break;
//      }//switch
//    }//if..
//
//    System.out.println("*****");
//    System.out.println("***** Thank you for using KEA - Goodbye *****");
//    run.finished = true;
//  }// end of main


	public static void main(String[] argv) {
	  Kea run;
	  run = new Kea();

	  System.out.println("***** Welcome to KEA (c) 2003 by PG 419, University of Dortmund, Germany *****");
	  System.out.println("***** ");
	  System.out.println("***** You are using KEA v" + Kea.VERSION);
	  System.out.println("***** ");
	  System.out.println("***** 1. Initialising KEA");

	  if (run.parser.splitCL(argv)) {
		if (run.parser.getExportStep() >= 0) {
		  run.exportFrequency = run.parser.getExportStep();
		}
		if (run.parser.getSaveStep() >= 0) {
		  run.saveFrequency = run.parser.getSaveStep();
		}

		// initialising Timecounter for savingperiods
		run.lastSaveTime = System.currentTimeMillis();

		//hard Programmbreak will be caught with this thread
		//Runtime.getRuntime().addShutdownHook(new BreakThread(run));
		BreakThread bt = new BreakThread(run);
		bt.start();

		switch (run.parser.getRunType()){
	  case RunTypes.RUN_ALGOSTATE:
			System.out.println("***** Requested KEA-Action: Continue Optimizationrun");
			run.initAlgoContinue();
			System.out.println("***** Exportfrequency: " + run.exportFrequency);
			System.out.println("***** Savefrequency: " + run.saveFrequency + " (sec)");
			run.startRun();
			run.currentRun.getAlgorithm().getProblem().cleanUp();
		break;
	  case RunTypes.RUN_ALGOANDPROB:
			System.out.println("***** Requested KEA-Action: New Optimizationrun");
			System.out.println("***** Exportfrequency: " + run.exportFrequency);
			System.out.println("***** Savefrequency: " + run.saveFrequency + " (sec)");
			run.initAlgoProb();
			run.startRun();
			run.currentRun.getAlgorithm().getProblem().cleanUp();
			break;
	  case RunTypes.RUN_COMPARISON:
			System.out.println("***** Requested KEA-Action: Comparing Optimization results");
			run.initComparison();
			run.startComparison();
		break;
	  case RunTypes.RUN_EVALUATION:
			System.out.println("***** Requested KEA-Action: Evaluation on Optimization results");
			run.initEvaluation();
			run.startEvaluation();
		break;
		  case RunTypes.RUN_MAKEDEFAULT:
			System.out.println("***** Requested KEA-Action: Create default Parameter File");
			run.initDefaultParameters();
			break;
		}//switch
	  }//if..
    
	  if (run.userBreak) {
		  run.stopRun();
	  }
		  else {run.finished = true;}
    
	  System.out.println("*****");
	  System.out.println("***** Thank you for using KEA - Goodbye *****");
	}// end of main




    /**
     * Wird ausgefuehrt, wenn der Benutzer einen Lauf vorzeitig beendet (userBreak==true).
     */
//    public void stopRun() {
//        if (!finished) {
//          try{
//            currentRun.save();
//            currentRun.export();
//
//            currentRun.getAlgorithm().getProblem().cleanUp();
//            System.out.println("Dateien vor Abbruch gespeichert!");
//            System.out.println("KEA ***** Programm stopped before work was regularly finished! *****");
//          }
//          catch (KeaCoreException ex) {
//            System.out.println("KEA aborded with following Problem: " + ex.getMessage());
//          }
//        }
//    }//end of stopRun


	public void stopRun() {
		if (!finished) {
			currentRun.save();
			//currentRun.export();

			currentRun.getAlgorithm().getProblem().cleanUp();
			System.out.println("Dateien vor Abbruch gespeichert!");
			System.out.println("KEA ***** Programm stopped before work was regularly finished! *****");
		}
	}//end of stopRun





    public String getParameterPath(){
      return this.parameterPath;
    }


    /**
     * Depending on the variable "saveFrequency" the value (true or false)
     * for saving is set and returned. It is assumed that "runtime" from
     * "algorithm" counts the minutes since the beginning of the optimization.
     *
     * @return If the currentRun should be saved.
     */
    public boolean checkSaveCondition() {
	long currentTime = System.currentTimeMillis();

	if (currentTime - lastSaveTime >= (saveFrequency * 1000)) {
      	    lastSaveTime = currentTime;
	    return true;
	}
	else return false;
    }


    /**
     * This method creates a parameter file with the default values of a
     * specific class from any module also the unique name of the class.
     * @param fileName	The file name to be created (adds "_default.param" automatically) which is
     * @param dep	Name of the associated problem
     */
    private boolean createDefaultParamFile(Parameters para, String depName){
      String fileName = para.getMembership();
      String fileNameToWrite = fileName;
      int index = fileNameToWrite.lastIndexOf(".");

      if (index != -1) {
        fileNameToWrite = fileNameToWrite.substring(index+1,fileNameToWrite.length());
      }

      try {
        if (depName != null) {
	  index = depName.lastIndexOf(".");
	  if (index != -1) {
	    depName = depName.substring(index+1,depName.length());
	  }
	  fileNameToWrite = fileNameToWrite + "_with_" + depName;
        }

        File temp = new File(this.parameterPath, fileNameToWrite +"_defaults.param");
	if(Kea.logger.debugOn())Kea.logger.logDebug(temp.getAbsolutePath(),this);

        BufferedWriter writer = new BufferedWriter(new FileWriter(temp));

	writer.write("###########################################################");
	writer.newLine();
	if (depName != null) {
          writer.write("#   This is a config file for " + fileName + " with " + depName);
        }
	else {
          writer.write("#   This is a config file for " + fileName);
        }
	writer.newLine();
	writer.write("#   This file was automatically generated by Kea");
	writer.newLine();
	writer.write("###########################################################");
	writer.newLine();
	writer.write(fileName);
	writer.newLine();

        if(logger.infoOn()) Kea.logger.logInfo("Creating file " + fileNameToWrite+"_defaults.param....................OK ", this);

        if (para.getParameterArray() != null) {
	  Parameter[] paramArray= para.getParameterArray();
	  for (int i=0;i<paramArray.length;i++) {
	    // get the parameter out of the array and get the 3 necessary attributes
	    Parameter p = paramArray[i];
	    String name = p.getName();
	    Object data = p.getData();
	    String tip = p.getToolTip();

            // now write the 3 attributes into one line
	    writer.write(name + "=" + data + " # " + tip);
	    writer.newLine();
          }
        }

	writer.close();

	if(logger.infoOn()) logger.logInfo("Writing to file " + fileNameToWrite+"_defaults.param..................OK ", this);

        return true;
      }
      catch (Exception e){
        logger.logError("While trying to create file " + fileName, e, this);
	return false;
      }
    } //createDefaultParamFile
}//Kea
