/*
 * Parameters.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.Serializable;

/**
 * Datencontainer, der Parameterobjekte beschreibt.
 * Der OptionHandler arbeitet auf Objekten dieser Klasse.
 * Bei jeder Klasse, die Parameter verwendet, sollte es eine
 * Dokumentation geben (z.B. Txt-Datei), in der beschrieben wird,
 * welche Bedeutung die einzelnen Eintr"age haben.
 *
 * Attribute:
 * - membership: Eindeutiger Verweis auf den Algorithmus/das Problem/...,
 *   zu dem das Parameterobjekt geh\uFFFDrt.
 *
 * -parameterDocu: Dokumentation zu den einzelnen ParameterNamen
 *
 * - parameterNames: Namen der im Parameterobjekt gespeicherten Parameter.
 *
 * - parameterValues: Werte der im Parameterobjekt gespeicherten Parameter.
 *   Muss gleiche L\uFFFDnge wie Vektor von parameterNames haben.
 * 
 * @author Stefan Kusper, Peter Senft, Norman Welp
 * @version 0.1.0
 */
public class Parameters implements Serializable {
	
    /**
     * This Value represents the class (algorithm, problem, etc.), which
     * parameters are stored within this Parameters.<p>
     * When creating the parameterset for your class you must store here the
     * full qualified class name of your class (e.g. kea.problems.kursawe)
     */
    private String membership;
    private Parameter[] parameterArray;
    private String saveFileName;
    private String exportFileName;

    public Parameters() {
    }

    public Parameters(int parameterZahl) {
	parameterArray = new Parameter[parameterZahl];
	for(int i=0;i<parameterArray.length;i++){
	    parameterArray[i] = new Parameter();
	}
    }

    public String getMembership(){
	return membership;
    }

    public void setMembership(String membership){
	this.membership = membership;
    }

    public void export() {
	// TODO: oder auch nicht?!
    }

    public String getSaveFileName(){
	return this.saveFileName;
    }

    public void setSaveFileName(String fileName) {
	this.saveFileName = fileName;
    }

    public String getExportFileName(){
	return this.exportFileName;
    }

    public void setExportFileName(String fileName) {
	this.exportFileName = fileName;
    }

    public void setParameterArray(Parameter[] params){
	parameterArray = params;
    }

    public Parameter[] getParameterArray(){
	return parameterArray;
    }

    public Parameter getParameterByName(String name){
	if (parameterArray != null) {
	    for (int i=0; i < parameterArray.length; ++i) {
		if (parameterArray[i].getName().equalsIgnoreCase(name)) {
		    return parameterArray[i];
		}
	    }
	}
        return null;
    }

    public boolean setParameter(Parameter param) {
	if (parameterArray != null ) {
	    for (int i = 0; i < parameterArray.length; ++i) {
		if (parameterArray[i].getName().equalsIgnoreCase(param.getName())) {
		    parameterArray[i] = param;
		    return true;
		}
	    }
	}
	return false;
    }

    public boolean addParameter(Parameter param) {
	boolean retval = false;

	// test if such a Parameter (by name) is already in the list.
	if (this.getParameterByName(param.getName()) == null ) {
	    retval = true;
	    /*
	     * not quite elegant, but it works :)
	     */

	    if (parameterArray != null) {
		Parameter newParameterArray[] = new Parameter[this.parameterArray.length + 1];
		System.arraycopy(this.parameterArray, 0,
				 newParameterArray, 0, this.parameterArray.length );
		newParameterArray[parameterArray.length] = param;
		this.parameterArray = newParameterArray;
	    }
	    else {
	    this.parameterArray = new Parameter[1];
	    this.parameterArray[0] = param;
	    }
	}
	else { // parameter is already in list and not added
	    retval = false;
	}

	return retval;
    }

/******************************************************************************/
    public boolean addParameter(String name, Value val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Object val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Boolean val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Integer val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Short val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Byte val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Character val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Double val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Long val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, Float val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, boolean val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, int val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, short val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, long val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, byte val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, char val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, double val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, float val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

    public boolean addParameter(String name, String val, String toolTip) {
        Parameter p = new Parameter(name, val, toolTip);
        return addParameter(p);
    }

/******************************************************************************/

    public String toString() {
	String retval = new String();

	retval = ">>>   " + getMembership(); // + "\n# Parameters:";
	if(parameterArray==null || parameterArray.length==0){
		return retval+"\n There are no parameters avaible.";
	}
	else{
	
		for (int i = 0; i < parameterArray.length; ++i) {
		    if (parameterArray[i] == null) {
			retval = retval + "\nnull";
		    }
		    else {
			retval = retval + "\n" + parameterArray[i].getName();
			// if u want the type use the following line
			// retval = retval + " = (" + parameterArray[i].getType().getTypeName() + "): ";
			retval = retval + " = " + parameterArray[i].getData();
			if (parameterArray[i].getRange() != null) {
			    retval = retval + " |";
			    for (int j = 0; j < parameterArray[i].getRange().length; ++j) {
				retval = retval + " ";
				retval = retval + parameterArray[i].getRange()[j].getMinVal();;
				retval = retval + "/" + parameterArray[i].getRange()[j].getMaxVal();
			    }
			}
			retval = retval + " # " + parameterArray[i].getToolTip();
		    }
		}
		return retval;
    	}
    }


}
