/*
 * Parser.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.Vector;

/**
 * This class generally handles the parsing of command line arguments and also the parsing of files.
 * It offers a public method to split a command line into logical parts and stores them internally.
 * Then you can get the passed arguments (names of files) and call it's file parse methods to update
 * parameters.
 *
 * @author Norman Welp
 * @version $Revision: 723 $
 */
public class Parser {


    /**
     * default constructor
     */
    public Parser() { }

    /**
     * This method is called to dump out command line help.
     */
    private void dumpOutHelp(){
		System.out.println();
		System.out.println("Usage: Kea <option> <value> [<option>] [<value>]");
		System.out.println("    where <option> can be -A (--algorithm), -P (--problem),");
		System.out.println("        -C (--comparison), -E (--evaluation), -CD (--createdefaults),");
		System.out.println("        -ES (--exportstep), -SS (--savestep), FN (--filename), -D (--display) ");
		System.out.println("    where <value> is a file or in case of -CD/-D the name of a class");
		System.out.println("    -D may have no value if used with optimizations/evaluations to enable a default display");
		System.out.println("    Values of -ES, -SS must be integers and are only used for optimizations.");
		System.out.println();
		System.out.println("Examples:");
		System.out.println("    kea -A myBinaryAlgoStateFile.sav");
		System.out.println("    (Will continue a paused and saved optimization run)");
		System.out.println();
		System.out.println("    kea -A myNSGA2.param -P myZdt6.param -SS 90 -ES 10 -FN nsga_zdt6 -D");
		System.out.println("    (Will start a new optimization run with default display enabled and with");
		System.out.println("    binary saving every 90 secs and textual protocol every 10th iteration");
        System.out.println("    and the resulting .ast file will by named nsga_zdt6)");
		System.out.println();
		System.out.println("    kea -C Hypervolume_defaults.param");
		System.out.println("    (Will do a comparison based on hypervolume and it's default param file)");
		System.out.println();
		System.out.println("    kea -E myParetoSet.param -D");
		System.out.println("    (Will do an evaluation based on ParetoSet and it's param file and displays the results)");
		System.out.println();
		System.out.println("    kea -CD kea.algorithms.populationBased.nsga2.NSGA2");
		System.out.println("    (Will create a param file with the default values for NSGA2)");
		System.out.println();
		System.out.println("For further information about options/values and their meanings and");
		System.out.println("the semantics of calls check the kea main documentation.");
		System.out.println("");
		System.out.println("Act wisely... :-)");

    }

    /**
     * Important method which should be called to split a command line string.
     * This method should be called before any other method of this class !
     * @param	argv	The arguments given by the command line String
     * @return	<b>true</b> if the command line string could be parsed successfully, <b>false</b> otherwise
     */
    public boolean splitCL(String[] argv) {

		// first we need to check the arguments for help calls or emptiness (which leeds to a help call :-))
		if (argv.length==0){
			dumpOutHelp();
			return false;
		}


		Vector params = new Vector(); // we need this vector for the command line data handling

		/* let's have a look if we find any of the possible KEA flags
		* We will temporarely use a vector to search for flags because it's easier to trim arrays that way
		*/

		// check for help
		boolean paramExportExists   = false;
		boolean paramNameExists     = false;
		boolean paramSaveExists     = false;
		boolean paramDisplayExists  = false;

		for (int i=0;i<argv.length;i++){
			if (argv[i].equals("--help") || argv[i].equals("-H")) {
				dumpOutHelp();
				return false;
			}
			//check for exportstep and savestep, if we get a match, convert to lower case (we need this for vector search later on)

			if(argv[i].equals("--filename")|| argv[i].equals("-FN")){
				if(paramNameExists){
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				paramNameExists=true;
				params.add("-filename");
			}

			if(argv[i].equals("--display") || argv[i].equals("-D")){
				if(paramDisplayExists){
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				paramDisplayExists=true;
				params.add("-display");
			}

			if(argv[i].equals("--exportstep") || argv[i].equals("-ES")){
				if(paramExportExists){
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				paramExportExists=true;
				params.add("-exportstep");
			}

			if(argv[i].equals("--savestep") || argv[i].equals("-SS")){
				if(paramSaveExists){
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				paramSaveExists=true;
				params.add("-savestep");
			}


			// no flag in this substring
			if( !(argv[i].equals("-ES")
				|| argv[i].equals("-FN")
			 	|| argv[i].equals("-SS")
				|| argv[i].equals("-D")
				|| argv[i].equals("--filename")
				|| argv[i].equals("--exportstep")
				|| argv[i].equals("--savestep")
				|| argv[i].equals("--display")))
			params.add(argv[i]);
		}

		if(Kea.logger.debugOn())Kea.logger.logDebug("Starting to split command line string...", this);
		try{
			int index;

			// check for user name
			index = params.indexOf("-filename");
			if(index!=-1){
				if(params.size()<=index+1){
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				else{
					fileName=(String)params.get(index+1);
					if(Kea.logger.debugOn())Kea.logger.logDebug("File name detected --> " +fileName, this);
					params.remove(index);
					params.remove(index);
					
				}
			}

			// check for exportStep
			index = params.indexOf("-exportstep");
			if(index!=-1){
				if(params.size()<=index+1){
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				else{
					exportStep=new Integer((String)params.get(index+1));
					if(Kea.logger.debugOn())Kea.logger.logDebug("Export step detected --> " +exportStep, this);
					params.remove(index);
					params.remove(index);
					//System.out.println("removed: " + params.remove(index));
					//System.out.println("removed: " + params.remove(index));
				}
			}

		// check for saveStep
		index = params.indexOf("-savestep");
		if(index!=-1){
			if(params.size()<=index+1){
				Kea.logger.logError("Invalid input ! Enter correct params !", this);
				dumpOutHelp();
				return false;
			}
			else{
				saveStep=new Integer((String)params.get(index+1));
				if(Kea.logger.debugOn())Kea.logger.logDebug("Save step detected --> " +saveStep, this);
				params.remove(index);
				params.remove(index);
				
			}
		}

			// check for display
		index = params.indexOf("-display");
		if(index!=-1){
			displayEnabled = true;
			if(params.size()<=index+1){
				
				displayName = "";
				if(Kea.logger.debugOn())Kea.logger.logDebug("Display detected. Using default display", this);
				params.remove(index);
			}
			else{
				displayName = (String)params.get(index+1);
				if(displayName.startsWith("-")){
					displayName="";
					if(Kea.logger.debugOn())Kea.logger.logDebug("Display detected. Using default display", this);
					params.remove(index);
				}
				else{
					if(Kea.logger.debugOn())Kea.logger.logDebug("Display detected --> " + displayName, this);
					params.remove(index);
					params.remove(index);

				}
			}
		}


		}
		catch(Exception e){
			Kea.logger.logError("Invalid input ! Enter correct params !", e, this);
			dumpOutHelp();
			return false;
		}

		// copy trimmed command line back into array
		argv = new String[params.size()];
		argv = (String[])params.toArray(argv);
		/*
		for(int q=0;q<argv.length;q++){
			System.out.println(argv[q]);
		}
		*/
		if(argv.length==0){
			Kea.logger.logError("Invalid input ! Enter correct params !", this);
			dumpOutHelp();
			return false;
		}
		boolean paramAExists = false;
		boolean paramPExists = false;
		boolean paramEExists = false;
		boolean paramCExists = false;
		boolean paramKEAExists = false;
		boolean paramBinRunExists = false;

		// if a createDefaults call occured create defaults and return true or false if an error

		if(argv[0].equals("-CD") || argv[0].equals("--createdefaults")){
			if(argv.length>1){
				defaultNames = new String[1];
				if(argv.length>2) defaultNames = new String[2];
				action=RunTypes.RUN_MAKEDEFAULT;
				defaultNames[0] = argv[1];
				if(argv.length>2){
					if(argv.length>3){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					defaultNames[1]=argv[2];
				}
			return true;//createDefaultsFile(argv[1]);
			}
			Kea.logger.logError("Invalid input ! Enter correct params !", this);
			dumpOutHelp();
			return false;
		}

		for(int i=0;i<argv.length;i++){

			// now parse the command line arguments

			if(argv[i].equals("-A") || argv[i].equals("--algorithm")){
				if(paramAExists) {
					Kea.logger.logError("Parameter A already exists ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				else {
					paramAExists=true;
					if(argv.length<=i+1){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					if(argv[i+1].endsWith(".param")){
						action=RunTypes.RUN_ALGOANDPROB;
						algoFileName=argv[i+1];
						if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+1], this);
					}
					if(argv[i+1].endsWith(".sav")){
						if(argv.length>i+2){
							Kea.logger.logError("Invalid input ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						action=RunTypes.RUN_ALGOSTATE;
						algoStateName=argv[i+1];
						if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+1], this);
						return true;
					}
					if(argv.length<=i+2){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false	;
					}

					if(!argv[i+2].equals("-P") || argv[i+2].equals("--problem")){
						Kea.logger.logError(" Wrong input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					else {
						if(argv.length<=i+3){
							Kea.logger.logError("Invalid input ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						if(argv.length>i+4){
							Kea.logger.logError("Invalid input ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+3], this);
						paramPExists=true;
						probFileName=argv[i+3];
						return true;
					}
				}
			}

			else
			// check for problem cl part
			if(argv[i].equals("-P") || argv[i].equals("--problem")){
				if(paramPExists) {
					Kea.logger.logError("Parameter P already exists ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				else {
					if(argv.length<=i+1){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					action=RunTypes.RUN_ALGOANDPROB;
					probFileName=argv[i+1];
					paramPExists=true;
					if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+1], this);
					if(argv.length<=i+2){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					if(!argv[i+2].equals("-A") || argv[i+2].equals("--algorithm")){
						paramAExists=true;
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					else {
						if(argv.length<=i+3){
							Kea.logger.logError("Invalid input ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						if(argv.length>i+4){
							Kea.logger.logError("Invalid input ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						algoFileName=argv[i+3];
						if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+3], this);
						return true;
					}
				}
			}
			else
				// check for evaluation cl part
				if(argv[i].equals("-E") || argv[i].equals("--evaluation")){
					if(paramEExists) {
					Kea.logger.logError("Parameter E already exists ! Enter correct params !", this);
					dumpOutHelp();
					return false;
				}
				else {
					if(argv.length<=i+1){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					if(argv.length>i+2){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					paramEExists=true;
					action=RunTypes.RUN_EVALUATION;
					evalFileName=argv[i+1];
					if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+1], this);
					return true;
				}
				}
				else
				// check for comparison cl part
				if(argv[i].equals("-C") || argv[i].equals("--comparison")  ){
					if(paramCExists) {
					Kea.logger.logError("Parameter C already exists ! Enter correct params !", this);
					dumpOutHelp();
					return false;
					}
					else {
					if(argv.length<=i+1){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					if(argv.length>i+2){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
					}
					paramCExists=true;
					action=RunTypes.RUN_COMPARISON;
					compFileName=argv[i+1];
					if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+1], this);
					return true;
					}
				}
				/*
				else
					// check for kea/tool cl part
					if(argv[i].equalsIgnoreCase("-KEA")) {
						if(paramCExists) {
							Kea.logger.logError("Parameter KEA already exists ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						else {
						if(argv.length<=i+1){
						Kea.logger.logError("Invalid input ! Enter correct params !", this);
						dumpOutHelp();
						return false;
						}
						if(argv.length>i+2){
							Kea.logger.logError("Invalid input ! Enter correct params !", this);
							dumpOutHelp();
							return false;
						}
						paramKEAExists=true;
						action=RunTypes.RUN_KEACONF;
						toolFileName=argv[i+1];
						if(Kea.logger.debugOn())Kea.logger.logDebug("Reference to file detected --> " + argv[i+1], this);
						return true;
						}
					}
				*/
					// in case none of the above parts was found the call was malformed
					else {
					Kea.logger.logError("Invalid input ! Enter correct params !", this);
					dumpOutHelp();
					return false;
					}

		}
		//logger.logDebug("Successfully split the command line string !");
		return true;
    }

    /**
     * This is the main method to parse a file for updating a given parameters object
     * @param file	The file to parse
     * @param paramObj	The parameters to update
     * @return The updated Parameters object
     */
    private Parameters fileParse(File file, Parameters paramObj) {
		KEALogger.getInstance("logger.kea").logDebug("paramObj ="+paramObj);
 	
	BufferedReader reader;
	String line;
	Vector paramVec = new Vector();
    KEALogger.getInstance("logger.kea").logDebug("file ="+file);
	Parameter[] paramArray = paramObj.getParameterArray();
 	try{
	    reader = new BufferedReader(new FileReader(file));
	    boolean firstLinePassed=false;
	    line = reader.readLine();
		if (line==null) throw new Exception();

	    while (line.startsWith("#") || line.equals("")){	// skip current line as long as it starts with '#'
			    line = reader.readLine();
			}


	    while(line!=null){	// as long as there are lines in the file...

			if(!firstLinePassed){

		    	if(Kea.logger.debugOn())Kea.logger.logDebug("Class name found: " + line, this);
		    	firstLinePassed=true;
		    	line=reader.readLine();
			}
			if(line==null) {
				return paramObj;
			}

			while (line.startsWith("#") || line.equals("")){	// skip current line as long as it starts with '#'
			    line = reader.readLine();
				if(line==null) {
					return paramObj;
				}

			}

			// now split the current line into parts to get the value and name out of it
			String value, name;
			int equals = line.indexOf("=");
			name = (line.substring(0,equals)).trim();
			String s = line.substring(equals+1);
			s = s.trim();
			if(s.indexOf("#")!=-1){
				value = s.substring(0, s.indexOf("#"));
				value = value.trim();
			}
			else{
		    	  value =s.trim();
			}

			Parameter p = paramObj.getParameterByName(name);
			if(p==null) throw new Exception();

			switch (p.getTypeInfo()){

				case Type.BOOLEAN:
					if(value.equalsIgnoreCase("true")) p.setData(true);
					if(value.equalsIgnoreCase("false")) p.setData(false);
					if(!(value.equalsIgnoreCase("false") || value.equalsIgnoreCase("true"))) throw new NumberFormatException();
					break;

				case Type.INTEGER:
					int in =new Integer(value).intValue();
					p.setData(in);
					break;
				case Type.SHORT :
					short sh = new Short(value).shortValue();
					p.setData(sh);
					break;
				case Type.LONG :
					long lo = new Long(value).longValue();
					p.setData(lo);
					break;
				case Type.BYTE :
					byte by = new Byte(value).byteValue();
					p.setData(by);
					break;
				case Type.CHAR :
					char ch = new Character(value.charAt(0)).charValue();
					p.setData(ch);
					break;
				case Type.DOUBLE :
					double dou = new Double(value).doubleValue();
					p.setData(dou);
					break;
				case Type.FLOAT :
					float fl = new Float(value).floatValue();
					p.setData(fl);
					break;
				case Type.STRING:
					p.setData(value);
					break;
			}

			// read next line to proceed
			line = reader.readLine();
	  	}
	    paramObj.setParameterArray(paramArray);
	    reader.close();
	}
	catch(Exception e){
		Kea.logger.logError("Invalid file format detected ! While parsing " + file.getName(), e, this);
		return null;
	}
	return paramObj;
    }


    /**
     * This method parses a param file for algorithms. The given Parameters
     * will be updated by the content of the file and then returned.
     *
     * @param paramObj	The parameters object to be updated
     * @return The updated parameter object
     */
    public Parameters clParseAlgo(Parameters paramObj) {
      File file = new File(paraPath + algoFileName);

      return fileParse(file, paramObj);
    }


    /**
     * This method parses a param file for comparisons. The given Parameters
     * will be updated by the content of the file and then returned.
     *
     * @param paramObj	The parameters object to be updated
     * @return The updated parameter object
     */
    public Parameters clParseComp(Parameters paramObj) {
      File file = new File(paraPath + compFileName);

      return fileParse(file, paramObj);
    }


    /**
     * This method parses a param file for evaluatons. The given Parameters
     * will be updated by the content of the file and then returned.
     *
     * @param paramObj	The parameters object to be updated
     * @return The updated parameter object
     */
    public Parameters clParseEvaluation(Parameters paramObj){
      File file = new File(paraPath + evalFileName);

      return fileParse(file, paramObj);
    }


    /**
     * This method parses a paramfile for problems. The given Parameters will
     * be updated by the content of the file and then returned.
     *
     * @param paramObj	The parameters object to be updated
     * @return The updated parameter object
     */
    public Parameters clParseProblem(Parameters paramObj) {
      File file = new File(paraPath + probFileName);

      return fileParse(file, paramObj);
    }


    /**
     * This method parses a param file for kea.
     * The given Parameters will be updated by the content of the file and then returned.
     * @param paramObj	The parameters object to be updated
     * @return The updated parameter object
     */
    public Parameters clParseTool(Parameters paramObj) {
	File file = new File(toolFileName);

	return fileParse(file, paramObj);
    }


    /**
     * This method extracts the class name of an algorithm from an algorithm
     * parameter file and returns it.
     *
     * @return The class name as a String
     */
    public String clParseAlgoName() {
      String line ;

      try {
        BufferedReader reader = new BufferedReader(new FileReader(paraPath + algoFileName));
	line = reader.readLine();
	while (line.startsWith("#") || line.equals("")) {
	  //nothing to do just go on reading
	  line = reader.readLine();
        }
      }
      catch (Exception e) {
        Kea.logger.logError("While parsing algorithm name : " , e, this);
	return null;
      }

      return line.trim();
    }


    /**
     * This method returns the name for creating a default Parameter File.
     *
     * @return The name as a String
     */
    public String[] clParseDefaultNames() {
      return defaultNames;
    }


    /**
     * This method extracts the class name of an comparison from a comparison
     * parameter file and returns it.
     *
     * @return The class name as a String
     */
    public String clParseCompName() {
      String line;

      try {
        BufferedReader reader = new BufferedReader(new FileReader(paraPath + compFileName));
	line = reader.readLine();
	while (line.startsWith("#") || line.equals("")) {
	  //nothing to do just go on reading
	  line = reader.readLine();
        }
      }
      catch (Exception e) {
        Kea.logger.logError("While parsing comparison name : " , e, this);
	return null;
      }

      return line.trim();
    }


    /**
     * This method extracts the class name of an evaluation from an evaluation
     * parameter file and returns it.
     *
     * @return The class name as a String
     */
    public String clParseEvalName() {
      String line ;

      try {
        BufferedReader reader = new BufferedReader(new FileReader(paraPath + evalFileName));
	line = reader.readLine();
	while (line.startsWith("#") || line.equals("")) {
          //nothing to do just go on reading
	  line = reader.readLine();
        }
      }
      catch (Exception e) {
        Kea.logger.logError("While parsing evaluation name : " , e, this);
	return null;
      }

      return line.trim();
    }


    /**
     * This method extracts the class name of a problem from a problem parameter
     * file and returns it.
     *
     * @return The class name as a String
     */
    public String clParseProblemName(){
      String line ;

      try {
        BufferedReader reader = new BufferedReader(new FileReader(paraPath + probFileName));
	line = reader.readLine();
	while (line.startsWith("#") || line.equals("")){
	  //nothing to do just go on reading
	  line = reader.readLine();
        }
      }
      catch (Exception e) {
        Kea.logger.logError("While parsing problem name : " , e, this);
        return null;
      }

      return line.trim();
    }


    /**
     * This method allows the caller to find out which action should be executed depending on the run type
     * @return Therun type (as an int)
     */
    public int getRunType(){
		return action;
    }

    /**
     * This method returns the name of the binary algo state file
     * @return The name as a String
     */
    public String getclAlgoState(){
	return algoStateName;
    }

	public String getFileName(){
		return fileName;
	}

	public int getSaveStep(){
		return saveStep.intValue();
	}

	public int getExportStep(){
		return exportStep.intValue();
	}
  	/**
     * This method allows the caller to find out whether or not the user enabled display
     * @return <b>true</b> if display was enabled, <b>false</b> otherwise
     */
	public boolean getDisplayEnabled(){
		if(displayEnabled) return true;
		else return false;
	}
	/**
	* This method returns the name of a display class the user wants to use for visualization
	* @return The name of the display class
	*/
	public String getDisplayName(){

		return displayName;
	}


    public void setParaPath(String newPath){
      this.paraPath = newPath;
    }


    private String algoStateName;

    private int action = -1;
    private boolean displayEnabled = false;
    private String[] defaultNames;
    private String algoFileName;
    private String probFileName;
    private String evalFileName;
    private String compFileName;
    private String toolFileName;
    private String displayName;
    private String defaultName;
    private String fileName = "";
    private Integer saveStep = new Integer(-1);
    private Integer exportStep =new Integer( -1);
    private String paraPath;

	/*
	public static void main (String args[]){
		Parser p =new Parser();
		p.splitCL(args);
		if(p.getRunType()==RunTypes.RUN_MAKEDEFAULT){
		String[] names = p.clParseDefaultNames();
			for(int i=0;i<names.length;i++){
				System.out.println(names[i]);
			}
		}
	}
	*/
}
