/*
 * Type.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea;

import java.io.Serializable;

/**
 * Type.java represents the type of values for a Value-object.
 * The types are used analogue to the standard java objects:
 * <ul>
 * 	<li> Boolean
 * 	<li> Integer
 * 	<li> Short
 * 	<li> Byte
 * 	<li> Character
 * 	<li> Long
 * 	<li> Double
 * 	<li> Float
 * 	<li> String
 * </ul>
 * 
 * @author Stefan Kusper, Peter Senft, Norman Welp
 * @version 0.1.0
 */
public class Type implements Serializable, Cloneable {

    /**
     * Internal used representation of the type as an integer.
     */
    private int typeInfo;

    /**
     * Array of human readable type names. For internal use only.
     */
    private static String typeName[] =
	{"Boolean", "Integer", "Short", "Long",
	"Byte", "Character", "Double", "Float", "String"};

    /**
     * Human readable definitions of the types and their integer
     * representations.
     */
    public static final int UNDEFINED = -1;
    public static final int BOOLEAN = 0;
    public static final int INTEGER = 1;
    public static final int SHORT = 2;
    public static final int LONG = 3;
    public static final int BYTE = 4;
    public static final int CHAR = 5;
    public static final int DOUBLE = 6;
    public static final int FLOAT = 7;
    public static final int STRING = 8;

    /**
     * Returns true, if the actual type is usable for our algorithms. That
     * means, it is a number type oder boolean.
     */
    public boolean isValid() {
      boolean retval = true;

      switch (this.typeInfo) {
        case Type.BOOLEAN :
          retval = true;
          break;
        case Type.BYTE :
          retval = true;
          break;
        case Type.CHAR :
          retval = false;
          break;
        case Type.DOUBLE :
          retval = true;
          break;
        case Type.FLOAT :
          retval = true;
          break;
        case Type.INTEGER :
          retval = true;
          break;
        case Type.LONG :
          retval = true;
          break;
        case Type.SHORT :
          retval = true;
          break;
        case Type.STRING :
          retval = false;
          break;
        case Type.UNDEFINED :
          retval = false;
          break;
        default :
          retval = false;
          break;
      }

      return retval;
    }

    /**
     * Returns true if the actual type is a number
     */
    public boolean isNumber() {
      boolean retval = true;

      switch (this.typeInfo) {
        case Type.BOOLEAN :
          retval = false;
          break;
        case Type.BYTE :
          retval = true;
          break;
        case Type.CHAR :
          retval = false;
          break;
        case Type.DOUBLE :
          retval = true;
          break;
        case Type.FLOAT :
          retval = true;
          break;
        case Type.INTEGER :
          retval = true;
          break;
        case Type.LONG :
          retval = true;
          break;
        case Type.SHORT :
          retval = true;
          break;
        case Type.STRING :
          retval = false;
          break;
        case Type.UNDEFINED :
          retval = false;
          break;
        default :
          retval = false;
          break;
      }

      return retval;
    }

    /**
     * Returns true if the acutal type is a boolean
     */
    public boolean isBoolean() {
      boolean retval = false;

      if (this.typeInfo == Type.BOOLEAN) {
        retval = true;
      }
      else {
        retval = false;
      }

      return retval;
    }

    /**
     * Standard constructor for this object. The type is set to Type.UNDEFINED.
     */
    public Type() {
        this.typeInfo = Type.UNDEFINED;
    }

    /**
     * Constructor with given integer representation of a type.
     * Checks are made during initialisation. If type is not defined the
     * its value changes to Type.UNDEFINED
     *
     * @param type - integer representation of the type.
     */
    public Type(int type) {
        this.setTypeInfo(type);
    }

    /**
     * Sets the type to the given value and checks for correct type values.
     * If undefined type value is given it is set to Type.UNDEFINED.
     *
     * @param typeInfo - integer representaiton of the type.
     */
    public void setTypeInfo(int typeInfo) {
        if ((typeInfo >= 0) && (typeInfo <= 8)) {
            this.typeInfo = typeInfo;
        }
        else {
            this.typeInfo = Type.UNDEFINED;
            return;
        }
    }

    /**
     * Returns the type value as an integer.
     *
     * @return typeInfo - integer representation of this type object.
     */
    public int getTypeInfo() {
        return typeInfo;
    }

    /**
     * Returns a human readable string representation of the type object.
     *
     * @return retval - string representation of the type object.
     */
    public String getTypeName() {
        String retval = "Undefined";
        if (typeInfo == Type.UNDEFINED)
            retval = "Undefined";
        else
            retval = typeName[typeInfo];
        return retval;
    }

    public Object clone() {
        Type retval = null;
        try {
            retval = (Type)super.clone();
        }
        catch (CloneNotSupportedException e) {
            System.out.println(e);
        }
        retval.setTypeInfo(this.typeInfo);
        return retval;
    }
}
