/*
 * AbstractAlgorithm.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms;

import java.io.Serializable;

import kea.Data;
import kea.Parameters;
import kea.problems.AbstractProblem;

/**
 * If you want to add an new algorithm to <code>KEA</code>, this algorithm
 * must inherit from this abstract class.
 *
 * @author Torsten Kohlen
 * @version 0.1.0
 */
public abstract class AbstractAlgorithm implements Serializable {
    /**
     * Holds the Parameter for the algorithm
     *
     * @see Parameters
     */
    protected Parameters param = null;

    /**
     * To interact with the problem an algorithem has to solve, this problem
     * is hold in this attribute. By using the problem method <i>score()</i>,
     * an algorithm gets the functionvalue evaluation for the recently computed
     * result. The algorithm get an Instance of the problem in the setup method.
     */
    protected AbstractProblem problem = null;

//*****************************************************************************/
    /**
     * The implementation of this method must prepare the algorithm to run
     * properly. After the call of setup, the algorithm must be ready to
     * iterate. Each new call of setup must reset the algorithm and all former
     * computed result are lost. Errors during setup can be reported through
     * exceptions.
     *
     * @param para The Parameters a user set to run this algorithm. If para is
     *             <b>null</b> the algorithm must use its default Parameters.
     *        prob The Problem this algorithm has to solve.
     */
    abstract public void setup(Parameters para, AbstractProblem prob);

    /**
     * The implementation of this method must check if the algorithm is
     * finished with its calcualtion. KEA needs to know when an algorithm got
     * his final results.
     */
    abstract public boolean checkStopCondition();

    /**
     * Implementing this method must do one iterationstep of the algorithm.
     * Hence, the algorithm has no steps it finishes after only one iteration.
     * This method bilds the core of the algorithm, by calling this method
     * all computing must be done. Each call produces an intermediate result.
     * This result can then be accessed by calling <i>getResult()</i>. KEA will
     * call this method until <i>checkStopCondition()</i> returns true.
     * Thereafter the final Result can accessed with <i>getResult()</i>, too
     *
     */
    abstract public void iterate();

    /**
     * The implementation of this method must achieve that the return is either
     * the current intermediate result of he algorithm, or when
     * <i>checkStopCondition()</i> retruns true, the final result of the
     * algorithm.
     *
     * @return Data[] A set of result-data, e.g. a single Value or population
     *                of pareto-optimal-values.
     * @see Data
     */
    abstract public Data[] getResults();

    /**
     * Notwendigkeit muss noch geprueft werde!
     */
    abstract public void setResults(Data[] myDataSet);

    /**
     * The implementataion of this method must return the current parameterset
     * for the algorithm. If there are no Parameters set in <i>setup()</i> a
     * call of this method MUST return the default Parameters of the algorithm.
     * <p>
     * KEA needs the default Parameters to create a Parameter-File, where the
     * user can reset the Parameters to his own needs.
     *
     * @see Parameters
     * @return The current or default set of Prameters for this algorithm. If
     *         the Parameters can not be returned because e.g. the problem is
     *         not set simply return <b>null</b>
     */
    abstract public Parameters getParameter();

    /**
     * This method returns the Problem which was set in <i>setup()</i>.
     *
     * @return The current value of problem.
     */
    public AbstractProblem getProblem(){
        return this.problem;
    }
}



