/*
 * DataSet.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.common;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Comparator;

import kea.Data;
import kea.Range;
import kea.Type;
import kea.Value;

/**
 * @author Stefan Kusper, Peter Senft, Norman Welp
 * @version $Revision: 723 $
 */
public class DataSet implements Cloneable, Serializable {
    protected Data[] data;
    protected Range[] decRanges;
    protected Type[] decTypes;
    protected Type[] objTypes;
    protected static Comparator comparator;

    /**
     * Standard Comparator (DataSetSorter) will be created.
     * @param size number of elements in this set
     */
    public DataSet(int size) {
        data = new Data[size];
        this.comparator = new DataSetSorter(0);
    }

    /**
     * Standard Comparator (DataSetSorter) will be created.
     * Number of elements in this set is set to 0.
     */
    public DataSet() {
        data = new Data[0];
        this.comparator = new DataSetSorter(0);
    }

    /**
     * Standard Comparator (DataSetSorter) will be created.
     * @param d the Data array to use in this object
     */
    public DataSet(Data[] d) {
        this.data = d;
        this.comparator = new DataSetSorter(0);
    }

    /**
     * This constructor merges two Data arrays.
     * Standard Comparator (DataSetSorter) will be created.
     * @param d1 a data array
     * @param d2 the other Data array
     */
    public DataSet(Data[] d1, Data[] d2) {
        this.data = d1;
        this.merge(new DataSet(d2));
        this.comparator = new DataSetSorter(0);
    }

    /**
     * This constructor can be used to create an DataSet object with a initial
     * size and a comparator for sorting.
     * @param size the number of Data objects that will be used in this set
     * @param c the comparator that will be used for sorting as default
     */
    public DataSet (int size, Comparator c) {
        data = new Data[size];
        comparator = c;
    }

    /**
     * @return the array of ranges for the decision vectors
     */
   	public Range[] getDecRanges(){
   		return(this.decRanges);
   	}

    /**
     * @return the array of types for the decision vectors
     */
   	public Type[] getDecTypes(){
   		return(this.decTypes);
   	}

    /**
     * @return the array of types for the objective vectors
     */
   	public Type[] getObjTypes(){
   		return(this.objTypes);
   	}

    /**
     * @return the data array
     */
   public Data[] getData() {
    	if (data!=null)
            return data;
    else return null; //use exception handling if needed:
    // else throw new Exception("DataSet.getData: reference is null");
   }

   public Data getData(int pos) {
    	if (data!=null && pos < data.length)
            return data[pos];
    else return null; //use exception handling if needed:
    // else throw new Exception("DataSet.getData: reference is null");
   }

   /**
    * Set the Data array in this set to the given array.
    * @param d array of Data objects
    */
   public void setData(Data[] d) {
        this.data = d;
   }

   /**
    * Set the decision value ranges array in this set to the given array.
    * @param d array of Range objects
    */
   public void setDecRanges(Range[] r) {
    this.decRanges = r;
   }

   /**
    * Set the decision value Type array in this set to the given array.
    * @param d array of Type objects
    */
    public void setDecTypes(Type[] t) {
        this.decTypes = t;
    }

   /**
    * Set the objective value Type array in this set to the given array.
    * @param d array of Type objects
    */
    public void setObjTypes(Type[] t) {
        this.objTypes = t;
    }

   /**
    * Remove an element from the set.
    * @param d the Data that should be removed
    */
	public void remove(Data d) {
        Data[] nData = new Data[this.data.length-1];
        int j = 0;
        for (int i = 0; i < data.length; ++i) {
            if(data[i]!=d){
                nData[j++] = this.data[i]; // j wird erst ausgewertet, dann inkementiert
            }
        }
        this.data = nData;
        System.gc();
    }

   /**
    * Remove an element from the set.
    * @param pos the position of the Data that should be removed
    * @return the removed Data
    */
	public Data remove(int pos) {
        Data retval = data[pos];
        Data[] nData = new Data[this.data.length-1];
        int j = 0;
        for (int i = 0; i < data.length; ++i) {
            if(i!=pos){
                nData[j++] = this.data[i]; // j wird erst ausgewertet, dann inkementiert
            }
            else retval = data[i];
        }
        this.data = nData;
        // call garbage collection ?
        return retval;
    }

	/**
     * Add an element to the set.
     * @param d the element to be inserted
     * @param pos the position where the element should be inserted
     */
    public void insert(Data d, int pos) {
        Data[] nData = new Data[0];
        if(pos < this.data.length) {
        int j=0;
        nData = new Data[this.data.length+1];
            for (int i = 0; i < data.length; i++) {
                if(i!=pos) {
                    nData[j++] = this.data[i];
                }
                else {
                    nData[j++] = d;
                    nData[j++] = this.data[i];
                }
            }
        }
        this.data = nData;
        //else throw new Exception("array index out of range");
    }

    /**
     * Add Data at the end of the set.
     * @param d the Data object to add at the end
     */
        public void append(Data d) {
        if (this.data != null) {
        Data[] nData = new Data[this.data.length+1];
        System.arraycopy(this.data,0,nData,0,this.data.length);
        nData[nData.length-1] = d;
        this.data = nData;
        }
        else {
            this.data = new Data[1];
            this.data[0] = d;
        }
    }

    /**
     * Add Data at the beginning of the set.
     * @param d the Data object to add at the beginning
     */
    public void prepend(Data d) {
        if (this.data != null) {
        Data[] nData = new Data[this.data.length+1];
        System.arraycopy(this.data,0,nData,1,this.data.length);
        nData[0] = d;
        this.data = nData;
        }
        else {
            this.data = new Data[1];
            this.data[0] = d;
        }
    }
    /**
     * Change the positions of to elements in the set.
     * @param pos_a position of the first element
     * @param pos_b position of the second element
     */
    public void exchange(int pos_a, int pos_b) {
        if (pos_a < this.data.length && pos_b < this.data.length) {
            Data d = this.data[pos_a];
            this.data[pos_a] = this.data[pos_b];
            this.data[pos_b] = d;
        }
    }

    /**
     * Merge two sets. Add all elements in s to this set and sort the new set.
     * @param s the DataSet to merge with this DataSet.
     */
	public void merge(DataSet s) {
        if (s.getData()!=null && this.getData()!=null) {
        	Data[] newData = new Data[s.getData().length + this.data.length];
            int i;
			for (i=0; i<this.getData().length; ++i) {
                newData[i] = this.data[i];
            }
            for(int j=0; j<s.getData().length; ++j) {
				newData[i+j] = s.getData()[j];
            }
            this.data = newData;
            this.sort();
        }
    }

    /**
     * Sorts the Data array, uses the Comparator given as a constructor
     * parameter. If no Comparator is given a DataSetSorter will be used and
     * the array will be sorted by the value of the objective vector.
     */
    public void sort() {
        if (this.data != null) return; // if there is nothing to sort...
        if (this.comparator != null)
            Arrays.sort(this.data, comparator);
        else Arrays.sort(this.data, new DataSetSorter(0));
    }

    /**
     * Sorts the array with the given Comparator. This will not change the
     * default Comparator in this object.
     * @param c The Comparator to use for sorting the Data array.
     */
    public void sort(Comparator c) {
        Arrays.sort(this.data, c);
    }

    /**
     * Using this is the same as writing "sort(new DataSetSorter(fNo));"
     * It's just shorter.
     * @param fNo index number of the entry in the objective vector
     */
    public void sort(int fNo) {
        Arrays.sort(data, new DataSetSorter(fNo));
    }

    /**
     * @return the string representation of this class
     */
    public String toString() {
        String s = new String();
        for (int i = 0; i < data.length; i++) {
            s += data[i].toString();
        }
        return s;
    }

    /**
     * use for testing only!
     */
    public static Value[] randomDoubleValue(int size) {
        Value[] v = new Value[size];
        for (int i = 0; i < v.length; i++) {
            v[i]=new Value(Math.random());
        }
        return v;
    }

    /**
     * @return a copy of this object
     */
    protected Object clone() throws CloneNotSupportedException {
        DataSet retval = null;
        try {
            retval = (DataSet)super.clone();

            Type[] n_decTypes = new Type[0];
            if (this.decTypes != null)
                n_decTypes = new Type[decTypes.length];
            for (int i = 0; i < n_decTypes.length; i++) {
                n_decTypes[i] = (Type)(this.decTypes[i].clone());
            }

            Type[] n_objTypes = new Type[0];
            if (this.objTypes != null)
                n_objTypes = new Type[objTypes.length];
            for (int i = 0; i < n_objTypes.length; i++) {
                n_objTypes[i] = (Type)(this.objTypes[i].clone());
            }

            Comparator n_comparator = this.comparator;

            Range[] n_decRanges = new Range[0];
            if (this.decRanges!=null)
                n_decRanges = new Range[decRanges.length];

            for (int i = 0; i < n_decRanges.length; i++) {
                n_decRanges[i] = (Range)(this.decRanges[i].clone());
            }
            Data[] n_data = new Data[data.length];
            for (int i = 0; i < this.data.length; i++) {
                n_data[i] = (Data)(data[i].clone());
            }
            retval.setData(n_data);
        }
        catch (CloneNotSupportedException e) {
            System.out.println(e);
        }
     return retval;
    }

    /**
     * Used for testing only!
     */
    public static void main(String[] argv) {
        DataSet ds = new DataSet(5);
        DataSet ds_clone;
        DataSet ds2 = new DataSet(5);
        for (int i = 0; i < ds.getData().length; i++) {
            ds.getData()[i] = new Data(randomDoubleValue(3), randomDoubleValue(3) );
        }
        for (int i = 0; i < ds2.getData().length; i++) {
            ds2.getData()[i] = new Data(randomDoubleValue(4), randomDoubleValue(4) );
        }
        Data dat1 = new Data(randomDoubleValue(4), randomDoubleValue(4) );
        Data dat2 = new Data(randomDoubleValue(4), randomDoubleValue(4) );

        System.out.println(dat1 + "\n" + dat2 + "\n");
        System.out.println("Teste insert und append:");
        ds.insert(dat1, 2);
        System.out.println(ds + "\n\n");
        ds.append(dat2);
        System.out.println(ds);
        ds.prepend((Data)dat2.clone());
        dat2.clear();
        System.out.println(ds);
/*
        System.out.println("Zufllige Daten:");
        System.out.println(ds);
        System.out.println("*************************************************");
        System.out.println("sortiere Array:");
        ds.sort(new DataSetSorter(1)); //funktioniert auch, wenn Zahl zu gross !?
        System.out.println(ds);
        System.out.println("*************************************************");
        System.out.println("Entferne Eintrage mit remove:");
        ds.remove(ds.getData(0));
        System.out.println(ds);
        System.out.println("*************************************************");
        System.out.println("Tausche 0 und 2");
        ds.exchange(0,2);
        System.out.println(ds);
        try {
            Object tmp;
            System.out.println("Erzeuge Kopie des Objektes");
            tmp = ds.clone();
            ds_clone = (DataSet)tmp;
            System.out.println("ndere Original");
            ds.getData()[0] = new Data(randomDoubleValue(1), randomDoubleValue(1) );
            //ds.remove(0);ds.remove(0);
            System.out.println("und gebe beides aus:");
            System.out.println(ds);
            System.out.println("- - - - - - - - - - - - - - - - - - - - - - -");
            System.out.println((DataSet)ds_clone);
        }
        catch (CloneNotSupportedException e) {
            System.out.println("Cloning error");
        }
*/
    } //end of main
};
