/*
 * OptSet.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.common;

import java.io.Serializable;
import java.util.Comparator;

import kea.Data;
import kea.Value;

public class OptSet extends DataSet implements Cloneable, Serializable {
    private Comparator domSort;
    private Comparator fctSort;
    public Data[] nonDominatedSet;

    /**
     * Some constructors with different numbers od parameters:
     * @param size the number of Data-objects in this DataSet
     * @param fNo the number of the objective function that should be used for sorting.
     * @param f the Comparator that will be used for objective function sorting.
     * @param d the Comparator that will be for finding the non dominated set.
     */
    public OptSet(int size, Comparator f, Comparator d) {
        super(size, f);
        this.domSort = d;
    }

    public OptSet(Data[] d) {
        super(d);
    }
	public OptSet(int size, Comparator f) {
        super(size,f);
        this.domSort = new OptDataDomination();
    }

	public OptSet(int size, int fNo) {
        super(size, new DataSetSorter(fNo));
        this.domSort = new OptDataDomination();
    }

	public OptSet(int size) {
        super(size);
        this.domSort = new OptDataDomination();
        this.fctSort = new DataSetSorter(0);
    }

    /**
     * deletes the non dominated set in this OptSet
     */
    public void clearNonDominated() {
        this.nonDominatedSet = null;
        System.gc();
    }

    /**
     * Sets the non dominated set in this OptSet.
     * @param the new dominated set in this OptSet
     */
    public void setNonDomSet(Data[] d) {
        this.nonDominatedSet = d;
    }

    /**
     * @return the comparator that is used to sort by objective function
     */
    public Comparator getFctComparator() {
        return this.fctSort;
    }

    /**
     * Sets the comparator that is used to sort by objective function
     * @param the new comparator
     */
    public void setFctComparator(Comparator c) {
        this.fctSort = c;
    }

    /**
     * @return the comparator that is used to sort by domination
     */
    public Comparator getDomComparator() {
        return this.domSort;
    }
    /**
     * Sets the comparator that is used to sort by domination
     * @param the new comparator
     */
    public void setDomComparator(Comparator c) {
        this.domSort = c;
    }

    /**
     * This method calculates the non-dominated set.
     * It uses the algorithm described in "Multi objective Optimization using
     * Evolutionary Algorithms", by Kalyanmoy Deb.
     * p38 "Approach 3: Kung et al.'s Efficient Method"
     *
     * @return an array containing the non dominated set
     */
    public Data[] nonDomSet() {
        sort(0); //sortiert aufsteigend!
        this.nonDominatedSet = front(this.data);
        return this.nonDominatedSet;
    }

    /**
     * Part of the algorithm that returns the non dominated set.
     */
    private Data[] merge2(Data[] left, Data[] right) {
        boolean rDom = false;
        boolean lDom = false;
        for (int i = 0; i < left.length; ++i) {
            for (int j = 0; j < right.length; ++j) {
                if (domSort.compare(left[i], right[j]) < 0) {
                    rDom = true;
                    //System.out.println("rDom = true");
                }
                if (domSort.compare(left[i], right[j]) > 0) {
                    lDom = true;
                    //System.out.println("lDom = true");
                }
            }
        }
        if (rDom) return right;
        if (lDom) return left;
        //System.out.println("merging : " + right[0] + left[0]);
        return new DataSet(left, right).getData();
    }

    /**
     * Part of the algorithm that returns the non dominated set.
     */
    private Data[] front(Data[] d) {
        Data[] left, right, retval;
        if (d.length <= 1)
            return d;
        else {
            left = new Data[(int)d.length/2];
            System.arraycopy(d,0,left,0,left.length);
            if (d.length % 1 == 0) {
                right = new Data[(int)d.length/2];
                System.arraycopy(d,left.length,right,0,right.length);
            }
            else {
                right = new Data[(int)d.length/2 + 1];
                System.arraycopy(d,left.length,right,0,right.length);
            }
        }
        retval = merge2(front(left), front(right));
        return retval;
    }

    /**
     * @return a copy of this object
     */
    public Object clone() {
        OptSet retval = null;
        try {
            retval = (OptSet)super.clone();
        }
        catch (CloneNotSupportedException e) {
            //System.out.println(e);
        }
        Data[] n_ndset = new Data[0];
        if (this.nonDominatedSet != null)
            n_ndset = new Data[this.nonDominatedSet.length];
        for (int i = 0; i < n_ndset.length; i++) {
            n_ndset[i] = (Data)(this.nonDominatedSet[i].clone());
        }
        retval.setNonDomSet(n_ndset);
        retval.setFctComparator(fctSort);
        retval.setDomComparator(domSort);
        return retval;
    }

    /**
     * use for testing only!
     */
    public static void main(String argv[]) {
        System.out.println("Beginne Test: OptSet");
        OptSet os = new OptSet(0);
        System.out.println("Initialisiere TestOptSet");
        for (int i=0; i<100; i++) {
            os.append(new Data(os.randomDoubleValue(5), os.randomDoubleValue(5)));
        }
        Value[] v1 = new Value[5];
        Value[] v2 = new Value[5];
        for(int i=0; i<5; ++i) {
            v1[i] = new Value(0.05);
            v2[i] = new Value(0.05);
        }
        os.append(new Data(v1,v1));
        os.append(new Data(v2,v2));
        System.out.println("Ausgabe");
        System.out.println(os);
        os.nonDomSet();
        System.out.println("Ausgabe der nicht dominierten Menge:");
        for (int i = 0; i < os.nonDominatedSet.length; i++) {
            System.out.print(os.nonDominatedSet[i]);
        }
        System.out.println("Test OptSet beendet");
    }
}; //end of class

