/*
 * Dops.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.dops;

import java.io.Serializable;
import java.util.Random;
import java.util.Vector;

import kea.Data;
import kea.Parameter;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.algorithms.AbstractAlgorithm;
import kea.algorithms.common.AlgoException;
import kea.problems.AbstractProblem;

/**
 * The dops algorithm.
 * The core class of dops.
 * 
 * @author Philipp Limbourg
 * @version 0.1.0
 * @see kea.algorithms.AbstractAlgorithm
 */

public class Dops extends AbstractAlgorithm implements Serializable {
  /**
   * Constructor of dops. Please set your default values here.
   */
  public Dops() {
    //Defaults:
    particleCount=20;
    cubesPerDimension=10000;
    generationsToRun=1000;
    repsize=100;
    w=0.4;
    c1=1;
    c2=1.2;
    printout=1;
    seekerPercentage=0.3;
    spacing=1;
    particleSpacing=1;
    pRepsize=1;
    reptype=0;
    threshold=10;
  }

    /**
     * Checks the stopCondition
     */
    public  boolean checkStopCondition() {
      if (iterationCount>=generationsToRun) {
	return true;
      }
      return false;
    }
    /**
     * Gets the current parameter settings of Mopso
     * @return Parameters The settings of Mopso
     */
    public Parameters getParameter() {
    Parameters params=new Parameters();
    Parameter[] param=new Parameter[14];
    param[0]= new Parameter();
    param[0].setName("Particles");
    param[0].setToolTip("Number of particles");
    param[0].setType(new Type(Type.INTEGER));
    param[0].setRange(1, Integer.MAX_VALUE);
    param[0].setData(new Integer(particleCount));

    param[1]= new Parameter();
    param[1].setName("Hypercubes");
    param[1].setToolTip("Hypercubes per dimension");
    param[1].setType(new Type(Type.INTEGER));
    param[1].setRange(1, Integer.MAX_VALUE);

    param[1].setData(new Integer(cubesPerDimension));

    param[2]= new Parameter();
    param[2].setName("Generations");
    param[2].setToolTip("Generations to run");
    param[2].setType(new Type(Type.INTEGER));
    param[2].setRange(1, Integer.MAX_VALUE);
    param[2].setData(new Integer(generationsToRun));

    param[3]= new Parameter();
    param[3].setName("Repository");
    param[3].setToolTip("Size of the repository");
    param[3].setType(new Type(Type.INTEGER));
    param[3].setRange(1, Integer.MAX_VALUE);

    param[3].setData(new Integer(repsize));

    param[4]= new Parameter();
    param[4].setName("Inertia");
    param[4].setToolTip("Weight of former speed");
    param[4].setType(new Type(Type.DOUBLE));
    param[4].setRange((double)0, (double)10);

    param[4].setData(new Double(w));

    param[5]= new Parameter();
    param[5].setName("Personal best");
    param[5].setToolTip("Weight of personal best");
    param[5].setType(new Type(Type.DOUBLE));
    param[5].setRange((double)0, (double)10);

    param[5].setData(new Double(c1));

    param[6]= new Parameter();
    param[6].setName("Repository weight");
    param[6].setToolTip("Weight of repository");
    param[6].setType(new Type(Type.DOUBLE));
    param[6].setRange((double)0, (double)10);
    param[6].setData(new Double(c2));

    param[7]= new Parameter();
    param[7].setName("Seekers");
    param[7].setToolTip("Ratio of seekers");
    param[7].setType(new Type(Type.DOUBLE));
    param[7].setRange((double)0,(double)1);
    param[7].setData(new Double(seekerPercentage));

    param[8]= new Parameter();
    param[8].setName("Printout");
    param[8].setToolTip("0 - Particles, 1 - Repository, 2 - Both");
    param[8].setType(new Type(Type.INTEGER));
    param[8].setRange(0, 2);
    param[8].setData(new Integer(printout));

    param[9]= new Parameter();
    param[9].setName("Spacing");
    param[9].setToolTip("0 - Parameter Space, 1 - Objective Space");
    param[9].setType(new Type(Type.INTEGER));
    param[9].setRange(0, 1);
    param[9].setData(new Integer(spacing));

    param[10]= new Parameter();
    param[10].setName("Particle spacing");
    param[10].setToolTip("0 - Parameter Space, 1 - Objective Space");
    param[10].setType(new Type(Type.INTEGER));
    param[10].setRange(0, 1);
    param[10].setData(new Integer(particleSpacing));

    param[11]= new Parameter();
    param[11].setName("Particle repository");
    param[11].setToolTip("Size of the particles' repository");
    param[11].setType(new Type(Type.INTEGER));
    param[11].setRange(1, Integer.MAX_VALUE);
    param[11].setData(new Integer(pRepsize));

    param[12]= new Parameter();
    param[12].setName("Repository type");
    param[12].setToolTip("0 - Fast, 1 - Slow but good, 2 - Randomized, 3 - Fast+Slow");
    param[12].setType(new Type(Type.INTEGER));
    param[12].setRange(0, 3);
    param[12].setData(new Integer(reptype));

    param[13]= new Parameter();
    param[13].setName("Explosion threshold");
    param[13].setToolTip("Number of Generations without dominating Particles found");
    param[13].setType(new Type(Type.INTEGER));
    param[13].setRange(1,Integer.MAX_VALUE);

    param[13].setData(new Integer(threshold));
    params.setParameterArray(param);
    params.setMembership(this.getClass().getName());
    defaultParams=new Parameters();

    for (int i = 0; i < param.length; i++) {
      defaultParams.addParameter((Parameter)param[i].clone());
    }

    return params;
    }
    /**
     * Sets parameters for Mopso
     * @param parameter The new settings
     */
    public void setParameter(Parameters parameter) {
                parametersCorrect(parameter);
		particleCount=((Number)parameter.getParameterByName("Particles").getData()).intValue();
		cubesPerDimension=((Number)parameter.getParameterByName("Hypercubes").getData()).intValue();
		generationsToRun=((Number)parameter.getParameterByName("Generations").getData()).intValue();
		repsize=((Number)parameter.getParameterByName("Repository").getData()).intValue();
		w=((Number)parameter.getParameterByName("Inertia").getData()).doubleValue();
		c1=((Number)parameter.getParameterByName("Personal best").getData()).doubleValue();
		c2=((Number)parameter.getParameterByName("Repository weight").getData()).doubleValue();
		seekerPercentage=((Number)parameter.getParameterByName("Seekers").getData()).doubleValue();
		printout=((Number)parameter.getParameterByName("Printout").getData()).intValue();
		spacing=((Number)parameter.getParameterByName("Spacing").getData()).intValue();
		particleSpacing=((Number)parameter.getParameterByName("Particle spacing").getData()).intValue();
		pRepsize=((Number)parameter.getParameterByName("Particle repository").getData()).intValue();
		reptype=((Number)parameter.getParameterByName("Repository type").getData()).intValue();
                threshold=((Number)parameter.getParameterByName("Explosion threshold").getData()).intValue();
    }

    /**
     * Startup Method of the algoritm
     * @return boolean true if setup was correct
     */
    public void setup(Parameters para, AbstractProblem prob) {
      setProblem(prob);
      setParameter(para);
      initialize();
    }
    /**
     * Steps one generation. Calculates the new position of each particle based on its actual, its best and a position of a particle in the repository.
     */
    public void iterate() {
    	for(int i=0;i< this.particles.length;i++){
    	
    //	this.problem.repair(this.particles[i].vectorify());
    	}
   
    if (count>0) {
      count--;
    }
      if (iterationCount<10 && spacing != 0) {
        updateBorders();
      }
    if (iterationCount==0) {
     }
     if (iterationCount>(int)generationsToRun*0.8 && changed==false && reptype==3) {
      changeRep();
      changed=true;
      System.out.println("Changed to NewRep");
     }
      int nohap=iterationCount-lastUp;
      if (nohap>threshold&&count==0) {
        count=threshold;
      }

      if ((iterationCount/100)*100==iterationCount||(nohap>50 && (nohap/10)*10==nohap)) {

      }

      double[] var=getVariance();
      double[] avg=getAverage();
      double[] variance=getVariance();
    for (int i = 0; i < particles.length; i++) {
    if (particles[i].role==Particle.seeker && count>0) {
        particles[i].w=(1/threshold)*count+w;
        particles[i].c1=(-1/threshold)*count+c1;
        particles[i].c2=(-1/threshold)*count+c2;
        }
      double[] speed= new double[dimx];
      Particle p=repository.selectFromRep();
      Range[] ranges=problem.getRanges();
      for (int j = 0; j < speed.length; j++) {
        speed[j]+=particles[i].w*particles[i].speed[j];
	speed[j]+=particles[i].c1*r.nextDouble()*(particles[i].pbestx[j]-particles[i].positionx[j]);
	speed[j]+=particles[i].c2*r.nextDouble()*(p.positionx[j]-particles[i].positionx[j]);
        if ((count==threshold&&nohap>threshold&&particles[i].role==particles[i].seeker)) {
        double factor=((Number)ranges[j].getMaxVal()).doubleValue()-((Number)ranges[j].getMinVal()).doubleValue();
	factor/=2;
        speed[j]+=factor*r.nextGaussian()*0.5;
        particles[i].positionx[j]=speed[j];
        speed[j]/=100;
        if ((iterationCount/100)*100==iterationCount&&i==1) {

        }

        }
      }
      particles[i].fly(speed);

    }
    boolean updated=repository.updateRep(particles);
    if (updated) {
          lastUp=iterationCount;
    }
iterationCount++;
    }

/**
 * Calculates the Average of all particle positions
 * @return double[] The average of each coordinate
 */
    private double[] getAverage() {
      double[] avg=new double[dimx];
      for (int i = 0; i < avg.length; i++) {
	avg[i]=0;
      }

      for (int i = 0; i < particles.length; i++) {
        for (int j = 0; j < avg.length; j++) {
          avg[j]+=particles[i].positionx[j];
        }
      }
        for (int j = 0; j < avg.length; j++) {
          avg[j]=avg[j]/particles.length;
        }
return avg;
}

/**
 * Calculates the Variance of all particle positions
 * @return double[] The Variance of each coordinate
 */
private double[] getVariance()  {
      double[] avg=getAverage();
      double[] variance=new double[dimx];

      for (int i = 0; i < particles.length; i++) {
        for (int j = 0; j < avg.length; j++) {
          variance[j]+=Math.abs(particles[i].positionx[j]-avg[j]);
        }
      }
        for (int j = 0; j < avg.length; j++) {
          variance[j]=variance[j]/particles.length;
        }
        return variance;
}

    /**
     * Returns the results of Mopso as stated in printout.
     * @return Data[] Either the current swarm, the repository or both.
     */
    public Data[] getResults() {
    Data[] erg;
    switch (printout) {
      case 0:
          erg = new Data[particles.length];
	  for (int i = 0; i < particles.length; i++) {
	  erg[i] = particles[i].toData();
	  }

      break;
      case 1:
      erg = new Data[repository.rep.size()];
      for (int i = 0; i < repository.rep.size(); i++) {
	erg[i] = ((Particle)repository.rep.get(i)).toData();
      }

	break;
      default:
      erg = new Data[repository.rep.size()+particles.length];
      for (int i = 0; i < particles.length; i++) {
	erg[i] = particles[i].toData();
      }

      for (int i = 0; i < repository.rep.size(); i++) {
	erg[i+particles.length] = ((Particle)repository.rep.get(i)).toData();
      }
	break;
    }
          return erg;
    }

    /**
     * Sets result. Not yet used.
     * @param d The results to be setted.
     */

    public void setResults(Data[] d) {
    }

    public AbstractProblem getProblem(){ return problem; }

/**
 * Sets the problem.
 * param problem The new Problem
 */
    public void setProblem(AbstractProblem problem){
      this.problem = problem;

      }

    public int getIterationCount(){
	return iterationCount;
    }

    public void setIterationCount(int iterationCount){
	this.iterationCount = iterationCount;
    }

    public int getRuntime(){
	return runtime;
    }
    public void setRuntime(int runtime){
	this.runtime = runtime;
    }

    public long getRandomSeed() {
        return randomSeed;
    }

    public void setRandomSeed(long seed) {
    }
    /**
     * Initializes each particles position uniform distributed in the search space. Sets maxborders/minborders of the search space. Initializes the random number with the random seed if set.
     */
    private void initialize() {
      changed=false;
      dimx=problem.getDimDecisionSpace();
      dimy=problem.getDimObjectiveSpace();
      this.edgelengthX = new double[dimx];
      this.edgelengthY = new double[dimy];
      this.maxborderY = new double[dimy];
      this.minborderY = new double[dimy];
      Range[] range=problem.getRanges();
      for (int i = 0; i < edgelengthX.length; i++) {
      edgelengthX[i]=(((Number)range[i].getMaxVal()).doubleValue() -((Number)range[i].getMinVal()).doubleValue())/cubesPerDimension;
      }
      if (reptype==0||reptype==3) {
        if (spacing==0) {
          repository=new Repository(false, dimx, repsize, edgelengthX, problem);
        }
        else {
          repository=new Repository(true, dimy, repsize, edgelengthY, problem);
        }
      }
      if (reptype==1) {
      if (spacing==0) {
        repository=new NewRep(false, dimx, repsize, edgelengthX, problem);
      }
      else {
        repository=new NewRep(true, dimy, repsize, edgelengthY, problem);
      }
      }
      if (reptype==2) {
      if (spacing==0) {
		NewRep r=new NewRep(false, dimx, repsize, edgelengthX, problem);
		r.randomized=true;
        repository=r;
      }
      else {
        NewRep r=new NewRep(true, dimy, repsize, edgelengthY, problem);
			r.randomized=true;
        repository=r;

      }
      }
      if (particleSpacing==0) {
	pOSpacing=false;
      }
      else {
	pOSpacing=true;
      }


      if (randomSeed!=0) {
        r=new Random(randomSeed);
      }
      else {
	r=new Random();
      }
      particles= new Particle[particleCount];
      for (int i = 0; i < particles.length; i++) {
	if (pOSpacing) {
	particles[i]=new Particle(dimx, problem.getDimObjectiveSpace(), problem, w, c1, c2, pOSpacing, pRepsize, edgelengthY);
	particles[i].evaluate();
	}
	else {
	particles[i]=new Particle(dimx, problem.getDimObjectiveSpace(), problem, w, c1, c2, pOSpacing, pRepsize, edgelengthX);
	particles[i].evaluate();
	}
      }
      for (int i = 0; i < dimy; i++) {
	maxborderY[i]=Double.MIN_VALUE;
	minborderY[i]=Double.MAX_VALUE;
      }
      repository.updateRep(particles);
      int seekers=(int)(particleCount*seekerPercentage);
      int i=0;
      for (; i < seekers; i++) {
	particles[i].role=Particle.seeker;
      }

    repository.updateRep(particles);
    for(int x =0;x< this.particles.length;x++)
    {
    //	this.problem.StartIndividuals(this.particles[x].vectorify());
    }
}

/**
 * As the y-ranges are not yet known on startup, this Method is called the first 10 generations
 * It refits the hypercubes to the objective space.
 */
public void updateBorders() {
boolean updateDense=false;
for (int j = 0; j < problem.getDimObjectiveSpace(); j++) {
for (int i = 0; i < particles.length; i++) {
if (this.maxborderY[j]<particles[i].positiony[j]) {
  this.maxborderY[j]=particles[i].positiony[j];
  updateDense=true;
  }
if (this.minborderY[j]>particles[i].positiony[j]) {
  this.minborderY[j]=particles[i].positiony[j];
  updateDense=true;
  }
}
}
for (int i = 0; i < edgelengthY.length; i++) {
  edgelengthY[i]=(maxborderY[i]-minborderY[i])/cubesPerDimension;
}

if (updateDense) {
  repository.setEdgelength(edgelengthY);
  if (pOSpacing) {
  for (int i = 0; i < particles.length; i++) {
    particles[i].rep.setEdgelength(edgelengthY);
  }
  }

}
}
/**
 * Changes the current repository to another type.
 */
private void changeRep() {
  Vector temp=repository.rep;

      if (spacing==0) {
        repository=new NewRep(false, dimx, repsize, edgelengthX, problem);
      }
      else {
        repository=new NewRep(true, dimy, repsize, edgelengthY, problem);
      }

    repository.rep=temp;
}
/**
 * Checks, if parameter settings are correct.
 * @parameter
 * @return boolean true, if all parameters are in range
 */
private boolean parametersCorrect(Parameters params) {
boolean erg=true;
Parameter[] p= params.getParameterArray();
Parameter temp;
String name;
for (int i = 0; i < p.length; i++) {
name=p[i].getName();
temp=defaultParams.getParameterByName(name);
  if (!(temp.inRange(p[i].getData()))) {
//  Kea.logger.logError("Illegal parameter setting: "+name+" = "+p[i].getData()+"! Default value is: "+ name+" = "+temp.getData());
  throw new AlgoException("Illegal parameter setting: "+name+" = "+p[i].getData()+"! Default value is: "+ name+" = "+temp.getData());
  }
}
return erg;
}

    private long randomSeed;
    private AbstractProblem problem;
    /*
     * Counts the iterations;
     */
    private int iterationCount;
    /*
     * The runtime
     */
    private int runtime;
    /**
     * Number of particles
     */
    private int particleCount;
    /**
     * The particle Swarm
     */
    public Particle[] particles;

    private double[] edgelengthX;
    /**
     * The estimated maxborders of the object space
     */
    private double[] maxborderY;
    /**
     * The estimated minborders of the object space
     */
    private double[] minborderY;
    /**
     * The edge lengths of a hypercube.
     */
    private double[] edgelengthY;
    /**
     * The inertia of a particle.
     */
    private double w;
    /**
     * The weight of a particles' personal best.
     */
    private double c1;
    /**
     * The weight of a particle in the repository
     */
    private double c2;
    /**
     * This specifies what to be returned by getResults()
     * printout == 0: Only the current Swarm is returned.
     * printout == 1: Only the repository is returned.
     * printout == 2: The current Swarm followed by the repository is returned.
     */
    private int printout;
    private int spacing;
    private int particleSpacing;
    private boolean pOSpacing;
    private int pRepsize;
    /**
     * The Dimension of the decision space
     */
    private int dimx;
    /**
     * The Dimension of the objective space
     */
    private int dimy;
    /**
     * Maximal size of the repository
     */
    private int repsize;
    /**
     * Number of cubes per dimension
     */
    private int cubesPerDimension;
    /**
     * Iterations to run
     */
    private int generationsToRun;
    /**
     * The random number generator.
     */
    private Random r;
    private double seekerPercentage;
    private int lastUp;
    private int count;
    /**
     * The repository
     */
    private Repository repository;
    private int reptype;
    private boolean changed;
    private int threshold;
    /**
     * Default settings
     */
    private Parameters defaultParams;
}
