/*
 * NewRep.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.dops;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import kea.problems.AbstractProblem;

public class NewRep extends Repository implements Serializable {

  public NewRep(boolean oSpace, int dim, int repsize, double[] edgelength, AbstractProblem problem) {

    this.dim=dim;
    this.oSpace=oSpace;
    this.problem=problem;
    this.r=new Random();
    this.rep=new Vector();
    this.repsize=repsize;
    this.edgelength=edgelength;
    randomizedArchive=1/Math.sqrt((double)repsize);
    this.randomized=false;
    }

  /**
   * Updates the repository after a new iteration step. Tests if a particle is pareto dominant and removes all dominated particles from the repository.
   */
  public boolean updateRep(Particle[] particles) {
    boolean ret=false;
    for (int j=0;j<particles.length;j++) {
      boolean add =true;
      for (int i = 0; i < rep.size(); i++) {
        Particle p=(Particle)rep.get(i);
        if (Particle.isDominant(p.positiony, particles[j].positiony)) {
	  add=false;
          //if (rep.size()<repsize) {
	  //  add=true;
          //}
	  //else {
          i=rep.size();
	  break;
          //}
        }
        if (Particle.isDominant(particles[j].positiony, p.positiony)) {
          ret=true;
          //if (rep.size()>=repsize) {
	    rep.remove(i);
	  //}
        }
      }
      if (add==true) {
         addToRep(particles[j].klone(), particles);
      }
    }
    return ret;
  }

    /**
     * Selects a random particle from the repository used in the calculation of the new particle positions. Selected by roulette wheel selection. Fitness is 10/Particles in the same hypercube. This is for a well-distributed pareto front.
     */
    public Particle selectFromRep() {
      double[] pfitness=new double[rep.size()];
      double p=r.nextDouble();
      int sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
	pfitness[i]=getFitness((Particle)(rep.get(i)));
	sumfit+=pfitness[i];
      }
      p=p*sumfit;
      sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
	sumfit+=pfitness[i];
	if (sumfit>=p) {
        return (Particle)rep.get(i);
	}
      }
      Particle part=new Particle(problem.getDimDecisionSpace(), problem.getDimObjectiveSpace(), problem, 0.4, 1, 1, oSpace, repsize, edgelength);
      part.evaluate();
      return part;
    }
    /**
     * Selects a particle to remove from the repository. Selected by roulette wheel selection. Fitness is Sum of Particles in the same hypercube. This is for a well-distributed pareto front.
     */
    public void removeFromRep(Particle[] particles) {
      double[] pfitness=new double[rep.size()];
      double p=r.nextDouble();
      double sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
      pfitness[i]=getFitnessRem((Particle)(rep.get(i)));
      sumfit+=pfitness[i];
      }
      p=p*sumfit;
      sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
      double basefit=pfitness[i];
      sumfit+=basefit;
	if (sumfit>=p) {
          rep.remove(i);
	  return;
	}
      }
    }
    /**
     * Adds a new dominant particle to the repository. If the size surpasses its maximum, a random particle is removed.
     */
    public void addToRep(Particle p, Particle[] particles) {
	rep.add(p);
	if (rep.size()>repsize) {
	  removeFromRep(particles);
	}
    }
  public double distance (Particle p1, Particle p2) {
    double erg=0;
    double temp=0;
    if (oSpace) {
      for (int i = 0; i < dim; i++) {
	temp=p1.positiony[i]-p2.positiony[i];
	temp=temp/edgelength[i];
	erg+=(temp*temp);
      }
    }
    else {
      for (int i = 0; i < dim; i++) {
	temp=p1.positionx[i]-p2.positionx[i];
	temp=temp/edgelength[i];
	erg+=(temp*temp);
      }
    }
    double exp=1/(double)dim;
    erg=Math.pow(erg, exp);
    return erg;
  }
  public double getFitness (Particle p) {
  double erg=0;
  double temp=0;
  for (int i = 0; i < rep.size(); i++) {
  if (randomized==false || r.nextDouble()<randomizedArchive) {
    temp=distance(p, (Particle)rep.get(i));
    if (temp!=0) {
    temp=1/Math.pow(temp, 3);
    erg+=temp;
    }
    }
  }
if (erg<0.0000001) {
 erg=0.0000001;
}
  erg=1/erg;
  return erg;

  }
public double getFitnessRem (Particle p) {
  double erg=1/getFitness(p);
  return erg;
}
    public void setEdgelength (double[] edgelength) {
    this.edgelength=edgelength;

    }
    public double[] getEdgelength() {
      return edgelength;
    }






    boolean oSpace;
    private Hashtable Space;
    private int dim;
    private int repsize;
    private double randomizedArchive;
    private Random r;
    private double[] edgelength;
    private AbstractProblem problem;
    public boolean randomized;
}
