/*
 * Mopso.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.mopso;

import java.io.Serializable;
import java.util.Hashtable;
import java.util.Random;
import java.util.Vector;

import kea.Data;
import kea.Parameter;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.algorithms.AbstractAlgorithm;
import kea.algorithms.common.AlgoException;
import kea.problems.AbstractProblem;

/**
 * The MOPSO algorithm as proposed in Carlos A. Coello Coello and Maximino Salazar Lechuga. MOPSO: A Proposal for
 * Multiple Objective Particle Swarm Optimization. In Congress on Evolutionary Computation
 * (CEC2002), volume 2, pages 10511056, Piscataway, New Jersey, May 2002.
 * IEEE Service Center. <br>
 * The core class of MOPSO.
 * 
 * @author Philipp Limbourg
 * @version 0.1.0
 * @see kea.algorithms.AbstractAlgorithm
 */
public class Mopso extends AbstractAlgorithm implements Serializable {
	
  /**
   * Constructor of MOPSO. Please set your default values here.
   */
  public Mopso() {
    //Defaults:
    particleCount=20;
    cubesPerDimension=10000;
    generationsToRun=1000;
    repsize=100;
    w=0.4;
    c1=1;
    c2=1.2;
    printout=1;
  }
  
    /**
     * Checks the stopCondition
     */
    public  boolean checkStopCondition() {
      if (iterationCount>=generationsToRun) {
	return true;
      }
      return false;
    }

    /**
     * Startup Method of the algoritm
     * @return boolean true if setup was correct
     */
    public void setup(Parameters para, AbstractProblem prob) {
   //   try {

      setProblem(prob);
      setParameter(para);
      initialize();
/*}
      catch (Exception ex) {
	return false;
      }
*/
    }
    
    /**
     * Gets the current parameter settings of Mopso
     * @return Parameters The settings of Mopso
     */
    public Parameters getParameter() {
    Parameters params=new Parameters();
    Parameter[] param=new Parameter[8];
    param[0]= new Parameter();
    param[0].setName("Particles");
    param[0].setToolTip("Number of particles");
    param[0].setType(new Type(Type.INTEGER));
    param[0].setRange(1, Integer.MAX_VALUE);
    param[0].setData(new Integer(particleCount));

    param[1]= new Parameter();
    param[1].setName("Hypercubes");
    param[1].setToolTip("Hypercubes per dimension");
    param[1].setType(new Type(Type.INTEGER));
    param[1].setRange(1, Integer.MAX_VALUE);
    param[1].setData(new Integer(cubesPerDimension));

    param[2]= new Parameter();
    param[2].setName("Generations");
    param[2].setToolTip("Generations to run");
    param[2].setType(new Type(Type.INTEGER));
    param[2].setRange(1, Integer.MAX_VALUE);
    param[2].setData(new Integer(generationsToRun));

    param[3]= new Parameter();
    param[3].setName("Repository");
    param[3].setToolTip("Size of the repository");
    param[3].setType(new Type(Type.INTEGER));
    param[3].setRange(1, Integer.MAX_VALUE);
    param[3].setData(new Integer(repsize));

    param[4]= new Parameter();
    param[4].setName("Inertia");
    param[4].setToolTip("Weight of former speed");
    param[4].setType(new Type(Type.DOUBLE));
    param[4].setRange((double)0, (double)10);
    param[4].setData(new Double(w));

    param[5]= new Parameter();
    param[5].setName("Personal best");
    param[5].setToolTip("Weight of personal best");
    param[5].setType(new Type(Type.DOUBLE));
    param[5].setRange((double)0,(double)10);
    param[5].setData(new Double(c1));

    param[6]= new Parameter();
    param[6].setName("Repository weight");
    param[6].setToolTip("Weight of repository");
    param[6].setType(new Type(Type.DOUBLE));
    param[6].setRange((double)0, (double)10);
    param[6].setData(new Double(c2));

    param[7]= new Parameter();
    param[7].setName("Printout");
    param[7].setToolTip("0 - Particles, 1 - Repository, 2 - Both");
    param[7].setType(new Type(Type.INTEGER));
    param[7].setRange(1, 2);
    param[7].setData(new Integer(printout));

    params.setParameterArray(param);
    params.setMembership(this.getClass().getName());
    defaultParams=new Parameters();
    for (int i = 0; i < param.length; i++) {
      defaultParams.addParameter((Parameter)param[i].clone());
    }
    return params;
    }
    
    /**
     * Sets parameters for Mopso
     * @param parameter The new settings
     */
    public void setParameter(Parameters parameter) {
    parametersCorrect(parameter);
		particleCount=((Number)parameter.getParameterByName("Particles").getData()).intValue();
		cubesPerDimension=((Number)parameter.getParameterByName("Hypercubes").getData()).intValue();
		generationsToRun=((Number)parameter.getParameterByName("Generations").getData()).intValue();
		repsize=((Number)parameter.getParameterByName("Repository").getData()).intValue();
		w=((Number)parameter.getParameterByName("Inertia").getData()).doubleValue();
		c1=((Number)parameter.getParameterByName("Personal best").getData()).doubleValue();
		c2=((Number)parameter.getParameterByName("Repository weight").getData()).doubleValue();
		printout=((Number)parameter.getParameterByName("Printout").getData()).intValue();
    }
    
    /**
     * Steps one generation. Initializes MOPSO if iterationcount=0. Calculates the new position of each particle based on its actual, its best and a position of a particle in the repository.
     */
    public void iterate() {

      if (iterationCount<10) {
        updateBorders();
      }
    for (int i = 0; i < particles.length; i++) {
      double[] speed= new double[dimx];
      Particle p=selectFromRep();
      for (int j = 0; j < speed.length; j++) {
        speed[j]+=w*particles[i].speed[j];
	speed[j]+=c1*r.nextDouble()*(particles[i].pbestx[j]-particles[i].positionx[j]);
	speed[j]+=c2*r.nextDouble()*(p.positionx[j]-particles[i].positionx[j]);
      }
      particles[i].fly(speed);

    }
    updateRep();
iterationCount++;
    }
    
  /**
   * Updates the repository after a new iteration step. Tests if a particle is pareto dominant and removes all dominated particles from the repository.
   */
    private void updateRep() {
    for (int j=0;j<particles.length;j++) {
    boolean add =true;
    for (int i = 0; i < rep.size(); i++) {
    Particle p=(Particle)rep.get(i);
      if (particles[j].equals(p)) {
	add=false;
      i=rep.size();
      break;
      }
      if (Particle.isDominant(p.positiony, particles[j].positiony)) {
      add=false;
      i=rep.size();
      break;
      }
      if (Particle.isDominant(particles[j].positiony, p.positiony)) {
      rep.remove(i);
      i--;
      }
      }
      if (add==true) {
      addToRep(particles[j].klone());
      }

     }
    }

    /**
     * Selects a random particle from the repository used in the calculation of the new particle positions. Selected by roulette wheel selection. Fitness is 10/Particles in the same hypercube. This is for a well-distributed pareto front.
     */
    private Particle selectFromRep() {
      double p=r.nextDouble();
      int sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
	sumfit+= 10/((Particle)(rep.get(i))).dense;
      }
      p=p*sumfit;
      sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
	sumfit+=10/((Particle)(rep.get(i))).dense;
	if (sumfit>=p) {
        return (Particle)rep.get(i);
	}
      }
      return new Particle(dimx, problem.getDimObjectiveSpace(), problem);
    }
    
    /**
     * Selects a particle to remove from the repository. Selected by roulette wheel selection. Fitness is Sum of Particles in the same hypercube. This is for a well-distributed pareto front.
     */
    private void removeFromRep() {
      double p=r.nextDouble();
      int sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
	sumfit+=((Particle)(rep.get(i))).dense;
      }
      p=p*sumfit;
      sumfit=0;
      for (int i = 0; i < rep.size(); i++) {
	sumfit+=((Particle)(rep.get(i))).dense;
	if (sumfit>=p) {
          Particle toKill=(Particle)rep.remove(i);
	  Hypercube cube=new Hypercube(edgelength, toKill.positiony);
	  Vector v=(Vector)hypercubes.remove(cube);
	  v.remove(toKill);
	  for (i = 0; i < v.size(); i++) {
	    ((Particle)v.get(i)).dense=v.size();
	  }
	  hypercubes.put(cube, v);
	  return;
	}
      }
    }
    
    /**
     * Adds a new dominant particle to the repository. If the size surpasses its maximum, a random particle is removed.
     */
    private void addToRep(Particle p) {
	rep.add(p);
	Hypercube cube=new Hypercube(edgelength, p.positiony);
	Vector v=(Vector)hypercubes.remove(cube);
	if (v==null) {
	  v=new Vector();
	}
	v.add(p);
	for (int i = 0; i < v.size(); i++) {
	  ((Particle)v.get(i)).dense=v.size();
	}
	hypercubes.put(cube, v);
	if (rep.size()>repsize) {
	  removeFromRep();
	}
    }
    
/**
 * As the y-ranges are not yet known on startup, this Method is called the first 10 generations
 * It refits the hypercubes to the objective space.
 */
public void updateBorders() {
boolean updateDense=false;
for (int j = 0; j < problem.getDimObjectiveSpace(); j++) {
for (int i = 0; i < particles.length; i++) {
if (this.maxborder[j]<particles[i].positiony[j]) {
  this.maxborder[j]=particles[i].positiony[j];
  updateDense=true;
  }
if (this.minborder[j]>particles[i].positiony[j]) {
  this.minborder[j]=particles[i].positiony[j];
  updateDense=true;
  }
}
}
for (int i = 0; i < edgelength.length; i++) {
  edgelength[i]=(maxborder[i]-minborder[i])/cubesPerDimension;
}

if (updateDense) {
  updateDense();
}
}

/**
 * Updates the dense of the particles in the repository.
 */
private void updateDense() {
 this.hypercubes=new Hashtable();
 for (int i = 0; i < rep.size();i++) {
        Particle p=(Particle)rep.get(i);
   	Hypercube cube=new Hypercube(edgelength, p.positiony);
	Vector v=(Vector)hypercubes.remove(cube);
	if (v==null) {
	  v=new Vector();
	}
	v.add(p);
	for (int j = 0; j < v.size(); j++) {
	  ((Particle)v.get(j)).dense=v.size();
	}
	hypercubes.put(cube, v);
 }

}

    /**
     * Returns the results of Mopso as stated in printout.
     * @return Data[] Either the current swarm, the repository or both.
     */
    public Data[] getResults() {
    Data[] erg;
    switch (printout) {
      case 0:
          erg = new Data[particles.length];
	  for (int i = 0; i < particles.length; i++) {
	  erg[i] = particles[i].toData();
	  }
      break;
      case 1:
      erg = new Data[rep.size()];
      for (int i = 0; i < rep.size(); i++) {
	erg[i] = ((Particle)rep.get(i)).toData();
      }

	break;
      default:
      erg = new Data[rep.size()+particles.length];
      for (int i = 0; i < particles.length; i++) {
	erg[i] = particles[i].toData();
      }

      for (int i = 0; i < rep.size(); i++) {
	erg[i+particles.length] = ((Particle)rep.get(i)).toData();
      }
	break;
    }
          return erg;
    }

    /**
     * Sets result. Not yet used.
     * @param d The results to be setted.
     */
    public void setResults(Data[] d) {
    }

    public AbstractProblem getProblem(){ return problem; }
    
/**
 * Sets the problem.
 * @param problem The new Problem
 */
    public void setProblem(AbstractProblem problem){
      this.problem = problem;

      }

    public int getIterationCount(){
	return iterationCount;
    }

    public void setIterationCount(int iterationCount){
	this.iterationCount = iterationCount;
    }

    public int getRuntime(){
	return runtime;
    }
    public Vector getRep(){
	return rep;
    }

    public void setRuntime(int runtime){
	this.runtime = runtime;
    }

    public long getRandomSeed() {
        return randomSeed;
    }

    public void setRandomSeed(long seed) {
    }
    
    /**
     * Initializes each particles position uniform distributed in the search space. Sets provisory maxborders/minborders of the objective space. Initializes the random number with the random seed if set.
     */
    private void initialize() {
      dimx=problem.getDimDecisionSpace();
      dimy=problem.getDimObjectiveSpace();
      this.edgelength = new double[dimy];
      this.maxborder = new double[dimy];
      this.minborder = new double[dimy];
      rep=new Vector();
      hypercubes=new Hashtable();
      if (randomSeed!=0) {
        r=new Random(randomSeed);
      }
      else {
	r=new Random();
      }
      particles= new Particle[particleCount];
      for (int i = 0; i < particles.length; i++) {
	particles[i]=new Particle(dimx, problem.getDimObjectiveSpace(), problem);
	problem.score(this.particles[i].toData().getDecVector());
      }
      Range[] ranges= problem.getRanges();
      for (int i = 0; i < dimy; i++) {
	maxborder[i]=Double.MIN_VALUE;
	minborder[i]=Double.MAX_VALUE;
      }
      updateRep();
}

/**
 * Checks, if parameter settings are correct.
 * @return boolean true, if all parameters are in range
 */
private boolean parametersCorrect(Parameters params) {
boolean erg=true;
Parameter[] p= params.getParameterArray();
Parameter temp;
String name;
for (int i = 0; i < p.length; i++) {
name=p[i].getName();
temp=defaultParams.getParameterByName(name);
  if (!(temp.inRange(p[i].getData()))) {
//  Kea.logger.logWarn("Illegal parameter setting: "+name+" = "+p[i].getData()+"! Resetted to default: "+ name+" = "+temp.getData());
//  p[i].setData(temp.getData());
//  erg=false;
  throw new AlgoException("Illegal parameter setting: "+name+" = "+p[i].getData()+"! Default value is: "+ name+" = "+temp.getData());
  }
}
return erg;
}
    private long randomSeed;
    private AbstractProblem problem;
    
    /**
     * Counts the iterations;
     */
    private int iterationCount;
    
    /**
     * The runtime
     */
    private int runtime;
    
    /**
     * Number of particles
     */
    private int particleCount;
    
    /**
     * The particle Swarm
     */
    public Particle[] particles;
    
    /**
     * The repository
     */
    private Vector rep;
    
    /**
     * The estimated maxborders of the object space
     */
    private double[] maxborder;
    
    /**
     * The estimated minborders of the object space
     */
    private double[] minborder;
    
    /**
     * The edge lengths of a hypercube.
     */
    private double[] edgelength;
    
    /**
     * The inertia of a particle.
     */
    private double w;
    
    /**
     * The weight of a particles' personal best.
     */
    private double c1;
    
    /**
     * The weight of a particle in the repository
     */
    private double c2;
    
    /**
     * This specifies what to be returned by getResults()
     * printout == 0: Only the current Swarm is returned.
     * printout == 1: Only the repository is returned.
     * printout == 2: The current Swarm followed by the repository is returned.
     */
    private int printout;
    
    /**
     * The Dimension of the decision space
     */
    private int dimx;
    
    /**
     * The Dimension of the objective space
     */
    private int dimy;
    
    /**
     * Maximal size of the repository
     */
    private int repsize;
    
    /**
     * Number of cubes per dimension
     */
    private int cubesPerDimension;
    
    /**
     * Iterations to run
     */
    private int generationsToRun;
    
    /**
     * A second representation of the repository. Its buckets symbolize the hypercubes in the objective space.
     * Each bucket contains the repository particles in that cube. Both this and Vector rep are used simultaneously.
     * As hash key a hypercube object is used.
     */
    private Hashtable hypercubes;
    
    /**
     * The random number generator.
     */
    private Random r;
    
    /**
     * Default settings
     */
    private Parameters defaultParams;
}
//Dreckshck by PL
