/*
 * Particle.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.mopso;

import java.io.Serializable;

import kea.Data;
import kea.Range;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * The MOPSO algorithm as proposed in Carlos A. Coello Coello and Maximino Salazar Lechuga. MOPSO: A Proposal for
 * Multiple Objective Particle Swarm Optimization. In Congress on Evolutionary Computation
 * (CEC2002), volume 2, pages 10511056, Piscataway, New Jersey, May 2002.
 * IEEE Service Center. <br>
 * This class is the data structure for a particle.
 * 
 * @author Philipp Limbourg
 * @version 0.1.0
 * @see kea.algorithms.AbstractAlgorithm
 */

public class Particle implements Serializable {
	
/**
 * Constructor
 * @param dimx Dimension of the decicision space
 * @param dimy Dimension of the objective space
 * @param problem The problem to optimize.
 */
  public Particle(int dimx, int dimy, AbstractProblem problem) {
  this.problem=problem;
  positionx=new double[dimx];
  positiony=new double[dimy];
  speed=new double[dimx];
  pbestx=new double[dimx];
  pbesty=new double[dimy];
  max=new double[dimx];
  min=new double[dimx];
  dense=1;
  initialize();
  }
  
  /**
   * Adds the given speed to the current position and evaluates the new position
   * @param Speed of the particle.
   * @return New position in the decision space.
   */
  public double[] fly (double[] speed) {
    if (speed.length!=positionx.length) {
    return positionx;
  }
  for (int i = 0; i < speed.length; i++) {
    if (max[i]<positionx[i]+speed[i] || min[i]>positionx[i]+speed[i]) {
    return positionx;
  }

  }

  for (int i = 0; i < speed.length; i++) {
  positionx[i]+=speed[i];
  }
evaluate();

if (isDominant (positiony, pbesty)) {
  for (int i = 0; i < positionx.length; i++) {
    pbestx[i]=positionx[i];
  }
    for (int i = 0; i < positiony.length; i++) {
      pbesty[i]=positiony[i];
      }

}
else if (isDominant (positiony, pbesty)) {
}
else {
    if (Math.random()<0.5) {
  for (int i = 0; i < positionx.length; i++) {
    pbestx[i]=positionx[i];
  }
    for (int i = 0; i < positiony.length; i++) {
      pbesty[i]=positiony[i];
      }
  }
}
return positionx;
  }
  
  /**
   * Static method. Tests for dominance.
   * @param dominant The particle checked for dominance
   * @param dominated The particle checked for domination
   * @return boolean true if dominant is dominant (-;
   */
  public static boolean isDominant (double[] dominant, double[] dominated) {
  boolean ret=false;
  for (int i = 0; i < dominant.length; i++) {
  if (dominant[i]!=dominated[i]) {
    ret=true;
  }
}
  for (int i = 0; i < dominant.length; i++) {
  if (dominant[i]>dominated[i]) {
    return false;
  }

  }
  return ret;
  }
  
/**
 * Parses a Value array to this.positiony
 * @param phaenotyp The Value array to be parsed
 */
private void arrify(Value[] phaenotyp) {
for (int i=0;i<phaenotyp.length;i++) {
positiony[i]=((Number)phaenotyp[i].getData()).doubleValue();
}
}

/**
 * Overwrites Object.equals
 * @param o The object to be compared with
 * @return true if o is a particle with the same position
 */
public boolean equals(Object o) {
  Particle p=(Particle)o;
  for (int i = 0; i < positionx.length; i++) {
  if (positionx[i]!=p.positionx[i]) {
    return false;
  }
  }
  return true;
}

/**
 * Creates a Value array from this.positionx
 * @return Value[] The Value array created
 */
private Value[] vectorify() {
Value[] genotyp = new Value[positionx.length];
for (int i=0;i<positionx.length;i++) {
genotyp[i]=new Value(positionx[i]);
}
return genotyp;
}

/**
 * Initializes a particle with a random position
 */
private void initialize() {
Range[] ranges = problem.getRanges();
for (int i = 0; i < positionx.length; i++) {
max[i]=((Number)ranges[i].getMaxVal()).doubleValue();
min[i]=((Number)ranges[i].getMinVal()).doubleValue();
positionx[i]=Math.random()*(max[i]-min[i])+min[i];
pbestx[i]=positionx[i];
}
evaluate();

}

/**
 * Creates a deep clone of this particle
 */
public Particle klone() {
  Particle klon=new Particle(positionx.length, positiony.length, problem);
  klon.dense=dense;
    for (int i = 0; i < positionx.length; i++) {
    klon.positionx[i]=positionx[i];
    klon.pbestx[i]=pbestx[i];
    klon.speed[i]=speed[i];
  }
    for (int i = 0; i < positiony.length; i++) {
    klon.positiony[i]=positiony[i];
    klon.pbesty[i]=pbesty[i];

      }

  return klon;
}

/**
 * Overwrites Object.toString
 * @return String position in the decision and the objective space.
 */
public String toString() {
  String erg="x: ";
  for (int i = 0; i < positionx.length; i++) {
    erg+=positionx[i]+" ";
  }
  erg+="y: ";
  for (int i = 0; i < positiony.length; i++) {
    erg+=positiony[i]+" ";
  }

  return erg;
}

/**
 * Checks if constraints are violated and evaluates the current position.
 */
private void evaluate() {
  Value[] genotyp=vectorify();
  if (problem.verify(genotyp)) {
  arrify(problem.score(genotyp));
  }
  else {
  for (int i = 0; i < positiony.length; i++) {
    positiony[i]=Double.MAX_VALUE;
  }
}
}

/**
 * Parses the particle to a Data object.
 * @return Data The Data representation of this particle.
 */
public Data toData() {
  Value[] x= new Value[positionx.length];
  Value[] y= new Value[positiony.length];
for (int i = 0; i < x.length; i++) {
  x[i]=new Value(positionx[i]);
}
for (int i = 0; i < y.length; i++) {
  y[i]=new Value(positiony[i]);
}
  Data erg = new Data(x, y);
  return erg;

}

/**
 * The current problem
 */
  public AbstractProblem problem;
  
  /**
   * The position in the decision space
   */
  public double[] positionx;
  
  /**
   * The position in the objective space
   */
  public double[] positiony;
  
  /**
   * The current speed
   */
  public double[] speed;
  
  /**
   * The best reached position (decision space)
   */
  public double[] pbestx;
  
  /**
   * The best reached position (objective space)
   */
  public double[] pbesty;
  
  /**
   * Number of particles in the same hypercube. Used only if this particle is in the repository.
   */
  public int dense;

  public double[] max;
  public double[] min;
}
