/*
 * NSGA2.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
 
package kea.algorithms.populationBased.nsga2;

import java.util.Random;
import java.util.Vector;

import kea.Data;
import kea.Parameter;
import kea.Parameters;
import kea.Value;
import kea.algorithms.AbstractAlgorithm;
import kea.algorithms.common.AlgoException;
import kea.problems.AbstractProblem;

/**
 *
 * @author Torsten Kohlen
 * @version 0.0.1
 */
public class NSGA2 extends AbstractAlgorithm {
  /**
   *
   */
  private int populationSize = 50;

  /**
   *
   */
  private int generations = 100;

  /**
   *
   */
  private long rndSeed;
  private int distributionCrossover = 20;
  private int distributionMutation = 20;
  private double crossoverProb = 0.7;
  private double mutationProb = 0.2;

  /**
   *
   */
  private NSGA2Population population;

  /**
   *
   */
  private Random rndGen;
  private int geneLength;
  private int genCount = 0;

//*****************************************************************************/

  /**
   *
   */
  public NSGA2() {
      // setting the default parameters for NSGA2
      rndSeed = System.currentTimeMillis();
      this.param = new Parameters();
      param.setMembership(this.getClass().getName());
      param.addParameter(new Parameter("populationSize"       , populationSize, "Size of the population to be used"));
      param.addParameter(new Parameter("generations"          , generations, ""));
      param.addParameter(new Parameter("crossoverProp"        , crossoverProb, ""));
      param.addParameter(new Parameter("mutationProp"         , mutationProb, ""));
      param.addParameter(new Parameter("distributionCrossover", distributionCrossover, ""));
      param.addParameter(new Parameter("distributionMutation" , distributionMutation, ""));
      param.addParameter(new Parameter("RandomSeed"           , rndSeed, ""));

      param.addParameter(new Parameter("frontRatio"             ,1    ,""));
      param.addParameter(new Parameter("chromosomeLength"       ,10   ,""));
      param.addParameter(new Parameter("selectionTyp"           ,1    ,""));
      param.addParameter(new Parameter("crossoverTyp"           ,1    ,""));
      population = new NSGA2Population();
      rndGen = new Random(rndSeed);
	
  }

  /**
   *
   */
  private void createInitialPopulation(){
    int ddim = problem.getDimDecisionSpace();
    int odim = problem.getDimObjectiveSpace();
   

    for (int p=0;p<populationSize;p++){
      Value[] gene = new Value[ddim];
      Value[] fit = new Value[odim];
      for (int j=0;j<ddim;j++){
        double upperLimit = ((Number)problem.getRanges()[j].getMaxVal()).doubleValue();
        double lowerLimit = ((Number)problem.getRanges()[j].getMinVal()).doubleValue();
        if ((upperLimit != Double.MAX_VALUE) || (lowerLimit != Double.MIN_VALUE)) {
          gene[j] = new Value((rndGen.nextDouble() * (upperLimit - lowerLimit) + lowerLimit));
        }
        else {
          gene[j] = new Value((1 / (2 * rndGen.nextDouble() - 1)));
        }
      }
      for (int j=0;j<odim;j++){
        fit[j] = new Value(0.0);
      }
      population.append(new NSGA2Individual(gene,fit));
    }
  } //end of createInitialPupulation


  public void setResults(Data[] parm) {
    /**@todo: implement this kea.algorithms.AbstractAlgorithm abstract method*/
  }


  /**
   *
   */
//  public void setup(Parameters para, AbstractProblem prob){
//    this.problem = prob;
//
//    if (param != null) {
//      populationSize = ((Number)((Parameter)param.getParameterByName("populationSize")).getData()).intValue();
//      generations    = ((Number)((Parameter)param.getParameterByName("generations")).getData()).intValue();
//      crossoverProb  = ((Number)((Parameter)param.getParameterByName("crossoverProp")).getData()).doubleValue();
//      mutationProb   = ((Number)((Parameter)param.getParameterByName("mutationProp")).getData()).doubleValue();
//
//      //parameter neu setzen!!!!!!!!!!!
//      population = new NSGA2Population();
//    }
//    createInitialPopulation();
//    calcFitness(population);
//  }
	public void setup(Parameters para, AbstractProblem prob) throws AlgoException{
	  this.problem = prob;

	  String errorMessage = "";
	  boolean retval = true;
	  if (param != null) {
		populationSize        = ((Number)((Parameter)param.getParameterByName("populationSize")).getData()).intValue();
		generations           = ((Number)((Parameter)param.getParameterByName("generations")).getData()).intValue();
		crossoverProb         = ((Number)((Parameter)param.getParameterByName("crossoverProp")).getData()).doubleValue();
		mutationProb          = ((Number)((Parameter)param.getParameterByName("mutationProp")).getData()).doubleValue();
		distributionCrossover = ((Number)((Parameter)param.getParameterByName("distributionCrossover")).getData()).intValue();
		distributionMutation  = ((Number)((Parameter)param.getParameterByName("distributionMutation")).getData()).intValue();
		rndSeed               = ((Number)((Parameter)param.getParameterByName("RandomSeed")).getData()).longValue();
    

		if (populationSize < 1) {
			errorMessage += "NSGA2.setup : population Ssize minimum is 1.\n";
			retval = false;
		}
		if (generations < 1 ) {
			errorMessage += "NSGA2.setup : number of generation minimum is 1.\n";
			retval = false;
		}
		if (crossoverProb < 0 || crossoverProb > 1)  {
			errorMessage += "SPEA2.setup : probability of recombination is between 0.0 and 1.0.\n";
			retval = false;
		}
		if (mutationProb < 0 || mutationProb > 1) {
			errorMessage += "SPEA2.setup : probability of recombination is between 0.0 and 1.0.\n";
			retval = false;
		}
		if (distributionCrossover < 1 || distributionCrossover > 500) {
			errorMessage += "SPEA2.setup : distributionCrossover is between 1 and 500.\n";
			retval = false;
		}
		if (distributionMutation < 1 || distributionMutation > 500) {
			errorMessage += "SPEA2.setup : distributionMutation is between 1 and 500.\n";
			retval = false;
		}          
		if (rndSeed < 0) {
			errorMessage += "SPEA2.setup : randomSeed minimum is 0.\n";
			retval = false;
		}
          
		if (!retval) {
		  errorMessage = "=======================================================\n" + errorMessage;
		  errorMessage = "\nNSGA2.setup : Setup failed due to the following errors:\n" + errorMessage;
		  errorMessage = errorMessage + "Please correct these errors first.\n";
		  throw new AlgoException(errorMessage);
		}
      
		//parameter neu setzen!!!!!!!!!!!
		population = new NSGA2Population();
	  }
	  createInitialPopulation();
	  calcFitness(population);
	}



  /**
   *
   */
  public Data[] getResults() {
    return population.getData();
  }


  /**
   *
   */
  public void iterate() {
    population.rank();
    population.assignCrodwingDistance();
  //  population.printRanking();
 //   population.printInfo();

    NSGA2Population matingPool = new NSGA2Population();
    matingPool = selection();
   // matingPool.printInfo();

    NSGA2Population newPopulation = new NSGA2Population();
  //  System.out.println(newPopulation.getSize());

    newPopulation = crossover(matingPool);
    //newPopulation.printInfo();
    newPopulation = mutation(newPopulation);
    calcFitness(newPopulation);

    population = buildNewPopulation(population, newPopulation);
    if (population == null) {
      System.out.println("Population ist NULL!!!!!!!!!!!!!!!!!!");
    }
   // System.out.println("*******************************************************");
   // System.out.println("*******************************************************");
    genCount++;
  }


  /**
   *
   */
  public Parameters getParameter() {
    /**@todo: implement this kea.algorithms.AbstractAlgorithm abstract method*/
    return this.param;
  }


  /**
   *
   */
  public boolean checkStopCondition() {
    if (genCount == generations) {
      return true;
    }
    else {
      return false;
    }
  }


  /**
   *
   */
  public AbstractProblem getProblem() {
    /**@todo: implement this kea.algorithms.AbstractAlgorithm abstract method*/
    return this.problem;
  }


  /**
   *
   */
  private NSGA2Population mutation(NSGA2Population mutPop) {
    double delta, deltaq, newGene, rndNumber, indi, xy, val;
    int dim = problem.getDimDecisionSpace();

//    System.out.println("Starting mutation on generation " + genCount);
//    System.out.println("--------------------------------------");
    for (int i=0;i<populationSize;i++){
      for (int j=0;j<dim;j++){
       // System.out.println("J: " + j);
        double upperLimit = ((Number)problem.getRanges()[j].getMaxVal()).doubleValue();
        double lowerLimit = ((Number)problem.getRanges()[j].getMinVal()).doubleValue();

        if (rndGen.nextDouble() < mutationProb){
          double y = ((Number)mutPop.getIndividual(i).getGene(j).getData()).doubleValue();
          if (y > lowerLimit) {
            if ((y - lowerLimit) < (upperLimit - y)) {
              delta = (y - lowerLimit) / (upperLimit - lowerLimit);
            }
            else {
              delta = (upperLimit - y) / (upperLimit - lowerLimit);
            }
            rndNumber = rndGen.nextDouble();
            indi = 1.0 / 21.0;
            xy = 1.0 - delta;
            if (rndNumber <= 0.5) {
              val = 2 * rndNumber + (1.0 - 2.0 * rndNumber) * Math.pow(xy, 21.0);
              deltaq = Math.pow(val, indi) - 1;
            }
            else {
              val = 2 * (1.0 - rndNumber) + 2 * (rndNumber - 0.5) * Math.pow(xy, 21.0);
              deltaq = 1 - Math.pow(val, indi);
            }
            newGene = y + deltaq * (upperLimit - lowerLimit);
          }
          else {
            newGene = rndGen.nextDouble();
            newGene = newGene * (upperLimit - lowerLimit) + lowerLimit;
          }
          //Mutationen zaehlen!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         // System.out.println("Mutation bei Ind " + i);
         // System.out.println("Gene " + j + " von: " + y + "\t => " + newGene);
          mutPop.getIndividual(i).getGene(j).setData(newGene);
        }//mutation
      }//for j
    }//for i
    return mutPop;
  }//end of mutation


  /**
   *
   */
//  private NSGA2Population crossover(NSGA2Population pool){
//    NSGA2Individual parent1, parent2, child1, child2;
//    NSGA2Population newPop = new NSGA2Population();
//    double oldGene1, oldGene2, lowerLimit, upperLimit, gval1, gval2, beta, alpha, rndNumber, betaq, exp;
//    Value[] newGene1, newGene2;
//    int dim, y = 0;
//
//    //System.out.println("Starting crossover on generation " + genCount);
//    //System.out.println("--------------------------------------");
//    dim = problem.getDimDecisionSpace();
//    for (int i=0;i<(populationSize/2);i++){
//      newGene1 = new Value[dim];
//      newGene2 = new Value[dim];
//      if (rndGen.nextDouble() <= crossoverProb) {
//        //do crossover for each Gene
//        //System.out.println("Crossover planed with Individuals " + y + " and " + (y+1));
//        for (int j=0;j<dim;j++) {
//          parent1 = (NSGA2Individual)pool.getData(y);
//          parent2 = (NSGA2Individual)pool.getData(y+1);
//          oldGene1 = ((Number)parent1.getGene(j).getData()).doubleValue();
//          oldGene2 = ((Number)parent2.getGene(j).getData()).doubleValue();
//          //test if current gene is used for crossover
//          if (rndGen.nextDouble() <= 0.5) {
//            //Anzahl der durchgefuehrten Crossover zaehlen!!!!!!
//            if (Math.abs(oldGene1 - oldGene2) > 0.000001) {
//              //System.out.println("Abs is OK");
//              upperLimit = ((Number)problem.getRanges()[j].getMaxVal()).doubleValue();
//              lowerLimit = ((Number)problem.getRanges()[j].getMinVal()).doubleValue();
//              if (oldGene2 > oldGene1){
//                gval2 = oldGene2;
//                gval1 = oldGene1;
//              }
//              else {
//                gval2 = oldGene1;
//                gval1 = oldGene2;
//              }
//              if ((gval1 - lowerLimit) > (upperLimit - gval2)) {
//                beta = 1 + (2 * (upperLimit - gval2) / (gval2 - gval1));
//              }
//              else {
//                beta = 1 + (2 * (gval1 - lowerLimit) / (gval2 - gval1));
//              }
//              exp = 21.0;
//              alpha = 2.0 - Math.pow((1.0 / beta), exp);
//              rndNumber = rndGen.nextDouble();
//
//              if (rndNumber <= (1.0 / alpha)){
//                alpha = alpha * rndNumber;
//                exp = 1.0 / 21.0;
//                betaq = Math.pow(alpha, exp);
//              }
//              else {
//                alpha = alpha * rndNumber;
//                alpha = 1.0 / (2.0 - alpha);
//                exp = 1.0 / 21.0;
//                betaq = Math.pow(alpha, exp);
//              }
//              newGene1[j] = new Value(0.5 * ((gval1 + gval2) - betaq * (gval2 - gval1)));
//              newGene2[j] = new Value(0.5 * ((gval1 + gval2) + betaq * (gval2 - gval1)));
//            }
//            else {//small distance
//              newGene1[j] = new Value(0.5 * ((oldGene1 + oldGene2) - 1.0 * (oldGene2 - oldGene1)));
//              newGene2[j] = new Value(0.5 * ((oldGene1 + oldGene2) + 1.0 * (oldGene2 - oldGene1)));
//            }
//          }
//          else {// no crosspver with gene j
//            newGene1[j] = new Value(oldGene1);
//            newGene2[j] = new Value(oldGene2);
//          }
//        }//for all gene's
//        newPop.append((NSGA2Individual)pool.getData(y).clone());
//        newPop.append((NSGA2Individual)pool.getData(y+1).clone());
//        newPop.getData(y).setDecVector(newGene1);
//        newPop.getData(y+1).setDecVector(newGene2);
//         // System.out.println("Gene " + j + " von " + y + ": " + oldGene1 + "\t => " + newGene1[j]);
//         // System.out.println("Gene " + j + " von " + (y+1) + ": " + oldGene2 + "\t => " + newGene2[j]);
//         // System.out.println();
//      }
//      else {//no crossover to do
//        newPop.append((NSGA2Individual)pool.getData(y).clone());
//        newPop.append((NSGA2Individual)pool.getData(y+1).clone());
//      }
//      y = y + 2;
//    }
//  //  System.out.println(newPop.getSize());
//    return newPop;
//  }//end of crossover
	private NSGA2Population crossover(NSGA2Population pool){
	  NSGA2Individual parent1, parent2, child1, child2;
	  NSGA2Population newPop = new NSGA2Population();
	  double oldGene1, oldGene2, lowerLimit, upperLimit, gval1, gval2, beta, alpha, rndNumber, betaq, exp;
	  Value[] newGene1, newGene2;
	  int dim, y = 0;

	  //System.out.println("Starting crossover on generation " + genCount);
	  //System.out.println("--------------------------------------");
	  dim = problem.getDimDecisionSpace();

	  for (int i=0;i<(populationSize/2+populationSize%2);i++){
		newGene1 = new Value[dim];
		newGene2 = new Value[dim];
		if (rndGen.nextDouble() <= crossoverProb) {
		  //do crossover for each Gene
		  //System.out.println("Crossover planed with Individuals " + y + " and " + (y+1));
		  for (int j=0;j<dim;j++) {
			parent1 = (NSGA2Individual)pool.getData(y);
			parent2 = (NSGA2Individual)pool.getData((y+1)%populationSize);
			oldGene1 = ((Number)parent1.getGene(j).getData()).doubleValue();
			oldGene2 = ((Number)parent2.getGene(j).getData()).doubleValue();
			//test if current gene is used for crossover
			if (rndGen.nextDouble() <= 0.5) {
			  //Anzahl der durchgefuehrten Crossover zaehlen!!!!!!
			  if (Math.abs(oldGene1 - oldGene2) > 0.000001) {
				//System.out.println("Abs is OK");
				upperLimit = ((Number)problem.getRanges()[j].getMaxVal()).doubleValue();
				lowerLimit = ((Number)problem.getRanges()[j].getMinVal()).doubleValue();
				if (oldGene2 > oldGene1){
				  gval2 = oldGene2;
				  gval1 = oldGene1;
				}
				else {
				  gval2 = oldGene1;
				  gval1 = oldGene2;
				}
				if ((gval1 - lowerLimit) > (upperLimit - gval2)) {
				  beta = 1 + (2 * (upperLimit - gval2) / (gval2 - gval1));
				}
				else {
				  beta = 1 + (2 * (gval1 - lowerLimit) / (gval2 - gval1));
				}
				exp = 21.0;
				alpha = 2.0 - Math.pow((1.0 / beta), exp);
				rndNumber = rndGen.nextDouble();

				if (rndNumber <= (1.0 / alpha)){
				  alpha = alpha * rndNumber;
				  exp = 1.0 / 21.0;
				  betaq = Math.pow(alpha, exp);
				}
				else {
				  alpha = alpha * rndNumber;
				  alpha = 1.0 / (2.0 - alpha);
				  exp = 1.0 / 21.0;
				  betaq = Math.pow(alpha, exp);
				}
				newGene1[j] = new Value(0.5 * ((gval1 + gval2) - betaq * (gval2 - gval1)));
				newGene2[j] = new Value(0.5 * ((gval1 + gval2) + betaq * (gval2 - gval1)));
			  }
			  else {//small distance
				newGene1[j] = new Value(0.5 * ((oldGene1 + oldGene2) - 1.0 * (oldGene2 - oldGene1)));
				newGene2[j] = new Value(0.5 * ((oldGene1 + oldGene2) + 1.0 * (oldGene2 - oldGene1)));
			  }
			}
			else {// no crosspver with gene j
			  newGene1[j] = new Value(oldGene1);
			  newGene2[j] = new Value(oldGene2);
			}
		  }//for all gene's
		  newPop.append((NSGA2Individual)pool.getData(y).clone());
		  newPop.getData(y).setDecVector(newGene1);
		  if (i < populationSize/2) {
			newPop.append((NSGA2Individual)pool.getData(y+1).clone());
			newPop.getData(y+1).setDecVector(newGene2);
		  }
		   // System.out.println("Gene " + j + " von " + y + ": " + oldGene1 + "\t => " + newGene1[j]);
		   // System.out.println("Gene " + j + " von " + (y+1) + ": " + oldGene2 + "\t => " + newGene2[j]);
		   // System.out.println();
		}
		else {//no crossover to do
		  newPop.append((NSGA2Individual)pool.getData(y).clone());
		  if (i < populationSize/2)
			newPop.append((NSGA2Individual)pool.getData(y+1).clone());
		}
		y = y + 2;
	  }
	//  System.out.println(newPop.getSize());
	  return newPop;
	}//end of crossover


  /**
   *
   */
  private NSGA2Population selection(){
    int p1, p2;
    NSGA2Population newPop = new NSGA2Population();
    NSGA2Individual parent1, parent2;

    //System.out.println("Starting selection on generation " + genCount);
    //System.out.println("--------------------------------------");
    for (int i=0;i<populationSize;i++){
      //select first parent Index
      p1 = (int)Math.floor(populationSize * rndGen.nextDouble());
      if (p1 == 0 ) {
        p1 = populationSize - i;// - 2;
      }
      if (p1 >= populationSize) {
        p1 = populationSize-1; //(populationSize - 4) / 2;
      }

      //select second parent Index
      p2 = (int)Math.floor(populationSize * rndGen.nextDouble());
      if (p2 == 0 ) {
        p2 = populationSize - i;
      }
      if (p2 >= populationSize) {
        p2 = populationSize-1;//(populationSize - 2) / 2;
      }
     // System.out.println("Tournament :"+ i + " || " + p1 + " vs. " + p2);
      parent1 = (NSGA2Individual)population.getIndividual(p1);
      parent2 = (NSGA2Individual)population.getIndividual(p2);

      //tournament
      if (parent1.rank > parent2.rank) {
        newPop.append((NSGA2Individual)parent2.clone());
       // System.out.println("Winner: " + p2);
      }
      else {
        if (parent1.rank < parent2.rank) {
          newPop.append((NSGA2Individual)parent1.clone());
         // System.out.println("Winner: " + p1);
        }
        else {
          if (parent1.cuboidLength < parent2.cuboidLength) {
            newPop.append((NSGA2Individual)parent2.clone());
           // System.out.println("Winner: " + p2);
          }
          else {
            newPop.append((NSGA2Individual)parent1.clone());
           // System.out.println("Winner: " + p1);
          }
        }
      }//tournament
    }//for i
 //   System.out.println("Selection on generation " + genCount + "finished");
//    problem.ObjectiveName();
    return newPop;
  }// end of selction


  /**
   *
   */
  private void calcFitness(NSGA2Population p){
    NSGA2Individual ind;
    for (int i=0;i<populationSize;i++){
      ind = p.getIndividual(i);
      if (ind != null){
        ind.setFitness(problem.score(ind.getGene()));
      
      }
      else {
      //FEHLER!!!!!!!!!
      }
    }
  }

  private NSGA2Population buildNewPopulation(NSGA2Population oldPop, NSGA2Population newPop){
    NSGA2Population totalPop = new NSGA2Population();
    Vector currentFront;

    //unify the two populations
    for (int i=0;i<populationSize;i++){
      totalPop.append((NSGA2Individual)oldPop.getData(i).clone());
      totalPop.append((NSGA2Individual)newPop.getData(i).clone());
    }

    totalPop.rank();
    totalPop.assignCrodwingDistance();

    //assign Individuals to nextPopulation
    NSGA2Population nextPop = new NSGA2Population();

//     System.out.println("Maxrank: " + totalPop.maxrank);
    for (int i=0;i<totalPop.maxrank;i++){
      totalPop.crowdingSort(i);
      currentFront = (Vector)totalPop.fronts.get(i);
      //System.out.println("groesse "+ i + ". Front: " + currentFront.size());
      for (int j=0;j<currentFront.size();j++){
        nextPop.append((NSGA2Individual)currentFront.get(j));
        if (nextPop.getSize() == populationSize) {
          break;
        }
      }
      if (nextPop.getSize() == populationSize) {
        break;
      }
    }
 //   totalPop.printRanking();
  //   System.out.println("newSize: " +nextPop.getSize());
    return nextPop;
  }

}
