/*
 * SPEA2Individual.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.spea2;

import java.io.Serializable;

import kea.Data;
import kea.Range;
import kea.Type;
import kea.Value;

/**
 * This class implements the individuals used by SPEA2.
 * 
 * @author Peter Senft
 * @version $Revision: 723 $
 */
public class SPEA2Individual implements Serializable, Cloneable {

  public SPEA2Individual() {
    this.speaRandom = new SPEA2Random();
    this.data = null;
    this.fitness = 0.0;
    this.strength = 0;
    this.raw = 0;
    this.decTypes = null;
    this.objTypes = null;
    this.decRanges = null;
    this.valid = false;
  }

  public SPEA2Individual(SPEA2Random speaRandom, Type[] decTypes, Type[] objTypes, Range[] decRanges) {
    this.initialize(speaRandom, decTypes, objTypes, decRanges);
  }

  public void initialize(SPEA2Random speaRandom, Type[] decTypes, Type[] objTypes, Range[] decRanges) {
    this.speaRandom = speaRandom;
    this.data = new Data(decTypes, objTypes, decRanges);
    this.fitness = 0.0;
    this.strength = 0;
    this.raw = 0;
    this.decTypes = decTypes;
    this.objTypes = objTypes;
    this.decRanges = decRanges;
    this.valid = false;
  }

  public Data getResults() {
    return this.data;
  }

  public void setResults(Data data) {
    this.data = data;
  }

  public void setFitness(double fitness) {
    this.fitness = fitness;
  }

  public double getFitness() {
    return this.fitness;
  }

  public void setStrength (int strength) {
    this.strength = strength;
  }

  public int getStrength () {
    return this.strength;
  }

  public void setRaw (int raw) {
    this.raw = raw;
  }

  public int getRaw() {
    return this.raw;
  }

  public void setDensity(double density) {
    this.density = density;
  }

  public double getDensity() {
    return this.density;
  }

  public boolean isValid() {
    return this.valid;
  }

  public void setValid() {
    this.valid = true;
  }

  public void setInValid() {
    this.valid = false;
  }

  public Object clone() {
    SPEA2Individual retval = null;
    // retval = (SPEA2Individual)super.clone();

    retval = new SPEA2Individual(this.speaRandom, this.decTypes, this.objTypes, this.decRanges);
    retval.setStrength(this.getStrength());
    retval.setRaw(this.getRaw());
    retval.setDensity(this.getDensity());
    retval.setFitness(this.getFitness());
    if (this.valid) {
      retval.setValid();
    }
    else {
      retval.setInValid();
    }
    // now set data
    Data newData = retval.getResults();
    Value[] newDecVector = newData.getDecVector();
    Value[] newObjVector = newData.getObjVector();
    // set decision vector
    for (int i = 0; i < this.decTypes.length; ++i) {
      Value newVal = newDecVector[i];
      if (this.data.getDecVector()[i].getData() != null) {
        switch (this.data.decVector[i].getTypeInfo()) {
          case Type.BOOLEAN     : newVal.setData(((Boolean)this.data.decVector[i].getData()).booleanValue());
                                  break;
          case Type.BYTE        : newVal.setData(((Byte)this.data.decVector[i].getData()).byteValue());
                                  break;
          case Type.CHAR        : newVal.setData(((Character)this.data.decVector[i].getData()).charValue());
                                  break;
          case Type.DOUBLE      : newVal.setData(((Double)this.data.decVector[i].getData()).doubleValue());
                                  break;
          case Type.FLOAT       : newVal.setData(((Float)this.data.decVector[i].getData()).floatValue());
                                  break;
          case Type.INTEGER     : newVal.setData(((Integer)this.data.decVector[i].getData()).intValue());
                                  break;
          case Type.LONG        : newVal.setData(((Long)this.data.decVector[i].getData()).longValue());
                                  break;
          case Type.SHORT       : newVal.setData(((Short)this.data.decVector[i].getData()).shortValue());
                                  break;
          case Type.STRING      : newVal.setData((String)this.data.decVector[i].getData());
                                  break;
        }
      }
      newDecVector[i] = newVal;
    }
    // set objective vector
    for (int i = 0; i < this.objTypes.length; ++i) {
      Value newVal = newObjVector[i];
//	  Kea.logger.logError("data size is  ................" + this.data.objVector.length);
//      Kea.logger.logError("data is..............."+ this.data.getObjVector()[i].getData().toString());
      if (this.data.getObjVector()[i].getData() != null) {
        switch (this.data.decVector[i].getTypeInfo()) {
          case Type.BOOLEAN     : newVal.setData(((Boolean)this.data.objVector[i].getData()).booleanValue());
                                  break;
          case Type.BYTE        : newVal.setData(((Byte)this.data.objVector[i].getData()).byteValue());
                                  break;
          case Type.CHAR        : newVal.setData(((Character)this.data.objVector[i].getData()).charValue());
                                  break;
          case Type.DOUBLE      : newVal.setData(((Double)this.data.objVector[i].getData()).doubleValue());
                                  break;
          case Type.FLOAT       : newVal.setData(((Float)this.data.objVector[i].getData()).floatValue());
                                  break;
          case Type.INTEGER     : newVal.setData(((Integer)this.data.objVector[i].getData()).intValue());
                                  break;
          case Type.LONG        : newVal.setData(((Long)this.data.objVector[i].getData()).longValue());
                                  break;
          case Type.SHORT       : newVal.setData(((Short)this.data.objVector[i].getData()).shortValue());
                                  break;
          case Type.STRING      : newVal.setData((String)this.data.objVector[i].getData());
                                  break;
        }
      }
      newObjVector[i] = newVal;
    }
    newData.setDecVector(newDecVector);
    newData.setObjVector(newObjVector);
    retval.setResults(newData);

    return retval;
  }

  public double distanceTo(SPEA2Individual ind) {
    double retval = 0.0;
    double thisval = 0.0;
    double indval = 0.0;

    // both arrays exist?
    if ((this.data.getObjVector() != null) && (ind.data.getObjVector() != null)) {
      // arrays of same size?
      if (this.data.getObjVector().length == ind.data.getObjVector().length) {
        for (int i = 0; i < this.data.getObjVector().length; ++i) {
          if ((this.data.getObjVector()[i].getData() != null) && (ind.data.getObjVector()[i].getData() != null)) {
            // get values from vector
            switch (this.data.getObjVector()[i].getTypeInfo()) {
              case Type.BOOLEAN   : if ( ((Boolean)this.data.getObjVector()[i].getData()).booleanValue() ) {
                                      thisval = 1.0;
                                    }
                                    else {
                                      thisval = 0.0;
                                    }
                                    if ( ((Boolean)ind.data.getObjVector()[i].getData()).booleanValue() ) {
                                      indval = 1.0;
                                    }
                                    else {
                                      indval = 0.0;
                                    }
                                    break;
              case Type.BYTE      : thisval = ((Byte)this.data.getObjVector()[i].getData()).doubleValue();
                                    indval = ((Byte)ind.data.getObjVector()[i].getData()).doubleValue();
                                    break;
              case Type.CHAR      : thisval = 0.0;
                                    indval = 0.0;
                                    break;
              case Type.DOUBLE    : thisval = ((Double)this.data.getObjVector()[i].getData()).doubleValue();
                                    indval = ((Double)ind.data.getObjVector()[i].getData()).doubleValue();
                                    break;
              case Type.FLOAT     : thisval = ((Float)this.data.getObjVector()[i].getData()).doubleValue();
                                    indval = ((Float)ind.data.getObjVector()[i].getData()).doubleValue();
                                    break;
              case Type.INTEGER   : thisval = ((Integer)this.data.getObjVector()[i].getData()).doubleValue();
                                    indval = ((Integer)ind.data.getObjVector()[i].getData()).doubleValue();
                                    break;
              case Type.LONG      : thisval = ((Long)this.data.getObjVector()[i].getData()).doubleValue();
                                    indval = ((Long)ind.data.getObjVector()[i].getData()).doubleValue();
                                    break;
              case Type.SHORT     : thisval = ((Short)this.data.getObjVector()[i].getData()).doubleValue();
                                    indval = ((Short)ind.data.getObjVector()[i].getData()).doubleValue();
                                    break;
              case Type.STRING    : thisval = 0.0;
                                    indval = 0.0;
                                    break;
              case Type.UNDEFINED : thisval = 0.0;
                                    indval = 0.0;
                                    break;
            }
            retval = retval + ((indval - thisval) * (indval - thisval));
          } else {
            retval = retval + Double.MAX_VALUE;
          }
        }
        retval = Math.sqrt(retval);
      }
      else {
        retval = Double.MAX_VALUE;
      }
    }
    else {
      retval = Double.MAX_VALUE;
    }

    return retval;
  }

  /**
   * Mutate this individual
   */
  public void mutate(double pMutation, int mutationType, double upperBound, double lowerBound, int scale) {
    Value[] decVals = this.data.getDecVector();
    for (int i = 0; i < this.data.getDecVector().length; ++i) {
      if (this.speaRandom.uniform01() < pMutation) {
        switch (decVals[i].getTypeInfo()) {
          case Type.BOOLEAN:
            decVals[i] = mutateBoolean(decVals[i], mutationType);
            break;
          case Type.DOUBLE:
            decVals[i] = mutateDouble(decVals[i], mutationType, upperBound, lowerBound, scale);
            break;
        }
      }
    }
    this.data.setDecVector(decVals);
    this.setInValid();
  }

  /**
   * Mutation for a double Value
   */
  private Value mutateDouble(Value toMutate, int mutationType, double upperBound, double lowerBound, int scale) {
    double retval = 0.0;
    double orgval = ((Double)toMutate.getData()).doubleValue();
    double minVal = Double.MIN_VALUE;
    double maxVal = Double.MAX_VALUE;
    // check for ranges
    if (toMutate.getRange() != null) {
      minVal = ((Double)toMutate.getRange()[0].getMinVal()).doubleValue();
      maxVal = ((Double)toMutate.getRange()[0].getMaxVal()).doubleValue();
    }
    // now the mutation
    switch (mutationType) {
      case 0: // ES
        retval = mutateDoubleES(orgval, minVal, maxVal);
        break;
      case 1: // GA (NSGA2)
        retval = mutateDoubleGA(orgval, minVal, maxVal);
        break;
      case 2: // Gaussian
        retval = mutateDoubleRandomAddGaussian(orgval, minVal, maxVal, upperBound, lowerBound, scale);
        break;
      case 3: // Random
        retval = mutateDoubleRandom(minVal, maxVal);
        break;
      case 4: // org + Random
        retval = mutateDoubleRandomAddUniform(orgval, minVal, maxVal, upperBound, lowerBound);
        break;
    }
    // set new value
    toMutate.setData(retval);

    // System.out.println("SPEA2Individual.mutate: " + orgval + " -> " + retval);

    return toMutate;
  }

  /**
   * ES mutation of a double
   */
  private double mutateDoubleES(double orgval, double minVal, double maxVal) {



    return orgval;
  }

  /**
   * GA mutation of a double
   */
  private double mutateDoubleGA(double orgval, double minVal, double maxVal) {
    double delta = 0.0;
    double deltaq = 0.0;
    double newGene = 0.0;
    double rndNumber = 0.0;
    double indi = 0.0;
    double xy = 0.0;
    double val = 0.0;
    int dim = this.data.decVector.length;

    // double y = orgval;
    if (orgval > minVal) {
      if ((orgval - minVal) < (maxVal - orgval)) {
        delta = (orgval - minVal) / (maxVal - minVal);
      }
      else {
        delta = (maxVal - orgval) / (maxVal - minVal);
      }
      rndNumber = this.speaRandom.uniform01();
      indi = 1.0 / 21.0;
      xy = 1.0 - delta;
      if (rndNumber <= 0.5) {
        val = 2 * rndNumber + (1.0 - 2.0 * rndNumber) * Math.pow(xy, 21.0);
        deltaq = Math.pow(val, indi) - 1;
      }
      else {
        val = 2 * (1.0 - rndNumber) + 2 * (rndNumber - 0.5) * Math.pow(xy, 21.0);
        deltaq = 1 - Math.pow(val, indi);
      }
      newGene = orgval + deltaq * (maxVal - minVal);
    }
    else {
      newGene = this.speaRandom.uniform01();
      newGene = newGene * (maxVal - minVal) + minVal;
    }

    return newGene;
  }

  /**
   * Random mutation of a double
   */
  private double mutateDoubleRandom(double minVal, double maxVal) {
    return this.speaRandom.uniformMinMaxDouble(minVal, maxVal);
  }

  /**
   * "Add a Random" mutation of a double
   */
  private double mutateDoubleRandomAddUniform(double orgval, double minVal, double maxVal, double upperBound, double lowerBound) {
    double retval = orgval;
    double addval = 0.0;
    double ranval = this.speaRandom.uniformMinMaxDouble(0, upperBound - lowerBound);

    if (this.speaRandom.flipP(0.5)) {
      // add
      retval = retval + ranval;
    }
    else {
      // sub
      retval = retval - ranval;
    }
    if (retval < minVal) {
      retval = minVal;
    }
    if (retval > maxVal) {
      retval = maxVal;
    }

    return retval;
  }

  /**
   * "Add a Random (gaussian)" mutation of a double
   */
  private double mutateDoubleRandomAddGaussian(double orgval, double minVal, double maxVal, double upperBound, double lowerBound, int scale) {
    double retval = orgval;
    double addval = 0.0;
    double ranval = this.speaRandom.gaussian() * (upperBound - lowerBound) / scale;

    retval = retval + ranval;
    if (retval < minVal) {
      retval = minVal;
    }
    if (retval > maxVal) {
      retval = maxVal;
    }

    return retval;
  }

  /**
   * Mutation for a Boolean Value
   */
  private Value mutateBoolean(Value toMutate, int mutationType) {
    boolean retval = false;
    boolean minVal = false;
    boolean maxVal = true;
    // check for ranges
    if (toMutate.getRange() != null) {
      minVal = ((Boolean)toMutate.getRange()[0].getMinVal()).booleanValue();
      maxVal = ((Boolean)toMutate.getRange()[0].getMaxVal()).booleanValue();
    }
    // now the mutation. Mutationtype does not matter here
    if (minVal == maxVal) {
      retval = minVal;
    }
    else {
      retval = this.speaRandom.flipP(0.5);
    }
    toMutate.setData(retval);

    return toMutate;
  }

  /**
   * Mate 2 individuals and produce one child
   */
  public SPEA2Individual mateWith(SPEA2Individual ind) {
    SPEA2Individual retval = new SPEA2Individual();
    Value[] newVal = new Value[this.data.getDecVector().length];
    Data newData = new Data(this.decTypes, this.objTypes, this.decRanges);

    retval = (SPEA2Individual)this.clone();

    for (int i = 0; i < newData.getDecVector().length; ++i) {
      if (this.speaRandom.flipP(0.5)) {
        newVal[i] = (Value)ind.getResults().getDecVector()[i].clone();
      }
      else {
        newVal[i] = (Value)this.getResults().getDecVector()[i].clone();
      }
    }
    newData.setDecVector(newVal);
    retval.setResults(newData);
    retval.setInValid();
    return retval;
  }

  /**
   * Mate 2 individuals and produce two children
   */
  public SPEA2Individual[] mateWith(double pRecombination, SPEA2Individual ind) {
    SPEA2Individual[] retval = new SPEA2Individual[2];
    retval[0] = new SPEA2Individual();
    retval[1] = new SPEA2Individual();

    Value[] newVal1 = new Value[this.data.getDecVector().length];
    Data newData1 = new Data(this.decTypes, this.objTypes, this.decRanges);
    Value[] newVal2 = new Value[this.data.getDecVector().length];
    Data newData2 = new Data(this.decTypes, this.objTypes, this.decRanges);

    retval[0] = (SPEA2Individual)this.clone();
    retval[1] = (SPEA2Individual)ind.clone();

    if (this.speaRandom.uniform01() < pRecombination) {
      for (int i = 0; i < newData1.getDecVector().length; ++i) {
        if (this.speaRandom.flipP(0.5)) {
          newVal1[i] = (Value)ind.getResults().getDecVector()[i].clone();
          newVal2[i] = (Value)this.getResults().getDecVector()[i].clone();
        }
        else {
          newVal1[i] = (Value)this.getResults().getDecVector()[i].clone();
          newVal2[i] = (Value)ind.getResults().getDecVector()[i].clone();
        }
      }
      newData1.setDecVector(newVal1);
      newData2.setDecVector(newVal2);
      retval[0].setResults(newData1);
      retval[1].setResults(newData2);
    }
    retval[0].setInValid();
    retval[1].setInValid();

    return retval;
  }

  /**
   * returns the boolean, whether this individual dominates the given
   * individual, concerning the optimiztionType
   */
  
  public boolean dominates(int optimizationType, SPEA2Individual ind) {
	// 0 - minimum / 1 - maximum
	boolean retval = true;

	if (this.data.objVector.length == ind.data.objVector.length) {
	  for (int i = 0; i < this.data.objVector.length; ++i) {
		if (retval == false) break;
		if (optimizationType == 0) {
		  // minimum
		  switch (this.data.getObjVector()[i].getTypeInfo()) {
			case Type.BOOLEAN   : if ((((Boolean)this.data.getObjVector()[i].getData()).booleanValue() == false) && (((Boolean)ind.data.getObjVector()[i].getData()).booleanValue() == true)) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.BYTE      : if (((Byte)this.data.getObjVector()[i].getData()).byteValue() < ((Byte)ind.data.getObjVector()[i].getData()).byteValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.CHAR      : retval = retval && false; // not implemented yet
								  break;
			case Type.DOUBLE    : if (((Double)this.data.getObjVector()[i].getData()).doubleValue() < ((Double)ind.data.getObjVector()[i].getData()).doubleValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.FLOAT     : if (((Float)this.data.getObjVector()[i].getData()).floatValue() < ((Float)ind.data.getObjVector()[i].getData()).floatValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.INTEGER   : if (((Integer)this.data.getObjVector()[i].getData()).intValue() < ((Integer)ind.data.getObjVector()[i].getData()).intValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.LONG      : if (((Long)this.data.getObjVector()[i].getData()).longValue() < ((Long)ind.data.getObjVector()[i].getData()).longValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.SHORT     : if (((Short)this.data.getObjVector()[i].getData()).shortValue() < ((Short)ind.data.getObjVector()[i].getData()).shortValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.STRING    : retval = retval && false; // not implemented yet
								  break;
			case Type.UNDEFINED : retval = retval && false;
								  break;
		  }
		}
		else {
		  // minimum
		  switch (this.data.getObjVector()[i].getTypeInfo()) {
			case Type.BOOLEAN   : if ((((Boolean)this.data.getObjVector()[i].getData()).booleanValue() == true) && (((Boolean)ind.data.getObjVector()[i].getData()).booleanValue() == false)) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.BYTE      : if (((Byte)this.data.getObjVector()[i].getData()).byteValue() > ((Byte)ind.data.getObjVector()[i].getData()).byteValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.CHAR      : retval = retval && false; // not implemented yet
								  break;
			case Type.DOUBLE    : if (((Double)this.data.getObjVector()[i].getData()).doubleValue() > ((Double)ind.data.getObjVector()[i].getData()).doubleValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.FLOAT     : if (((Float)this.data.getObjVector()[i].getData()).floatValue() > ((Float)ind.data.getObjVector()[i].getData()).floatValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.INTEGER   : if (((Integer)this.data.getObjVector()[i].getData()).intValue() > ((Integer)ind.data.getObjVector()[i].getData()).intValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.LONG      : if (((Long)this.data.getObjVector()[i].getData()).longValue() > ((Long)ind.data.getObjVector()[i].getData()).longValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.SHORT     : if (((Short)this.data.getObjVector()[i].getData()).shortValue() > ((Short)ind.data.getObjVector()[i].getData()).shortValue()) {
									retval = retval && true;
								  }
								  else {
									retval = retval && false;
								  }
								  break;
			case Type.STRING    : retval = retval && false; // not implemented yet
								  break;
			case Type.UNDEFINED : retval = retval && false;
								  break;
		  }
		}
	  }
	}
	else { // sizes don't mach
	  retval = false;
	}

	return retval;
  }

  
  
//  public boolean dominates(int optimizationType, SPEA2Individual ind) {
//    // 0 - minimum / 1 - maximum
//    boolean retval = true;
//
//    if (this.data.objVector.length == ind.data.objVector.length) {
//      for (int i = 0; i < this.data.objVector.length; ++i) {
//        if (optimizationType == 0) {
//          // minimum
//          switch (this.data.getObjVector()[i].getTypeInfo()) {
//            case Type.BOOLEAN   : if ((((Boolean)this.data.getObjVector()[i].getData()).booleanValue() == false) && (((Boolean)ind.data.getObjVector()[i].getData()).booleanValue() == true)) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.BYTE      : if (((Byte)this.data.getObjVector()[i].getData()).byteValue() < ((Byte)ind.data.getObjVector()[i].getData()).byteValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.CHAR      : retval = retval && false; // not implemented yet
//                                  break;
//            case Type.DOUBLE    : if (((Double)this.data.getObjVector()[i].getData()).doubleValue() < ((Double)ind.data.getObjVector()[i].getData()).doubleValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.FLOAT     : if (((Float)this.data.getObjVector()[i].getData()).floatValue() < ((Float)ind.data.getObjVector()[i].getData()).floatValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.INTEGER   : if (((Integer)this.data.getObjVector()[i].getData()).intValue() < ((Integer)ind.data.getObjVector()[i].getData()).intValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.LONG      : if (((Long)this.data.getObjVector()[i].getData()).longValue() < ((Long)ind.data.getObjVector()[i].getData()).longValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.SHORT     : if (((Short)this.data.getObjVector()[i].getData()).shortValue() < ((Short)ind.data.getObjVector()[i].getData()).shortValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.STRING    : retval = retval && false; // not implemented yet
//                                  break;
//            case Type.UNDEFINED : retval = retval && false;
//                                  break;
//          }
//        }
//        else {
//          // minimum
//          switch (this.data.getObjVector()[i].getTypeInfo()) {
//            case Type.BOOLEAN   : if ((((Boolean)this.data.getObjVector()[i].getData()).booleanValue() == true) && (((Boolean)ind.data.getObjVector()[i].getData()).booleanValue() == false)) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.BYTE      : if (((Byte)this.data.getObjVector()[i].getData()).byteValue() > ((Byte)ind.data.getObjVector()[i].getData()).byteValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.CHAR      : retval = retval && false; // not implemented yet
//                                  break;
//            case Type.DOUBLE    : if (((Double)this.data.getObjVector()[i].getData()).doubleValue() > ((Double)ind.data.getObjVector()[i].getData()).doubleValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.FLOAT     : if (((Float)this.data.getObjVector()[i].getData()).floatValue() > ((Float)ind.data.getObjVector()[i].getData()).floatValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.INTEGER   : if (((Integer)this.data.getObjVector()[i].getData()).intValue() > ((Integer)ind.data.getObjVector()[i].getData()).intValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.LONG      : if (((Long)this.data.getObjVector()[i].getData()).longValue() > ((Long)ind.data.getObjVector()[i].getData()).longValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.SHORT     : if (((Short)this.data.getObjVector()[i].getData()).shortValue() > ((Short)ind.data.getObjVector()[i].getData()).shortValue()) {
//                                    retval = retval && true;
//                                  }
//                                  else {
//                                    retval = retval && false;
//                                  }
//                                  break;
//            case Type.STRING    : retval = retval && false; // not implemented yet
//                                  break;
//            case Type.UNDEFINED : retval = retval && false;
//                                  break;
//          }
//        }
//      }
//    }
//    else { // sizes don't mach
//      retval = false;
//    }
//
//    return retval;
//  }

  /**
   * initializes the decision vector with random values in the given
   * range. the objective vector must initialized in a seperate step!
   */
  public void initRandom() {
    Object minVal = null;
    Object maxVal = null;
    Object ranVal = null;
    // initialize decision space
    for (int i = 0; i < this.data.getDecVector().length; ++i) {
      // find minimum and maximum
      if (this.data.getDecVector()[i].getRange() == null) {
        // no ranges defined
        switch (this.data.getDecVector()[i].getTypeInfo()) {
          case Type.BOOLEAN   : minVal = new Boolean(false);
                                maxVal = new Boolean(true);
                                break;
          case Type.BYTE      : minVal = new Byte(Byte.MIN_VALUE);
                                maxVal = new Byte(Byte.MAX_VALUE);
                                break;
          case Type.CHAR      : minVal = new Character(Character.MIN_VALUE);
                                maxVal = new Character(Character.MAX_VALUE);
                                break;
          case Type.DOUBLE    : minVal = new Double(Double.MIN_VALUE);
                                maxVal = new Double(Double.MAX_VALUE);
                                break;
          case Type.FLOAT     : minVal = new Float(Float.MIN_VALUE);
                                maxVal = new Float(Float.MAX_VALUE);
                                break;
          case Type.INTEGER   : minVal = new Integer(Integer.MIN_VALUE);
                                maxVal = new Integer(Integer.MAX_VALUE);
                                break;
          case Type.LONG      : minVal = new Long(Long.MIN_VALUE);
                                maxVal = new Long(Long.MAX_VALUE);
                                break;
          case Type.SHORT     : minVal = new Short(Short.MIN_VALUE);
                                maxVal = new Short(Short.MAX_VALUE);
                                break;
          case Type.STRING    : minVal = new String("UNDEFINED MIN");
                                maxVal = new String("UNDEFINED MAX");
                                break;
          case Type.UNDEFINED : minVal = null;
                                maxVal = null;
                                break;
        }
      }
      else {
        // ranges defined
        switch (this.data.getDecVector()[i].getTypeInfo()) {
          case Type.BOOLEAN   : minVal = new Boolean(((Boolean)this.data.getDecVector()[i].getRange()[0].getMinVal()).booleanValue());
                                maxVal = new Boolean(((Boolean)this.data.getDecVector()[i].getRange()[0].getMaxVal()).booleanValue());
                                break;
          case Type.BYTE      : minVal = new Byte(((Byte)this.data.getDecVector()[i].getRange()[0].getMinVal()).byteValue());
                                maxVal = new Byte(((Byte)this.data.getDecVector()[i].getRange()[0].getMaxVal()).byteValue());
                                break;
          case Type.CHAR      : minVal = new Character(((Character)this.data.getDecVector()[i].getRange()[0].getMinVal()).charValue());
                                maxVal = new Character(((Character)this.data.getDecVector()[i].getRange()[0].getMaxVal()).charValue());
                                break;
          case Type.DOUBLE    : minVal = new Double(((Double)this.data.getDecVector()[i].getRange()[0].getMinVal()).doubleValue());
                                maxVal = new Double(((Double)this.data.getDecVector()[i].getRange()[0].getMaxVal()).doubleValue());
                                break;
          case Type.FLOAT     : minVal = new Float(((Float)this.data.getDecVector()[i].getRange()[0].getMinVal()).floatValue());
                                maxVal = new Float(((Float)this.data.getDecVector()[i].getRange()[0].getMaxVal()).floatValue());
                                break;
          case Type.INTEGER   : minVal = new Integer(((Integer)this.data.getDecVector()[i].getRange()[0].getMinVal()).intValue());
                                maxVal = new Integer(((Integer)this.data.getDecVector()[i].getRange()[0].getMaxVal()).intValue());
                                break;
          case Type.LONG      : minVal = new Long(((Long)this.data.getDecVector()[i].getRange()[0].getMinVal()).longValue());
                                maxVal = new Long(((Long)this.data.getDecVector()[i].getRange()[0].getMaxVal()).longValue());
                                break;
          case Type.SHORT     : minVal = new Short(((Short)this.data.getDecVector()[i].getRange()[0].getMinVal()).shortValue());
                                maxVal = new Short(((Short)this.data.getDecVector()[i].getRange()[0].getMaxVal()).shortValue());
                                break;
          case Type.STRING    : minVal = new String((String)this.data.getDecVector()[i].getRange()[0].getMinVal());
                                maxVal = new String((String)this.data.getDecVector()[i].getRange()[0].getMinVal());
                                break;
          case Type.UNDEFINED : minVal = null;
                                maxVal = null;
                                break;
        }
      }
      // get random data in range
      switch (this.data.getDecVector()[i].getTypeInfo()) {
        case Type.BOOLEAN   : if (((Boolean)minVal).booleanValue() == ((Boolean)maxVal).booleanValue()) {
                                ranVal = (Boolean)minVal;
                              }
                              else {
                                ranVal = new Boolean(this.speaRandom.flipP(0.5));
                              }
                              break;
        case Type.BYTE      : ranVal = new Byte((new Double((this.speaRandom.uniform01() * (((Byte)maxVal).byteValue() - ((Byte)minVal).byteValue())) + ((Byte)minVal).byteValue())).byteValue());
        case Type.CHAR      : ranVal = new Character(Character.MIN_VALUE); // Not implemented yet
                              break;
        case Type.DOUBLE    : ranVal = new Double((this.speaRandom.uniform01() * (((Double)maxVal).doubleValue() - ((Double)minVal).doubleValue())) + ((Double)minVal).doubleValue());
                              break;
        case Type.FLOAT     : ranVal = new Float((new Double((this.speaRandom.uniform01() * (((Float)maxVal).floatValue() - ((Float)minVal).floatValue())) + ((Float)minVal).floatValue())).floatValue());
                              break;
        case Type.INTEGER   : ranVal = new Integer((new Double((this.speaRandom.uniform01() * (((Integer)maxVal).intValue() - ((Integer)minVal).intValue())) + ((Integer)minVal).intValue())).intValue());
                              break;
        case Type.LONG      : ranVal = new Long((new Double((this.speaRandom.uniform01() * (((Long)maxVal).longValue() - ((Long)minVal).longValue())) + ((Long)minVal).longValue())).longValue());
                              break;
        case Type.SHORT     : ranVal = new Short((new Double((this.speaRandom.uniform01() * (((Short)maxVal).shortValue() - ((Short)minVal).shortValue())) + ((Short)minVal).shortValue())).shortValue());
                              break;
        case Type.STRING    : ranVal = new String("NOT YET DEFINED");
                              break;
        case Type.UNDEFINED : ranVal = null;
                              break;
      }
      this.data.getDecVector()[i].setData(ranVal);
    }
  }

  private boolean valid;
  private SPEA2Random speaRandom;
  private Data data;
  private double fitness;
  private int strength;
  private int raw;
  private double density;
  private Type[] decTypes;
  private Type[] objTypes;
  private Range[] decRanges;
}
