/*
 * SPEA2Population.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.spea2;

import java.io.Serializable;

import kea.Data;
import kea.Range;
import kea.Type;
import kea.problems.AbstractProblem;

/**
 * This class implements the population used by SPEA2.
 * 
 * @author Peter Senft
 * @version $Revision: 723 $
 */
public class SPEA2Population implements Serializable, Cloneable {

  public SPEA2Population() {
    this.speaRandom = new SPEA2Random();
    this.objTypes = null;
    this.decTypes = null;
    this.decRanges = null;
    this.population = null;
    this.problem = null;
  }

  /**
   * mostly used constructor with given size, SPEA2Random and the objective, decision types
   */
  public SPEA2Population(int size, SPEA2Random speaRandom, Type[] decTypes, Type[] objTypes, Range[] decRanges, AbstractProblem problem) {
    this.decTypes = decTypes;
    this.objTypes = objTypes;
    this.decRanges = decRanges;
    this.setRandom(speaRandom);
    this.problem = problem;
    this.setSize(size);
  }

  /**
   * sets the size of the population and initialises every individual with
   * the decTypes and objTypes and SPEA2Random if not null.
   */
  public void setSize(int size) {
    if (size == 0) {
      this.population = null;
    }
    else {
      this.population = new SPEA2Individual[size];
      if (this.decTypes != null && this.objTypes != null) {
        for (int i = 0; i < size; ++i) {
          this.population[i] = new SPEA2Individual(this.speaRandom, this.decTypes, this.objTypes, this.decRanges);
        }
      }
    }
  }

  public int getSize() {
    if (this.population == null) {
      return 0;
    }
    else {
      return this.population.length;
    }
  }

  public void setRandom(SPEA2Random speaRandom) {
    this.speaRandom = speaRandom;
  }

  public void initRandom() {
    if (this.population != null) {
      for (int i = 0; i < this.population.length; ++i) {
        // initialize decision space
        this.population[i].initRandom();
      }
    }
  }

  public void mutate(double pMutation, int mutationType, double upperBound, double lowerBound, int scale) {
    for (int i = 0; i < this.population.length; ++i) {
      this.population[i].mutate(pMutation, mutationType, upperBound, lowerBound, scale);
    }
  }

  public void score() {
    for (int i = 0; i < this.population.length; ++i) {
      if (this.population[i].getResults().getDecVector()[0].getData() != null) {
        if (!this.population[i].isValid()) {
          this.population[i].getResults().setObjVector(this.problem.score(this.population[i].getResults().getDecVector()));
          this.population[i].setValid();
        }
      }
    }
  }

  public void setResults(Data[] population) {
    for (int i = 0; i < population.length; ++i) {
      this.population[i].setResults(population[i]);
    }
  }

  public Data[] getResults() {
    Data[] retval = new Data[this.population.length];

    for (int i = 0; i < this.population.length; ++i) {
      retval[i] = this.population[i].getResults();
    }

    return retval;
  }

  public SPEA2Individual getAt(int pos) {
    SPEA2Individual retval = null;
    if (pos >= 0 && pos < this.population.length) {
      retval = this.population[pos];
    }
    else {
      retval = null;
    }
  return retval;
  }

  public void setAt(int pos, SPEA2Individual ind) {
    if (pos >= 0 && pos < this.population.length) {
      this.population[pos] = ind;
    }
  }

  private AbstractProblem problem;
  private SPEA2Individual[] population;
  private SPEA2Random speaRandom;
  private Type[] decTypes;
  private Type[] objTypes;
  private Range[] decRanges;
}
