/*
 * SPEA2Random.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.algorithms.populationBased.spea2;

import java.io.Serializable;
import java.util.Random;

/**
 * This class implements the random number generator used by SPEA2.
 * 
 * @author Peter Senft
 * @version $Revision: 723 $
 */
public class SPEA2Random implements Serializable, Cloneable {

  /**
   * use the current time as seed
   */
  public SPEA2Random() {
    this.seed = System.currentTimeMillis();
    this.random = new Random(this.seed);
  }

  /**
   * use the argument as seed
   */
  public SPEA2Random(long seed) {
    this.seed = seed;
    this.random = new Random(this.seed);
  }

  /**
   * sets the seed to the given value and reinitializes the
   * random object
   */
  public void setSeed(long seed) {
    this.seed = seed;
    this.random = new Random(this.seed);
  }

  /**
   * returns the given seed
   */
  public long getSeed() {
    return this.seed;
  }

  /**
   * return the next uniformly distributed random
   * double in ]0;1[
   */
  public double uniform01() {
    double retval;

    do {
      retval = this.random.nextDouble();
      if (retval == 0) {
      }
    } while (retval == 0.0);

    return retval;
  }

  /**
   * return the next uniformly distributed random
   * double in ]min;max[
   */
  public double uniformMinMaxDouble(double min, double max) {
    return (this.uniform01() * (max - min)) + min;
  }

  /**
   * return the next uniformly distributed random
   * integer in [min;max[
   */
  public int uniformMinMax(int min, int max) {
    return ((this.random.nextInt(max - min)) + min);
  }

  /**
   * return the next uniformly distributed random
   * integer in [0;max[
   */
  public int uniform0Max(int max) {
    return this.random.nextInt(max);
  }

  /**
   * returns the next double value, which is gaussian (normally) distributed
   * with mean 0.0 and a standard derivation of 1.0
   */
  public double gaussian() {
    return this.random.nextGaussian();
  }

  /**
   * flip an unfair coin (head is up with probability P),
   * return true if head is up and false otherwise
   */
  public boolean flipP(double P) {
    boolean retval = false;
    if (this.uniform01() < P) {
      retval = true;
    }
    else {
      retval = false;
    }
    return retval;
  }

  /**
   * Random seed
   */
  private long seed;

  /**
   * Real random object. In this case the standard java Random-object.
   */
  private Random random;
}
