/*
 * AttainmentSurfaces.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.comparisons.attainmentSurfaces;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.text.DecimalFormat;
import java.util.Date;
import java.util.StringTokenizer;

import kea.CompResult;
import kea.Datacontainer;
import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.comparisons.AbstractComparison;
import kea.comparisons.common.CompException;

/**
 * Compares  pareto sets. assumes minimising on each objective.
 * <p>
 * <center>
 *     Mostats5.c Copyright (C) 2000, David Corne and Joshua Knowles
 * </center>
 * <p>
 * For further informations have a look at the included file Mostats5.c.
 * <p>
 *
 * Before you are able to execute this program make sure that your PATH or
 * LD_LIBRARY_PATH points to the kea.comparisons directory, and that
 * JNIMostats5.dll or lib{Sun/Lin}JNIMostats5.so is available in there, respective.
 * If the files aren't there use the following commands.  To compile c-code
 * under Windows try for example the \'cygwin\'-environment.
 * <p>
 * <ol>
 * <li> javah -jni kea.comparisons.attainmentSurfaces.AttainmentSurfaces
 * <li> g++ -c Mostats5.c                         (Windows),
 * <li> g++ -shared Mostats5.o -o JNIMostats5.dll  (Windows),<br>
 * respective <br>
 * <li> gcc -G -I/< ... >/jdk/< ... >/include/[solaris/,linux/]
 *          Mostats5.c -o lib{Sun/Lin}JNIMostats5.so        (Unix/Linux).
 * </ol>
 * 
 *
 * @author Miroslaw Dragan
 * @version $Revision: 723 $
 */
public class AttainmentSurfaces extends AbstractComparison{
    
    /* - Static initializer -
     * Ensures that the AttainmentSurfaces-library is loaded for future use.
     * Therefore, the library path must be set properly before.
     */   
    
    
    private AttainmentSurfaces(String forStatic){}
    
    static
    {
	String asPath = 
	    Datacontainer.replaceAll
	    ( new File
		( new AttainmentSurfaces("").getClass()
		    .getResource(".").getFile()).getAbsolutePath(), "%20", " ")
	    + File.separator;
	
	String osName =  System.getProperty("os.name").trim().substring(0,3);
	String libHead = (osName.equals("Win")) ? "" : "lib" + osName;
	String libTail = (osName.equals("Win")) ? "dll" : "so";
	
	String libraryName = libHead + "JNIMostats5." + libTail;

	try
	    {
		System.load(asPath + libraryName);
	    }
	catch(UnsatisfiedLinkError e)
	    { throw new CompException
		// Kea.logger.logError
		    ( "\nW A R N I N G:\n"
		      + "Kea was unable to set the path to your attainmentSurfaces-dir auto-\n"
		      + "matically or a file is missing.  AttainmentSurfaces requires that your\n\n"
		      + "\tLD_LIBRARY_PATH (Unix/Linux), PATH (Windows)\n\n"
		      + "points to the kea.comparisons.attainmentSurfaces dir.  There should\n"
		      + "be your " + libraryName + ".  If\n"
		      + "not, read in the documentation of this file how to create it.\n"

		      );
		     
	    }
    }
    
    
    /**
     * <code>AttainmentSurfaces()</code> does the startup initialization
     * of all parametes to their defaults.
     */
    
    
    public AttainmentSurfaces(){
	
	param = new Parameters();
	param.setMembership(this.getClass().getName());
	Parameter[] p = new Parameter[5];
	p[0]= new Parameter("Files", "", "Ers-Files: file1 | file2 | ... _xor_ dir to compare ");
	p[1]= new Parameter("Number of objectives", "", "Number of objectives");
	p[2]= new Parameter("Gridpoints", "500", "Recommended: Number of objectives d=2 => Gridpoints g=500, d=3 => g=19, d=4 => g=7");
	p[3]= new Parameter("GroupResults", false , "true if comparing more then one run of each algorithm, the runs have to be located in subdirectories");
	p[4]= new Parameter("Shift","1000","Shifts all the points by shift away from the origin in each dimension");
	
	
	param.setParameterArray(p);
	
    };
    
    /**
     *<code>prepareDataFile</code> prepares the input textfile for C-code Mostats5.c
     *
     */
    
    
    protected String prepareDataFile(File[] fileNames){
	
	// temp file for Knowles' c-code. Will be deleted on exit of the program
	String tempFileName="AttainmentSurfaces" + new Date().getTime()+".dat";
	String thisLine;
	new File(tempFileName).deleteOnExit();
	
	
	
	if(fileNames.length > 1){
	    
	    if(groupResultsText.equals("true")){
		
		try{
		    PrintWriter printer =
			new PrintWriter
			    ( new BufferedWriter( new FileWriter(tempFileName,true)));
		    
		   
		    
		    for(int j=0; j<fileNames.length; j++){
			File actualFile = new File(fileNames[j].getPath());
			
			
			// we need to check if the file is a directory. If it's true, we list all ers-files in File[]
			if(actualFile.isDirectory()){
			    File[] files = actualFile.listFiles();
			    
			    
			    
				
			    for(int i=0; i<files.length; i++){
				 
				 if (files[i].getName().endsWith("ers")){
				BufferedReader reader = new BufferedReader(new FileReader(files[i]));
				thisLine = reader.readLine();
				
				while(!thisLine.startsWith("###")) thisLine = reader.readLine();
				thisLine= reader.readLine();
				
				
				while(thisLine!=null){
				    int ind = thisLine.indexOf("/");
				    				    
								    if(ind!=-1){
					
					StringTokenizer st = new StringTokenizer(thisLine.substring(ind + 1, thisLine.length()));
					while (st.hasMoreTokens()){
					    String stringForDouble = st.nextToken();
					    Double db = new Double(stringForDouble);
					    double temp = db.doubleValue();
					    String resultedString = formatDouble(temp);
					    printer.print(resultedString+" ");
					    int k = st.countTokens();
					    if(k==0)printer.print("\n");
					    
					    					
										}
					
					
					
				    }
				    thisLine= reader.readLine();
				}
				
				if(i != files.length-1){printer.println("1000000");}
			    }
			    }
			    printer.println("2000000");
			    printer.flush();
			}
		   }
		      
		      printer.close();
	  } 
		
		catch(IOException e) {
		    throw new CompException("IOError occured while creating " + tempFileName);
		}
	    }
	    
	    
	    
	    else{
		if(groupResultsText.equals("false")){
		    try{
			PrintWriter printer =
			    new PrintWriter
				( new BufferedWriter( new FileWriter(tempFileName,true)));
					
			
			for(int i=0; i<fileNames.length; i++){
			    
			    if(fileNames[i].getName().endsWith(".ers")){
			    BufferedReader reader = new BufferedReader(new FileReader(fileNames[i]));
			    thisLine = reader.readLine();
			    
			    while(!thisLine.startsWith("###")) thisLine = reader.readLine();
			    thisLine= reader.readLine();
			    
			    
			    while(thisLine!=null){
				int ind = thisLine.indexOf("/");
				
				if(ind!=-1){
				    
				    StringTokenizer st = new StringTokenizer(thisLine.substring(ind + 1, thisLine.length()));
				    while (st.hasMoreTokens()){
					String stringForDouble = st.nextToken();
					Double db = new Double(stringForDouble);
					double temp = db.doubleValue();
					String resultedString = formatDouble(temp);
					printer.print(resultedString+" ");
					int k = st.countTokens();
					if(k==0)printer.print("\n");
					
					
				    }
				}
				thisLine= reader.readLine();
			    }
			    
			    printer.println("2000000"); }
			}
			
			printer.flush();
			
			
			
			printer.close();
		       
			
			
		    } 
		    catch(IOException e) {
		    throw new CompException("IOError occured while creating " + tempFileName);
		    }
		    
		    
		    
		    
		    
		    
		}
	    }
	    
	}
	else {{Kea.logger.logWarn
 	 (" less then two directories",this);}
	  
	}
	
	return tempFileName;
	
    }
    
    protected static String formatDouble(double value){
	DecimalFormat df = new DecimalFormat("#0.0000000000");
	String formattedDouble = df.format(value);
	return formattedDouble;
    }
    
    
    public static native double[] calculate(String file,String noObjectives, String gridPoints, String output, String shif, double[] resultFromC);
    
    /**
     * <code>compare</code> starts the comparison with the given parameters.   
     * @return <code>CompResult</code>, i.e. the result of a comparison.
     * @see kea.CompResult
     */
    
    
    public CompResult compare(){
	
	
	
	String tempFile = new String();
	File sourceDir;
	File[] fileNames;
	
	
	// we split the entry in param and check if it's a file-array or directory
	String[] compFiles=Datacontainer.split((String)param.getParameterByName("Files").getData(), "|");
	if(compFiles.length==1){
	
	 
	sourceDir = new File(compFiles[0]);
	if(sourceDir.isDirectory()!=true){Kea.logger.logWarn
 	 ("File is not a directory",this);}
	fileNames = sourceDir.listFiles();
	}
	
	else {
	groupResultsText="false";
	fileNames = new File[compFiles.length];
	for (int i=0 ; i<fileNames.length; i++){
	fileNames[i]=new File(compFiles[i]);}
	}
	

	tempFile=prepareDataFile(fileNames);
	
	
	
       
	
	double[] resultFromC;
	resultFromC = new double[0]; 
	output="0";
	double[] res=calculate(tempFile, nObjectives, gridPoints, output, shift, resultFromC);
	
	
	
	comp = new CompResult();
	
	
		
		if (!this.getFileName().equals("")){
	    resultFileName = this.getFileName();
	}
	
	
	
	comp.setSetOfParameters(new Parameters[]{param});
	
	
	// we have to prepare the outcome from c-code for compresult	
	String[] results;
	results = new String[res.length/2];
	
	if(groupResultsText.equals("true")){
	    for(int i=0; i<res.length/2;i++){
		results[i]=fileNames[i].getAbsolutePath()+": "+res[i]+" "+res[(res.length/2)+i];
		
	    }
	}
	if(groupResultsText.equals("false")){
	int validFiles=0;
	
	for(int i=0; i<fileNames.length; i++)
	{if (fileNames[i].getName().endsWith(".ers")) validFiles++; }
	
	
	
	    if(validFiles == 2){
		results[0]=fileNames[0].getAbsolutePath()+" "+res[0]+" % better or equal"+" "+res[1]+" % better" ; 
		
		
	    }
	   
	    else {int j=0;int i=0;
	  
	    while(i<fileNames.length){
	    if(fileNames[i].getName().endsWith(".ers")){
	    results[j]=fileNames[i].getAbsolutePath()+" : "+res[j]+"  "+res[(res.length/2)+j];j++;}i++;
	    }
	    }
	}
	
	
	comp.setExportFileName(resultFileName);
	
	comp.setCompResult(results);
	
	
	return comp;
	
    };
    
    
    
    /**
     * <code>getParameter</code> returns the <code>Parameters</code>-object of this class.
     *
     * @return <code>Parameters</code> of AttainmentSurfaces
     */
    
    
    
    
    
    public Parameters getParameter(){
	
	return param;
	
    };
   
    /**
     * <code>setParameter</code> sets the Parameters-object of this class.
     *
     * @param parameter <code>Parameters</code> of AttainmentSurfaces
     */
    
    public void setParameter(Parameters parameter){
	
	Parameter[] p = parameter.getParameterArray();
	filesToComp = parameter.getParameterByName("Files").getData().toString();
	nObjectives = parameter.getParameterByName("Number of objectives").getData().toString();
	gridPoints = parameter.getParameterByName("Gridpoints").getData().toString();
	groupResultsText = parameter.getParameterByName("GroupResults").getData().toString();
	shift = parameter.getParameterByName("Shift").getData().toString();
	
    };
   
    private String resultFileName;  
    private String groupResultsText;
    private boolean groupResults;
    private File[] files;
    private Parameters param;
    private CompResult comp;
    private File[] fileNames;
    private String sourceDir;
    private String filesToComp;
    private String nObjectives;
    private String gridPoints;
    private String output;
    private String shift;
    private String tempFileName;
}


