/*
 * Hypervolume.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.comparisons.hypervolume;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.util.StringTokenizer;
import java.util.Vector;

import kea.CompResult;
import kea.Data;
import kea.Datacontainer;
import kea.EvalResult;
import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.Value;
import kea.comparisons.AbstractComparison;
import kea.comparisons.common.ShiftResults;
import kea.evaluations.common.GrepHeader;

/**
  * This class computes the hypervolume for a set of nondominated Points and is based on a C-Code from Zitzler.
  * (It is not allowed to give a set of points with dominated ones or negative values.)
  * This set has to be
  * a result of an optimzation run. The objective functions of the optimization
  * problem have to be minimizing ones.
  * (For minimizing there has to be a reference point for one nondominated set.) If you do not give one, the
  * reference point is zero and the nondominated set with the smaller volume is the better one.
  *
  * @author Marion Scheel
  * @version $Revision: 723 $
  */
public class Hypervolume extends AbstractComparison{

  private Parameter[] p;

  private int numberOfPoints;

  private int numberOfObjectives;

  private String files;

  private String file1;

  private String file2;

  private String run;

  private String name;

  private int noP;

  private int noO;
  private double[] offset;


	/**
    * Creates the metric which computes the hypervolume created by the nondominated points and initializes the paramaters.
    */
  public Hypervolume() {
    param = new Parameters();
    param.setMembership(this.getClass().getName());
    param.addParameter(new Parameter("Files", "", "The result(s) (divorced by \"|\") you want to assess."));
    param.addParameter(new Parameter("referencePoint", "", "The point, so that the hypervolume can be computed."));
    //param.addParameter(new Parameter("resulting file name", this.getClass().getName(), "The name of the file containing the result(s) ending with .crs."));
  }

  /**
   * This method fetches the data (the nondominated points) from binary files.
   */
  private double[][] fetchData(File file) {


      if (Kea.logger.infoOn()){
        Kea.logger.logInfo("is fetching Data from "+file.toString(), this);
      }

    String filename = file.getName();
    double[][] front = {{}};

    if (filename.endsWith(".sav")) {

      EvalResult evalResult = new EvalResult();
      evalResult = (EvalResult)EvalResult.load(filename);

      front = new double[1][1];
      int k, l;
      Data[] ersTemp = evalResult.getEvalResult();
      k = ersTemp.length;

      for (int i=0; i < k; i++) {

        Value[] temp = ersTemp[i].getObjVector();
        l = temp.length;
        front = new double[k][l];

        for (int j=0; j<l; j++) {
	  front[i][j] = ((Double)(temp[j].getData())).doubleValue();
        }
      }
    }
    if (filename.endsWith(".ers")) {

      int k, l;
      String currentLine = "";
      Vector vec = new Vector();
      front = new double[0][0];
      noP = 0;
      noO = 0;

      try {
        BufferedReader in = new BufferedReader(new FileReader(file));
      // Skip Header.
        while(in.ready()) {
          currentLine = in.readLine();
          if (currentLine != null){
          // Search for end of header.
            if (currentLine.indexOf("###") != -1){
              if (Kea.logger.debugOn()) {
                Kea.logger.logDebug("end of header!", this);
              }
              break;
            }
          }
        }

        //Kea.logger.logInfo("Line = "+currentLine, this);
        while (in.ready()){
          currentLine = in.readLine().trim();
          if (currentLine != null) {
          // Skip empty lines.
            if (!currentLine.equals("")) {
	      break;
	    }
	  }
        }

      // All empty lines up to the first individuum or end of file skipped.
        if (Kea.logger.debugOn()) {
          Kea.logger.logDebug("First not empty line = " + currentLine, this);
        }

        while (in.ready()){
        //currentLine = in.readLine();
          if (currentLine != null) {
            //while (!currentLine.equals("")) {
            while (!currentLine.equals("")) {	    
              currentLine = currentLine.trim();
              int i = currentLine.indexOf('/');
	      //Kea.logger.logInfo("Index des /: "+i, this);
	      if (i>-1) {
              vec.addElement(currentLine.substring(i+1));
	      }
	      Kea.logger.logInfo("objective vector: "+(String)vec.lastElement(), this);
              currentLine = in.readLine();
	      if (currentLine == null) break;//???
	    }
          }
        }
 // Remove evalCount etc. ...		
        k = vec.size();
        noP = k;
        StringTokenizer tokenizer = new StringTokenizer((String)vec.elementAt(0), "  ");

        l = tokenizer.countTokens();
        front = new double[k][l];
        for(int a = 0; a < k; a++){
          StringTokenizer toke = new StringTokenizer((String)vec.elementAt(a), " ");
	  double[] doubles = new double[toke.countTokens()];
	  int d = toke.countTokens();
	  for (int c = 0; c < d; c++) {
	    String st = toke.nextToken();
	    doubles[c] = Double.parseDouble(st);
	  }
          for (int b = 0; b < l; b++){
	    front[a][b] = doubles[b];
	    Kea.logger.logInfo("front["+a+"]["+b+"] = "+front[a][b], this);
	  }
        }
      }
      catch (Exception e){}
    }
    return front;
  }

  private boolean dominates (double[] point1, double[] point2, int objectives) {

    int i;
    boolean betterInAnyObjective = false;

    for (i= 0; i < objectives && point1[i] >= point2[i]; i++) {
      if(point1[i] > point2[i]) {betterInAnyObjective = true;}
    }
      if ((i >= objectives) && betterInAnyObjective) {return true;}
      else {return false;}
  }

  private void swap(double[][] front, int i, int j){

    double[] temp;

    temp = front[i];
    front[i] = front[j];
    front[j] = temp;
  }

  /** This method computes the threshold which determines the extent the volume has reached so far.*/
  private double computeNoDoubleVolume (double[][] front, int points, int objective){

    double minValue, value;
    minValue = front[0][objective];

    for (int i=0; i < points; i++) {
      value = front[i][objective];
      if (value < minValue) {minValue = value;}
    }
    return minValue;
  }

  /** This method prevent to compute twofold determined spaces. Thus, some points
    * may cover parts of the same space.*/
  private int reduceNondominatedSet(double[][] front, int points, int objectives, double threshold) {

    int n;

    n = points;
    for (int i=0; i < n; i++) {
      if(front[i][objectives] <= threshold) {
        n--;
        swap(front, i, n);
      }
    }
    return n;
  }

  /** This method sorts the nondominated points for handling the computation of the
    * hypervolume better.*/
  private int filterNondominatedSet(double[][] front, int points, int objectives){

    int i, j, n;

    n=points;
    i=0;
    j=0;

    while (i < n) {
      j = i+1;
      while (j < n) {
        if (dominates(front[i], front[j], objectives)) {
          n--;
	  swap(front, j, n);
        }
        else if (dominates(front[j], front[i], objectives)) {
          n--;
          swap(front, i, n);
          i--;
          break;
        }
        else j++;
      }
      i++;
    }
    return n;
  }

  /** This method computes the hypervolue recursively with the help of the other methods.*/
  private double calculateHypervolume(double[][] front, int points, int objectives) {

    if (Kea.logger.infoOn()){
      Kea.logger.logInfo("is calculating the hypervolume ...", this);
    }
    int n;
    double volume, distance;

    volume = 0;
    distance = 0;
    n = points;

    try{
      while (n > 0) {

        int nondominatedPoints;
        double tempVolume, tempDistance;
        nondominatedPoints = filterNondominatedSet(front, n, objectives-1);
        tempVolume = 0;

        if (objectives < 3) {
	  if (nondominatedPoints < 1) {throw new NumberFormatException();}
	  tempVolume = front[0][0];
	  if (tempVolume < 0){
	    Kea.logger.logError("Not all points are positive. The volume is not comparable!", this);
	  }
	}

        else {tempVolume = calculateHypervolume(front, nondominatedPoints, objectives-1);}

        tempDistance = computeNoDoubleVolume(front, n, objectives-1);

//	if (Kea.logger.infoOn()){
	  Kea.logger.logInfo(volume+" + "+tempVolume+" * ("+tempDistance+" - "+distance+")", this );
//	}
        volume += tempVolume * (tempDistance - distance);
	if (Kea.logger.infoOn()){
          Kea.logger.logInfo("Volumen "+volume, this);
	}
	distance = tempDistance;
        n = reduceNondominatedSet(front, n, objectives-1, distance);
      }
    }
    catch (Exception NumberFormatException){}

    return volume;
  }

  public CompResult compare(){

    Parameters evalParams1 = new Parameters();
    Parameters evalParams2 = new Parameters();
    files = (String)(param.getParameterByName("Files")).getData();

    // If the problem is the Kursawe Testfunction, the input data gets shiftet
    // to get only positve objective values.
    String[] fileNames = Datacontainer.split(files, "|");
    String newValueFiles = "";
    for (int i = 0; i < fileNames.length; i++){
      ShiftResults testAndShift = new ShiftResults(fileNames[i]);
      if (testAndShift.test()){
        double[] shiftValues = new double[]{11,8};
        newValueFiles = newValueFiles + testAndShift.shift(shiftValues);
        Kea.logger.logWarn("The used Problem was the Kursawe Testproblem. This problem produces negative Objective values. This metric can't handle negative values, so the results got shifted.", this);
      }
    }
    if (!newValueFiles.equals("")){
      newValueFiles = newValueFiles.substring(0, newValueFiles.length()-2);
      files = newValueFiles;
    }

    CompResult compResult = new CompResult();
    Parameters[] allParams;

    /* (dk) write the exportFileName given by Kea or user */
    compResult.setExportFileName(getFileName());


    if(fileNames.length==0
       || fileNames.length>2)
      {
	Kea.logger.logError("Number of files should be 1 or 2",this);
	return null;
      }
    file1 = fileNames[0];
    file2 = (fileNames.length==2) ?  fileNames[1] : null;
      
    if(Kea.logger.infoOn()){
       Kea.logger.logInfo("file1 " + file1, this);
       if (file2 != null) {
         Kea.logger.logInfo("file2 " + file2, this);
       }
    }
    

    double[][] front1 = this.fetchData(new File(file1));
    turnAndShift(front1);
    double result1 = calculateHypervolume(front1, front1.length, front1[0].length);
    String[] resultArray = (file2 == null) ? new String[1] : new String[2];

    resultArray[0] = Double.toString(result1);
    if(Kea.logger.infoOn()){
      Kea.logger.logInfo("result1: "+resultArray[0]);
    }

    if (file2 != null){
      double[][] front2 = this.fetchData(new File(file2));
      turnAndShift(front2);
      double result2 = calculateHypervolume(front2, front2.length, front2[0].length);
      resultArray[1] = Double.toString(result2);
      if(Kea.logger.infoOn()){
        Kea.logger.logInfo("result2: "+resultArray[1]);
      }

    }

    Parameters[] params1 = GrepHeader.makeParams(file1);
    Parameters[] params2 = {};
    int limit = params1.length;  // space needed for file1-parameters

    if (fileNames.length == 2){
      params2 = GrepHeader.makeParams(file2);
      limit += (params2.length-1);
      // + space needed for file2-parameters
      // without problem parameters
    }

    limit++;  // leave space for the comparison-parameters
    allParams = new Parameters[limit];
 
    /* Start with problem-parameters of file1 */
   System.arraycopy(params1, 1,
		     allParams, 0, 1);
   /* followed by algo- and corresponding evaluation-parameters */
   System.arraycopy(params1, 0,
		     allParams, 1, 1);
   System.arraycopy(params1, 2,
		     allParams, 2, params1.length-2);

    if (fileNames.length == 2){
      /* algo- and corresponding evaluation-parameters of file2 */
      System.arraycopy(params2, 0,
		       allParams, params1.length, 1);
      System.arraycopy(params2, 2,
		       allParams, params1.length + 1, params2.length-2);
   }

    //System.out.println(param.toString());

    allParams[limit-1] = param;

    compResult.setSetOfParameters(allParams);

    compResult.setCompResult(resultArray);

    if (Kea.logger.infoOn()){
	if (file2 == null){
	  if (Kea.logger.infoOn())
	    Kea.logger.logInfo("CompResult:"+resultArray[0], this);
	}
	else if (Kea.logger.infoOn()){
	    Kea.logger.logInfo("CompResult:"+resultArray[0]
			       +", "+resultArray[1],this);
	}
    }


    return compResult;
  }


  public void setParameter(Parameters parameter){

    files = parameter.getParameterByName("Files").toString();
    //name = (String)(param.getParameterByName("resulting file name")).getData();
    //name = name.trim();
    StringTokenizer tokenizer = new StringTokenizer((String)(param.getParameterByName("referencePoint")).getData(), "  ");
    offset = new double[tokenizer.countTokens()];
	  int d = tokenizer.countTokens();
	  for (int i = 0; i < d; i++) {
	    String st = tokenizer.nextToken();
	    offset[i] = Double.parseDouble(st);
	    System.out.println(offset[i]);
	    }       
    
  }

  public Parameters getParameter(){
    return param;
  }

  public void setReferencePoint(String refpoint){
  }
  public void turnAndShift(double[][] front) {
  for (int i=0; i<front.length;i++) {
  for (int j=0; j<front[0].length;j++) {
  front[i][j]=front[i][j]*(-1);
  front[i][j]=front[i][j]+offset[j];
  }
  }
  }
}
