/* Rmetrics.cc  (C) Joshua Knowles 4th April 2002

** extended by Daniel Krmer 10th Januar 2003
** You can find all changes if you search for "(dk)".
** Partly, original code is outcommented, marked with "(dk) before"


This program calculates Hansen and Jaszkiewicz's R1, R2 and R3 measures 
for evaluating/comparing the quality of sets of nondominated vectors
generated by a multiobjective search algorithm. For more information
see sections below.

** Please contact me - Joshua Knowles - if you have any comments, suggestions
or questions regarding this program or metrics for measuring nondominated sets. 
My email address is jknowles@ulb.ac.be

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version. 

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details. 

   The GNU General Public License is available at:
      http://www.gnu.org/copyleft/gpl.html
   or by writing to: 
        The Free Software Foundation, Inc., 
        675 Mass Ave, Cambridge, MA 02139, USA.  


USAGE:
------
To Compile:
g++ Rmetrics.cc -o Rmetrics

To Run:
./Rmetrics -f <data_file> [-s <s>] [-k <n_k>] [-i1 <ideal1>]...[-i5 <ideal5>]
           [-n1 <nadir1>]...[-n5 <nadir5>] [-r <rho>] [-m <method>] [-h]

where all command line parameters are optional except for the data_file
and they have the following meanings and conventions:

<data_file> is any file whose format on each line is:
<v_1> <v_2>..<v_n_k> <alg_name> <run>
where v_1...v_n_k are the integer (now: double (dk)) components of the vector,
then there is an algorithm name, then there is a run number. All separators should
be a space. 
IMPORTANT: this code assumes minimization so if you are maximizing, please multiply 
all numbers in your data-file by -1.
((dk) not longer:) IMPORTANT: this code only accepts integer data points at present.)
*/ 
/*
<s> is an integer parameter which controls the number of different weight vectors 
that are used by the R metrics. An appropriate value of s depends on the value
of n_k, the number of objectives. The number of evenly distributed weight vectors
is given by the equation |wv| = (s + n_k - 1)^C_(n_k - 1) where n^C_k is the "choose"
function i.e. the number of ways of choosing n items from k. Default values of
s are:

n_k=2, s=500
n_k=3, s=30
n_k=4, s=12
n_k=5, s=8

<n_k> is the number of objectives i.e. the dimension of the vectors. At present a
maximum of 5 for n_k is allowed. The default value is 2.

<ideal1>..<ideal5> are the co-ordinates of an ideal point in the objective space.
If no value is set for ideal1 then the ideal point is calculated from all the data
in the file. When comparing two sets of results, however, it is VERY IMPORTANT that
the same ideal point is used. The ideal point must be better in all dimensions than
the best value of any point in each dimension.

<nadir1>..<nadir5> are the co-ordinates of a nadir (or worst) point in the objective space.
If no value is set for nadir1 then the nadir point is calculated from the data
in the file. When comparing two sets of results, however, it is VERY IMPORTANT that
the same ideal point is used. The nadir point must be worse in all dimensions than
the worst value of any point in each dimension.

<rho> is used in the calculation of the augmented Tchebycheff function. It controls the 
degree to which the Thebycheff function is augmented by a weighted sum function. Values
in the range 0.001 to 0.1 are normal.

<method> is an integer in 1..3 which defines whether the weighted sum (1), Thebycheff (2), or
augmented Thebycheff function (3) is used in the calculation of the R1 and R2 metrics.

-h causes this set of instructions to be printed.

OUTPUT:
-------

The program should recognise the number of algorithms and runs of each algorithm
that are in the data file. It should display this information. Then it calculates
the mean value of the distance of the vectors in each data set from the ideal point
and displays these numbers as follows:

0.34456 ACO 1
0.32234 ACO 2
0.35674 PAES 1
0.33232 PAES 2

for example where the number is the mean distance and the algorithm and run are given,
so here two algorithms, ACO and PAES, with two runs each must have been in the input
data file.

Following this, the R1, R2, and R3 metrics are calculated between pairs of 
algorithms for each pair of matching numbered runs. In other words, if there
are two runs of ACO and three of PAES, then there will be results for 
ACO run 1 vs PAES run 1 and for ACO run 2 vs PAES run 2, and that is all.


FURTHER INFORMATION:
--------------------

This code calculates the R1, R2 and R3 metrics of a set of points as described in
Hansen and Jaszkiewicz 1998:

@techreport{hansen98evaluating,
    author = "Michael Pilegaard Hansen and Andrzej Jaszkiewicz",
    title = "Evaluating the quality of approximations to the non-dominated set",
    number = "IMM-REP-1998-7",
    year = "1998",
    url = "citeseer.nj.nec.com/hansen98evaluating.html"}

These metrics were reviewed in the PhD thesis of Joshua Knowles and also in 
a paper by Knowles and Corne published in the 2002 Congress on 
Evolutionary Computation. These can
both be downloaded from http://iridia.ulb.ac.be/~jknowles/publications.html

*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
/* #include "random.h" */           // (dk): not used in this code


/* extended for use with JNI (dk) */
#include <jni.h>
#include "kea_comparisons_rmetrics_Rmetrics.h"

#define JNIRETURN(x,args...) sprintf(natReturnHelp,x,##args), strcat(natReturn, natReturnHelp), fprintf(stdout, natReturnHelp)
/*-------------------------*/

#define LARGE 2147000000.0
#define MAX_POINTS 20000
#define MAX_ALGS 12                // (dk) before:  MAX_ALGS 5
#define MAX_RUNS 30
#define MAX_LINES 501
#define MAX_K 5
#define GDFALSE 0
#define GDTRUE 1

FILE *fp;
double weighted_sum(double *wv, double *x, double *ideal, double *nadir);
double Tchebycheff(double *wv, double *x, double *ideal, double *nadir);
double augTcheby(double *wv, double *x, double *ideal, double *nadir);

/* (dk) before:
double weighted_sum(double *wv, int *x, int *ideal, int *nadir);
double Tchebycheff(double *wv, int *x, int *ideal, int *nadir);
double augTcheby(double *wv, int *x, int *ideal, int *nadir);
*/

long getlong(FILE * f, long *valaddr, int stopateol);
double getdouble(FILE * file, double *valaddr, int stopateol);
int getint(FILE * f, int *valaddr, int stopateol);
void get_file(char *filename);
void create_vectors(double **wv, const int s, const int k);
inline void int2kary(int x, const int basek, const int digits, int *kary);
inline int mypow(int x, int exp);
int choose(const int n, const int r);
int factorialdiff(const int n, const int r);
int factorial(const int n);
void print_help();

/* extended for use with JNI */
int main(int argc, char *argv[]);
bool natCalled = false;
/* boolean, to prevent constructing of the following return-String*/
char natReturnHelp[100];
char natReturn[16384]; /* choose a higher number for comparing more runs */
/* ------------------------------------------------------ */

int n_steps=10000;
int n_k=0;
int n_points[MAX_ALGS][MAX_RUNS];

/* long seed=3459243; */                        // (dk): not used in this code

double ideal[MAX_K];                            // (dk) before:  int ideal[MAX_K];
double point[MAX_ALGS][MAX_RUNS][MAX_POINTS][MAX_K]; // (dk) before: int point...
int method = 1;
double rho=0.01; // parameter controlling the augmented Tchebycheff metric
int s=0;
double nadir[MAX_K];                           // (dk) before:  int nadir[MAX_K];
int n_runs[MAX_ALGS];
char algname[MAX_ALGS][100];
int n_algs=0;
double min[MAX_ALGS][MAX_RUNS][MAX_LINES];

double **wv;


/* (dk) This code is needed to linke Java- and C-code 
 * It shouldn't affect the normal use of this class without Java,
 * since the main method still exists.  Furthermore the variable
 * natCalled prevents to execute "JNI-code", since it's set to
 * false as standard and is only true if the following lines were
 * executed.
 */

JNIEXPORT jstring JNICALL Java_kea_comparisons_rmetrics_Rmetrics_nativeCompArr
(JNIEnv *jEnv, jobject jObj, jobjectArray jObjArr)
{
  /* Point out that the program was called by means of JNI */
  natCalled = true;

  /* fprintf(stdout,"\n\nRmetrics.cc#nativeCompArr: preparing call of Rmetrics.main\n\n");
   * fflush(stdout);
   */
  
  /* The given jStrings shall now be transfered into a "char *argv[]"*/
  jsize jArgc = jEnv->GetArrayLength(jObjArr);
  char* jArgv[jArgc];

  int i;
  for (i = 0; i<jArgc; i++){
    jArgv[i]=(char*)jEnv->GetStringUTFChars((jstring)jEnv->GetObjectArrayElement(jObjArr,i),NULL);
  }

  main(jArgc,jArgv);
  
  /*  fprintf(stdout,"\n\nRmetrics.cc#nativeCompArr: freeing memory ...\n");*/
  for(i=0; i<jArgc; i++){
    jEnv->ReleaseStringUTFChars((jstring)jEnv->GetObjectArrayElement(jObjArr,i),jArgv[i]);
  }

  fflush(stdout);

  return jEnv->NewStringUTF(natReturn);
}

/* --------------------------------------------- */

int main(int argc, char *argv[])
{
  int a,b;
  int least_runs;
  int i,j,k;
  int run,alg;
  double total_wv;
  double val, minim;
  int best;
  bool ff = GDFALSE;

  double aval,bval,atotal1,atotal2,atotal3;

  double total_val=0.0;
  double mean_val;

  char filename[100];

  /* (dk) Did the JNI-call reach the main-method?
     fprintf(stderr,"\nmain-method reached\n");
     
     for(i=0; i<argc; i++){
     fprintf(stderr,"%s\n",argv[i]);
     fflush(stderr);
     }
     ------------------------------------------ */




  for(k=0;k<5;k++)
    {
      ideal[k]=-1;
      nadir[k]=-1;
    }
  
  /*  if(argc>1){
      sprintf(filename,"%s", argv[1]);
      }
      else
      {
      fprintf(stderr,"Input file needed as an argument. Exiting now.\n");
      exit(1);
      }
  */  //(dk)
  if(argc==1){
    fprintf(stderr,"Missing arguments");
    exit(1);
  }
  else if(argc==2)
    {
      if(strcmp("-h", argv[1])==0)
	{
	  print_help();
	  exit(0);
	}
      else
	fprintf(stderr, "Undefined command line parameter entered. Do \"./Rmetrics -h\" for help with parameters.\n");
      exit(1);
    }
  
  
  else if((argc>2)&&(argc%2==1))
    {
      for(i=1;i<argc-1;i+=2)
	{
	  if(strcmp("-f", argv[i])==0)
	    {
	      sprintf(filename,"%s", argv[i+1]);
	      ff=GDTRUE;
	    }
	  else if(strcmp("-s", argv[i])==0)
 	    s = atoi(argv[i+1]);
	  else if(strcmp("-k", argv[i])==0)
 	    n_k = atoi(argv[i+1]);
	  else if(strcmp("-i1", argv[i])==0)
	    {
	      ideal[0] = (double) atof(argv[i+1]);

	      /* (dk) before: ideal[0] = atoi(argv[i+1]);
	       * all lines containing atof were changed, too
	       */

	      // fprintf(stderr,"%f, ",ideal[0]);
	    }
	  else if(strcmp("-i2", argv[i])==0)
	    {
	      ideal[1] = (double) atof(argv[i+1]);
	      // fprintf(stderr,"%f, ",ideal[0]);
	    }
	  else if(strcmp("-i3", argv[i])==0)
 	    ideal[2] = (double) atof(argv[i+1]);
	  else if(strcmp("-i4", argv[i])==0)
 	    ideal[3] = (double) atof(argv[i+1]);
	  else if(strcmp("-i5", argv[i])==0)
 	    ideal[4] = (double) atof(argv[i+1]);
	  else if(strcmp("-m", argv[i])==0)
	    method = atoi(argv[i+1]);
	  else if(strcmp("-r", argv[i])==0)
	    rho = (double) atof(argv[i+1]);
	  else if(strcmp("-n1", argv[i])==0)
	    {
	      nadir[0] = (double) atof(argv[i+1]);
	      // fprintf(stderr,"%f, ",ideal[0]);
	    }
	  else if(strcmp("-n2", argv[i])==0)
	    {
	      nadir[1] = (double) atof(argv[i+1]);
	      // fprintf(stderr,"%f.\n",ideal[0]);
	    }
	  else if(strcmp("-n3", argv[i])==0)
	    nadir[2] = (double) atof(argv[i+1]);
	  else if(strcmp("-n4", argv[i])==0)
	    nadir[3] = (double) atof(argv[i+1]);
	  else if(strcmp("-n5", argv[i])==0)
	    nadir[4] = (double) atof(argv[i+1]);
	  else
	    {
	      fprintf(stderr, "Undefined command line parameter entered. Do \"./Rmetrics -h\" for help with parameters.\n");
	      exit(1);
	    }
	}
    }
  
  if(!ff)
    exit(1);

  if(n_k==0)
    n_k=2;

  if(s==0)
    {
      if(n_k==2)
	s=500;
      else if(n_k==3)
	s=30;
      else if(n_k==4)
	s=12;
      else if(n_k==5)
	s=8;
    }

  get_file(filename);
  
  //exit(0);

  /*  for(i=0;i<n_points;i++)
      {
      for(k=0;k<n_k;k++)
      fprintf(stdout, "%d ",point[i][k]);
      fprintf(stdout, "\n");
      }
  */

  //fprintf(stdout,"Calculating choose(%i, %i)\n",s+n_k-1,n_k-1);
  //fflush(stdout);
  n_steps = choose(s+n_k-1,n_k-1);


  /* added natCalled and JNIRETURN (dk) */
  if(!natCalled) fprintf(stdout, "Number of weight vectors will be %d\n", n_steps);
  else JNIRETURN("Number of weight vectors will be %d\n", n_steps);

  if(n_steps>MAX_LINES)
    {
      fprintf(stderr,"Maximum number is currently %d. Exiting\n", MAX_LINES);
      exit(1);
    }

  /* added natCalled and JNIRETURN (dk) */
  if(!natCalled)
    {
      if(method==1)
	printf("Will be using weighted sum evaluation\n");
      else if (method==2)
	printf("Will be using Tchebycheff evaluation\n");
      else
	printf("Will be using augmented Tchebycheff evaluation\n");
    }
  else
    {
      if(method==1)
	JNIRETURN("Will be using weighted sum evaluation\n");
      else if (method==2)
	JNIRETURN("Will be using Tchebycheff evaluation\n");
      else
	JNIRETURN("Will be using augmented Tchebycheff evaluation\n");
     }
      
  /* (dk) show the informations, which are already available */
  fflush(stdout);

  
  if((wv = (double **)malloc(n_steps * sizeof(double *)))==NULL)
    {
      fprintf(stdout, "Malloc error. Exiting\n");
      exit(1);
    }
  for(i=0;i<n_steps;i++)
    {
      if((wv[i] = (double *)malloc(n_k * sizeof(double)))==NULL)
	{
	  fprintf(stdout, "Malloc error. Exiting\n");
	  exit(1);
	}
    }

  create_vectors(wv,s,n_k);

  for(alg=0;alg<n_algs;alg++)
    {
      for(run=0;run<n_runs[alg];run++)
	{
	  total_val=0.0;
	  for(i=0;i<n_steps;i++)
	    {
	      minim=LARGE;
	      for(j=0;j<n_points[alg][run];j++)
		{
		  if(method==1)
		    val=weighted_sum(wv[i], point[alg][run][j], ideal, nadir);
		  else if(method==2)
		    val=Tchebycheff(wv[i], point[alg][run][j], ideal, nadir);
		  else
		    val= augTcheby(wv[i], point[alg][run][j], ideal, nadir);
		  if(val<minim)
		    {
		      best=j;
		      min[alg][run][i]=minim=val;
		    }
		}
	      // fprintf(stderr,"min=%lf, wv=%lf %lf, best=%d %d\n", min, wv[i][0], wv[i][1], point[best][0], point[best][1]);
	      total_val+=minim;
	    }
	  mean_val = total_val/n_steps;
	  
	  /* added natCalled and JNIRETURN (dk) */
	  
	  if(!natCalled) fprintf(stdout, "Mean_val:  %lf %s %d\n", mean_val, algname[alg], run+1);
	  else JNIRETURN("Mean_val:  %lf %s %d\n", mean_val, algname[alg], run+1);
	}
    }

  
  /*********   JNI extended version (dk)  **********
   * use of JNIRETURN in combination with natCalled *
   **************************************************/
  
   
  if(!natCalled) fprintf(stdout, "R1/R2/R3 metric:\n");
  else JNIRETURN("R1/R2/R3 metric:\n");
  for(a=0;a<n_algs-1;a++)
    {
      for(b=a+1;b<n_algs;b++)
	{
	  if(n_runs[a]<=n_runs[b])
	    least_runs=n_runs[a];
	  else
	    least_runs=n_runs[b];
	  
	  for(run=0;run<least_runs;run++)
	    {
	      atotal1=0.0;
	      atotal2=0.0;                      // (dk) before: see below "R2 metric"
	      atotal3=0.0;                      // (dk) before: see below "R3 metric"
	      for(i=0;i<n_steps;i++)  
		{
		  aval=min[a][run][i];
		  bval=min[b][run][i];


		  if(aval < bval)
		    atotal1+=1.0;
		  else if (bval == aval)
		    atotal1+=0.5;

		  // (dk) following two lines are extracted from below 
		  atotal2+=min[b][run][i]-min[a][run][i];                              // (dk)
		  atotal3+=(min[b][run][i]-min[a][run][i])/(min[b][run][i]+0.0000001); // (dk)
		}
	      atotal1/=n_steps;
	      atotal2/=n_steps;                    // (dk)
	      atotal3/=n_steps;                    // (dk)
	      /* (dk) changed %lf to %.17E */
	      if(!natCalled) fprintf(stdout, "alg %s versus alg %s on run %d = \t%lf/%lf/%lf\n", algname[a],
				     algname[b], run+1, atotal1, atotal2, atotal3);
	      else JNIRETURN("alg %s versus alg %s on run %d = \t%lf/%lf/%lf\n", algname[a], algname[b],
			     run+1, atotal1, atotal2, atotal3);
	    }
	}
    }

  /* (dk) the following lines are grouped in the above part */
    
  /* if(!natCalled) fprintf(stdout, "R1 metric:\n");
     else JNIRETURN("R1 metric:\n");
     for(a=0;a<n_algs-1;a++)
     {
     for(b=a+1;b<n_algs;b++)
     {
     if(n_runs[a]<=n_runs[b])
     least_runs=n_runs[a];
     else
     least_runs=n_runs[b];
     
     for(run=0;run<least_runs;run++)
     {
     atotal=0.0;
     for(i=0;i<n_steps;i++)
     {
     aval=min[a][run][i];
     bval=min[b][run][i];
     
     if(aval < bval)
     atotal+=1.0;
     else if (bval == aval)
     atotal+=0.5;
     }
     atotal/=n_steps;
     // (dk) changed %lf to %.17E
     if(!natCalled) fprintf(stdout, "alg %s versus alg %s on run %d = %lf\n", algname[a], algname[b], run+1, atotal);
     else JNIRETURN("alg %s versus alg %s on run %d = %lf\n", algname[a], algname[b], run+1, atotal);
     }
     }
     }
      
     if(!natCalled)  fprintf(stdout, "R2 metric:\n"); 
     else JNIRETURN("R2 metric:\n");
     for(a=0;a<n_algs-1;a++)
     {
     for(b=a+1;b<n_algs;b++)
     {
     if(n_runs[a]<=n_runs[b])
     least_runs=n_runs[a];
     else
     least_runs=n_runs[b];
     
     for(run=0;run<least_runs;run++)
     {
     atotal2=0.0;
     for(i=0;i<n_steps;i++)
     {
     atotal2+=min[b][run][i]-min[a][run][i];
     }
     atotal2/=n_steps;
     // (dk) changed %lf to %.17E 
     if(!natCalled) fprintf(stdout, "alg %s versus alg %s on run %d = %lf\n", algname[a], algname[b], run+1, atotal2);
     else JNIRETURN("alg %s versus alg %s on run %d = %lf\n", algname[a], algname[b], run+1, atotal2);
     }
     }
     }
     
     if(!natCalled) fprintf(stdout, "R3 metric:\n");
     else JNIRETURN("R3 metric:\n");
     for(a=0;a<n_algs-1;a++)
     {
     for(b=a+1;b<n_algs;b++)
     {
     if(n_runs[a]<=n_runs[b])
     least_runs=n_runs[a];
     else
     least_runs=n_runs[b];
     
     for(run=0;run<least_runs;run++)
     {
     atotal3=0.0;
     for(i=0;i<n_steps;i++)
     {
     atotal3+=(min[b][run][i]-min[a][run][i])/(min[b][run][i]+0.0000001);
     }
     atotal3/=n_steps;
     // (dk) changed %lf to %.17E 
     if(!natCalled) fprintf(stdout, "alg %s versus alg %s on run %d = %lf\n", algname[a], algname[b], run+1, atotal3);
     else JNIRETURN("alg %s versus alg %s on run %d = %lf\n", algname[a], algname[b], run+1, atotal3);
     }
     }
     }
  */

}

void get_file(char *filename)
{
  int a,b,c;
  int i,j,k;
  int total_points;
  double value[MAX_K];   // (dk) before:  int value[MAX_K];
  char alg[100];

  int algnum;
  int scan;
  int run;
  
  bool algflag;
  bool set;
  bool nset;
  bool get;

  if(nadir[0]==-1.0)
    nset=0;
  else
    nset=1;

  if(ideal[0]==-1.0)
    set=0;
  else
    set=1;

  if(!set)
    for(k=0;k<n_k;k++)
      ideal[k]=LARGE;

  for(i=0;i<MAX_ALGS;i++)
    {
      n_runs[i]=0;
      for(j=0;j<MAX_RUNS;j++)
	n_points[i][j]=0;
    }

  if(fp=fopen(filename, "rb"))
    {
      j=0;
      while(!feof(fp))
	{
	  for(i=0;i<n_k;i++)
	    { 
	      /* not needed any more (dk) */
	      //get=getint(fp, &value[i], GDFALSE);

	      get=fscanf(fp,"%lf", &value[i]);
	      /* outcommented; %.17E instead of %d (dk) */
	      // fprintf(stderr, "%lf ", value[i]);

	    }
	  // fprintf(stderr,"\n"); 
	  scan=0;
	  
	  /* (dk) Warn and exit, if errors occur while reading objective vectors */
	  if(!get && !feof(fp))
	    {
	      fprintf(stderr,"[CC: Rmetrics.get_file] ERROR:  At least one double value is ");
	      fprintf(stderr,"missing.\nMake sure, that each objective vector ");
	      fprintf(stderr,"contains at most the char \'E\' as in \n<double>E+-<int>; ");
	      fprintf(stderr,"even \"NaN\" is not allowed.\n");
	      exit(-1);
	    }

	  /* (dk) outcommented, because of condition above */
	  // if(get)
	  //  {
	  scan=fscanf(fp, "%s %d", alg, &run);
	  
	  /* (dk) Stop if no arguments were found, i.e. scan==-1 */
	  if(scan==-1) break;
	  
	  /* outcommented (dk) */
	  //fprintf(stderr,"scan=%d: alg = %s; run = %d\n",scan, alg, run);
	  
	  if(scan==2)
	    {
	      algflag=0;
	      
	      /* check if identified algo is already known */
	      for(a=0;a<n_algs;a++)
		if(strcmp(alg,algname[a])==0)
		  {
		    algnum=a;
		    algflag=1;
		    a=n_algs;
		    // fprintf(stdout, "algnum=%d\n",algnum);
		  }
	      /* otherwise "add" it to the others */
	      if(!algflag)
		{		 
		  algnum=n_algs;
		  n_algs++;
		  sprintf(algname[algnum],alg);
		  // fprintf(stdout, "algnum=%d\n",algnum);
		}
	      
	      if(run>n_runs[algnum])
		n_runs[algnum]=run;
	      
	      for(i=0;i<n_k;i++)
		{
		  point[algnum][run-1][n_points[algnum][run-1]][i]=value[i];
		  if(!set)
		    if(value[i]<ideal[i])
		      ideal[i]=value[i];
		  if(!nset)
		    if(value[i]>nadir[i])
		      nadir[i]=value[i];
		  
		}
	      (n_points[algnum][run-1])++;
	      // fprintf(stdout, "number of points = %d\n", n_points[algnum][run-1]);
	      j++;
	    }
	  else 
	    {
	      /* (dk) Added some lines for better understanding */
	      fprintf(stderr, "ERROR: missing algorithm and run information\n");
	      fprintf(stderr, "Check your files whether the first String in\n");
	      fprintf(stderr, "every line is the name of the algorithm!\n");
	      fprintf(stderr, "Those lines will be ingnored (in a future version)!\n\n");
	      exit(1);
	    }
	  //}
	}
      fclose(fp);
    }
  else
    {
      fprintf(stderr,"Couldn't open file. Exiting now.\n");
      exit(1);
    }  

  /*  for(i=0;i<MAX_ALGS;i++)
    for(j=0;j<2;j++)
      fprintf(stdout, "n_points=%d\n", n_points[i][j]);
  */

  total_points=j;

  /*******   extended for use with JNI (dk)   *******
   *    again: natCalled and JNIRETURN were used    *
   **************************************************/
  if(!natCalled){
    fprintf(stderr,"Number of algorithms = %d\n", n_algs);
    
    for(i=0;i<n_algs;i++)
      {
	for(j=0;j<n_runs[i];j++)
	  fprintf(stderr, "Number of points in alg %s run %d = %d\n", algname[i], j+1, n_points[i][j]);
      }
    fprintf(stderr,"Number of points read = %d\n", total_points);
    
    fprintf(stderr,"Nadir = ");
    
    for(i=0;i<n_k;i++)
      /* (dk) before %lf instead of %.17E */

      fprintf(stderr,"%lf ", nadir[i]);
    fprintf(stderr,"\n");
    
    fprintf(stderr,"Ideal = ");
    for(i=0;i<n_k;i++)
      fprintf(stderr,"%lf ", ideal[i]);
    fprintf(stderr,"\n");
  }
  else { 
    JNIRETURN("Number of algorithms = %d\n", n_algs);
    
    for(i=0;i<n_algs;i++)
      {
	JNIRETURN("Number of runs of algorithm %s = %d\n",algname[i], n_runs[i]);
	
	for(j=0;j<n_runs[i];j++)
	  JNIRETURN("Number of points in alg %s run %d = %d\n", algname[i], j+1, n_points[i][j]);
      }
    JNIRETURN("Number of points read = %d\n", total_points);
    
    JNIRETURN("Nadir = ");
    
    for(i=0;i<n_k;i++)
      JNIRETURN("%lf ", nadir[i]);
    JNIRETURN("\n");
    
    JNIRETURN("Ideal = ");
    for(i=0;i<n_k;i++)
      JNIRETURN("%lf ", ideal[i]);
    JNIRETURN("\n");
  }
}

/* ------------------------------------------------------------------- */

double weighted_sum(double *wv, double *x, double *ideal, double *nadir)
{
  int i;
  double total=0.0;

  for(i=0;i<n_k;i++)
    {
      total += wv[i]*((double)(x[i]-ideal[i])/(double)(nadir[i]-ideal[i])); 
    }
  return(total);
}

double Tchebycheff(double *wv, double *x, double *ideal, double *nadir)
{
  int i;
  double val;
  double max=0.0;

  for(i=0;i<n_k;i++)
    {
      val = wv[i]*((double)(x[i]-ideal[i])/(double)(nadir[i]-ideal[i])); 
      if(val>max)
	max=val;
    }
  return(max);
}

double augTcheby(double *wv, double *x, double *ideal, double *nadir)
{
  double result;
  result = Tchebycheff(wv, x, ideal, nadir) + rho*weighted_sum(wv,x,ideal,nadir);
  return(result);
}

/***************************************************************/
/* Get the next number from the input: put it in the location  */
/* addressed by second argument. This function returns 0 on    */
/* EOF. If stopateol is true, it returns -1 when it hits \n    */
/* (after which some other procedure has to read past the \n), */
/* otherwise it continues looking for the next number.         */
/* A number has an optional sign, perhaps followed by digits,  */
/* perhaps followed by a decimal point, perhaps followed by    */
/* more digits. There must be a digit somewhere for it to count*/
/* as a number. So it would read any of:                       */
/*  -.5                                                        */
/*  -0.5                                                       */
/*  -.5.7                                                      */
/* as minus-a-half. In the last case, it would read .7 next    */
/* time around.                                                */
/*   There doesn't seem to be a neat and reliable way to do    */
/* all this, including stopateol, using scanf?                 */
/***************************************************************/

/* (dk) outcommented the following lines, because there was only
 * one position were it was needed and this is outcommented, too.
 * Maybe it's right, that scanf can't identify the end of line,
 * but this shouldn't be a problem here since each line ends with
 * <algoName> <runNumber>.  If we hit <algoName> normally an error
 * occurs, because it isn't a double as expected.
 */

/* double getdouble(FILE * file, double *valaddr, int stopateol)
   {
   int c;
   int found = GDFALSE, indecimal = GDFALSE;
   int sign = +1;
   double n = 0.0, p = 1.0;
   
   // First find what looks like start of a number - the first digit. 
   // And note any sign and whether we just passed a decimal point.   
   do {
   c = fgetc(file);
   if (c == EOF)
   return (0);
   else if (stopateol && c == '\n')
   return (-1);
   else if (c == '+' || c == '-') {
   sign = (c == '+') ? +1 : -1;
   c = fgetc(file);
   if (c == EOF)
   return (0);
   else if (stopateol && c == '\n')
   return (-1);
   }
   if (c == '.') {
   indecimal = GDTRUE;
   c = fgetc(file);
   if (c == EOF)
   return (0);
   else if (stopateol && c == '\n')
   return (-1);
   }
   if (c >= '0' && c <= '9') {
   found = GDTRUE;
   } else {
   sign = +1;
   indecimal = GDFALSE;
   }
   } while (!found);
   
   // Now we've got digit(s) ...
   do {
   n = 10.0 * n + c - '0';
   p = 10.0 * p;
   c = fgetc(file);
   
   if ((c < '0') || (c > '9')) {
   found = GDFALSE;
   // We've run out. If we already saw a decimal point, return now 
   if (indecimal) {
   if (c != EOF)
   ungetc(c, file);
   *valaddr = sign * n / p;
   return (1);
   } else
   p = 1.0;
   }
   } while (found);
   
   // We ran out and we didn't see a decimal point, so is this a decimal? 
   if (c != '.') {
   // No, give it back to caller  
   if (c != EOF)
   ungetc(c, file);
   *valaddr = sign * n;
   return (1);
   } else {
   // It is. Step past it, carry on hoping for more digits 
   c = fgetc(file);
   while (c >= '0' && c <= '9') {
   n = 10.0 * n + c - '0';
   p = p * 10.0;
   c = fgetc(file);
   }
   // We've run out of digits but we have a number to give 
   if (c != EOF)
   ungetc(c, file);
   *valaddr = sign * n / p;
   return (1);
   }
   }
   
   // Use getdouble() above but convert result to int. 
   int getint(FILE * f, int *valaddr, int stopateol)
   {
   int r;
   double x;
   r = (int)getdouble(f, &x, stopateol);
   *valaddr = (int) x;
   return (r);
   }
   
   // Use getdouble() above but convert result to long. 
   long getlong(FILE * f, long *valaddr, int stopateol)
   {
   int r;
   double x;
   r = (long)getdouble(f, &x, stopateol);
   *valaddr = (long) x;
   return (r);
   }
*/

void create_vectors(double **wv, const int s, const int k)
{
  int i,j;
  int c=0;
  int a;
  int num, sum;
  int count[k];
  double *near;
  num = choose(s+k-1,k-1);
  //  fprintf(stdout, "num=%d\n",num);


  /*
  if((near = (double *)malloc(k * sizeof(double)))==NULL)
    {
      fprintf(stdout, "Malloc error. Exiting\n");
      exit(1);
    }
  */

  /*
  for(i=0;i<num;i++)
    {
      for(j=0;j<k;j++)
	{
	  fprintf(stdout, "%.2lf,", wv[i][j]);
	}
    }
  fprintf(stdout, "\n");
  */
  i=0;
  while(i<mypow(s+1,k))
    {
      sum=0;
     
      int2kary(i,s+1,k,count);
      for(j=0;j<k;j++)
	{
	  sum+=count[j];
	}
      if(sum==s)
	{
	  //	  fprintf(stdout, "%d=",c+1);
	  for(j=0;j<k;j++)
	    {
	      wv[c][j]=(double)count[j]/(double)s;
	      //	      fprintf(stdout, "%.2f,",wv[c][j]);	      
	    }
	  // fprintf(stdout, "\n");
	  /*  for(a=0;a<30;a++)
	    {
	      near_wv(wv[c], k, s, near);
	      // fprintf(stdout, "near %d=",a+1);
	      for(j=0;j<k;j++)
		{
		  fprintf(stdout, "%f ",near[j]);	      
		} 
	      fprintf(stdout, "\n");
	    }
	  */
	  c++;
	}
      
      i++;
    }
}

inline void int2kary(int x, const int basek, const int digits, int *kary)
{
  int i;
  int val;
  if (x>=mypow(basek,digits))
    {
      fprintf(stdout, "Number in int2kary() too large. Exiting.\n");
      exit(-1);
    }
  val=digits-1;

  for(i=0;i<digits;i++)
    kary[i]=0;

  i=0;
  while(x)
    {
      if(x>=mypow(basek,val))
	{
	  kary[i]+=1;
	  x-=mypow(basek,val);
	}
      else 
	{
	  val-=1;
	  i++;
	}
    }
}

inline int mypow(int x, int exp)
{
  int i;
  int var=1;
  for(i=0;i<exp;i++)
    var*=x;
  return(var);
}

int choose(const int n, const int r)
{
  return(factorialdiff(n,r)/factorial(r));
}

int factorialdiff(const int n, const int r)
{
  //this function calculates n!/(n-r)!
  int i;
  int result=1;
  for(i=n;i>(n-r);i--)
    {
      result*=i;
      if(result < 0) {
        fprintf(stderr,"too big number in Rmetrics.cc#factorialdiff!\n");
        exit(1);
      }	
    }
  return(result);
  
}

int factorial(const int n)
{
  if(n>11)
    {
      fprintf(stderr,"too big number to factorial!\n");
      exit(1);
    }
  int i;
  int result=1;
  for(i=n;i>1;i--)
    {
      result*=i;
    }
  return(result);
}

void print_help()
{
  fprintf(stdout, "No help-output available, because multi-line string literal are deprecated");
  /*
    ------
    To Compile:
    g++ Rmetrics.cc -o Rmetrics
    
    To Run:
    ./Rmetrics [-f <data_file>] [-s <s>] [-k <n_k>] [-i1 <ideal1>]...[-i5 <ideal5>] [-n1 <nadir1>]...[-n5 <nadir5>] [-r <rho>] [-m <method>] [-h]
    
    where all command line parameters are optional except for the data_file
    and they have the following meanings and conventions:
    
    <data_file> is any file of vectors with one vector per line and each component of
    the vector separated by a comma, a space, or a tab. IMPORTANT: this code assumes
    minimization so if you are maximizing, please multiply all numbers in your data-file
    by -1.
    
    <s> is an integer parameter which controls the number of different weight vectors 
    that are used by the R metrics. An appropriate value of s depends on the value
    of n_k, the number of objectives. The number of evenly distributed weight vectors
    is given by the equation |wv| = (s + n_k - 1)^C_(n_k - 1) where n^C_k is the \"choose\"
    function i.e. the number of ways of choosing n items from k. Default values of
    s are:
    
    n_k=2, s=500
    n_k=3, s=30
    n_k=4, s=12
    n_k=5, s=8
    
    <n_k> is the number of objectives i.e. the dimension of the vectors. At present a
    maximum of 5 for n_k is allowed. The default value is 2.
    
    <ideal1>..<ideal5> are the co-ordinates of an ideal point in the objective space.
    If no value is set for ideal1 then the ideal point is calculated from the data
    in the file. When comparing two sets of points, however, it is VERY IMPORTANT that
    the same ideal point is used. The ideal point must be better in all dimensions than
    the best value of any point in each dimension.
    
    <nadir1>..<nadir5> are the co-ordinates of a nadir (or worst) point in the objective space.
    If no value is set for nadir1 then the nadir point is calculated from the data
    in the file. When comparing two sets of points, however, it is VERY IMPORTANT that
    the same ideal point is used. The nadir point must be worse in all dimensions than
    the worst value of any point in each dimension.
    
    <rho> is used in the calculation of the augmented Tchebycheff function. It controls the 
    degree to which the Thebycheff function is augmented by a weighted sum function. Values
    in the range 0.001 to 0.1 are normal.
    
    <method> is an integer in 1..3 which defines whether the weighted sum (1), Thebycheff (2), or
    augmented Thebycheff function (3) is used in the calculation of the R1 and R2 metrics.
    
    -h causes this set of instructions to be printed.
    
    OUTPUT:
    -------
    
    The program should recognise the number of algorithms and runs of each algorithm
    that are in the data file. It should display this information. Then it calculates
    the mean value of the distance of the vectors in each data set from the ideal point
    and displays these numbers as follows:
    
    0.34456 ACO 1
    0.32234 ACO 2
    0.35674 PAES 1
    0.33232 PAES 2
    
    for example where the number is the mean distance and the algorithm and run are given,
    so here two algorithms, ACO and PAES, with two runs each must have been in the input
    data file.
    
    Following this, the R1, R2, and R3 metrics are calculated between pairs of 
    algorithms for each pair of matching numbered runs. In other words, if there
    are two runs of ACO and three of PAES, then there will be results for 
    ACO run 1 vs PAES run 1 and for ACO run 2 vs PAES run 2, and that is all.
    
    
    FURTHER INFORMATION:
    --------------------
    
    This code calculates the R1, R2 and R3 metrics of a set of points as described in
    Hansen and Jaszkiewicz 1998:
    
    @techreport{hansen98evaluating,
    author = \"Michael Pilegaard Hansen and Andrzej Jaszkiewicz\",
    title = \"Evaluating the quality of approximations to the non-dominated set\",
    number = \"IMM-REP-1998-7\",
    year = \"1998\",
    url = \"citeseer.nj.nec.com/hansen98evaluating.html\"}
    
    These metrics were reviewed in the PhD thesis of Joshua Knowles and also in 
    a paper by Knowles and Corne published in the 2002 Congress on 
    Evolutionary Computation. These can
    both be downloaded from http://iridia.ulb.ac.be/~jknowles/publications.html");
  */
}
