/*
 * Rmetrics.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.comparisons.rmetrics;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.Date;
import java.util.HashMap;

import kea.CompResult;
import kea.Datacontainer;
import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.comparisons.AbstractComparison;
import kea.comparisons.common.CompException;
import kea.comparisons.common.FilesToFile;
import kea.evaluations.common.GrepHeader;

/**
 * Compares given results of algorithms using an JNI-adapted version, which
 * now also allows double-values in decision vectors, of
 * <p>
 * <center>
 *     Rmetrics.cc (C) Joshua Knowles 4th April 2002.
 * </center>
 * <p>
 * This program calculates Hansen and Jaszkiewicz's R1, R2 and R3.
 * For further informations have a look at the included file Rmetrics.cc.
 * <p>
 *
 * Before you are able to execute this program make sure that your PATH or
 * LD_LIBRARY_PATH points to the kea.comparisons directory, and that
 * JNIRmetrics.dll or lib{Sun/Lin}JNIRmetrics.so is available in there, respective.
 * If the files aren't there use the following commands.  To compile c-code
 * under Windows try for example the \'cygwin\'-environment.
 * <p>
 * <ol>
 * <li> javah -jni kea.comparisons.rmetrics.Rmetrics
 * <li> g++ -c Rmetrics.cc                         (Windows),
 * <li> g++ -shared Rmetrics.o -o JNIRmetrics.dll  (Windows),<br>
 * respective <br>
 * <li> gcc -G -I/< ... >/jdk/< ... >/include/[solaris/,linux/]
 *          Rmetrics.o -o lib{Sun/Lin}JNIRmetrics.so        (Unix/Linux).
 * </ol>
 * 
 *
 * @author PG 419, University of Dortmund (dk)
 * @version $Revision: 723 $
 */   
public class Rmetrics extends AbstractComparison
{
  private static final String GZEI = Datacontainer.GZIP_END_IDENTIFIER; //"# A A A"
  private static final String KCI = Datacontainer.KEA_CLASS_IDENTIFIER; //">>>"
  private static final String HEI = Datacontainer.HEADER_END_IDENTIFIER;//"###"
  
  /* - Static initializer -
   * Ensures that the rmetrics-library is loaded for future use.
   * Therefore, the library path must be set properly before.
   * Otherwise it's impossible(?) to find the library.
   */

  private Rmetrics(String forStatic){}
  
  static
  {
    String rmPath = 
      Datacontainer.replaceAll
      ( new File
	( new Rmetrics("").getClass()
	  .getResource(".").getFile()).getAbsolutePath(), "%20", " ")
      + File.separator;

    String osName =  System.getProperty("os.name").trim().substring(0,3);
    String libHead = (osName.equals("Win")) ? "" : "lib" + osName;
    String libTail = (osName.equals("Win")) ? "dll" : "so";

   String libraryName = libHead + "JNIRmetrics." + libTail;

    try
      {
	System.load(rmPath + libraryName);
      }
    catch(UnsatisfiedLinkError e)
      {
	throw new CompException
	  //Kea.logger.logError
	  ( "\nW A R N I N G:\n"
	    + "Kea was unable to set the path to your rmetrics-dir automtically\n"
	    + "or a file is missing.  Rmetrics requires that your\n\n"
	    + "\tLD_LIBRARY_PATH (Unix/Linux), PATH (Windows)\n\n"
	    + "points to the kea.comparisons.rmetrics directory.  There should\n"
	    + "be your " + libraryName + ".  If\n"
	    + "not, read in the documentation of this file how to create it.\n"
	    //, e 
	    //, new Rmetrics("")
	    );
	//System.exit(1);
      }
  }
  
  /**
   * <code>rmParams</code> holds the parameters for the comparison.  It will
   * be used instead of <code>param</code> of AbstractComparison.<p>
   *
   * @see kea.AlgoState
   * @see kea.EvalResult
   * @see kea.Parameters
   */

  private Parameters[] returnParams;
  /* the following Rmetrics Parameters must be
   * appended to returnParams before returning!
   */
  private Parameters rmParams;


  /**
   * <code>Rmetrics()</code> does the startup initialization
   * of all parametes to their defaults.
   */
  
  public Rmetrics()
  {
    rmParams = new Parameters();
    
    rmParams.setMembership(this.getClass().getName());
    
    Parameter[] pa = new Parameter[17];
    pa[0] = new Parameter("Files","","file1 | file2 | ... _xor_ dir to compare");
    /* use prepareDataFile() to create a suitable file out of the given files
     * and decide whether to keep it after the comparison with "delTmpFile"
     */
    pa[1] = new Parameter("groupResults",false,"group results in direct " 
			  + "subdirs in one file each");
    pa[2] = new Parameter("delTmpFile", true,
			   "delete C-code-specific temp-file after comparison");
    /* More detailed descriptions for the next parameters in Rmetrics.cc */
    pa[3]  = new Parameter("s", 0,
			   "|{weight vectors}|=choose(s+n_k-1,n_k-1); def. 0");
    pa[3].setRange(1,500);
    pa[4]  = new Parameter("n_k", 2, "2 <= |objectives| <= 5");
    pa[4].setRange(2,5);
    pa[5]  = new Parameter("ideal1", -1.0,
			   "co-ordinates of an ideal point in obj-space");
    pa[6]  = new Parameter("ideal2", -1.0,
			   "***  \'-1\' forces RMetrics to calculate");
    pa[7]  = new Parameter("ideal3", -1.0,
			   "***  the co-ordinate for the respective");
    pa[8]  = new Parameter("ideal4", -1.0,
			   "***  direction automatically.");
    pa[9]  = new Parameter("ideal5", -1.0,"");
    pa[10]  = new Parameter("nadir1", -1.0,
			   "co-ordinates of a nadir point in obj-space");
    pa[11] = new Parameter("nadir2", -1.0, "");
    pa[12] = new Parameter("nadir3", -1.0,
			   "***  Don't use \'-1\' if you want comparable");
    pa[13] = new Parameter("nadir4", -1.0,
			   "***  results for several starts of RMetrics");
    pa[14] = new Parameter("nadir5", -1.0,"");
    pa[15] = new Parameter
      ( "method",1,
	"1=weighted sum, 2=Tchebycheff, 3=augmented Tch.");
    pa[15].setRange(1,3);
    pa[16] = new Parameter
      ( "rho", 0.01,
	"used for method 3; \\in[0.001,0.1]");

    rmParams.setParameterArray(pa);
  }
  
  /**
   * <code>nativeCompArr</code> allows to address the native
   * Rmetrics.cc-file, which must be in reach as
   * <ul>
   * <li>libJNIRmetrics.so-file (Linux) or
   * <li>JNIRmetrics.dll (Windows).
   * </ul>
   */

  private native String nativeCompArr(String[] argv);
  
  
  /**
   * <code>compare</code> starts the comparison with the given parameters
   * and files.  This is is done in three steps. 
   *  
   * In a <em>first step</em> <code>compare</code> does all preparations for
   * a comparison.  So it starts methods to get the parameters and
   * result-data out of the different files which shall be compared.
   * Those are combine in a new file the comparison itself works on.<br>
   * 
   * In a <em>second step</em> the native comparison metric <code>nativeCompArr</code>
   * is called which returns its results as String-object.<br>
   * 
   * <em>Finally</em>, all data received so far will be stored and returned in a
   * <code>CompResult</code>-object. 
   *
   *
   * @return <code>CompResult</code>, i.e. the result of a comparison.
   * @see kea.CompResult
   */
  
  
  public CompResult compare() //throws CompException
  {
    String compDataFile;
    compDataFile =
      prepareDataFile
      ( Datacontainer.split((String)rmParams.getParameterByName("Files").getData()
	      , "|")
	);

    /* life is easier with JDK >= 1.4; simply use the following lines */
    /* compDataFile =
       prepareDataFile(((String)rmParams.getParameterByName("Files").
       getData()).split("|"));
    */
    
    /* Remove the file containing the collected data, which should be
     * compared; if you want to have it, "delTmpFile" must be 'false'
     */

    File tmpDataFile =  new File(compDataFile);
    if(rmParams.getParameterByName("delTmpFile").getData().toString().equals("true"))
      tmpDataFile.deleteOnExit();
    
    
    /* Don't forget to append the Parameters of Rmetrics
     * at the end of Parameters[] returnParams.
     */
    Parameters[] tmpParams =
      new Parameters[returnParams.length+1];
    System.arraycopy(returnParams, 0,
		     tmpParams, 0, returnParams.length);
    tmpParams[returnParams.length] = rmParams;
    returnParams = tmpParams;


    /*   ---   Return the result of the comparison   ---   */
    CompResult rmCompResult = new CompResult();
    
    rmCompResult.setSetOfParameters(returnParams);
    
    /* Kea garanties that getfileName() returns a suitable name */
    rmCompResult.setExportFileName(getFileName());
    rmCompResult.setSaveFileName(getFileName());
     

    /* Maybe worth to throw an exception, too? */
    if (tmpDataFile.length()==0)
      {
	Kea.logger.logWarn
	  ("compare() couldn't find any data to compare.",this);
	rmCompResult.setCompResult(new String[] {"Nothing to compare"});
	return rmCompResult;
      }
    // maybe also reasonable, if only one result was found ...
    
    try{
      rmCompResult.setCompResult
	(new String[]
	    {nativeCompArr
	     (new String[]
		 { "31",  /*---   A r r a y   l e n g t h   ---*/
		   "-f",  compDataFile,
		   "-s",  rmParams.getParameterByName("s").getData().toString(),
		   "-k",  rmParams.getParameterByName("n_k").getData().toString(),
		   "-i1", rmParams.getParameterByName("ideal1").getData().toString(),
		   "-i2", rmParams.getParameterByName("ideal2").getData().toString(),
		   "-i3", rmParams.getParameterByName("ideal3").getData().toString(),
		   "-i4", rmParams.getParameterByName("ideal4").getData().toString(),
		   "-i5", rmParams.getParameterByName("ideal5").getData().toString(),
		   "-n1", rmParams.getParameterByName("nadir1").getData().toString(),
		   "-n2", rmParams.getParameterByName("nadir2").getData().toString(),
		   "-n3", rmParams.getParameterByName("nadir3").getData().toString(),
		   "-n4", rmParams.getParameterByName("nadir4").getData().toString(),
		   "-n5", rmParams.getParameterByName("nadir5").getData().toString(),
		   "-r",  rmParams.getParameterByName("rho").getData().toString(),
		   "-m",  rmParams.getParameterByName("method").getData().toString()
	       }
	    )     
	  }
       );
    }
    catch(Exception e)
      {
      	System.out.println("This is the error message: "+e);
      }
    return rmCompResult;
  }
  

  /*
   * almost no error-handling for the following at the moment
   */
  
  /**
   * <code>prepareDataFile</code> converts the evalResults and
   * AlgoStates, to which the parameter "Files" refers,
   * to an comparison-method-conform datafile-format.
   * <p>
   *  -- Subdirs will be either ignored or their result-files be
   * grouped in a single file for each subdir if parameter
   * "groupResults" is "true"; therefore make sure that only
   * results of the same type are in subdirs if "groupResults"
   * is "true".  --
   * <p>
   * Input-result-lines looking like<br>
   * <center>   < dec_vector > / < obj_vector >   </center>
   * will be converted to output-result-lines looking like
   * <center>   < obj_vector > < alg_name > < run_number >.  </center>
   * <p>
   * Therefore it tries to find the name of the algorithm in the
   * header of each file containing results and to differ algorithms
   * which use different parameters.  Algorithms with the same name
   * and parameters in different files count as different runs of an
   * algorithm and will lead to following lines in the created
   * file e.g.:
   * <ul>
   * <center> < obj-vector > < algo_i > < 1 > </center>
   * <center> < obj-vector > < algo_i > < 2 > </center>
   * <center>               ...               </center>
   * <center> < obj-vector > < algo_i > < n > </center>
   * </ul>
   * <p>
   * Furthermore, <code>prepareDataFile</code> prepares returnParams[]
   * so it contains the <code>Parameters</code> of the problem and 
   * all algorithms + corresponding evaluations when the method quits.
   * Only the Rmetrics-Parameteres are left to be appended.
   * <em>Warning:  Not implemented (yet)</em>:<br>
   * It would be helpful to warn if different problems, problem-
   * or evaluation-parameters were used, because, in this case,
   * a comparison often doesn't make sense.<br>
   * 
   * In case of problems needs to be checked:
   * <ul>
   *  + kea.problems. ... <br>
   *  + any exchangable variables of the problem <br>
   * </ul>
   * (In case of evaluations needs to be checked: <br>
   * <ul>
   *  + kea.evaluation. ... <br>
   *  + ...?) <br>
   * </ul>
   *
   * @param fileNames A <code>String[]</code> containing the fileNames
   *                  or the directory-name of the input-files.
   * @return A <code>String</code> with the name of the file the
   *         comparison-method works with.
   */
  protected String prepareDataFile(String[] fileNames) //throws CompException
  {
    String tmpPath = getFileName();
    tmpPath = (tmpPath.lastIndexOf(File.separator)==-1) ?
      "" : tmpPath.substring(0,tmpPath.lastIndexOf(File.separator)+1);
    String tempFileName= tmpPath + "rm" + new Date().getTime()+".dat";
    HashMap hashFileNames = new HashMap();
    returnParams = new Parameters[]{};
    boolean exportProbParams = true;
    String algoName;
    String line;
    
    if(fileNames.length==1)
      {
	File[] filePaths = {new File(fileNames[0])};
	if (filePaths[0].isDirectory()){
	  
	  if (rmParams.getParameterByName("groupResults").
	      getData().toString().equals("true"))
	    FilesToFile.groupResultsInSubDirsOf(fileNames[0]);
	  
	  filePaths = filePaths[0].listFiles();
	  fileNames = new String[filePaths.length];
	  for (int i=0; i < filePaths.length ;i++){
	    fileNames[i] = filePaths[i].getPath();
	  }
	}
      }

    /* sorts the fileNames-Array by name */    
    java.util.Arrays.sort(fileNames);
    
    try
      {
	PrintWriter outPrint =
	  new PrintWriter
	  ( new BufferedWriter
	    ( new FileWriter(tempFileName,true)));
	BufferedReader inRead;
	

	/* counts how many of the fileNames belong
	 * to files which can be compared
	 */
	int nrOfGoodFiles = 1;
	
	  for(int i=0; i<fileNames.length; i++){
	    
	  //  System.out.println("\n Entered iteration "+i+"/"+(fileNames.length-1));
	  //  System.out.println("fileName "+fileNames[i]);
	  /*
	   * ignore directories in fileNames[]
	   * (so it's non-recursive!)
	   */
	  if(new File(fileNames[i]).isDirectory()
	     /*
	      * If you comment out the next line missing
	      * files will cause a FileNotFoundException
	      *
	      */
	     //|| !(new File(fileNames[i])).exists()
	     )
	    continue;

	  inRead = new BufferedReader
	    (new FileReader(fileNames[i]));

	  if(Kea.logger.debugOn())
	    Kea.logger.logDebug("Try to interpret file "+fileNames[i]+" now.");
	  

	  /* tries to identify algonames ...*/

	  Parameters[] sourceParams = GrepHeader.makeParams(fileNames[i]);
	  algoName = "unknown";

	  if(sourceParams != null)
	    {
	      if (exportProbParams)
		{
		  /* Let the problem-parameters in the Parameters[], but
		   * change place with the algo-parameters (cmp. to order
		   * given in AlgoState for save(...) and export(...)).
		   * (By now, returnParams is an "empty" array, because
		   * it seems to be better to append the Rmetrics-Parameters
		   * later.)
		   */
		  exportProbParams = false;
		  Parameters[] tmpParams =
		    new Parameters[sourceParams.length];
		  System.arraycopy(sourceParams, 1,
				   tmpParams, 0, 1);
		  System.arraycopy(sourceParams, 0,
				   tmpParams, 1, 1);
		  System.arraycopy(sourceParams, 2,
				   tmpParams, 2, sourceParams.length-2);
		  returnParams = tmpParams;
		}
	      else
		/* Since the problem should be the same for all algorithms,
		 * don't take its parameters again
		 */
		{
		  Parameters[] tmpParams =
		    new Parameters[returnParams.length+sourceParams.length-1];
		  System.arraycopy(returnParams, 0,
				   tmpParams, 0, returnParams.length);
		  System.arraycopy(sourceParams, 0,
				   tmpParams, returnParams.length, 1);
		  System.arraycopy(sourceParams, 2,
				   tmpParams, returnParams.length+1, sourceParams.length-2);
		  returnParams = tmpParams;
		}
	      
	      /* find out the name of the algorithm, to identify every objectiv-vector
	       * in the c-specific-tempfile at the end of the corresponding line.
	       */

	      for (int parIdx=0; parIdx<sourceParams.length; parIdx++)
		{
		  if ((algoName = sourceParams[parIdx].getMembership())
		      .indexOf("kea.algo")!=-1)
		    {
		      algoName = algoName.substring
			( algoName.lastIndexOf('.')+1,algoName.length()).trim();
		      if(Kea.logger.debugOn())
			Kea.logger.logDebug("Name: "+algoName);		
		      break;
		    }
		}

	    }

	  /* otherwise it isn't very likely to find the HEADER_END_IDENTIFIER,
	   * (see below) which marks the start of the results. Therefore:
	   */
	  else
	    {
	      Kea.logger.logWarn
		("Wrong file-format in " + fileNames[i]
		 + ".  No header found.  File will be ignored.", this);
	      continue;
	    }

	  
	  
	  /* Here the above described conversion takes place.
	   *
	   * Tries to find out the number of a run corresponding to 
	   * it's name ..., as far as this is possible.  Different
	   * .ers with the same algoname momentarily count as 
	   * different runs of an algorithm;
	   * 
	   * T O D O:
	   *  - take another criterium for this; best would be to
	   *    take the parameters of the algorithms and use them
	   *    to decide if algorithms are different
	   *   
	   * (Of course, it would be easier to just print out all
	   *  parameters and let decide the users themselves ;))
	   *  
	   *  -> this should be possible with new GZIP-Header
	   * Combine it in a class FindComparables, because this will
	   * be useful for (almost) every comparison and use a boolean
	   * to decide whether to use it or not ...  
	   */
	  
	  /* Leave out header while scanning the file
	   */

	  line = inRead.readLine();
	  while(!line.startsWith(HEI))
	    {
	      line = inRead.readLine();
	    }

	  if(Kea.logger.debugOn())
	    Kea.logger.logDebug
	      ("End EvalResult-Header; read results",this);
	  // Write further lines of the file which should contain the
	  // data to compare, written in Rmetrics-format
	  
	  //outPrint.println(fileNames[i]);

	  Integer nameInt = (Integer)hashFileNames.get(algoName);
	  int runNumber = 1;
	  boolean resultsEmpty = true;

	  line = inRead.readLine();
	  while(line!=null)
	    {
	      /* Find the index of the separator between
	       * decision- and objective-vector
	       */
	      int ind = line.indexOf("/");
	      if(Kea.logger.debugOn())
		Kea.logger.logDebug(line,this);

	      /* Ignore comments (1) and broken lines (2, 3)*/
	      if(line.startsWith("#")
		 || line.indexOf("NaN")!=-1
		 || ind == -1)
		{
		  line = inRead.readLine();
		  continue;
		}

	      if(resultsEmpty == true)
		{
		  /* This enables the Rmetrics to identify
		   * several runs of one algorithm and increase
		   * the corresponding number in the c-specific
		   * temp-file.
		   * (For the moment it has no effect, because of
		   * the "Out of time solution", which ensures that
		   * that algonames exists only once in the HashMap.) 
		   */
		  if (nameInt!=null){ 
		    runNumber = nameInt.intValue()+1;
		  }
		  if (Kea.logger.debugOn())
		    Kea.logger.logDebug("runNumber: "+runNumber, this);
		  
		  
		  /* O u t   o f   t i m e   s o l u t i o n:
		   * realizes a compare each file with each other ...
		   * _Attention:_ this only works for max 12 result files
		   * according to the restrictions in Rmetrics.cc--MAX_ALGS
		   *
		   * (If you don't want to use this you've to make sure
		   * that not only the name of an algorithm but all his
		   * parameters are the same.  Otherwise you can't be sure
		   * that you really have several runs of _one_ algorithm
		   * and increasing the runNumer (see above) is correct.)
		   */
		  /*---------------------*/
		  algoName = algoName + "." + (nrOfGoodFiles++);
		  /*---------------------*/
		  
		  hashFileNames.put(new String(algoName),new Integer(runNumber));
		  resultsEmpty = false;
		}

	      outPrint.println
		( line.substring(ind + 1,line.length()).trim()
		  + "\t" + algoName + " " + runNumber);
	      
	      line = inRead.readLine();
	    }
	  outPrint.flush();
	  // System.out.println("Finished reading file " + fileNames[i]);
	}
	outPrint.close();
      }

    /* search for FileNotFoundException to find the line,
     * which could prevent them to occur
     */
    catch(FileNotFoundException e)
      {
	throw new CompException
	  ("\nonly file-names   or   the name of a directory, which contains files\n"
	   + "(handling of subdirs depends on \'GroupResults\'-parameter),\n"
	   + "are allowed as input\n");
      }
    catch(IOException e)
      {
	throw new CompException
	  ("IOError occured while working with " + tempFileName +"\n");
      }
    return tempFileName;
  }
  
  /**
   * <code>getParameter</code> returns the <code>Parameters</code>-object of this class.
   *
   * @return <code>Parameters</code> of Rmetrics
   */
  public Parameters getParameter()
  {
    return this.rmParams;
  }
  
  /**
   * <code>setParameter</code> sets the Parameters-object of this class.
   *
   * @param parameter <code>Parameters</code> of Rmetrics
   */
  public void setParameter(Parameters parameter)
  {
    this.rmParams=parameter;
  }
  
  
  /*
   * <code>main</code> only for testing this class
   */
  
   public static void main(String[] argv)
     {
       Parameter newPa = 
	 new Parameter("Files","toCompare",//"rm_exaCh.dat,rm_exbCh.dat",
		       "data-file-name(s; ,-separated)");
       Rmetrics nrm = new Rmetrics();
       nrm.rmParams.setParameter(newPa);
       nrm.compare().export();
     }
  
}
