/*
 * FilterEvaluationCount.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.evaluations;

import java.io.File;
import java.util.Vector;

import kea.EvalResult;
import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.evaluations.common.DataChooser;
import kea.evaluations.common.EvalException;
import kea.evaluations.common.FileName;
import kea.evaluations.common.GenAfterEvalCount;
import kea.evaluations.common.GrepHeader;
import kea.problems.AbstractProblem;

/**
 * It is used to reduce Data of .ast Files. The user can choose a generation by
 * giving a number of evaluations that at least must have been made (the first
 * generation that has at least that much evaluations is chosen), decision variables and objective
 * functions he wants to copy in a new reduced .ers file.
 *
 * @author Tim Hohm
 * @version $Revision: 723 $
 */
public class FilterEvaluationCount extends AbstractEvaluation {

  /**
   * The Parameter file for this evaluation.
   */
  private Parameters param;
  private Filter chooser;

  /**
   * Name of the evaluation method.
   */
  private String method;

  /**
   * Name of the file that is getting evaluated.
   */
  private String run;

  /**
   * If run contains more than one .ast name in each vector entry there is one of them.
   */
  private Vector runs;

  /**
   * The generations that will get sorted out from the <code>run</code> file.
   */
  private String generations;

  /**
   * The decision variables which are to be chosen.
   */
  private String decVar;

  /**
   * The objective variables which are to be chosen.
   */
  private String objVar;

  /**
   * Name of the resulting .ers file.
   */
  private String name;

  /**
   * The <code>EvalResult</code> that is to be returned by the evaluate method.
   */
  private EvalResult result;

  /**
   * The min value of evaluations that have to be made to choose the generation.
   */
  private String amountOfEvals;

  /**
   * All adjustable or automatically setted paramters are made.
   */
  public FilterEvaluationCount() {
    this.setFileName("");
    name = "";
    param = new Parameters();
    param.setMembership(this.getClass().getName());
    param.addParameter(new Parameter("Files", "", "The runs you want to evaluate. Divide multiple runs by '|'."));
    param.addParameter(new Parameter("Generations", "", "Is filled out automatically."));
    param.addParameter(new Parameter("Decision Variables", "", "Decision variables you want to keep. Divide variables by ' ' or ','. An Empty entry means that you choose all variables."));
    param.addParameter(new Parameter("Objective Variables", "", "Objective variables you want to keep. Divide variables by ' ' or ','. An Empty entry means that you choose all variables."));
    param.addParameter(new Parameter("DimDecSpace", "", "Is filled out automatically."));
    param.addParameter(new Parameter("DimObjSpace", "", "Is filled out automatically."));
    param.addParameter(new Parameter("EvaluationCount", "1", "Number of evaluations that at least must have been made. Only the first matching generation is chosen."));
  }

  /**
   * At first the generation is found that has at least the right number of evaluations.
   * Afterwards this generation is chosen and the corresponding <code>EvalResult</code> is made.
   *
   * @return The resulting <code>EvalResult[]</code> containing the <code>EvalResult</code>s.
   */
  public EvalResult[] evaluate() {

    // If there is no name or an invalid name in the parameter file a new name is made.
    runs = DataChooser.resolveStringMid(run);
    File denoter = new File(this.getFileName());
    String path = denoter.getAbsolutePath().substring(0, denoter.getAbsolutePath().indexOf(denoter.getName()));
    if (!this.getFileName().equals("")){
      name = this.getFileName();
    }
    if (name.equals("")){
      for (int i = 0; i < runs.size(); i++){
        File file = new File((String)runs.elementAt(i));
        String it;
        it = path + FileName.produceName("Filter") + "$" + file.getName().substring(0, file.getName().length()-4);
        name = (i==0) ? it :  name + "|" + it;
      }
      Kea.logger.logWarn("User given .ers name invalid or not available. It gets overwritten by System with " + name+".", this);
    }
    Vector names = DataChooser.resolveStringMid(name);
    if (names.size()!=runs.size()){
      name = "";
      for (int i = 0; i < runs.size(); i++){
        File file = new File((String)runs.elementAt(i));
        String it;
        it = path + FileName.produceName("Filter") + "$" + file.getName().substring(0, file.getName().length()-4);
        name = (i==0) ? it : name + "|" + it;
      }
      Kea.logger.logWarn("User given .ers name invalid or not available. It gets overwritten by System with " + name+".", this);
    }
    names = DataChooser.resolveStringMid(name);

    EvalResult[] returnArray = new EvalResult[runs.size()];
    for (int i = 0; i < runs.size(); i++){
      // With help of <code>GenAfterEvalCount</code> the right generation is found and the Generations parameter
      // Generations parameter gets set.
      try{
        Integer.parseInt(amountOfEvals);
      }
      catch(NumberFormatException lala){
        param.setParameter(new Parameter("Generations", "0", "Generation chosen."));
        throw new EvalException("Parameter EvaluationCount corrupted! No evaluation possible!");
      }

      GenAfterEvalCount sortOut = new GenAfterEvalCount((String)runs.elementAt(i));
      generations = sortOut.makeOutGen(amountOfEvals);
      param.setParameter(new Parameter("Generations", generations, "Generation chosen."));

      // An instance of <code>Filter</code> is made which seperates the right generation.
      chooser = new Filter();
      Parameters filterParam = chooser.getParameter();
      filterParam.setParameter(new Parameter("Files", (String)runs.elementAt(i), ""));
      filterParam.setParameter(new Parameter("Generations", generations, ""));
      filterParam.setParameter(new Parameter("Decision Variables", decVar, ""));
      filterParam.setParameter(new Parameter("Objective Variables", objVar, ""));
      chooser.setParameter(filterParam);
      EvalResult[] results = chooser.evaluate();
      EvalResult result = results[0];

      // With the returned Data[] the <code>EvalResult</code> is instanciated and returned.
      param.setParameter(new Parameter("DimDecSpace", chooser.getNewDimDec(), "Dimension of decision space."));
      param.setParameter(new Parameter("DimObjSpace", chooser.getNewDimObj(), "Dimension of objective space."));

      result.setAstParameters(GrepHeader.makeParams((String)runs.elementAt(i)));
      result.setEvalParameter(param);
	  AbstractProblem c = null; 
	  result.setIndividuum(c.IndividuumRepair(chooser.getIndividuum()));
      
      //result.setIndividuum(chooser.getIndividuum());
      result.setExportFileName((String)names.elementAt(i));
      returnArray[i] = result;
    }
    return returnArray;
  }

  public void setParameter(Parameters parameter) {
    run = (String)(param.getParameterByName("Files")).getData();
    decVar = (String)(param.getParameterByName("Decision Variables")).getData();
    objVar = (String)(param.getParameterByName("Objective Variables")).getData();
    amountOfEvals = (String)(param.getParameterByName("EvaluationCount")).getData();
  }

  public Parameters getParameter() {
    return param;
  }

}
