/*
 * FilterIterationCount.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.evaluations;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

import kea.Data;
import kea.EvalResult;
import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.evaluations.common.DataChooser;
import kea.evaluations.common.EvalException;
import kea.evaluations.common.FileName;
import kea.evaluations.common.GrepHeader;
import kea.problems.AbstractProblem;

/**
 * It is used to reduce Data of .ast Files. The user can choose a generation by
 * giving a number of iterations that must have been made (the first
 * generation that has that much iterations is chosen), decision variables and objective
 * functions he wants to copy in a new reduced .ers file.
 *
 * @author Tim Hohm
 * @version $Revision: 723 $
 */
public class FilterIterationCount extends AbstractEvaluation {

  /**
   * The Parameter file for this evaluation.
   */
  private Parameters param;
  private DataChooser chooser;

  /**
   * Name of the evaluation method.
   */
  private String method;

  /**
   * Name of the file that is getting evaluated.
   */
  private String run;

  /**
   * If run contains more than one .ast name in each vector entry there is one of them.
   */
  private Vector runs;

  /**
   * The generations that will get sorted out from the <code>run</code> file.
   */
  private String generations;

  /**
   * The decision variables which are to be chosen.
   */
  private String decVar;

  /**
   * The objective variables which are to be chosen.
   */
  private String objVar;

  /**
   * Name of the resulting .ers file.
   */
  private String name;

  /**
   * Data[] that is needed to instanciate the <code>EvalResult</code>.
   */
  private Data[] results;

  /**
   * New dimension of the decision variable space.
   */
  private String newDimDec;

  /**
   * New dimension of the objective variable space.
   */
  private String newDimObj;

  /**
   * This is needed at the <code>EvalResult</code> to know wheter a <code>Data</code> entry is an
   * individuum or an empty line or a status line (iterationCount line or evaluationCount line).
   */
  private Vector individuum;

  /**
   * The Iiteration that is to be choose.
   */
  private String iteration;

  /**
   * All adjustable or automatically setted paramters are made.
   */
  public FilterIterationCount() {
    this.setFileName("");
    name = "";
    param = new Parameters();
    param.setMembership(this.getClass().getName());
    param.addParameter(new Parameter("Files", "", "The runs you want to evaluate. Divide multiple runs by '|'."));
    param.addParameter(new Parameter("Generations", "", "Is filled out automatically."));
    param.addParameter(new Parameter("Decision Variables", "", "Decision variables you want to keep. Divide variables by ' ' or ','. An Empty entry means that you choose all variables."));
    param.addParameter(new Parameter("Objective Variables", "", "Objective variables you want to keep. Divide variables by ' ' or ','. An Empty entry means that you choose all variables."));
    param.addParameter(new Parameter("DimDecSpace", "", "Is filled out automatically."));
    param.addParameter(new Parameter("DimObjSpace", "", "Is filled out automatically."));
    param.addParameter(new Parameter("IterationCount", "1", "Number of evaluations that at least must have been made. Only the first matching generation is chosen."));
  }

  /**
   * At first the generation is found that has at least the right number of iterations.
   * Afterwards this generation is chosen and the corresponding <code>EvalResult</code> is made.
   *
   * @return The resulting <code>EvalResult[]</code> containing the <code>EvalResult</code>s.
   */
  public EvalResult[] evaluate() {

    // If there is no name or an invalid name in the parameter file a new name is made.
    runs = DataChooser.resolveStringMid(run);
    File denoter = new File(this.getFileName());
    String path = denoter.getAbsolutePath().substring(0, denoter.getAbsolutePath().indexOf(denoter.getName()));
    if (!this.getFileName().equals("")){
      name = this.getFileName();
    }
    if (name.equals("")){
      for (int i = 0; i < runs.size(); i++){
        File file = new File((String)runs.elementAt(i));
        String it;
        it = path + FileName.produceName("Filter") + "$" + file.getName().substring(0, file.getName().length()-4);
        name = (i==0) ? it : name + "|" + it;
      }
      Kea.logger.logWarn("User given .ers name invalid or not available. It gets overwritten by System with " + name+".", this);
    }
    Vector names = DataChooser.resolveStringMid(name);
    if (names.size()!=runs.size()){
      name = "";
      for (int i = 0; i < runs.size(); i++){
        File file = new File((String)runs.elementAt(i));
        String it;
        it = path + FileName.produceName("Filter") + "$" + file.getName().substring(0, file.getName().length()-4);
        name = (i==0) ? it : name + "|" + it;
      }
      Kea.logger.logWarn("User given .ers name invalid or not available. It gets overwritten by System with " + name+".", this);
    }
    names = DataChooser.resolveStringMid(name);

    EvalResult[] returnArray = new EvalResult[runs.size()];
    for (int i = 0; i < runs.size(); i++){
      this.findGeneration((String)runs.elementAt(i));
      chooser = new DataChooser((String)runs.elementAt(i));
      results = chooser.choose(generations, decVar, objVar);

      newDimDec = chooser.getNewDimDec();
      newDimObj = chooser.getNewDimObj();
      param.setParameter(new Parameter("DimDecSpace", newDimDec, "Dimension of decision space."));
      param.setParameter(new Parameter("DimObjSpace", newDimObj, "Dimension of objective space."));

      EvalResult result = new EvalResult();
      result.setAstParameters(GrepHeader.makeParams((String)runs.elementAt(i)));
//      individuum = chooser.getIndividuum();
//      result.setIndividuum(individuum);
		AbstractProblem c = null; 
	    result.setIndividuum(c.IndividuumRepair(chooser.getIndividuum()));
      
      result.setEvalParameter(param);
      result.setEvalResult(results);
      result.setExportFileName((String)names.elementAt(i));
      returnArray[i] = result;
    }
    return returnArray;
  }

  public void setParameter(Parameters parameter) {
    run = (String)(param.getParameterByName("Files")).getData();
    decVar = (String)(param.getParameterByName("Decision Variables")).getData();
    objVar = (String)(param.getParameterByName("Objective Variables")).getData();
    iteration = (String)(param.getParameterByName("IterationCount")).getData();
  }

  public Parameters getParameter() {
    return param;
  }

  /**
   * Sorts out the generation with the right iterationCount. If there is no right
   * generation it sets the String to "0".
   */
  private void  findGeneration(String theName){
    int genCount = 0;
    boolean stop = false;
    String curLine = "";
    try{
      BufferedReader in = new BufferedReader(new FileReader(theName));
      // Skip Header.
      while(in.ready()) {
        curLine = in.readLine();
        if (curLine != null){
          // Search for end of header.
          if (curLine.indexOf("###") != -1){
            if (Kea.logger.debugOn()) {
              Kea.logger.logDebug("Ende Header!", this);
            }
            break;
          }
        }
      }
      while (in.ready()){
        if (stop){
          break;
        }
        curLine = in.readLine();
        if (curLine != null) {
          // Skip empty lines.
          while (curLine.equals("")) {
            curLine = in.readLine();
            if (curLine == null)break;
          }
          // All empty lines up to the first individuum or end of file skipped.
          if (Kea.logger.debugOn()) {
            Kea.logger.logDebug("First not empty line = " + curLine, this);
          }
          // Increase generationcounter.
          genCount ++;

          while (!curLine.equals("")) {
            // If the current line is the iterationCount line ...
            if (curLine.indexOf("iterationCount") != -1){
              curLine = curLine.substring(curLine.indexOf("=") +2);
              int f = Integer.parseInt(curLine);
              if (f == Integer.parseInt(iteration)){
                stop = true;
                break;
              }
            }
            curLine = in.readLine();
            if (curLine == null)break;
          }
          if (stop){
            break;
          }
        }
      }
    }
    catch(FileNotFoundException exc){
      throw new EvalException("File "+theName+" not found. No operation on this file possible.");
    }
    catch (IOException k){
      throw new EvalException("An IO-Error occured while reading run data. Please try again.");
    }
    if (stop){
      generations = Integer.toString(genCount);
    }else{
      generations = "0";
      throw new EvalException("There is no generation with iterationCount "+iteration+" in the file "+run+".");
    }
  }
}
