/*
 * DataChooser.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.evaluations.common;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.Vector;

import kea.Data;
import kea.Datacontainer;
import kea.Kea;
import kea.Value;

/**
 * @author Tim Hohm
 * @version $Revision: 723 $
 */
public class DataChooser {

  private String run;
  private int dimDec;
  private int dimObj;
  private int newDimDec;
  private int newDimObj;
  private Vector generationSet;
  private int individuumCounter = 0;
  private Vector individuum = new Vector();
  private Data[] result;

  public DataChooser(String run) {
    this.run = run;
    generationSet = new Vector();
  }

  /**
   *
   * @param three strings which contain information about which generations, objectives and decisionvariables should be chosen.
   */
  public Data[] choose(String generations, String decVar, String objVar){
    if (Kea.logger.debugOn()) {
      Kea.logger.logDebug("Params von choose gens/dec/obj = " + generations + " ; " + decVar + " ; " + objVar, this);
    }
    scanHeader();
    if (Kea.logger.debugOn()) {
      Kea.logger.logDebug("Header gescannt!", this);
    }
    Vector g = resolveString(generations);
    g = CheckParams.checkGens(g, this);
    if (Kea.logger.debugOn()) {
      Kea.logger.logDebug("Generations: " + g.toString(), this);
    }
    chooseGenerations(g);
    if (Kea.logger.debugOn()) {
      Kea.logger.logDebug("Generations chosen!", this);
    }
    Vector d = resolveString(decVar);
    Vector o = resolveString(objVar);
    d = CheckParams.checkVars(d, this, dimDec);
    o = CheckParams.checkVars(o, this, dimObj);
    if (Kea.logger.debugOn()) {
      Kea.logger.logDebug("Decs/Objs" + d.toString() + " / " + o.toString(), this);
    }
    chooseVar(d, o);
    if (Kea.logger.debugOn()) {
	Kea.logger.logDebug("Variables and Objectives chosen!", this);
    }
    // The decision variable vector is found at index 0, the objective vector at index 1.
    return (result);
  }

  /**
   * This method scans the header for the occurence of the DimObjSpace and DimDecSpace parameter
   * and fills the initialises the coresponding dimDec and dimObj variable of this class.
   */
  private void scanHeader(){
    //Read file.
    try {
      BufferedReader in = new BufferedReader(new FileReader(run));
      while(in.ready()) {
        String curLine = in.readLine();
        if (curLine != null) {
          // Search for the DimDecSpace variable.
          if (curLine.indexOf("DimDecSpace") != -1){
            if (curLine.indexOf("#") != -1){
              curLine=curLine.substring(curLine.indexOf("=")+1, curLine.indexOf("#"));
            }else{
              curLine=curLine.substring(curLine.indexOf("=")+1);
            }
            curLine = curLine.trim();
            dimDec = Integer.parseInt(curLine);
          }
          // Search for the DimObjSpace variable.
          if (curLine.indexOf("DimObjSpace") != -1){
            if (curLine.indexOf("#") != -1){
              curLine=curLine.substring(curLine.indexOf("=")+1, curLine.indexOf("#"));
            }else{
              curLine=curLine.substring(curLine.indexOf("=")+1);
            }
            curLine = curLine.trim();
            dimObj = Integer.parseInt(curLine);
          }
          // If the header is finished stop the search routine without searching the run data.
          if (curLine.indexOf("###") != -1)break;
        }
      }
      if (Kea.logger.debugOn()) {
	Kea.logger.logDebug("DimensionDecSpace = " + dimDec, this);
        Kea.logger.logDebug("DimensionObjSpace = " + dimObj, this);
      }
    }catch (FileNotFoundException exc) {
      throw new EvalException("File "+run+" not found. No operation on this file possible.");
    }
    catch (IOException k){
      throw new EvalException("An IO-Error occured while reading run data. Please try again.");
    }
  }

  /**
   * This method wraps the results of
   * Datacontainer.split(String,String) in a vector.
   * <b>Attention:</b> It expects a "|" as separator.
   *
   * @param what <code>String</code> which should get converted to an vector.
   */
  public static Vector resolveStringMid(String what){
    String[] tmpStrings = Datacontainer.split(what, "|");
    Vector indices = new Vector();
    indices.clear();
    for(int i = 0; i < tmpStrings.length; i++)
      {
	indices.addElement(tmpStrings[i]);
      }
    return indices;
  }

  /**
   * This method wraps the results of
   * Datacontainer.split(String,String) in a vector.
   * <b>Attention:</b> It expects a comma, a whitespace or a "|" as separator.
   *
   * @param what <code>String</code> which should get converted to an vector.
   */
  public static Vector resolveString(String what){
    what = Datacontainer.replaceAll(what, " ", "|");
    what = Datacontainer.replaceAll(what, ",", "|");
    return resolveStringMid(what);
  }





  /**
   * This Method generates a vector of individuums which build the generations the user wanted to filter.
   *
   * @params generation vector created by the resolve method.
   */
  private void chooseGenerations(Vector vec){
    if (!vec.isEmpty()){
      // If the generation to get chosen is the last use getNumberOfLastGeneration() to get
      // total number of generations and add the that number to vec.
      if (((String)vec.firstElement()).equals("last")){
        int last = getNumberOfLastGeneration();
        if (last != 0){
          vec.addElement(Integer.toString(last));
        }
        vec.remove("last");
      }
    }
    if (!vec.isEmpty()){
      int i = 0;
      int genCount = 0;
      int whichGen;
      String curLine;
      try {
        BufferedReader in = new BufferedReader(new FileReader(run));
        // Skip Header.
        while(in.ready()) {
          curLine = in.readLine();
          if (curLine != null){
            // Search for end of header.
            if (curLine.indexOf("###") != -1){
              if (Kea.logger.debugOn()) {
                Kea.logger.logDebug("Ende Header!", this);
              }
              break;
            }
          }
        }
        while (in.ready()){
          curLine = in.readLine();
          if (curLine != null) {
            // Skip empty lines.
            while (curLine.equals("")) {
              curLine = in.readLine();
              if (curLine == null)break;
            }
            // All empty lines up to the first individuum or end of file skipped.
            if (Kea.logger.debugOn()) {
              Kea.logger.logDebug("First not empty line = " + curLine, this);
            }
            // Increase generationcounter.
            genCount ++;
            // If there are no more generations to be added, an Exception is thrown while attempting
            // to read the next element in the vector ...
            try{
              whichGen = Integer.parseInt((String)vec.elementAt(i));
            }
            // ... then this whole operation gets canceled.
            catch (ArrayIndexOutOfBoundsException noMoreGensToBeAdded){
              break;
            }
            // If the current generation is to be chosen ...
            if (genCount == whichGen){
              // ... and it is not the first generation, fill in an empty
              // line to seperate the generations ...
              if (genCount > 1) {
                generationSet.addElement("");
              }
              // ... and add the individuums.
              while (!curLine.equals("")) {
                // If the current line is not the iterationCount line, runtime line and not the evaluationCount line ...
                if ((curLine.indexOf("iterationCount") == -1) && (curLine.indexOf("evaluationCount") == -1) && (curLine.indexOf("runtime") == -1)){
                  // ... remove '/' charcter between objective- and decisionspace
                  // and increase individuumCounter.
                  individuumCounter++;
                  curLine = curLine.replace('/', ' ');
                }
                generationSet.addElement(curLine);
                if (Kea.logger.debugOn()) {
                  Kea.logger.logDebug("Line added to generationSet = " + curLine, this);
                }
                curLine = in.readLine();
                if (curLine == null)break;
              }
              i++;
            }else{
              // Skip not used generation.
              while (!curLine.equals("")) {
                curLine = in.readLine();
                if (curLine == null)break;
              }
            }
            if (curLine == null)break;
          }
        }
      }
      catch (FileNotFoundException exc) {
        throw new EvalException("File "+run+" not found. No operation on this file possible.");
      }
      catch (IOException k){
        k.printStackTrace();
      }
    }
  }

  private void chooseVar(Vector d, Vector o){
    Vector tempDV = new Vector();
    Vector tempOV = new Vector();
    Vector variablesDV = new Vector();
    Vector variablesOV = new Vector();
    String curLine;
    while (generationSet.size() > 0){
      // Get the first element and remove it from the vector.
      curLine = (String)generationSet.firstElement();
      generationSet.remove(0);
      // If the current line is the evaluationCount line: fill it in and start over with the next line.
      if (curLine.indexOf("evaluationCount") != -1){
        String it = curLine.trim();
        Value[] evalCount = new Value[1];
        evalCount[0] = new Value(it);
        tempDV.addElement(evalCount);
        tempOV.addElement(evalCount);
        individuumCounter++;
        individuum.addElement(new Integer(1));
      }
      // If the current line is the iterationCount line: fill it in and start over with the next line.
      if (curLine.indexOf("iterationCount") != -1){
        String it = curLine.trim();
        Value[] iterCount = new Value[1];
        iterCount[0] = new Value(it);
        tempDV.addElement(iterCount);
        tempOV.addElement(iterCount);
        individuumCounter++;
        individuum.addElement(new Integer(1));
      }
      // If the current line is the runtime line: fill it in and start over with the next line.
      if (curLine.indexOf("runtime") != -1){
        String it = curLine.trim();
        Value[] runtime = new Value[1];
        runtime[0] = new Value(it);
        tempDV.addElement(runtime);
        tempOV.addElement(runtime);
        individuumCounter++;
        individuum.addElement(new Integer(1));
      }
      if ((curLine.indexOf("iterationCount") == -1) && (curLine.indexOf("evaluationCount") == -1) && (curLine.indexOf("runtime") == -1)) {
        // If current generationSet element was empty fill in an empty line in the
        // temp vector. This garantees that generations still are seperated by an empty line ...
        if (curLine.equals("")){
          String it = "empty";
          Value[] nada = new Value[1];
          nada[0] = new Value(it);
          tempDV.addElement(nada);
          tempOV.addElement(nada);
          individuumCounter++;
          individuum.addElement(new Integer(2));
        }else{
          // ... otherwise build the variables vector.
          variablesDV.clear();
          variablesOV.clear();
          int counter = 0;
          while (!curLine.equals("")){
            curLine = curLine.trim();
            // For objectives and decisionvariables there are different variable vectors.
            if (counter < dimDec){
              if (curLine.indexOf(" ") != -1){
                String it = curLine.substring(0, curLine.indexOf(" "));
                variablesDV.addElement(it);
                counter++;
                curLine = curLine.substring(curLine.indexOf(" ")+1);
              }else{
                variablesDV.addElement(curLine);
                counter++;
                break;
              }
            }else{
              if (curLine.indexOf(" ") != -1){
                String it = curLine.substring(0, curLine.indexOf(" "));
                variablesOV.addElement(it);
                counter++;
                curLine = curLine.substring(curLine.indexOf(" ")+1);
              }else{
                variablesOV.addElement(curLine);
                counter++;
                break;
              }
            }
          }
          // Add chosen design variables.
          Value[] entryDV = new Value[d.size()];
          for (int j=0; j < d.size(); j++){
            String element = (String)d.elementAt(j);
            int index = Integer.parseInt(element) - 1;
            try {
              Double dou = new Double((String)variablesDV.elementAt(index));
              Value it = new Value(dou.doubleValue());
              entryDV[j] = it;
            }catch(NumberFormatException didi){
              try{
                boolean bo = (Boolean.valueOf((String)variablesDV.elementAt(index))).booleanValue();
                Value it = new Value(bo);
                entryDV[j] = it;
              }catch(NumberFormatException dada){
                throw new EvalException("The decision variables are non doubles and non booleans. Evaluation aborted.");
              }
            }
          }
          // Add chosen objectives.
          Value[] entryOV = new Value[o.size()];
          for (int j=0; j < o.size(); j++){
            String element = (String)o.elementAt(j);
            int index = Integer.parseInt(element) - 1;
            try {
              Double dou = new Double((String)variablesOV.elementAt(index));
              Value it = new Value(dou.doubleValue());
              entryOV[j] = it;
            }catch(NumberFormatException didi){
                try{
                boolean bo = (Boolean.valueOf((String)variablesDV.elementAt(index))).booleanValue();
                Value it = new Value(bo);
                entryOV[j] = it;
              }catch(NumberFormatException dada){
                throw new EvalException("The decision objective are non doubles and non booleans. Evaluation aborted.");
              }
            }
          }
          individuum.addElement(new Integer(0));
          // First the decision vector ...
          tempDV.addElement(entryDV);
          // ... then the objective vector.
          tempOV.addElement(entryOV);
        }
      }
    }
    newDimObj = o.size();
    newDimDec = d.size();
    generationSet.clear();
    if (Kea.logger.debugOn()) {
      Kea.logger.logDebug("Anzahl Individuen = Anzahl Eintrage in temp Vektoren? " + (tempOV.size()==
      tempDV.size()||tempOV.size()==individuumCounter) , this);
    }
    result = new Data[individuumCounter];
    for (int i=0; i<individuumCounter; i++){
      result[i] = new Data((Value[])tempDV.elementAt(i), (Value[])tempOV.elementAt(i));
    }
  }

  public int getNumberOfLastGeneration(){
    int last = 0;

    try {
      String curLine;
      BufferedReader in = new BufferedReader(new FileReader(run));
      // Skip Header.
      while(in.ready()) {
        curLine = in.readLine();
        if (curLine != null){
          // Search for end of header.
          if (curLine.indexOf("###") != -1){
            break;
          }
        }
      }
      while (in.ready()){
        curLine = in.readLine();
        if (curLine != null) {
          // Skip empty lines.
          while (curLine.equals("")) {
            curLine = in.readLine();
            if (curLine == null)break;
          }
          // All empty lines up to the first individuum or end of file skipped.
          // Increase generationcounter.
          last ++;
          // Skip not used generation.
          while (!curLine.equals("")) {
            curLine = in.readLine();
            if (curLine == null)break;
          }
          if (curLine == null)break;
        }
      }
    }
    catch (FileNotFoundException exc) {
      throw new EvalException("File "+run+" not found. No operation on this file possible.");
    }
    catch (IOException k){
      k.printStackTrace();
    }
    return last;
  }

  public String getNewDimObj(){
    String dim = new Integer(newDimObj).toString();
    return (dim);
  }

  public String getNewDimDec(){
    String dim = new Integer(newDimDec).toString();
    return (dim);
  }

  public Vector getIndividuum(){
    return individuum;
  }
}
