/*
 * AbstractProblem.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems;

import java.io.Serializable;
import java.util.Vector;

import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.Value;

/**
 * If you want to add a specific problemclass to <code>KEA</code>, you must inherit from this
 * abstract class and implement the abstract methods to run the new problem with <code>KEA</code>. 
 * All methods are needed by the algorithm to handle the problem. All these methods characterize 
 * a special problem.<br>
 *
 * @author Torsten Kohlen
 * @version 0.1.0
 */
public abstract class AbstractProblem implements Serializable{
    
    /**
     * Holds the Parameter for the problem.
     */
    protected Parameters param;

    /**
     * Number of times method score() was called. Is needed to assess the quality of
     * an algorithm.
     */
    private int evalCount = 0;

    /**
     * An algorithm can call this method, if it wants a computed decision Value
     * to be calculated by the problem. Input is a decision vector x and this method
     * must calculate the f(x) vector representing the problem.
     *
     * @param A Value[] calculated by a multiobjective algorithm as possible solution for the problem.
     * @return The corresponding function Value f(x).
     */
    public abstract Value[] score(Value[] genotyp);

    /**
     * This method must return the dimension of the decision-Space of the problem.
     *
     * @return dimension of the decision space as Integer.
     */
    public abstract int getDimDecisionSpace();

    /**
     * This method must return the dimension of the objective-Space of the problem.
     *
     * @return dimension of the objective space as Integer.
     */
    public abstract int getDimObjectiveSpace();
    /**
     * 
     * @return
     */
	public abstract Value[] StartIndividuals(Value[] genotype);
	public abstract Value[] generatNewInd(Value[] genotype);
	public abstract void  ObjectiveName();
        
    public abstract Value[] repair(Value[] genotype);
    public abstract Vector  IndividuumRepair(Vector Individiuum);
    /**
     * Implemtation of this method gives the limits for each values from a decision vector.
     * These limits must not be set for all of them. If there are no limits for a single value,
     * you must use MIN_VALUE or MAX_VALUE.
     *
     * @see Value
     * @return Array with ranges for all values of a decision vector.
     */
    public abstract Range[] getRanges();

    /**
     * Implemtation of this method gives the type for each value from a decision vector.
     * These can be e.g. Integer, Binary, etc.
     *
     * @see Type
     * @return Array with types for all values of a decision vector.
     */
    public abstract Type[] getDecTypes();

    /**
     * Implemtation of this method gives the type for each value from a objective vector.
     * These can be e.g. Integer, Binary, etc.
     *
     * @see Type
     * @return Array with types for all values of a objective vector.
     */
    public abstract Type[] getObjTypes();

    /**
     * This method must test a given solution, if it violates any constraints of the problem.
     *
     * @param Vector the calculated value of an algorithm.
     * @return boolean true, if the value violates no constraints, false else.
     */
    public abstract boolean verify(Value[] toVerify);

    /**
     * Gives the set of parameters nedded to run a problem properly.
     *
     * @see Parameters
     * @return The set of Prameters, combined in the Parameters-class.
     */
    public abstract Parameters getParameter();

    /**
     * Sets the parameters for the problem.
     *
     * @see Parameters
     * @param Parameters The new parameterset for the problem.
     */
    public abstract void setParameter(Parameters parameter);

    /**
     * Gives the value of evaulCount.
     *
     * @return int
     */
    public int getEvalCount(){
	return evalCount;
    }

    /**
     * Sets value of evalCount.
     *
     * @param int value to be set as evalCount.
     */
    public void setEvalCount(int evalCount){
	this.evalCount = evalCount;
    }
    
    /**
     * This method is called by KEA after an optimization run is finished.
     * If a problem needs to perform some final clean up operations, the
     * appropriate code should be placed here.
     */
    public void cleanUp() {}
    
}// end of class
