/*
 * DTLZ3.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.Vector;

import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * Calculates DTLZ3.
 * <ul>
 * <li> DTLZ3 has many local Pareto-optimal fronts.
 * <li> The global Pareto-optimal front corresponds to \vec{x}_M = (0.5, ..., 0.5).
 * </ul>
 * For further information see: <br>
 * Kalyanmoy Deb, Lothar Thiele, Marco Laumanns, and Eckart Zitzler.
 * Scalable Test Problems for Evolutionary Multi-Objective Optimization.
 * Technical Report 112, Computer Engineering and Networks Laboratory
 * (TIK), Swiss Federal Institute of Technology (ETH), Zurich, Switzerland,
 * 2001.
 *
 * @author PG 419, University of Dortmund (hp)
 * @version $Revision: 414 $
 */
public class DTLZ3 extends AbstractProblem {

	/**
	 * Contains the parameters for this problem.
	 */
	Parameters param;

	/**
	 * Sets all adjustable parameters to their defaults.
	 */
	public DTLZ3() {
		param = new Parameters();
		param.setMembership(this.getClass().getName());
		param.addParameter(new Parameter("k", 10, "= |\\vec(x)_M|"));
		param.addParameter(new Parameter("M", 2, "number of objectives"));
	}

	/**
	 * Calculates the dimension of decision space.
	 *
	 * @return dimension of decision space
	 */
	public int getDimDecisionSpace() {
		// Get value of 'M'.
		Parameter para_M = param.getParameterByName("M");
		int M = ((Integer) para_M.getData()).intValue();
		// Get value of 'k'.
		Parameter para_k = param.getParameterByName("k");
		int k = ((Integer) para_k.getData()).intValue();
		return M + k - 1;
	}

	/**
	 * Returns an <code>int</code> representing the dimension of objective space.
	 *
	 * @return dimension of objective space
	 */
	public int getDimObjectiveSpace() {
		Parameter para_M = param.getParameterByName("M");
		return ((Integer) para_M.getData()).intValue();

	}

	/**
	 * Returns an array of <code>Type</code> objects describing the decision variable types.
	 *
	 * @return a <code>Type[]</code> containing exactly one <code>Type</code> object for each decision variable
	 */
	public Type[] getDecTypes() {
		int n = getDimDecisionSpace();
		Type[] types = new Type[n];
		for (int i = 0; i <= n - 1; i++) {
			types[i] = new Type(Type.DOUBLE);
		}
		return types;
	}

	/**
	 * Returns an array of <code>Type</code> objects describing the objective variable types.
	 *
	 * @return a <code>Type[]</code> containing exactly one <code>Type</code> objects for each objective variable
	 */
	public Type[] getObjTypes() {
		int M = getDimObjectiveSpace();
		Type[] types = new Type[M];
		for (int i = 0; i <= M - 1; i++) {
			types[i] = new Type(Type.DOUBLE);
		}
		return types;
	}

	/**
	 * Returns an array of <code>Range</code> objects describing
	 * the allowed ranges for each descision variable.
	 *
	 * @return an array of <code>Range</code> objects for each decision variable
	 */
	public Range[] getRanges() {
		int n = getDimDecisionSpace();
		Range[] ranges = new Range[n];
		// Set Range for x_1, ..., x_n to [0,1].
		for (int i = 0; i <= n - 1; i++) {
			ranges[i] = new Range(new Double(0), new Double(1));
		}
		return ranges;
	}
    
    
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			return individuum;
		}
    
    
	/**
	 * Takes an array of decision variable values
	 * and calculates the corresponding array of objective values.
	 *
	 * @param genotype an array of decision variable values which should be scored
	 * @return the corresponding array of objective values
	 */
	public Value[] score(Value[] genotype) {
		// Create return-array.
		Value[] score = new Value[getDimObjectiveSpace()];
		// Initialize k, M, n.
		int k = ((Integer) param.getParameterByName("k").getData()).intValue();
		int M = getDimObjectiveSpace();
		int n = getDimDecisionSpace();
		// Calculate g(\vec(x)_M).
		double g;
		double sum = 0;
		for (int i = M; i <= n; i++) {
			double x_i = ((Number) genotype[i - 1].getData()).doubleValue();
			double addend =
				Math.pow(x_i - 0.5, 2) - Math.cos(20 * Math.PI * (x_i - 0.5));
			sum = sum + addend;
		}
		g = 100 * (k + sum);
		// Calculate f_i(\vec{x}).
		for (int i = 1; i <= M; i++) {
			double f_i;
			// Calculate product of cos factors.
			double cos_prod = 1;
			for (int j = 1; j <= M - i; j++) {
				double x_j = ((Number) genotype[j - 1].getData()).doubleValue();
				double cos_fac = Math.cos(x_j * Math.PI / 2);
				cos_prod = cos_prod * cos_fac;
			}
			// Calculate f_i.
			// No sin factor in f_1 results in special treatment.
			if (i == 1) {
				f_i = (1 + g) * cos_prod;
			}
			// f_2, ..., f_M can be processes n the same way.
			else {
				// Multiply witht (1+g) and sin factor.
				double x = ((Number) genotype[M - i].getData()).doubleValue();
				f_i = (1 + g) * cos_prod * Math.sin(x * Math.PI / 2);
			}
			// Store result f_i.
			score[i - 1] = new Value(f_i);
		}
		setEvalCount(getEvalCount() + 1);
		// Logging ...
		if (Kea.logger.debugOn()) {
			Kea.logger.logDebug("EvalCount = " + this.getEvalCount(), this);
			Vector w = new Vector();
			for (int i = 0; i < genotype.length; i++) {
				w.addElement(genotype[i].getData());
			}
			Kea.logger.logDebug("x = " + w.toString(), this);
			Kea.logger.logDebug("g(\\vec(x)_M) = " + g, this);
			Vector v = new Vector();
			for (int i = 0; i < score.length; i++) {
				v.addElement(score[i].getData());
			}
			Kea.logger.logDebug("f(x) = " + v.toString(), this);
		}
		return score;
	}

	/**
	 * Returns <code>false</code>, if the given array of decision variable values
	 * violates any of the problem's constraints.
	 *
	 * @param toVerify a <code>Value[]</code> that should be verified
	 * @return <code>true</code> or <code>false</code>, depending on the violation of constraints
	 */
	public boolean verify(Value[] toVerify) {
		int n = getDimDecisionSpace();
		// Check length.
		if (toVerify.length != n) {
			if (Kea.logger.debugOn())
				Kea.logger.logDebug("Verify failed: Vector has wrong length.", this);
			return false;
		} else {
			// Check ranges.
			for (int i = 0; i <= n - 1; i++) {
				if ((((Number) toVerify[i].getData()).doubleValue() < 0)
					|| (((Number) toVerify[i].getData()).doubleValue() > 1)) {
					if (Kea.logger.debugOn())
						Kea.logger.logDebug(
							"Verify failed: " + toVerify[i] + " not in allowed range.",
							this);
					return false;
				}
			}
			return true;
		}
	}

	/**
	 * Returns the <code>Parameters</code> object of this class.
	 * @return the <code>Parameters</code> object of this class
	 */
	public Parameters getParameter() {
		return param;
	}

	/**
	 * Sets a new <code>Parameters</code> object for this problem.
	 * @param param the <code>Parameters</code> object that will be set
	 */
	public void setParameter(Parameters p) {
		try {
			int k = ((Number) p.getParameterByName("k").getData()).intValue();
			param.setParameter(new Parameter("k", k, "= |\\vec(x)_M|"));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("k set to " + k, this);
			}
		} catch (Exception e) {
			param.setParameter(new Parameter("k", 10, "= |\\vec(x)_M|"));
			Kea.logger.logWarn("Received wrong value for k, k set to 10.", this);
		}
		try {
			int M = ((Number) p.getParameterByName("M").getData()).intValue();
			param.setParameter(new Parameter("M", M, "number of objectives"));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("M set to " + M, this);
			}
		} catch (Exception e) {
			param.setParameter(new Parameter("M", 2, "number of objectives"));
			Kea.logger.logWarn("Received wrong value for M, M set to 2.", this);
		}
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		//return null;
	}

} // DTLZ3
