/*
 * GSphere.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.Vector;

import kea.Parameter;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * Calculates the generic sphere problem. <br> 
 * For further information see: Eckart Zitzler, Marco Laumanns, and Lothar Thiele. 
 * SPEA2: Improving the Strength Pareto Evolutionary Algorithm. Technical Report 103, 
 * Computer Engineering and Networks Laboratory (TIK), Swiss Federal Institute of 
 * Technology (ETH) Zurich, Gloriastrasse 35, CH-8092 Zurich, Switzerland, May 2001. <br> 
 * Attributes: <ol>
 * 	<li> evalCount: Inherited from AbstractProblem 
 * 	<li> n: Dimension of the decision space
 * 	<li> m: Dimension of the objective space 
 * 	<li> x: Array of decision variables 
 * 	<li> x: Array of objective variables 
 * </ol>
 *
 * @author Philipp Limbourg
 * @version 0.1.0
 * @see AbstractProblem
 *
 */
public class GSphere extends AbstractProblem {
    /** Constructor, sets default values for n, m */
    public GSphere() {
        initialize(10, 2);
    }
    
    
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	
    
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			return individuum;
		}
    
    /**
     * Evaluates the given decision vector
     * @param genotyp the decision vector
     * @return Vector the objective vector
     */
    public Value[] score(Value[] genotyp) {
        evalCount++;
        if (verify(genotyp) == false) {
            return vectorifyMax();
        }
        else {
            f();
            return vectorify();
        }
    }

    /** Liefert die Dimension des Entscheidungsvariablenraums. */
    public int getDimDecisionSpace() {
        return n;
    }

    /** Liefert die Dimension des Zielraums. */
    public int getDimObjectiveSpace() {
        return m;
    }

    /** Liefert die Grenzen einer Variablen des Entscheidungsvariablenraums zur?ck. Mu? nicht unbedingt definiert sein. */
    public Range[] getRanges() {
        Range[] result = new Range[n];
        for (int i = 0; i < n; i++) {
            result[i] = new Range(new Double(-1000), new Double(1000));
        }
        return result;
    }

    /** Liefert den Typ einer Variablen im Entscheidungsvariablenraum zur?ck (Int, Bin?r, Float...). */
    public Type[] getDecTypes() {
        Type[] result = new Type[n];
        for (int i = 0; i < n; i++) {
            result[i] = new Type(6);
        }
        return result;
    }

    /** The function to evaluate */
    public void f() {
        for (int j = 0; j < m; j++) {
            y[j] = (x[j] - 1) * (x[j] - 1);
            for (int i = 0; i < n; i++) {
                if (i != j) {
                    y[j] += x[i] * x[i];
                }
            }
        }
    }

    /**
     * Returns false if constraints are violated, else true
     * @param genotyp the decision vector to verify
     * @return boolean true if no constraints are violated
     */
    public boolean verify(Value[] genotyp) {
        arrify(genotyp);
        for (int j = 0; j < m; j++) {
            if (x[j] < -1000 || x[j] > 1000) {
                return false;
            }
        }
        return true;
    }

    /** Liefert den Typ einer Variablen im Entscheidungsvariablenraum zur?ck (Int, Bin?r, Float...). */
    public Type[] getObjTypes() {
        Type[] result = new Type[m];
        for (int i = 0; i < m; i++) {
            result[i] = new Type(6);
        }
        return result;
    }

    public int getEvalCount() {
        return evalCount;
    }

    public void setEvalCount(int evalCount) {
        this.evalCount = evalCount;
    }

    /**
     * Parses the genotyp vector to x[]
     * @param genotyp vector to parse
     */
    private void arrify(Value[] genotyp) {
        for (int i = 0; i < genotyp.length; i++) {
            x[i] = ((Number)genotyp[i].getData()).doubleValue();
        }
    }

    /**
     * Parses y[] to the objective vector
     * @return Vector the objective vector
     */
    private Value[] vectorify() {
        Value[] phaenotyp = new Value[m];
        for (int i = 0; i < m; i++) {
            phaenotyp[i]=new Value(y[i]);
        }
        return phaenotyp;
    }

    /**
     * Returns a vector with maximal contents (if constraints are violated)
     * @return Vector the objective vector
     */
    private Value[] vectorifyMax() {
        Value[] phaenotyp = new Value[m];
        for (int i = 0; i < m; i++) {
            phaenotyp[i]=new Value(Double.MAX_VALUE);
        }
        return phaenotyp;
    }

    /**
     * Initializes the function
     * @param n dimension of decision space
     * @param m dimension of objective space
     */
    private void initialize(int n, int m) {
        this.n = n;
        this.m = m;
        x = new double[n];
        y = new double[m];
        evalCount = 0;
    }

    /** Setzt die Parameter DimDecisionSpace & DimObjectiveSpace */
    public Parameters getParameter() {
        Parameters params = new Parameters();
        Parameter[] param = new Parameter[2];
        param[0] = new Parameter();
        param[0].setName("DimDecisionSpace");
        param[0].setToolTip("dimension of the decision space");
        param[0].setRange(1, Integer.MAX_VALUE);
        param[0].setType(new Type(Type.INTEGER));
        param[0].setData(new Integer(n));
        param[1] = new Parameter();
        param[1].setName("DimObjectiveSpace");
        param[1].setToolTip("dimension of the objective space");
        param[1].setRange(1, Integer.MAX_VALUE);
        param[1].setType(new Type(Type.INTEGER));
        param[1].setData(new Integer(m));
        params.setParameterArray(param);
	params.setMembership(this.getClass().getName());
        return params;
    }

    /** ... fuer's Optionhandler-Interface ... */
    public void setParameter(Parameters params) {
        Parameter[] param = params.getParameterArray();
        int n = ((Number)param[0].getData()).intValue();
        int m = ((Number)param[1].getData()).intValue();
        initialize(n, m);
    }

    /** Z?hlt die Aufrufe von AbstractProblem.evaluate(). Wert wird in AlgoRun gespeichert. */
    private int evalCount;

    /** objective space dimension */
    private int m;

    /** decision space dimension */
    private int n;

    /** decision vector */
    private double[] x;

    /** objective vector */
    private double[] y;

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		return null;
	}


	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		return null;
	}


	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		//return null;
	}
}
