/*
 * Kursawe.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.Vector;

import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * Calculates the Kursawe function.
 * <ul>
 * <li> Kursawe's decision space can have an arbitrary number of dimensions,
 * it's objective space is two-dimensional.
 * <li> Pareto-optimal solutions lie in several disconnected
 * and unsymmetric areas in decision space.
 * <li>The Pareto optimal front consists of several disconnected curves.
 * </ul>
 *
 * @author PG 419, University of Dortmund (hp)
 * @version $Revision: 414 $
 */
public class Kursawe extends AbstractProblem {

	/**
	 * Contains the parameters for this problem.
	 */
	private Parameters param;

	/**
	 * Sets all adjustable parameters to their defaults.
	 */
	public Kursawe() {
		param = new Parameters();
		param.setMembership(this.getClass().getName());
		param.addParameter(new Parameter("n", 2, "dimension of decision space"));
		param.addParameter(
			"minRange",
			-100,
			"Lower border of decision space (for all dimensions).");
		param.addParameter(
			"maxRange",
			100,
			"Upper border of decision space (for all dimensions).");
	}

	/**
	 * Calculates the dimension of decision space.
	 *
	 * @return dimension of decision space
	 */
	public int getDimDecisionSpace() {
		Parameter n = param.getParameterByName("n");
		return ((Integer) n.getData()).intValue();
	}

	/**
	 * Returns an <code>int</code> representing the dimension of objective space.
	 * @return Since objective space is two-dimensional for Kursawe, this method always returns <code>2</code>.
	 */
	public int getDimObjectiveSpace() {
		return 2;
	}

	/**
	 * Returns an array of <code>Type</code> objects describing the decision variable types.
	 *
	 * @return a <code>Type[]</code> containing exactly one <code>Type</code> object for each decision variable
	 */
	public Type[] getDecTypes() {
		int n = getDimDecisionSpace();
		Type[] types = new Type[n];
		for (int i = 0; i <= n - 1; i++) {
			types[i] = new Type(Type.DOUBLE);
		}
		return types;
	}

	/**
	 * Returns an array of <code>Type</code> objects describing the decision variable types.
	 *
	 * @return a <code>Type[]</code> containing exactly two <code>Type</code> objects, one for each objective
	 */
	public Type[] getObjTypes() {
		Type[] types = new Type[2];
		types[0] = new Type(Type.DOUBLE);
		types[1] = new Type(Type.DOUBLE);
		return types;
	}

	/**
	 * Returns an array of <code>Range</code> objects describing
	 * the allowed ranges for each descision variable.
	 *
	 * @return an array of <code>Range</code> objects for each decision variable
	 */
	public Range[] getRanges() {
		int n = getDimDecisionSpace();
		double minRange =
			((Number) param.getParameterByName("minRange").getData()).doubleValue();
		double maxRange =
			((Number) param.getParameterByName("maxRange").getData()).doubleValue();
		Range[] ranges = new Range[n];
		for (int i = 0; i <= n - 1; i++) {
			ranges[i] = new Range(new Double(minRange), new Double(maxRange));
		}
		return ranges;
	}
    
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			return individuum;
		}
    
	/**
	 * Takes an array of decision variable values
	 * and calculates the corresponding array of objective values.
	 *
	 * @param genotype a descision vector which should be scored
	 * @return the corresponding objective vector
	 */
	public Value[] score(Value[] genotype) {
		int n = getDimDecisionSpace();
		// Calculate f_1(x_0, ..., x_{n-1}).
		double f1 = 0;
		for (int i = 0; i <= n - 2; i++) {
			double x_i = ((Number) genotype[i].getData()).doubleValue();
			double x_iPlusOne = ((Number) genotype[i + 1].getData()).doubleValue();
			double sqrt = Math.sqrt(Math.pow(x_i, 2) + Math.pow(x_iPlusOne, 2));
			double exp = Math.exp((-0.2) * sqrt);
			f1 = f1 + (-10) * exp;
		}
		// Calculate f_2(x_0, ..., x_{n-1}).
		double f2 = 0;
		for (int i = 0; i <= n - 1; i++) {
			double xi = ((Number) genotype[i].getData()).doubleValue();
			double abs_xi = Math.abs(xi);
			double pow_abs_xi = Math.pow(abs_xi, 0.8);
			double sin_xi = Math.sin(xi);
			double pow_sin_xi = Math.pow(sin_xi, 3);
			double addend_xi = pow_abs_xi + 5 * pow_sin_xi;
			f2 = f2 + addend_xi;
		}
		// Increase evaluation counter.
		setEvalCount(getEvalCount() + 1);
		// Return the results.
		Value[] score = new Value[getDimObjectiveSpace()];
		score[0] = new Value(f1);
		score[1] = new Value(f2);
		// Logging ...
		if (Kea.logger.debugOn()) {
			Kea.logger.logDebug("EvalCount = " + this.getEvalCount(), this);
			Vector dec = new Vector();
			for (int i = 0; i <= n - 1; i++) {
				double xi = ((Number) genotype[i].getData()).doubleValue();
				dec.addElement(new Double(xi));
			}
			Kea.logger.logDebug("x = " + dec.toString(), this);
			Kea.logger.logDebug("f_1(x) = " + f1 + " f_2(x) = " + f2, this);
		}
		return score;
	}

	/**
	 * Returns <code>false</code>, if the given array of decision variable values
	 * violates any of the problem's constraints. In the case of
	 * Kursawe only the ranges are checked.
	 *
	 * @param toVerify a <code>Value[]</code> that should be verified
	 * @return <code>true</code>, if the vector's coordinates are in the allowed range
	 */
	public boolean verify(Value[] toVerify) {
		// Has toVerify the correct length?
		int n = this.getDimDecisionSpace();
		if (toVerify.length != n) {
			if (Kea.logger.debugOn())
				Kea.logger.logDebug("Verify failed: Vector has wrong length.", this);
			return false;
		}
		// Check if coordinates have the correct range.
		Range[] ranges = this.getRanges();
		for (int i = 0; i < n; i++) {
			if (((Number) ranges[i].getMinVal()).doubleValue()
				>= ((Number) toVerify[i].getData()).doubleValue()
				|| ((Number) ranges[i].getMaxVal()).doubleValue()
					<= ((Number) toVerify[i].getData()).doubleValue()) {
				if (Kea.logger.debugOn())
					Kea.logger.logDebug(
						"Verify failed: " + toVerify[i] + " not in allowed range.",
						this);
				return false;
			}
		}
		return true;
	}

	/**
	 * Returns the <code>Parameters</code> object of this class.
	 * @return the <code>Parameters</code> object of this class
	 */
	public Parameters getParameter() {
		return param;
	}

	/**
	 * Sets a new <code>Parameters</code> object for this problem.
	 * @param param the <code>Parameters</code> object that will be set
	 */
	public void setParameter(Parameters p) {
		try {
			int n = ((Number) p.getParameterByName("n").getData()).intValue();
			param.setParameter(new Parameter("n", n, "dimension of decision space"));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("n set to " + n, this);
			}
		} catch (Exception e) {
			param.setParameter(new Parameter("n", 2, "dimension of decision space"));
			Kea.logger.logWarn("Received wrong value for n, n set to 2.", this);
		}
		try {
			double minRange =
				((Number) p.getParameterByName("minRange").getData()).doubleValue();
			param.setParameter(
				new Parameter(
					"minRange",
					minRange,
					"Lower border of decision space (for all dimensions)."));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("minRange set to " + minRange, this);
			}
		} catch (Exception e) {
			param.setParameter(
				new Parameter(
					"minRange",
					-100,
					"Lower border of decision space (for all dimensions)."));
			Kea.logger.logWarn("Received wrong minRange, minRange set to -100.", this);
		}
		try {
			double maxRange =
				((Number) p.getParameterByName("maxRange").getData()).doubleValue();
			param.setParameter(
				new Parameter(
					"maxRange",
					maxRange,
					"Upper border of decision space (for all dimensions)."));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("maxRange set to " + maxRange, this);
			}
		} catch (Exception e) {
			param.setParameter(
				new Parameter(
					"maxRange",
					100,
					"Upper border of decision space (for all dimensions)."));
			Kea.logger.logWarn("Received wrong maxRange, maxRange set to 100.", this);
		}
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		//return null;
	}

} // Kursawe
