/*
 * Lotz2.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.LinkedList;
import java.util.Vector;

import kea.Kea;
import kea.Parameter;
import kea.Range;
import kea.Type;
import kea.Value;

/**
 * Title:         Testfunction Lotz <br>
 * Description:   This Problem is a generalisation of the LEADINGONES-function.
 *                The object of this leading ones trailing zeroes function is to maximize
 *                the ones at the beginning of the bitstring and also to maximize the
 *                zeroes in the ending of the string. <br><br>
 * Function:
 *                Lotz:{0,1}^n \rightarrow N^2 <br>
 *
 *                Lotz(x1,...,xn)=(\sum{i=1}{n}{\prod{j=1}{i}{xj}},\sum{i=1}{n}{\prod{j=i}{n}1-xj}) <br><br>
 *
 * Additional Information:
 *                Naturaty the aim of the optimizer should be to maximize the Objective Function.
 *                Because most optimizers expect to minimize the Objective Function, we multiply the
 *                computated results of both of the functions f1 and f2 with -1.
 *                (Compare this in the method <code>score()</code>.) <br><br>
 *
 *                Nobody wants to confuse the optimizer...;-)
 *
 *
 * @author Stephan Sigg
 * @version $Revision: 414 $
 */
public class Lotz2 extends kea.problems.AbstractProblem {
    /**
     * Contains the parameters for this problem.
     */
    kea.Parameters parameters;
    /**
     * Constructor
     * Initializes the entries for the Parameter-Object
     */
    public Lotz2() {
	parameters = new kea.Parameters();
	parameters.setMembership(this.getClass().getName());
	kea.Parameter[] params  = new kea.Parameter[1];
	params[0] = new Parameter("n",20,"Length of the decision variable vector.");
	parameters.setParameterArray(params);
    }
    
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			return individuum;
		}
    
    /**
     * computes the Objective Space Values from the decision Space Values.
     * @param genotyp contains the Values of the decision space Values
     * @return results returns the Objective Space Values. In this case 2 Values
     */
    public kea.Value[] score(Value[] genotyp){
	Value[] results = new Value[2];
	if (verify(genotyp)){
	    if(genotyp != null){
		int f1 = 0;
		int f2 = 0;
		// computation of f1
		boolean noBreak = true;
		for (int i=0; i< genotyp.length;i++){
		    if( noBreak && ((Boolean)genotyp[i].getData()).booleanValue())
			f1++;
		    else
			noBreak = false;
		}
		// computation of f2
		noBreak = true;
		for (int i=genotyp.length - 1; i> -1 ;i--){
		    if( noBreak && !((Boolean)genotyp[i].getData()).booleanValue() )
			f2++;
		    else
			noBreak = false;
		}
		// Increase evaluation counter.
		setEvalCount(getEvalCount() + 1);
		// in order to guarantee that the results are never negative, we add n(i.e. genotyp.length) onto f1 and f2
		results[0] = new kea.Value(f1*(-1) + genotyp.length);
		results[1] = new kea.Value(f2*(-1) + genotyp.length);
	    }
	    else{
		Kea.logger.logDebug("please initialize your Value-Array before trying to compute any results.",this);
		results[0] = new kea.Value(Integer.MAX_VALUE);
		results[1] = new kea.Value(Integer.MAX_VALUE);
	    }
	}
	else{
	    results[0] = new kea.Value(Integer.MAX_VALUE);
	    results[1] = new kea.Value(Integer.MAX_VALUE);
	}
       	return (results);
    }// End of Score
    /**
     * @return Returns the dimension of the Decision Space
     */
    public int getDimDecisionSpace(){
	Parameter n = parameters.getParameterByName("n");
	if (n != null) {
	    return ((Number)n.getData()).intValue();
	}
	Kea.logger.logError("getDimDecisionSpace(): A problem occurred.", this);
	return -1;
    }
    /**
     * @return Returns the Dimension of the Objectiove Space
     */
    public int getDimObjectiveSpace(){
      return(2);
    }
    /**
     * returns the borders of a Variable in the decision space.
     * @return ranges Array of Ranges. These Ranges are all set to Double.Min_VALUE and Double.Max_VALUE.
     */
    public kea.Range[] getRanges(){
	int n = getDimDecisionSpace();
	Range[] ranges = new Range[n];
	for (int i = 0; i < n; i++) {
	    ranges[i] = new Range(new Boolean(true),new Boolean(false));
	}
	return ranges;
    }
    /**
     * @return types Array of Types containing type of the variables in the decision space.
     *               (Int, Binary, Float,...)
     *               these Specifications are coded as described in the Type-object.
     */
    public kea.Type[] getDecTypes(){
	int n = getDimDecisionSpace();
	Type[] types = new Type[n];
	for (int i = 0; i < n; i++) {
	    types[i] = new Type(Type.BOOLEAN);
	}
	return types;
    }
    /**
     * controls whether or not a to be tested result contains illegal Values.
     * @param toVerify Array of Values containing the decision space Values.
     * @return types returns a boolean, that is set to false, when a to be tested decision space
     *               Vector contains illegal Values.
     */
    public boolean verify(Value[] toVerify){
        Boolean bool = new Boolean(true);
	if (toVerify == null || toVerify.length != getDimDecisionSpace()){
	    Kea.logger.logDebug("please set the input values correctly",this);
	    return false;
	}
	for(int i=0; i < toVerify.length;i++){
	    if(toVerify[i].getData().getClass() != bool.getClass()){
		Kea.logger.logInfo("Hey, you are working on bit strings - remember ? Therefore please compose the input string exclusively from boolean objects",this);
		return false;
	    }
	}// of for
	return(true);
    }// of verify
    /**
     * @return Returns the Value of a Variable in the Objective Space (Int, Float...).
     */
    public kea.Type[] getObjTypes(){
	kea.Type[] types = new kea.Type[2];
        types[0] = new kea.Type(kea.Type.INTEGER);
	types[1] = new kea.Type(kea.Type.INTEGER);
	return(types);
    }
    public kea.Parameters getParameter(){
      return(this.parameters);
    }
    public void setParameter(kea.Parameters parameters){
	if (parameters.getParameterByName("n") != null){
	    Integer integer = new Integer(0);
	    Class cla;
	    cla = parameters.getParameterByName("n").getClass();
	    if (cla != integer.getClass()){
		Kea.logger.logInfo("Wrong Parameter setting. Unable to change default parameter setting.",this);
	    }
	    else{
		if (parameters != null)
		    this.parameters = parameters;
		else
		    Kea.logger.logInfo("Received new parameters containing: null. Unable to change default parameter setting",this);
	    }
	}
	else
	    Kea.logger.logInfo("Cannot find parameter with label 'n'. Therefor unable to change default parameter setting.",this);
    }

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#writestring()
	 */
	public LinkedList writestring() {
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		//return null;
	}
}









