/*
 * Mokp.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.LinkedList;
import java.util.Vector;

import kea.Kea;
import kea.Parameter;
import kea.Range;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * Title:        Multiple Objective Knapsack problem <br>
 * Description:  This problem is the multi dimensional version of the NP-hard Knapsack Problem.
 *               As described for example in
 *               "Eckard Zitzler. Evolutionary algorithms for Multiobjective Optimization:
 *               Methods and Appilcations. PhD thesis, Swiss Federal Institute of Technology (ETH), Zurich, Switzerland,
 *               November 1999" <br><br>
 *               This variant allows n Knapsacks and m tools which are described with <br>
 *                          p_{i,j} = Usability of tool j for Knapsack i <br>
 *                          w_{i,j} = Weight of tool j for Knapsack i <br>
 *                          c_i = capacity of Knapsack i <br><br>
 *
 *               The aim is to find a result X = (x_1,...,x_m) $/in/{0,1/}^n. <br>
 *               This result is to respect the Capacity borders <br><br>
 *
 *                         $ e_i(X)=/sum_{j=1}^{m}{w_{i,j}}*x_j /leq c_i$ where $(1 /leq i /leq k)$ <br><br>
 *
 *               and which maximizes $f(X) = (f_1(X),...,f_k(X))$, where <br>
 *                         $ f_(X)=/sum_{j=1}^{m}{p_{i,j}*x_j}$ <br><br>
 *
 *               $x_j = 1$ represents that tool j is chosen.
 *
 *
 * @author Stephan Sigg
 * @version $Revision: 723 $
 */
public class Mokp extends AbstractProblem{
    // new Parametervalue. contains all Parameterentries of this function.
  kea.Parameters parameters;
    //constructor
    public Mokp() {
	int k;
	int l;
	parameters = new kea.Parameters();
	parameters.setMembership(this.getClass().getName());
	kea.Parameter[] params  = new kea.Parameter[125];
	params[0] = new Parameter("n",3,"Number of Knapsacks");
	params[1] = new Parameter("m",20,"Number of tools");
	int arraycounter = 2;
	for (int i=0; i < 3;i++){
	    k= i+1;
	    for (int j = 0; j< 20; j++){
		l=j+1;
		params[arraycounter] = new Parameter("w"+k+l,2*k+l,"Weight of tool "+l+" for Knapsack " +k);
		params[arraycounter + 60] = new Parameter("p"+k+l,2*l+k,"usefullness of tool "+l+" for Knapsack " +k);
		arraycounter = arraycounter + 1;
		Kea.logger.logInfo("w"+k+l+"="+2*k+l+"Weight of tool "+l+" for Knapsack " +k,this);
		Kea.logger.logInfo("p"+k+l+"= "+2*l+k+"usefullness of tool "+l+" for Knapsack " +k,this);
	    }
	    params[122 + i] = new Parameter("c"+k,70 * k ,"Maximal Weight for Knapsack " +k);
	}
	parameters.setParameterArray(params);
    }
    
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			// TODO Auto-generated method stub
			return individuum;
		}
    
    
    /**
     * computes the Objective Space Values from the decision Space Values.
     * naturally the aim of this function is to maximize the usability of the knapsacks. Many algorithms on the other hand
     * normally minimize the objective function.
     * Therefore we have multiplied the objective function with -1 so that it now has to be minimized. Please keep this
     * in mind when testing your algo on this function.
     * @param genotyp contains the Values of the decision space Values
     * @return results returns the Objective Space Values. In this case 2 Values
     */
  public kea.Value[] score(kea.Value[] genotyp) {
      int k;
      int l;
      int maxuse = 0;
      //Kea.logger.logDebug("method call: score(kea.Value[] genotyp)",this);
      int n = ((Number)parameters.getParameterByName("n").getData()).intValue();
      kea.Value[] results = new kea.Value[n];
      if (verify(genotyp)){
	  for (int z=0;z<n;z++){
	      results[z] = new Value(0);
	  }
	  if (genotyp != null){
	      for (int i = 0;i< n; i++){
		  k= i+1;
		  for (int j = 0;j< ((Number)parameters.getParameterByName("m").getData()).intValue(); j++){
		      l= j+1;
		      maxuse = maxuse + ((Number)parameters.getParameterByName("p"+k+l).getData()).intValue();
		      results[i] = new Value(((Number)results[i].getData()).intValue() + (-1)*(((Number)parameters.getParameterByName("p"+k+l).getData()).intValue() * ((Number)genotyp[j].getData()).intValue()));
		  }
		  // in order to get non negative results we add the sum of all pij to resutls[i] 
		  results[i] = new Value(((Number)results[i].getData()).intValue() + maxuse);
		  maxuse = 0;
	      }
	  }
	  else{
	      Kea.logger.logDebug("please initialize your input Values before starting to evaluate.",this);
	  }
      }
      else{
	  for (int i = 0;i< n; i++){
	      results[i] = new Value(Double.MAX_VALUE);
	  }
      }
      return(results);
  }//end of score
    /**
     * Returns the dimension of the decision Space
     * @return this.parameters.getParameterArray().length the dimension of the Decision Space
     */
    public int getDimDecisionSpace(){
	Parameter m = parameters.getParameterByName("m");
	if (m != null) {
	    return ((Number) m.getData()).intValue();
	}
	Kea.logger.logError("getDimDecisionSpace(): A problem occurred.", this);
	return -1;
    }
    /**
     * @return returns the dimension of the objective Space.
     */
    public int getDimObjectiveSpace(){
	Parameter n = parameters.getParameterByName("n");
	if (n != null) {
	    return ((Number) n.getData()).intValue();
	}
	Kea.logger.logError("getDimDecisionSpace(): A problem occurred.", this);
	return -1;
    }
    /**
     * returns the borders of a Variable in the decision space.
     * if no borders are specified, MIN_VALUE and MAX_VALUE of the depending Datatype are to be returned.
     * @return ranges Array of Ranges. these ranges are set to integervalues of 0 or 1
     */
    public kea.Range[] getRanges(){
	//Kea.logger.logDebug("method call: getRanges()",this);
	int m = getDimDecisionSpace();
	Range[] ranges = new Range[m];
	for (int i = 0; i < m; i++) {
	    ranges[i] = new Range(new Double(0),new Double(2));
	}
	return ranges;
    }
    /**
     * @return types Array of Types containing type of the variables in the decision space.
     *               (Int, Float,...)
     *               these Specifications are coded as described in the Type-object.
     */
    public kea.Type[] getDecTypes(){
      kea.Type[] types = new kea.Type[getDimDecisionSpace()];
      for (int i=0;i < getDimDecisionSpace(); i++){
        types[i] = new kea.Type(kea.Type.INTEGER);
      }
      return(types);
    }
    /**
     * controls whether or not a to be tested result contains illegal Values.
     * @param toVerify Array of Values containing the decision space Values.
     * @return types returns a boolean, that is set to false, when a to be tested decision space
     *               Vector contains illegal Values.
     */
    public boolean verify(kea.Value[] toVerify){
	int k;
	int l;
	//Kea.logger.logDebug("method call: verify(kea.Value[] toVerify",this);
	int ci = 0;
	if (toVerify == null){
	    Kea.logger.logDebug("please set the input values correctly",this);
	    return false;
	}
	if (toVerify.length != ((Number)parameters.getParameterByName("m").getData()).intValue()){
	    Kea.logger.logInfo("the vector of input variables has not the sufficient length."+((Number)parameters.getParameterByName("m").getData()).intValue()+" entries are rezquired. Your input vector has length "+toVerify.length+". Please correct this error.",this );
	    return false;
	}
	for (int i=0; i < toVerify.length;i++){
	    if (((Number)toVerify[i].getData()).intValue() > 1 || ((Number)toVerify[i].getData()).intValue() < 0){
		Kea.logger.logInfo("The input Vector you used is not completely composed of Binary values. Only Integer values 0 or 1 are allowed. please correct this error.",this);
		return  false;
	    }
	}// of for
	for (int i=0;i<((Number)parameters.getParameterByName("n").getData()).intValue();i++){
	    k=i+1;
	    ci = 0;
	    for (int j = 0; j< ((Number)parameters.getParameterByName("m").getData()).intValue();j++){
		l=j+1;
		ci = ci + ((Number)parameters.getParameterByName("w"+k+l).getData()).intValue() * ((Number)toVerify[j].getData()).intValue();
	    }

	    if (ci > ((Number)parameters.getParameterByName("c"+k).getData()).intValue()){
		Kea.logger.logInfo("Your input vector is violating the constraints",this);
		return false;
	    }
	}
	return(true);
    }// of verify
    /**
     * @return types Array of Types, containing the types of the variables in the objective space (Int, Float...).
     *               coded as specified in the class Type
     */
    public kea.Type[] getObjTypes(){
	kea.Type[] types = new kea.Type[((Number)parameters.getParameterByName("n").getData()).intValue()];
	for (int i = 0; i<types.length; i++){
	    types[i] = new kea.Type(kea.Type.DOUBLE);
	}
	return types;
    }

    public kea.Parameters getParameter(){
	return(parameters);
    }
    public void setParameter(kea.Parameters parameters){
	int k = 0;
	int l = 0;
	boolean verify = true;
	Integer integer = new Integer(0);
	Class cla;
        // test of parameters
	if (parameters == null){
	    verify = false;
	    Kea.logger.logInfo("Received new parameters containing: null. Unable to change default parameter setting",this);
	}
        // test of the 'n' parameter
	if (parameters.getParameterByName("n") == null){
	    verify = false;
	    Kea.logger.logInfo("Cannot find parameter with label 'n'. Therefor unable to change default parameter setting.",this);
	}
	cla = parameters.getParameterByName("n").getClass();
	if (cla != integer.getClass()){
	    Kea.logger.logInfo("Wrong Parameter setting. Unable to change default parameter setting.",this);
	    verify = false;
	}
	// test of the 'm' parameter...
	if (parameters.getParameterByName("m") == null){
	    verify = false;
	    Kea.logger.logInfo("Cannot find parameter with label 'm'. Therefor unable to change default parameter setting.",this);
	}
	cla = parameters.getParameterByName("m").getClass();
	if (cla != integer.getClass()){
	    Kea.logger.logInfo("Wrong Parameter setting. Unable to change default parameter setting.",this);
	    verify = false;
	}
        // test of the Wij and pij parameters
	if (verify){
	    for (int i = 0; i < ((Number)parameters.getParameterByName("n").getData()).intValue();i++){
		k= i+1;
		for (int j = 0; j< ((Number)parameters.getParameterByName("m").getData()).intValue();j++){
		    l=j+1;
		    if (parameters.getParameterByName("w"+k+l) == null){
			verify = false;
			Kea.logger.logInfo("Cannot find parameter with label w"+k+l+". Therefor unable to change default parameter setting.",this);  
		    }
		    cla = parameters.getParameterByName("w"+k+l).getClass();
		    if (cla != integer.getClass()){
			Kea.logger.logInfo("Wrong Parameter setting. Parameter with label w"+k+l+" is no Integer.Unable to change default parameter setting.",this);
			verify = false;
		    }
                    // test the pij parameter 
		    if (parameters.getParameterByName("p"+k+l) == null){
			verify = false;
			Kea.logger.logInfo("Cannot find parameter with label p"+k+l+". Therefor unable to change default parameter setting.",this);  
		    }
		    cla = parameters.getParameterByName("p"+k+l).getClass();
		    if (cla != integer.getClass()){
			Kea.logger.logInfo("Wrong Parameter setting. Parameter with label p"+k+l+" is no Integer.Unable to change default parameter setting.",this);
			verify = false;
		    }
		}
		// test of the Ck parameters
		if (parameters.getParameterByName("c"+k) == null){
			verify = false;
			Kea.logger.logInfo("Cannot find parameter with label c"+k+". Therefor unable to change default parameter setting.",this);  
		}
		cla = parameters.getParameterByName("c"+k).getClass();
		if (cla != integer.getClass()){
		    Kea.logger.logInfo("Wrong Parameter setting. Parameter with label c"+k+" is no Integer.Unable to change default parameter setting.",this);
		    verify = false;
		}
	    }
	}
	// set parameters
	if (verify){
	    this.parameters = parameters;
	}
    }

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#writestring()
	 */
	public LinkedList writestring() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		// TODO Auto-generated method stub
		//return null;
	}
}
