/*
 * Quagliarella.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.LinkedList;
import java.util.Vector;

import kea.Kea;
import kea.Parameter;
import kea.Range;
import kea.Value;
import kea.problems.AbstractProblem;
/**
 * Title:        Testfunction of Quagliarella <br>
 * Description:  The Pareto-optimal set of the function is not connected in
 *               the decision space. The pareto-optimal front is also
 *               disconnected. <br><br>
 *
 * Implemented function: <br><br>
 *
 *    minimize  F             = (f1(x1,...,xn),f2(x1,...,xn)),<br>
 *    wheras    f1(x1,...,xn) = \sqrt{\frac{A1}{n}} <br>
 *    and       f2(x1,...,xn) = \sqrt{\frac{A2}{n}} <br><br>
 *
 *    further is <br>
 *              A1 = \sum{i=1}{n}{(xi * xi) - 10\cos{(2 * \pi * xi)} + 10} <br>
 *    and <br>
 *              A2 = \sum{i=1}{n}{(xi-1,5)*(xi-1,5))-10\cos{(2*\pi*(xi-1,5))}+10} <br>
 *    and also <br>
 *              -5,12 \leq xi \leq 5,12
 *
 * @author Stephan Sigg
 * @version $Revision: 723 $
 */
public class Quagliarella extends AbstractProblem{
    // new Parametervalue. contains all Parameterentries of this function.
  kea.Parameters parameters;
    //constructor
    public Quagliarella() {
	parameters = new kea.Parameters(1);
	Parameter[] params = new Parameter[1];
	parameters.setMembership(this.getClass().getName());
	params[0] = new Parameter("n",16,"Dimension of the Decision Space");
	parameters.setParameterArray(params);
    }
    
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			// TODO Auto-generated method stub
			return individuum;
		}
    
    /**
     * computes the Objective Space Values from the decision Space Values.
     * @param genotyp contains the Values of the decision space Values
     * @return results returns the Objective Space Values. In this case 2 Values
     */
  public kea.Value[] score(kea.Value[] genotyp) {
      // initialize the variables containing the results
      // Kea.logger.logInfo("method score called",this);
      double a1 = 0;
      double a2 = 0;
      double f1 = 0;
      double f2 = 0;
      double actValue = 0;
      kea.Value[] results = new kea.Value[2];
      if (verify(genotyp)){
	      for (int i=0;i< genotyp.length;i++){
		  actValue = ((Number)genotyp[i].getData()).doubleValue();
		  a1 = a1 + (actValue * actValue - 10 * Math.cos(2 * Math.PI * actValue) + 10);
	      }//end of for (comuptation of a1)
	      for (int i=0;i< genotyp.length;i++){
		  actValue = ((Number)genotyp[i].getData()).doubleValue();
		  actValue = actValue - 1.5;
		  a2 = a2 + (actValue * actValue - 10 * Math.cos(2*Math.PI*actValue)+10);
	      }//end of for (computation of a2)

		  f1 = Math.sqrt(a1/genotyp.length);
	      f2 = Math.sqrt(a2/genotyp.length);
	      results[0] = new kea.Value(f1);
	      results[1] = new kea.Value(f2);
	      // Increase evaluation counter.
	      setEvalCount(getEvalCount() + 1);
      }
      else{
	  results[0] = new kea.Value(Double.MAX_VALUE);
	  results[1] = new kea.Value(Double.MAX_VALUE);
      }
      return(results);
  }//end of score

    /**
     * Returns the dimension of the decision Space
     * @return this.parameters.getParameterArray().length the dimension of the Decision Space
     */
    public int getDimDecisionSpace(){
	Parameter n = parameters.getParameterByName("n");
	if (n != null) {
	    return ((Number) n.getData()).intValue();
	}
	Kea.logger.logError("getDimDecisionSpace(): A problem occurred.", this);
	return -1;
    }
    /**
     * @return 2 returns the dimension of the objective Space(i.e. 2).
     */
    public int getDimObjectiveSpace(){
	return(2);
    }
    /**
     * returns the borders of a Variable in the decision space.
     * if no borders are specified, MIN_VALUE and MAX_VALUE of the depending Datatype are to be returned.
     * @return ranges Array of Ranges. These Ranges are all set to Double.Min_VALUE and Double.Max_VALUE.
     */
    public kea.Range[] getRanges(){
	//Kea.logger.logDebug("method call : getRanges()",this);
	int n = getDimDecisionSpace();
	Range[] ranges = new Range[n];
	for (int i = 0; i < n; i++) {
	    ranges[i] = new Range(new Double(-5.12),new Double(5.12));
	}
	return ranges;
    }
    /**
     * @return types Array of Types containing type of the variables in the decision space.
     *               (Int, Binary, Float,...)
     *               these Specifications are coded as described in the Type-object.
     */
    public kea.Type[] getDecTypes(){
      kea.Type[] types = new kea.Type[getDimDecisionSpace()];
      for (int i=0;i < getDimDecisionSpace(); i++){
        types[i] = new kea.Type(kea.Type.DOUBLE);
      }
      return(types);
    }
    /**
     * controls whether or not a to be tested result contains illegal Values.
     * @param toVerify Array of Values containing the decision space Values.
     * @return types returns a boolean, that is set to false, when a to be tested decision space
     *               Vector contains illegal Values.
     */
    public boolean verify(kea.Value[] toVerify){
	//Kea.logger.logDebug("method call : verify(kea.Value[] toVerify)",this);
	if (toVerify == null || toVerify.length != getDimDecisionSpace()){
	    Kea.logger.logDebug("please set the input values correctly",this);
	    return false;
	}
	for (int i=0; i < toVerify.length;i++){
	    if (((Number)toVerify[i].getData()).doubleValue() > 5.12 || ((Number)toVerify[i].getData()).doubleValue() < -5.12){
		Kea.logger.logDebug("The input values have to be in [-5.12,5.12]. Please correct this Error.",this);
		return  false;
	    }
	}// of for
	return(true);
    }// of verify
    /**
     * @return types Array of Types, containing the types of the variables in the objective space (Int, Binary, Float).
     *               coded as specified in the class Type
     */
    public kea.Type[] getObjTypes(){
	kea.Type[] types = new kea.Type[2];
	types[0] = new kea.Type(kea.Type.DOUBLE);
	types[1] = new kea.Type(kea.Type.DOUBLE);
	return types;
    }

    public kea.Parameters getParameter(){
	return(parameters);
    }
    public void setParameter(kea.Parameters parameters){
	if (parameters.getParameterByName("n") != null){
	    Integer integer = new Integer(0);
	    Class cla;
	    cla = parameters.getParameterByName("n").getClass();
	    if (cla != integer.getClass()){
		Kea.logger.logInfo("Wrong Parameter setting. Unable to change default parameter setting.",this);
	    }
	    else{
		if (parameters != null)
		    this.parameters = parameters;
		else
		    Kea.logger.logInfo("Received new parameters containing: null. Unable to change default parameter setting",this);
	    }
	}
	else
	    Kea.logger.logInfo("Cannot find parameter with label 'n'. Therefor unable to change default parameter setting.",this);
    }

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#writestring()
	 */
	public LinkedList writestring() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		// TODO Auto-generated method stub
		//return null;
	}
}

