/*
 * Schaffer.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.LinkedList;
import java.util.Vector;

import kea.Kea;
import kea.Parameter;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * Calculates Schaffer(1).
 * <ul>
 * <li>Schaffer's decision space is one-dimensional, 
 *     it's objective space is two-dimensional.
 * <li>Pareto optimal solutions lie in [0, 2].
 * <li>Schaffer's pareto optimal front is a connected convex curve.
 * </ul>
 *
 * @author PG 419, University of Dortmund (hp)
 * @version $Revision: 723 $
 */
public class Schaffer extends AbstractProblem {

	/**
	 * Contains the parameters for this problem.
	 */
	private Parameters param;

	/**
	 * Creates a new instance of <code>Schaffer</code>.
	 */
	public Schaffer() {
		param = new Parameters();
		param.setMembership(this.getClass().getName());
		param.addParameter("minRange", -100, "Lower border of decision space.");
		param.addParameter("maxRange", 100, "Upper border of decision space.");
	}

	/**
	 * Returns an <code>int</code> representing the dimension of decision space.
	 * @return Since decision space is one-dimensional for Schaffer(1), this method always returns <code>1<code>.
	 */
	public int getDimDecisionSpace() {
		return 1;
	}

	/**
	 * Returns an <code>int</code> representing the dimension of objective space.
	 * @return Since objective space is two-dimensional for Schaffer(1), this method always returns <code>2<code>.
	 */
	public int getDimObjectiveSpace() {
		return 2;
	}

	/**
	 * Returns an array of <code>Type</code> objects describing the decision variable types.
	 * @return a <code>Type[]</code> containing exactly one <code>Type</code> object
	 */
	public Type[] getDecTypes() {
		Type[] types = new Type[1];
		types[0] = new Type(Type.DOUBLE);
		return types;
	}

	/**
	 * Returns an array of <code>Type</code> objects describing the objective variable types.
	 * @return a <code>Type[]</code> containing exactly two <code>Type</code> objects (one for the first, one for the second objective)
	 */
	public Type[] getObjTypes() {
		Type[] types = new Type[2];
		types[0] = new Type(Type.DOUBLE);
		types[1] = new Type(Type.DOUBLE);
		return types;
	}

	/**
	 * Returns an array of <code>Range</code> objects describing
	 * the allowed ranges for each descision variable.
	 *
	 * @return an array of <code>Range</code> objects for each decision variable
	 */
	public Range[] getRanges() {
		double minRange =
			((Number) param.getParameterByName("minRange").getData()).doubleValue();
		double maxRange =
			((Number) param.getParameterByName("maxRange").getData()).doubleValue();
		Range[] ranges = new Range[1];
		ranges[0] = new Range(new Double(minRange), new Double(maxRange));
		return ranges;
	}
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			// TODO Auto-generated method stub
			return individuum;
		}
	
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	
	
	

	/**
	 * Takes an array of decision variable values
	 * and calculates the corresponding array of objective values.
	 *
	 * @param genotype a descision vector which should be scored
	 * @return the corresponding objective vector
	 */
	public Value[] score(Value[] genotype) {
		double x = ((Number) genotype[0].getData()).doubleValue();
		// Calculate f_1(x).
		double f1 = Math.pow(x, 2);
		// Calculate f_2(x).
		double f2 = Math.pow(x - 2, 2);
		// Increase evaluation counter.
		setEvalCount(getEvalCount() + 1);
		// Logging ...
		if (Kea.logger.debugOn()) {
			Kea.logger.logDebug("EvalCount = " + this.getEvalCount(), this);
			Kea.logger.logDebug(
				"x = " + x + " f_1(x) = " + f1 + " f_2(x) = " + f2,
				this);
		}
		// Return the results.
		Value[] score = new Value[2];
		score[0] = new Value(f1);
		score[1] = new Value(f2);
		return score;
	}

	/**
	 * Returns <code>false</code>, if the given array of decision variable values
	 * violates any of the problem's constraints. In the case of
	 * Schaffer(1) only the ranges are checked.
	 *
	 * @param toVerify a <code>Value[]</code> that should be verified
	 * @return <code>true</code>, if the vector's coordinate is in the allowed range
	 */
	public boolean verify(Value[] toVerify) {
		// Has toVerify the correct length?
		// Excactly one coordinate is needed.
		if (toVerify.length != 1) {
			if (Kea.logger.debugOn())
				Kea.logger.logDebug("Verify failed: Vector has wrong length.", this);
			return false;
		}
		// Check if the coordinate has the correct range.
		Range range = (this.getRanges())[0];
		if (((Number) range.getMinVal()).doubleValue()
			>= ((Number) toVerify[0].getData()).doubleValue()
			|| ((Number) range.getMaxVal()).doubleValue()
				<= ((Number) toVerify[0].getData()).doubleValue()) {
			if (Kea.logger.debugOn())
				Kea.logger.logDebug(
					"Verify failed: " + toVerify[0] + " not in allowed range.",
					this);
			return false;
		}
		return true;
	}

	/**
	 * Returns the <code>Parameters</code> object of this class.
	 * @return the (empty) <code>Parameters</code> object of this class
	 */
	public Parameters getParameter() {
		return param;
	}

	/**
	 * Sets a new <code>Parameters</code> object for this problem.
	 * @param param the <code>Parameters</code> object that will be set
	 */
	public void setParameter(Parameters p) {
		try {
			double minRange =
				((Number) p.getParameterByName("minRange").getData()).doubleValue();
			param.setParameter(
				new Parameter("minRange", minRange, "Lower border of decision space."));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("minRange set to " + minRange, this);
			}
		} catch (Exception e) {
			param.setParameter(
				new Parameter("minRange", -100, "Lower border of decision space."));
			Kea.logger.logWarn(
				"Received wrong minRange, minRange set to -100.",
				this);
		}
		try {
			double maxRange =
				((Number) p.getParameterByName("maxRange").getData()).doubleValue();
			param.setParameter(
				new Parameter("maxRange", maxRange, "Upper border of decision space."));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug("maxRange set to " + maxRange, this);
			}
		} catch (Exception e) {
			param.setParameter(
				new Parameter("maxRange", 100, "Upper border of decision space."));
			Kea.logger.logWarn("Received wrong maxRange, maxRange set to 100.", this);
		}

	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#writestring()
	 */
	public LinkedList writestring() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		// TODO Auto-generated method stub
		//return null;
	}

} // Schaffer
