/*
 * Zdt6.java  -  This file is part of KEA.
 *
 * Copyright (C) 2002/03 PG 419, Department of Computer Science, 
 * University of Dortmund.          eMail: <peter.senft@udo.edu>
 */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the KEA License Agreement which is based on 
 * the GNU General Public License as published by the Free Software 
 * Foundation; either version 2 of the License, or (at your option) 
 * any later version.  For details see License.txt.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */

package kea.problems.testproblems;

import java.util.LinkedList;
import java.util.Vector;

import kea.Kea;
import kea.Parameters;
import kea.Range;
import kea.Type;
import kea.Value;
import kea.problems.AbstractProblem;

/**
 * Title:        Testfunction ZDT6 <br>
 * Description:  The results of this Problem are not equaly spreaded across the objective Space.
 *               This is true for the
 *               not convex Pareto-optimal solutions as in the Objective Space.
 *               Near the optimum are many results of the magnitude of f1(x1) = 1.
 *               Many the not optimal solutions are far away from the optimal Solutions. <br><br>
 *
 * Implemented function: <br><br>
 *
 * minimize    f1(x1)        = 1-exp(-4x1) sin^6(6\pix1), <br>
 * minimize    f2(x1,...,xm) = g(x2,...,xm) h(f1(x1),g(x2,...,xm)), <br><br>
 *
 * whereas     g(x2,...,xm) = 1+9(( \frac{\sum{m}{i=2} xi)} {(m-1)^{0,25}}), <br>
 * and         h(f1,g)      = 1-(\frac{f1}{g})^2 <br><br>
 *
 * further is  m = 10 <br>
 * and         xi in [0,1] <br><br>
 *
 * The non konvex frontier is reached for g(x2,...,xm) = 1
 *
 * @author Stephan Sigg
 * @version $Revision: 723 $
 */
public class Zdt6 extends AbstractProblem {
	// new Parametervalue. This Value is initialized in the constructor.
	Parameters parameters;


	/**
	 * constructor
	 * the Parameters-Object is initialized here.
	 */
	public Zdt6() {
		parameters = new Parameters();
		parameters.setMembership(this.getClass().getName());
		parameters.addParameter("n", 10, "dimension of decision space");
	}
	/**
		 * Takes an array of concretely variable values and repair him by means of ???
		 * @param genotype an array of decision variable values which should be scored
		 * @return the corresponding array of objective values
		 */
		public Value[] repair(Value[]genotype)
		{
			return genotype;
		}
	/* (non-Javadoc)
		 * @see kea.problems.AbstractProblem#IndividuumRepair(java.util.Vector)
		 */
		public Vector IndividuumRepair(Vector individuum) {
			// TODO Auto-generated method stub
			return individuum;
		}

	/**
	 *
	 * computes the Objective Space Values from the decision Space Values.
	 *
	 * @param genotyp contains the Values of the decision space Values
	 * @return results returns the Objective Space Values. In this case 2 Values
	 */
	public Value[] score(Value[] genotyp) {
		//Kea.logger.logDebug("method call: score(Value[] genotyp)",this);
		Value[] results = new Value[2];
		double f1 = 0;
		double f2 = 0;
		double g = 0;
		double h = 0;
		double sum = 0;
		double value = ((Number) genotyp[0].getData()).doubleValue();
		double m = genotyp.length;
		if (verify(genotyp)) {
			// Computation of f1
			f1 =
				1
					- (Math.exp((-4) * value)
						* Math.pow((Math.sin(6 * Math.PI * value)), 6));
			// Computation of the partial function g
			value = 0;
			for (int i = 1; i < genotyp.length; i++) {
				value = ((Number) genotyp[i].getData()).doubleValue();
				sum = sum + value;
			}
			// Computataion of the partial functions g and h
			g = 1 + (9 * Math.pow((sum / (m - 1)), 0.25));
			if (Kea.logger.debugOn()) {
				Kea.logger.logDebug(
					"Calc. g(x) = " + g + " - Pareto-optimal: g(x)=1",
					this);
			}
			h = 1 - Math.pow((f1 / g), 2);
			// Computation of the function f2
			f2 = g * h;
			// Set return values.
			results[0] = new Value(f1);
			results[1] = new Value(f2);
			// Set evaluation counter,
			this.setEvalCount(this.getEvalCount() + 1);
		} else {
			results[0] = new Value(Double.MAX_VALUE);
			results[1] = new Value(Double.MAX_VALUE);
		}
		return results;
	} // End of score

	/**
	 * retruns the dimension of the decision space as integer Value
	 * @return returns the dimension of the decision space
	 */
	public int getDimDecisionSpace() {
		if (parameters.getParameterByName("n") != null) {
			// Kea.logger.logDebug("dimension of the Decision Space:"+((Number)parameters.getParameterByName("n").getData()).intValue(),this);
			return (
				((Number) parameters.getParameterByName("n").getData())
					.intValue());
		} else {
			Kea.logger.logError(
				"Please set the Dimension of the Decision Space correctly before trying to get access.",
				this);
			return -1;
		}
	}

	/**
	 * returns the dimension of the Objective Space as integer Value(i.e. 2)
	 * @return 2 ;-)
	 */
	public int getDimObjectiveSpace() {
		return (2);
	}

	/**
	 * Returns the Boundaries of every Variable of the Decision-Space as Array of Range.
	 * @return the Array of Ranges is to be returned.
	 */
	public kea.Range[] getRanges() {
		//Kea.logger.logDebug("method call: getRanges()",this);
		int range = getDimDecisionSpace();
		Range[] ranges = new Range[range];
		for (int i = 0; i < range; i++) {
			ranges[i] = new Range(new Double(0.0), new Double(1.0));
		}
		return (ranges);
	}

	/**
	 * Computes the typ of all Variables in the Decision-space as Array of Type.
	 * The encoding is described in the Class kea.Type.java
	 * @return returns this type-Array.
	 */
	public Type[] getDecTypes() {
		Type[] types = new Type[getDimDecisionSpace()];
		for (int i = 0; i < getDimDecisionSpace(); i++) {
			types[i] = new Type(Type.DOUBLE);
		}
		return (types);
	}

	/**
	 * Returns false, if a to be tested input (i.e. a Vector containing Variables of the
	 * decision space) contains not legal Values.
	 * @param toVerify Array of Values containing the to be tested decision space enries
	 * @return the Boolean result is returned and is false, if at least one input is not legal.
	 */
	public boolean verify(Value[] toVerify) {
		//Kea.logger.logDebug("method call: verify(Value[] toVerify)",this);
		if (toVerify == null || toVerify.length != getDimDecisionSpace()) {
			Kea.logger.logDebug("please set the input values correctly", this);
			return false;
		}
		for (int i = 0; i < toVerify.length; i++) {
			if (((Number) toVerify[i].getData()).doubleValue() > 1
				|| ((Number) toVerify[i].getData()).doubleValue() < 0) {
				return false;
			}
		} // of for
		return (true);
	} // of verify

	/**
	  * Returns the Value of a Variable in the Objective Space (Int, Binaer, Float...).
	  * @return retruns the type array containing the type of the variables in objective space.
	  */
	public Type[] getObjTypes() {
		Type[] types = new Type[2];
		types[0] = new Type(Type.DOUBLE);
		types[1] = new Type(Type.DOUBLE);
		return types;
	}

	public Parameters getParameter() {
		return this.parameters;
	}

	public void setParameter(Parameters parameters) {
		if (parameters.getParameterByName("n") != null) {
			Double doub = new Double(0.0);
			Class cla;
			cla = parameters.getParameterByName("n").getClass();
			if (cla != doub.getClass()) {
				Kea.logger.logError(
					"Wrong Parameter setting. Unable to change default parameter setting.",
					this);
			} else {
				if (parameters != null)
					this.parameters = parameters;
				else
					Kea.logger.logInfo(
						"Received new parameters containing: null. Unable to change default parameter setting",
						this);
			}
		} else
			Kea.logger.logInfo(
				"Cannot find parameter with label 'n'. Therefor unable to change default parameter setting.",
				this);
	}
	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#StartIndividuals(kea.Value[])
	 */
	public Value[] StartIndividuals(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}
	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#generatNewInd(kea.Value[])
	 */
	public Value[] generatNewInd(Value[] genotype) {
		// TODO Auto-generated method stub
		return null;
	}
	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#writestring()
	 */
	public LinkedList writestring() {
		// TODO Auto-generated method stub
		return null;
	}
	/* (non-Javadoc)
	 * @see kea.problems.AbstractProblem#ObjectiveName()
	 */
	public void ObjectiveName() {
		// TODO Auto-generated method stub
		//return null;
	}

}
