/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on 24.06.2004
 */
package org.nobeltje.algo;


import java.util.Iterator;
import java.util.List;
import java.util.Properties;

import org.nobeltje.testfkt.TestFunctions;
import org.nobeltje.tools.Individual;
import org.nobeltje.tools.RandomPG;
import org.nobeltje.tools.io.AlgoOutput;
import org.nobeltje.tools.io.ArgsInputBeautifier;
import org.nobeltje.tools.io.PropertiesPG;
import org.nobeltje.exceptions.MalInputFormatException;
import org.nobeltje.exceptions.NoIndividualException;

/**
 * <h2>ANLEITUNG FUER DIESE KLASSE UND DIE ERBENDEN ALGOS:</h2><br>
 * 
 * Jeder Algo sollte von AbstractAlgo erben (mit "extend AbstractAlgo" hinter dem Klassennamen).
 * Im Konstruktur sollte der Algo dann "super(args);" stehen haben (args ist dann die Variable,
 * die die Kommandozeilenparameter enthaelt, wie bei der main()). Damit wird der Konstruktor
 * von AbstractAlgo aufgerufen und viele Dinge fuer die Algos bereits erledigt. <br><br>
 * Den Algos, die von dieser abstrakten Klasse erben, werden direkt viele Variablen und 
 * Objekte zur Verfuegung gestellt. Wenn diese auch verwendet werden sollen, so duerfen sie
 * in der erbenden Klasse NICHT ueberschrieben, also neu deklariert oder definiert werden: 
 * <br><br>
 * - <b>test</b>: Objekt der Testfunktion (bereits fertig eingebunden)<br>
 * - <b>rand</b>: Objekt fuer Zufallszahlen vom Typ RandomPG (bereits, ggf. mit Random-Seed, erstellt) <br>   
 * - <b>maxGenerations</b>: int, Anzahl der Iterationen/Generationen, ueber die der Algo laufen soll <br>
 * - <b>specificParameters</b>: java.util.Properties, mit Parametern fuer Algo <br>
 *   - sowohl aus Kommandozeile und Config-Datei, wobei die Kommandozeile Vorrang hat
 *   (Fuer einfachen Zugriff auf Properties gibt es die Methoden
 *    ("defaultwert" wird zurueckgegeben, wenn Fehler auftritt): <br>
 *			static int ArgsInputBeautifier.getIntProperty(properties, feldname, defaultwert, abbrechen wenn fehler) <br>
 *			static double ArgsInputBeautifier.getDoubleProperty(properties, feldname, defaultwert, abbrechen wenn fehler) <br>
 *			static String ArgsInputBeautifier.getStringProperty(properties, feldname, defaultwert, abbrechen wenn fehler) <br>
 * - <b>outputFileNamePrefix</b>: String, Praefix fuer die Ausgabedateien 
 * 	(daran wird fuer eine Ausgabe eine Suffix, z.B. ".best.res" angehaengt) <br>
 * - <b>debug</b>: boolean, Debug-Meldungen ausgeben (z.B. Konsolen-LiveOutput) <br> <br>
 * 
 * <b>Anmerkung zu den Parametern</b>: In der Datei muessten natuerlich die "-" weggelassen werden,
 *  also waere eine Zeile in der Parameterdatei z.B. "g 200" <br><br>
 * <b>Notwendige Parameter</b> bei Aufruf entweder in Parameter-Datei oder ueber die Kommandozeile: <br>
 * -t Testfunktion <br>
 * -g Anzahl Generationen	-> maxGenerations <br> <br>
 * 
 * <b>Optionale Parameter</b>: <br>
 * -c Name Config-Datei <br>
 * -r Random-Seed <br>
 * -o Namespraefix Ausgabedateien	-> outputFileNamePrefix <br>
 * -d Debug-Infos ausgeben <br> <br>
 * -live int, Abstand, wieviele Evaluationen zwischen 2 liveOutputs liegen soll. 
 *   <br>Es werden dann .best.res- und .best.in-Dateien geschrieben, bei denen im Dateinamen "_evaluations=Y" ergaenzt wird (z.B. "g=100_live=10_evaluations=60.best.in").
 * 
 * 
 * @author $Author: noot $
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 */

public abstract class AbstractAlgo implements AlgoInterface {
	
	/**
	* Fuer Sammlung aller Kommandozeilenparameter, die fur den Algo bestimmt sind; 
	* Speicherung von Strings, wo jeweils 2 miteinander assoziiert sind (Schluessel + Wert)
    * specificParameters = Algo-spezifische Parameter
	*/ 
	protected PropertiesPG specificParameters = new PropertiesPG(); 
	
	/**
	* Fuer Sammlung aller Kommandozeilenparameter fuer die Testfunktion; 
	* Speicherung von Strings, wo jeweils 2 miteinander assoziiert sind (Schluessel + Wert)
    * specificTestfktParameters = Testfunktions-spezifische Parameter
	*/ 
	protected PropertiesPG specificTestfktParameters = new PropertiesPG(); 
	
	/**
	 * Name der Testfkt-Klasse
	 */
	private String classnameTestfkt;
	
	/**
	 * Objekt der Testfunktion
	 */
	protected TestFunctions test;
	
	/**
	 * Maximale Anzahl der Generationen (exogen)
	 */
	protected int maxGenerations; 
	
	/**
	 * Debug-Ausgaben, wie LiveOutput, etc. ausgeben?
	 */
	protected boolean debug = false;
	
	/**
	 * Dateinamenprefix fuer Ausgabedateien
	 */
	protected String outputFileNamePrefix = ""; 
	
	/**
	 * Zufallszahlen-Objekt
	 */
	protected RandomPG rand; 
	
	/**
	 * Zaehler fuer Evaluations-Aufrufe (Testfkt und Simulatoren)
	 */
	protected int evaluationCount = 0;
	
	/**
	 * Evaluations-Intervall zwischen liveOutput 
	 * (alle evaluationsOutputIntervall Evaluationen den liveOutput ausgeben)
	 */
	protected int evaluationsOutputIntervall = 0;
	
	
	/**
	 * Einfacher leerer Standard-Konstruktor() 
	 */
	public AbstractAlgo() {
		super();
	}

	
	/**  
	 * Einfacher Standard-Konstruktor, den jede erbende Klasse mit "super();" aufrufen sollte (s. Klassenbeschreibung) 
	 * Wertet die wichtigsten uebergebenen Parameter, die auch in dieser Klasse
	 * deklariert werden, aus und instanziiert Testfunktion.
	 * 
	 * @param args Parameter, die main() ueber die Kommandozeile erhalten hat
	 */
	public AbstractAlgo(String[] args) {
		
		java.util.Properties prop = new Properties();
		String configFileName = "";

		/* Argumente in Tabelle/property umwandeln */
		try {
			prop = ArgsInputBeautifier.beautify(args);
		} catch (MalInputFormatException e) {
			e.printStackTrace();
			System.exit(1);	// Direkter Abbruch des Programms			
		}
		
		try {
			
			/* Algo-Parameter aus Config-Datei und nochmal Kommandozeile (ueberschreibt Dateiparameter) holen */
			if (prop.containsKey("c")) {
				configFileName = prop.getProperty("c") + ".param";
				// Parameter aus Datei einlesen
				specificParameters = (PropertiesPG)ArgsInputBeautifier.getArgsFromFile(configFileName);
			}

			/* Testfkt-Parameter aus Config-Datei und nochmal Kommandozeile (ueberschreibt Dateiparameter) holen */
			if (prop.containsKey("ttc")) {
				configFileName = prop.getProperty("ttc") + ".param";
				// Parameter aus Datei einlesen
				specificTestfktParameters = (PropertiesPG)ArgsInputBeautifier.getArgsFromFile(configFileName);
			}

			// Parameter aus Kommandozeile erneut einlesen und Datei-Parameter ggf. ueberschreiben
			/* Auskommentiert wegen Umstellung auf 2 Properties-Objekte fur algo+testfkt
			 specificParameters = (PropertiesPG)ArgsInputBeautifier.beautify(specificParameters, args);
			 */			 
			ArgsInputBeautifier.beautify(specificParameters, specificTestfktParameters, args);

			/* Variabeln aus Properties-Tabelle setzen */
			// Name der Testfunktion
			classnameTestfkt = specificParameters.getStringProperty("t", "", true);
			// Maximale Anzahl Generationen
			maxGenerations = specificParameters.getIntProperty("g", 0, true);
			// Evaluationen-Intervall zwischen liveOutput 
			evaluationsOutputIntervall = specificParameters.getIntProperty("live", 0, false);
			// Debug-Meldungen ausgeben?
			if (specificParameters.containsKey("d") && (specificParameters.getProperty("d")).equals("true")) {
				debug=true;
			} else {
				debug= false; 
			}
			// Random-Seed
			if (specificParameters.containsKey("r")) {
				rand = new RandomPG(specificParameters.getIntProperty("r", 0, true));
			} else {
				rand = new RandomPG();
			}
			// Praefix fuer Namen der Ausgabedateien
			if (specificParameters.containsKey("o")) {
				outputFileNamePrefix = specificParameters.getProperty("o");
			} else {
				outputFileNamePrefix = AlgoOutput.getFileNameByDate();
			}
			
		} catch (Exception e1) {
			System.err.println("Eingabe ist nicht komplett oder fehlerhaft");
			e1.printStackTrace();
		}
		
		/* Testklasse dynamisch einbinden */
		try {
			// Pfandangaben zur Testfkt-Klasse zu Packetangaben aendern
			classnameTestfkt = classnameTestfkt.replace('/', '.');
			classnameTestfkt = classnameTestfkt.replace('\\', '.');
			// Testklasse anhand des Namens anbinden
			Class testClass = Class.forName(classnameTestfkt);
			// Objekt vom Typ Objekt aus Testklasse erzeugen  
			Object testObject = testClass.newInstance();  
			// Objekt nun zu Interface TestFunktions casten, damit die Methoden zur Verfuegung stehen
			test = (TestFunctions)testObject;  
		}
		/* Fehler bei nicht gefundener Testklasse ausgeben und Abbrechen */
		catch (ClassNotFoundException e){
			System.out.println("\nFEHLER: Angegebene Testklasse konnte nicht gefunden werden.\n\n");
			e.printStackTrace();
			System.exit(1);	// Direkter Abbruch des Programms
		}
		/* Fehler bei unzulaessiger Testklasse ausgeben und Abbrechen */
		catch (ClassCastException e){
			 System.out.println("\n FEHLER: Angegebene Testklasse ist nicht zulaessig (Cast-Fehler). \n\n");
			 e.printStackTrace();
			 System.exit(1);	// Direkter Abbruch des Programms
		} 
		/* Fehler ausgeben und Abbrechen, wenn Testklasse nicht instanziiert werden kann */
		catch (InstantiationException e2) {
	 		 e2.printStackTrace();
			 System.exit(1);	// Direkter Abbruch des Programms
		}
		/* Noch nen Fehler.. */
		catch (IllegalAccessException e2) {
			 e2.printStackTrace();
			 System.exit(1);	// Direkter Abbruch des Programms
		}
		
	}
	

	/**
	 * Start-Routine des Algos. Wird von main-Methode der Algo-Klasse aufgerufen.
	 * Dieser uebernimmt die gesamte Berechnung des Algos, d.h. ruft alle n?tigen
	 * Methoden auf.
	 * Hieraus wird hoechstens der LiveOutput aufgerufen, der 'grosse' Output jedoch
	 * von der Routine gestartet, die auch diese Methode 'startAlgo()' aufruft.
	 * 
	 * Bitte am Ende der Methode den Aufruf test.endAlgo() einfuegen, damit der 
	 * Temperierbohrungssimulator ohne Fehlermeldung verlassen werden kann.
	 *   
	 * @return boolean Erfolgswert, wird aber glaube ich meist einfach auf true gesetzt
	 * @throws Exception
	 */
	public abstract boolean startAlgo() throws Exception;
	
	
	/**
	 * Routine fuer LiveOutput auf Konsole der Algorithmen (Generation, Objekt- und Strategievariablen, Fitnesswerte)
	 *  
	 * @param archive Liste auszugebender Individuen
	 * @param currentGeneration Nummer der aktuellen Generation als Ueberschrift
	 */
	public void liveOutput(List archive, int currentGeneration) {
		 try 
		 {
			StringBuffer row=new StringBuffer();
			
			row.delete(0,row.length());
			System.out.println(row.toString());
			System.out.println("#### Archive ####");
			row.append("*** INTERATION "+ currentGeneration);
			System.out.println(row.toString());
			
			for (Iterator iter = archive.iterator(); iter.hasNext();)
			{
				Individual ind = (Individual)iter.next();
				System.out.println("--------------- " + ind.getGeneration() + " ---------------");
				/* Objektvariablen */
				List obj = ind.getObjectVariables();
				row.delete(0,row.length());
				row.append("Obj: ");
				for (Iterator objIter = obj.iterator(); objIter.hasNext();)
				{
					Double value = (Double) objIter.next();
					row.append(value.toString()+" ");
				}
				System.out.println(row.toString());

				/* Strategievariablen */
				List str = ind.getStrategyVariables();
				row.delete(0,row.length());
				row.append("Str: ");
				for (Iterator strIter = str.iterator();strIter.hasNext();)
				{
					Double value = (Double) strIter.next();
					row.append(value.toString()+" ");
				}
				System.out.println(row.toString());

				/* Fitness */
				List fit = ind.getFitness();
				row.delete(0,row.length());
				row.append("Fit: ");
				for (Iterator fitIter = fit.iterator();fitIter.hasNext();)
				{
					Double value = (Double) fitIter.next();
					row.append(value.toString()+" ");
				}
				System.out.println(row.toString());
			}
		 }
		 catch (Exception e) 
		 {
			 e.printStackTrace();
		 }
		 System.out.println("--------------------");
	}
	
	
	/**
	 * Methode, die die Algos nach jeder Evaluation aufrufen sollen.<br>
	 * Hierin wird das interne Attribut this.evaluationCount um 1 erhoeht.
	 * Wenn dies dann ein Vielfaches des Attributs evaluationsOutputIntervall ist, 
	 * werden die Objektvariablen und Fitnesswerte der uebergebenen Individuen in Datei geschrieben.
	 *
	 * @param obj List Individuen fuer liveOutput
	 */
	protected void EvaluationBasedLiveOutput(List individuals) {
		if (evaluationsOutputIntervall > 0) {
			/* Evaluation-Count inkrementieren */
			evaluationCount++;
			/* Pruefen, ob liveOutput erfolgen soll */
			if ((evaluationCount % evaluationsOutputIntervall) == 0)  {
				/*
				// Auskommentert, weil ueberholt
				StringTokenizer prefixToken = new StringTokenizer(outputFileNamePrefix,"_");
				StringBuffer newOutputFileNamePrefix = new StringBuffer(prefixToken.countTokens());
				while (prefixToken.hasMoreElements()) {
					String prefix = prefixToken.nextToken();
					if ( !prefix.startsWith("g=") ) {
						newOutputFileNamePrefix.append(prefix);
						newOutputFileNamePrefix.append("_");
					}						
				}
				*/
				try {
					/* Archiv (Objektvariablen) in Textdatei ausgeben */
					AlgoOutput.writeFitness(individuals, outputFileNamePrefix+"_evaluations="+evaluationCount, false);
					/* Archiv (Fitnesswerte) in Textdatei ausgeben */
					AlgoOutput.writeObjectVariables(individuals, outputFileNamePrefix+"_evaluations="+evaluationCount, false);
				}
				catch (NoIndividualException e) {
					System.out.println("Fehler beim Evaluationen-basierten liveOutput:");
					e.printStackTrace();
				}
				
				
			}
		}
	}
	
	
}
	
	
	

