/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on Jun 16, 2004
 */
package org.nobeltje.algo;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.nobeltje.tools.Individual;
import org.nobeltje.tools.ParetoArchive;
import org.nobeltje.tools.io.AlgoOutput;


/**
 * <p>Dies ist eine (mue/rho kappa lambda) - Evolutionsstrategie mit sigma Selbstadaption</p>
 * Implementiert von Daniel Saltmann, Kay Thielmann und Igor Tipura
 * 
 * Folgende Parameter erwartet der Algorithmus:
 * <ul>
 * 	<li>-mue 15 groesse der Elterngeneration</li>
 * 	<li>-rho 3 Anzahl der Eltern, die in die Rekombination eines Offspring-Individuums einfliessen</li>
 * 	<li>-lambda 100 Groesse der Offspringgeneration</li>
 * 	<li>-kappa ? Lebensdauer eines Individuums</li>
 * 	<li>-tau1 0.1</li>
 * 	<li>-tau2 0.1</li>
 * 	<li>-testfkt testfkt.FonsecaF1 Angabe der Testfunktion</li>
 * 	<li>-g 1000 Anzahl der Generationen</li>
 * 	<li>-archive 100 Anzahl der Individuen die maximal ausgespuckt werden</li>
 * 	<li>-d fuer Debugausgaben</li>
 * 	<li>-r ? Random-Seed, nicht notwenig, Zahl variabel waehlbar!</li>
 * </ul>
 * 
 * 
 * <br><b>Benoetigte Testfunktionsaufrufe:</b>
 * <br>- Initialisierung : mue
 * <br>- Pro Generation  : lambda * #Generationen
 * 
 * 
 * @author $Author: noot $
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 */

public class MueRhoLES extends AbstractAlgo {
	
    /**
     * Anzahl der Individuen in der Elterngeneration (exogen)
     */
	private int mue;
	
	/**
	 * Anzahl der Eltern eines Individuums (exogen)
	 */
	private int rho;
	
	/**
	 * Anzahl der Individuen in der Off0pringgeneration (exogen)
	 */
	private int lambda;
	
	/**
	 * Lebensdauer eines Indiviuums (1=Komma-Selektion, 0=Plus-Selektion) (exogen)
	 */
	private int kappa; 
	
	/**
	 * Standardabweichung fuer den festen Teil der Mutation der Strategievariablen (exogen)
	 */
	private double tau1;
	
	/** 
	 * Standardabweichung fuer den variablen Teil der Mutation der Strategievariablen (exogen)
	 */
	private double tau2; 
	
	/**
	 * Maximale Anzahl der Individuen im Archiv (exogen)
	 */
	private int globalArchiveSize; 
	
	/**
	 * Archiv-Objekt: Sammlung aller nichtdominierten Individuen
	 */
	private List globalArchive; 
	

	/**
	 * Leerer Konstruktor, der nur den Konstruktor der Superklasse (AbstractAlgo) aufruft. (Wird fuer Help-Klasse benoetigt.)
	 */
	public MueRhoLES() {
	    super();
	}
	
	
	/**
	 * Konstruktor: ruft den Konstruktor der Superklasse (AbstractAlgo) auf und...
	 * liesst Algo-spezifische Parameter aus specificParameters ein
	 * 
     * @param args String-Array der Kommandozeilenparameter
	 */
	public MueRhoLES (String [] args) {
		
		/* Konstruktor von AbstractAlgo aufrufen um wichtigste Dinge zu setzen */
		super(args);
		
		/* Algo-Spezifische Variablen setzen */
		globalArchive = new ArrayList();		  
		globalArchiveSize = specificParameters.getIntProperty("archive", 0, false);
		mue = specificParameters.getIntProperty("mue", 0, true);
		rho = specificParameters.getIntProperty("rho", 0, true);
		lambda = specificParameters.getIntProperty("lambda", 0, true);
		tau1 = specificParameters.getDoubleProperty("tau1", 0, true);
		tau2 = specificParameters.getDoubleProperty("tau2", 0, true);
		kappa = specificParameters.getIntProperty("kappa", 0, true);
		
	}


	/** 
	 * Gibt eine Kurzbeschreibung des Algos zurueck.
	 * 
	 * (non-Javadoc)
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
	    String out;

	    out =  "\n\nKlasse MueRhoLES\n";
	    out += "=================\n";
	    out += "\n";
	    out += "Beschreibung:\n";
	    out += "Implementation einer MUE / RHO KAPPA LAMBDA Evolutionsstrategie" +
	    		" mit SIGMA-Selbstadaption.\n";
	    out += "Aus einer Elternpopulation mit MUE Individuen wird eine Offspringgeneration" +
	    		" mit LAMBDA Individuen generiert. Jedes Kind wird aus RHO Individuen rekombiniert" +
	    		" und anschliessend mutiert.\n";
	    out += "Ausserdem hat jedes Individuum fuer jede Objektvariable" +
	    		" eine Strategievariable, die die Mutation der Objektvariable als Standardabweichung" +
	    		" beeinflusst. Die Mutation der Strategievariablen erfolgt logarithmisch, kombiniert" +
	    		" aus einem Wert, der bei einem Mutationsschritt fuer alle Strategievariablen" +
	    		" eines Indiviuums gleich ist und einem anderen, der fuer jede Strategievariable" +
	    		" unabhaengig gewaehlt wird.\n";
	    out += "Die Selektion erfolgt mittels Einteilung in Fronten nicht dominierter Individuen und" +
	    		" laeuft sehr aehnlich der vom NSGA-II ab. Die Lebenszeit eines Individuums" +
	    		" ist durch KAPPA begrenzt (0=Plus-Selektion, 1=Komma-Selektion)\n";
	    out += "Ausgegeben werden nur nicht dominierte Elemente.\n";
	    out += "\n";
	    out += "Kommandozeilenparameter:\n";
	    out += "-t: Klasse der Testfunktion (z.B. testfkt.FonsecaF1)\n";
	    out += "-g (int): Anzahl der Generationen/Iterationen\n";
	    out += "-r (int): optional, Random-Seed (Initialisierung der Zufallsklasse)\n";
	    out += "-o: optional, Praefix fuer die Dateinamen der Ausgabe (An diese wird noch" +
	    		" das Kuerzel fuer die Art der Ausgabedatei angehaengt, z.B. .best.res);" +
	    		" Wird der Wert nicht gesetzt, so wird ein Praefix aus Datum und Uhrzeit gewaehlt\n";
	    out += "-d: optional, wenn gesetzt, werden Debug-Ausgaben angezeigt\n";
		out += "-live -> Abstand, wieviele Evaluationen zwischen 2 liveOutputs liegen soll.\n";
	    out += "-archive (int): optional, Maximalgroesse des globalen Archives, welches am Ende ausgegeben wird; 0=unbegrenzt (Standardwert)\n";
	    out += "-mue (int): Anzahl der Individuen der Elternpopulation\n";
	    out += "-rho (int): Anzahl der Eltern eines Individuums\n";
	    out += "-lambda (int): Anzahl der Individuen der Offspringpopulation\n";
	    out += "-tau1 (int): Standardabweichung fuer Mutation der Strategievariablen (fuer alle" +
	    		" Strategievariablen eines Individuums)\n";
	    out += "-tau2 (int): Standardabweichung fuer Mutation der Strategievariablen (fuer jede" +
		" einzelne Strategievariablen eines Individuums)\n";
	    out += "-kappa (int): Maximales Alter eines Individuums  (0=Plus-Selektion, 1=Komma-Selektion)\n";
	    out += "\n\n";
	    
	    return out;
	}
	
	
/**
 * <p>Startroutine des Algorithmus</p>
 * <p>
 * Die Startgeneration wird generiert. Daraus werden durch Rekombination und Mutation
 * Individuen fuer die Offspringgeneration erzeugt.<br />
 * Daraus wird durch die Selektion die neue Elterngeneration bestimmt.<br />
 * Das wird wiederholt solange currentGenerationCount kleiner als maxGenerations.<br />
 * </p>
 *
 * @return gibt <code>true</code> zur?ck bis die maximale Generationszahl erreicht wird
 * @throws exception
 *
 * (non-Javadoc)
 * @see algo.AbstractAlgo#startAlgo()
 */
	public boolean startAlgo() throws Exception {
		test.init(specificTestfktParameters);
		
		// Initialisierung
		int currentGenerationCount = 0;
		
		// Startgeneration generieren
		List parentGeneration = generateFirstParentGeneration();
		List offspringGeneration = new ArrayList(); 

		do {
			currentGenerationCount++;
			offspringGeneration.clear();
			
			for (int i=1; i<=lambda; i++) {
				List marriagedParents = marriage(parentGeneration, rho);
				Individual ind = recombinate (marriagedParents, currentGenerationCount);
				mutateStrategyVariables(ind);
				mutateObjectVariables(ind);
				/* Funktionswerte/Fitness setzen */
				ind.setFitness ((ArrayList)test.execute((List)ind.getObjectVariables()));
	  			
				/* Evaluationsbasierter LiveOutput */
	  			EvaluationBasedLiveOutput(globalArchive);
				
	  			offspringGeneration.add(ind);
			}
			
			/* Selektion */
			parentGeneration = select(offspringGeneration, parentGeneration, currentGenerationCount);
			/* Mini-Live-Output (Punkte) wenn kein ausfuehrlicher LiveOutput (debug=false) */
			if (!debug) {
				System.out.print(".");
				if (currentGenerationCount % 100 == 0) System.out.println(currentGenerationCount);
			} 
		}
		while (currentGenerationCount < maxGenerations);
		System.out.println("");

		globalArchive = (ArrayList)test.postProcessing(globalArchive);
		test.finalize();
		
		return true;
	}
	
	
	/**
	 * Generiert eine erste Elternpopulation und gibt diese zurueck 
	 * (gibt ggf. eine Exception von Methoden der Testfunktion weiter)
	 *  
	 * @return List Erste zufaellig generierte Parent-Generation 
	 * @throws Exception bei irgendwelchen, z.B. Cast-Problemen oder Testfunktions-Fehlern
	 */
	private List generateFirstParentGeneration() throws Exception {
		List result = new ArrayList(); // Liste fuer Start-Generation initialisieren 
		
		// Grenzen fuer initiale Objektvariablen aus Testfunktion holen
		List bounds = test.getOptimalBounds();
		
		// Schleife ueber alle MUE benoetigten Individuen 
		for (int i=0; i<mue; i++) {
			List objectValues = new ArrayList(); // Liste fuer Objektvariablen eines Ind. initialisieren
			List strategyValues = new ArrayList(); // Liste fuer Strategievariablen eines Ind. initialisieren
			// Schleife ueber die Dimensionen (Anzahl Objektvariablen) des Individuums
			for (int j=0; j<test.getDimension(); j++) {
				double min = ((Double)bounds.get(2*j)).doubleValue(); // Min-Grenze fuer ObjVar aus Testfkt
				double max = ((Double)bounds.get(2*j+1)).doubleValue(); // Max-Grenze fuer ObjVar aus Testfkt
				// Objektvariable setzen: Minimum + ZufallsIntegerZahl (max-min) + Double-Wert zwischen 0 und 1
				objectValues.add(new Double(rand.nextDouble(min, max)));
				// Strategievariable setzen: ZufallsDoubleZahl zwischen 0,0001 und 5 ([0,0001,5[)
				strategyValues.add(new Double(rand.nextDouble(0.0001, 5)));
			}
			// Neues Individuum-Objekt erzeugen, dafuer Testfkt aufrufen und Generation 0 waehlen
			Individual ind = new Individual ((ArrayList)objectValues, (ArrayList)strategyValues, (ArrayList) test.execute(objectValues), 0);
//			Individual ind = new Individual ((ArrayList)objectValues, (ArrayList)strategyValues, new ArrayList(), 0);
						// Individuum zur Genereation hinzufuegen
			result.add(ind);
		}
		return result;
	}
	
	
	/**
	 * Hochzeitsfunktion fuer die Rekombination
	 * Aus der uebergebenen Parent-Generation werden zufaellig "count" viele Individuen
	 * ausgesucht und als Liste zurueckgegeben.
	 * 
	 * @param parentGeneration Parent-Generation, aus der "count" viele Individuen gewaehlt werden sollen 
	 * @param count Anzahl der Eltern eines Individuums
	 * @return ArrayList Liste der Indiviuen, aus denen ein neues Ind. rekombiniert werden soll
	 */
	private List marriage (List parentGeneration, int count) {
		List marriedParents = new ArrayList ();
		for (int i=0; i<count; i++) {
			marriedParents.add (parentGeneration.get(rand.nextInt(parentGeneration.size())));
		}
		return marriedParents;		
	}
	
		
	/**
	 * Diskrete Recombination
	 * Fuer jedes Objekt/Strategievariablenpaar wird zufaellig eins
	 * von rho Eltern ausgewaehlt, dessen Wert dann fuer das neue Individuum uebernommen wird
	 * Die Fitness wird noch nicht gesetzt, da das Individuum erst noch durch die Mutation
	 * durch muss.
	 * 
	 * @param selectedParents Fuer Rekombination ausgewaehlte Eltern
	 * @param currentGeneration Nummer der aktuellen Generation fuer das neue Ind.
	 * @return Individual Rekombiniertes Individuum
	 */
	private Individual recombinate (List selectedParents, int currentGeneration) {
		List objectVariables = new ArrayList();
		List strategyVariables = new ArrayList();
		//Schleife ueber die Anzahl der Objektvariablen (Dimension)
		for (int i = 0 ; i<test.getDimension() ; i++) {
			//Auswahl eines Individuums aus den rho Eltern
			Individual ind = (Individual) selectedParents.get(rand.nextInt(rho));
			//einfuegen der i-ten Objekt- und Strategievariablen in die entsprechenden
			//Listenisten des neuen Individuums 
			objectVariables.add(i,ind.getObjectVariables().get(i));
			strategyVariables.add(i,ind.getStrategyVariables().get(i));			
		}
		//Rueckgabe des rekombinierten Indiviuums mit hochgezaehlter Generation!
		//aber OHNE FITNESSWERT !!!!!!!!!!!
		Individual result = new Individual (objectVariables, strategyVariables, null, currentGeneration);
		return result;
	}


	/**
	 * Mutieren der Strategievariablen durch Multiplikation mit e^(tau1RandomValue + tau2*Nk(0,1))
	 * D.h.: tau1RandomValue ist ein mit tau1 gewichteter Zufallswert, der fuer jede Strategievarieble
	 * eines Individuums gleich bleibt. 
	 * Der dazu addierte Teil tau2*Nk(0,1) ist ein normalverteilter Wert mit Erwartungswert 0 und
	 * Standartabweichung tau2, der fuer jede Strategievariable neu bestimmt wird.
	 *  
	 * @param ind Individuum, dessen Strategievariablen mutiert werden sollen
	 */
	private void mutateStrategyVariables (Individual ind) {
		/* Neue Strategievariablen */
	    List newStrategyVariables = new ArrayList();
		
		/* Teil der Standardabweichung, der fuers ganze Indiviuum fest ist */
		double tau1RandomValue = tau1*rand.nextGaussian();
		
		for (Iterator iter = ind.getStrategyVariables().iterator(); iter.hasNext();) {
			double oldValue = ((Double) iter.next()).doubleValue();
			double tau2RandomValue = tau2*rand.nextGaussian();
			double newValue = -1;
			newValue = oldValue * (Math.exp(tau1RandomValue + tau2RandomValue));
			newStrategyVariables.add (new Double (newValue));
			/* Die Strategievariablen sollen groesser als 0 sein.
			 * Falls sie jedoch mal kleiner gleich 0 werden (was laut den Betreuern nicht
			 * vorkommen kann) Fehlermeldung ausgeben */
			if (newValue <= 0) {
			    System.out.println("!!!! FEHLER, DER NICHT AUFTRETEN DARF: Eine Strategievariable ist nach Mutation: "+newValue);
			}
		}
		ind.setStrategyVariables(newStrategyVariables);
	}
	
	
	/**
	 * Mutieren der Objektvariablen
	 * durch Addition eines mit der entsprechenden Strategievariablen gewichteten
	 * Gauss-verteilten Zufallswert (Erwartungswert 0)
	 * 
	 * @param ind Individuum, dessen Objektvariablen mutiert werden sollen
	 */
	private void mutateObjectVariables (Individual ind) {
		List newObjectVariables = new ArrayList();
		for (int i=0; i<test.getDimension(); i++) {
			double oldValue = ((Double)ind.getObjectVariables().get(i)).doubleValue();
			double strategySigma = ((Double)ind.getStrategyVariables().get(i)).doubleValue();
			newObjectVariables.add (new Double (oldValue + (strategySigma * rand.nextGaussian())) );
		}
		ind.setObjectVariables(newObjectVariables);
	}
	
	
	/**
	 * Selektion: Es sollen aus lambda Indiviuduen nur mue viele ausgewaehlt werden, die die neue Elternpopulation bilden. 
	 * Diese wird zurueckgegeben.
	 * Hierbei kommt die Selektion nach Alter mittels kappa (Maximal-Alter der Individuen ist kappa) und 
	 * die Fronteneinteilung aus NSGA-II zum Einsatz.
	 * 
	 * @param offspringGeneration Nachkommenspopulation 
	 * @param parentGeneration Elternpopulation 
	 * @param currentGenerationCount Nummer der aktuellen Generation fuer Lebenszeitberechnung mit kappa
	 * @return List Neue Elternpopulation
	 */
	private List select (List offspringGeneration, List parentGeneration, int currentGenerationCount) {
		/* Initialisiere neue leere Population, die spaeter zur neuen Elternpopulation wird */
		List result = new ArrayList();
		/* Mische Eltern- und Offspringgeneration und selektiere spaeter mit Kappa die zu alten raus */
		parentGeneration.addAll(offspringGeneration);
		
		/* Kappa-Selektion (kappa=1 == Komma-Selektion) - zu alte Individuen werden geloescht */
		if (kappa>0) {
			for (int i=parentGeneration.size()-1; i>=0; i--) {
				if ( ((Individual)parentGeneration.get(i)).getGeneration() + kappa - 1 < currentGenerationCount ) {
					parentGeneration.remove(i);
				}
			}
		} else {
			// + Selektion, also keine Individuen loeschen
		}
		
		/* Hilfsvariable: Die erste Front (nicht dominierten Individuen) soll zum globalen
		 * Archiv hinzugefuegt werden. Nach dem Hinzufuegen wird first=false gesetzt, damit die
		 * weiteren Fronten nicht mehr hinzugefuegt werden. */ 
		boolean first = true;
		
		/* Hauptschleife zum Hinzufuegen:
		 * Die nicht dominierten Elemente/Individuen bilden die erste Front, also die Paretofront.
		 * Dann werden diese Ind. zur neuen Elternpopulation (result) hinzugefuegt und aus der
		 * alten Population geloescht. Dort werden wieder die nicht dominierten Ind. gesucht, die
		 * die naechste Front darstellen. Auch diese werden hinzugefuegt und geloescht.
		 * Sollten mehr Ind. in einer Front sein, als fuer die neue Elterngeneration noch benoetigt
		 * werden (insgesamt mue viele), so werden zufaellig so viele ausgewaehlt, wie benoetigt
		 * werden. (s. NSGA-II-Selektion). */
		while (result.size() < mue) {
			/* Initialisieren der Front */
			List front = new ArrayList();
			
			/* Jedes Ind. wird einzeln hinzugefuegt, falls es nicht dominiert wird.
			 * Die dominierten Elemente der Front werden automatisch geloescht.
			 * Als Archivgroesse wird mue+lambda gewaehlt, weil zu aus vielen Ind. spaeter
			 * zufaellig selektiert wird. */ 
			for (Iterator iParents = parentGeneration.iterator(); iParents.hasNext();) {
				Individual current = (Individual)iParents.next();
				front = ParetoArchive.refreshArchive(mue+lambda, (ArrayList)front, current);				
			}
		
			/* Die nun zur aktuellen Front gehoerenden Ind. werden aus der alten Population geloescht */ 
			parentGeneration.removeAll(front);
			
			/* Nicht dominierten Individuen zum globalen Archiv hinzuf?gen und dort alle
			 *  dominierten Individuen loeschen */
			if (first) {
				for (Iterator iterFront = front.iterator(); iterFront.hasNext();) {
					globalArchive = ParetoArchive.refreshArchive(globalArchiveSize, (ArrayList)globalArchive, (Individual)iterFront.next());
				}
				first = false;
			}
			
			/* Pruefen der Front-Groesse: sind in der Front mehr Ind. als noch benoetigt? */
			if (front.size() > (mue-result.size())) {
				/* Die Front hat mehr Indvidiuen als benoetigt.
				 * Es werden zufaellig so viele Individuen ausgewaehlt, wie noch benoetigt werden.
				 *  wird zur neuen Elternpopulation hinzugefuegt */
				result.addAll(marriage(front, mue-result.size()));
			} else {
				/* Die gesamte Front wird zur neuen Elternpopulation hinzugefuegt */
				result.addAll(front);
			}
			
		}

		/* nur debug-ausgaben */
		if (debug && (currentGenerationCount % 1 == 0)) {
			liveOutput(result,currentGenerationCount);
		}
		
		if (debug && (currentGenerationCount % 1 == 0)) {
			liveOutput(globalArchive,currentGenerationCount);
		}
		
		/* Rueckgabe */
		return result;
		
	}
	

	/**
	 * Mainmethode.
	 * 
	 * Hier wird ein neues Objekt erzeugt, mit der Methode startAlgo aufgerufen.
	 * Nach dem Durchlauf werden die Fitnesswerte der Individuum im  Archiv in eine
	 * bestimmte Datei geschrieben, ebenso die Objektvariablen der Archivindividuen.
	 * main-
	 * 
     * @param args String-Array der Kommandozeilenparameter
	 * @throws Exception
	 */
	public static void main(String[] args) throws Exception {
		
	   	// Algorithmus-Objekt erzeugen	
		MueRhoLES mrl = new MueRhoLES(args);
	   	// Algorithmus starten
		mrl.startAlgo();
	   	/* Archiv (Objektvariablen) Textdatei ausgeben */
		AlgoOutput.writeFitness(mrl.globalArchive, mrl.outputFileNamePrefix, false);
		/* Archiv (Fitnesswerte) in Textdatei output.fitness ausgeben */
		AlgoOutput.writeObjectVariables(mrl.globalArchive, mrl.outputFileNamePrefix, false);
		
	}
}
