/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /**
 * Created on June 17, 2004
 *
 */
package org.nobeltje.algo;

import java.io.FileWriter;
import java.io.IOException;
import java.text.NumberFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.nobeltje.tools.Individual;
import org.nobeltje.tools.ParetoArchive;
import org.nobeltje.tools.io.AlgoOutput;
import org.nobeltje.exceptions.InvalidBoundException;


/**
 * Raeuber-Beute-Algorithmus
 *
 *
 * <br><b>Benoetigte Testfunktionsaufrufe:</b>
 * <br>- Initialisierung : sizeX * sizeY
 * <br>- Pro Generation  : OBJECTIVES * #Generationen
 * 
 *   
 * @author $Author: noot $
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 * 
 * @see <a href="ftp://ftp.tik.ee.ethz.ch/pub/people/laumanns/diplomathesis.pdf">Diplomarbeit M. Laumanns</a>
 */

public class PredatorPrey extends AbstractAlgo{

	/**
	 * x-Ausdehnung des Populations-Graphen , wird per Kommandozeile uebergeben
	 */
	private int SIZE_X;  			 
	
	/**
	 * y-Ausdehnung des Populations-Graphen, wird per Kommandozeile uebergeben
	 */
	private int SIZE_Y;			    
	
	/**
	 * Anzahl der Ziele der Fitnessfunktion, wird anhand der Testfunktion ermittelt
	 */
	private int OBJECTIVES;			
	
	/**
	 * Selektionsmenge, die Menge der Individuen, aus denen der Raeuber sein
	 * "Opfer" auswaehlt
	 */
	private ArrayList selectionSet;				
	
	/**
	 * Dominierte Individuen der Selektionsmenge, diese werden bei der 
	 * Selektionsmethode der Dominiertheit als potentielle Opfer des Raeubers
	 * behandelt.
	 */
	private ArrayList dominatedSelectionSet;
	
	/**
	 * Rekombinationsmenge, diese Menge bildet durch Rekombination ein neues Individuum.
	 */
	private ArrayList recombinationSet;			
	
	/**
	 * Unser Populationsgraph, ein Torus, auf dem die Beute-Individuen leben 
	 * und sich der Raeuber bewegt.
	 */
	private PopulationGraph population;			
	
	/**
	 * Liste mit allen Raeubern
	 */
	public ArrayList predators;
	
	/**
	 * Die Individuen, die am Schluss ausgegeben werden (die besten n Individuen).
	 */
	private ArrayList archive;					

	
	/**
	 * Groesse der Individuen-Menge, die ausgegeben wird (kleiner als
	 *  die Populationsgroesse).
	 */
	private int archiveSize;
	
	/**
	 * Varianz fuer Mutation, wird als Aufrufparameter dem Algo von der Kommandozeile uebergeben
	 */
	private double variance;					
	
	
	/**
	 * Die Art der Selektion, wird als Integerwert (0, 1, ...) von der Kommandozeile her uebergeben 
	 */
	private char selectionMethod;
	
	/**
	 * Zaehler fuer die aktuelle Generationsnummer
	 */
	private int currentGeneration = 0; 
	
	/**
	 * Debugging-Variable fuer Ausgaben innerhalb der Selektion.
	 */
	private boolean debugSel = false;
	
	/**
	 * leerer Konstruktor
	 * @see AbstractAlgo
	 *
	 */
	public PredatorPrey() {
	}
	
	/**
	 * Konstruktor von PredatorPrey
	 * 
	 * @see AbstractAlgo
	 * @param args Die Kommandozeilenparameter
	 * 
	 */
	public PredatorPrey(String[] args){
		
		// Konstruktor der Super-Klasse AbstractAlgo
		super(args);		
		
		// Hier werden die Kommandozeilenparameter von Parser abgeholt
		try {
			test.init(specificTestfktParameters);
			// Anzahl der Ziele der Testfunktion 
			OBJECTIVES = test.getNumberOfObjectives();
			// Parameter des Aufrufs 
			SIZE_X = specificParameters.getIntProperty("sizeX", 32, false);
			SIZE_Y = specificParameters.getIntProperty("sizeY", 32, false);
			archiveSize = specificParameters.getIntProperty("archive", 200, false);
			variance = specificParameters.getDoubleProperty("variance", 0.5, false);
			selectionMethod = specificParameters.getStringProperty("selection", "0", false).charAt(0);
		}
		catch (Exception e) {
			/* Fehler beim Parsen ausgeben und Abbrechen */
			e.printStackTrace();
			System.exit(1);	// Direkter Abbruch des Programms
		}
		
		// Initialisiere die benoetigten Verwaltungs-Listen
		selectionSet = new ArrayList();
		dominatedSelectionSet = new ArrayList();
		recombinationSet = new ArrayList();
		
		// Initialisiere Population-Graphen
		population = new PopulationGraph(SIZE_X, SIZE_Y);
		
		// Initialisiere Liste der Raeuber
		predators = new ArrayList();
		// erzeuge fuer jedes Ziel einen Raeuber und fuege ihn in die Liste ein 
		for (int i=0; i<OBJECTIVES; i++) {
			predators.add(i, new Predator(i, new ArrayList()));
		}
		
		// Initialisiere das Archiv fuer die Ausgabe der besten Loesungen
		archive = new ArrayList();

		// Debug-Ausgaben
		if (debug){
			System.out.println("*** Eingabe: ***"); 
			System.out.println("SIZE_X: "+SIZE_X);
			System.out.println("SIZE_Y: "+SIZE_Y);
			System.out.println("Archiv: "+archiveSize);
			System.out.println("Varianz: "+variance);
			System.out.println("Testfunktion: wird in AbstractAlgo geparsed");
			System.out.println("Generationenzahl: "+maxGenerations);
			System.out.println("Ausgabedatei: "+outputFileNamePrefix);
			System.out.println("Selection: "+selectionMethod);
		}
	}// Ende: Konstruktor
	
	
	/**
	 * Ruft startAlgo auf.
	 * 
	 * @param args
	 * @throws Exception
	 */
	public static void main (String args[]) throws Exception {

		// erzeuge Instanz der Klasse
		PredatorPrey  algoPredatorPrey = new PredatorPrey(args);
		
		// starte Algorithmus
		algoPredatorPrey.startAlgo();
		
		// veraltete klasseneigene Methode zur Ausgabe
		//algoPredatorPrey.writeArchiveFitness2File();
		
		// Ausgabe der Fitnesswerte der Individuen im Archiv in die Datei <b>outputfFilenamePrefix</b>.best.res
		AlgoOutput.writeFitness(algoPredatorPrey.archive, algoPredatorPrey.outputFileNamePrefix, false);
		// Ausgabe der Genome der Individuen im Archiv in die Datei <b>outputfFilenamePrefix</b>.best.in
		AlgoOutput.writeObjectVariables(algoPredatorPrey.archive, algoPredatorPrey.outputFileNamePrefix, false);
		
		// Signal: Algo ist fertig
		System.out.println("Algorithmus PredatorPrey wurde erfolgreich beendet");	
	}// main
	
	
	/* *
	 * Startmethode des Algorithmus.
	 * Aufgerufen von main.
	 * Ruft init() auf.
	 * Ruft evol() <b>maxGenerations</b>-mal auf.
	 * 
	 *  (non-Javadoc)
	 * 
	 * @see algo.AbstractAlgo#startAlgo()
	 * @throws Exception
	 */
	public boolean startAlgo() throws Exception {

		test.init(specificTestfktParameters);
		
		// Initialisierung
		init();
		
		// Evolution
		for (currentGeneration = 1; currentGeneration <= maxGenerations; currentGeneration++){
			evol();
		}
		
		// schreibt eine ausgewaehlte Anzahl von nicht-dominierten Individuen in ein Archiv, das
		// durch einen Aufruf in der main-Methode ausgegeben wird.
		getBestIndividuals();
			
	    archive = (ArrayList)test.postProcessing(archive);
		test.finalize();
		
		return true;
	}// Ende: startAlgo
	
	
	/**
	 * Initialisierung der benoetigten Objekte.
	 * Aufgerufen von startAlgo().
	 */
	private void init() {
		
		// initialisiere Knoten des Graphen mit Individuen (Prey-Objekte)
		
		// Optimale Grenzen fuer die Koordinatenwerte der Individuen von der Testfunktion uebernehmen
		List bounds = test.getOptimalBounds();
		
		// Zufallszahl
		double randomNumber;
		
		// Ober-/Untergrenzen fuer "vernuenftige" Startwerte der Individuen-Koordinaten
		double lowerBound, upperBound;
		
		for (int i=0; i<SIZE_X; i++) {
			for(int j=0; j<SIZE_Y; j++) {
				
				// neues Individuum
				Prey ind = new Prey();
				
				// Geburts-Generation
				ind.setGeneration(currentGeneration);
				
				// Objekt-Variablen des Individuums 
				ArrayList variableValues = new ArrayList();
				for (int k=0; k<test.getDimension(); k++) {
					lowerBound = ((Double)bounds.get(2*k)).doubleValue();
					upperBound = ((Double)bounds.get(2*k+1)).doubleValue();
					try {
						randomNumber = rand.nextDouble(lowerBound, upperBound);
						// Objekt-Variablen
						variableValues.add(new Double(randomNumber));
					} catch (InvalidBoundException e1) {
						e1.printStackTrace();
					}
				}
				ind.setObjectVariables(variableValues);
				
				// Fitness des Individuums
				try {
					// Fitness des Individuums setzen mit Konvertierung: ArrayList->List und wieder zurueck
					ind.setFitness((ArrayList)test.execute(Collections.synchronizedList(ind.getObjectVariables())));
				} catch (Exception e) {
					e.printStackTrace();
					System.exit(1);	// Direkter Abbruch des Programms
				}
				
				// Position des Individums im Graphen
				ArrayList position = new ArrayList();
				position.add(new Integer(i));
				position.add(new Integer(j));
				ind.setPosition(position);
				
				// Individum in Graphen einfuegen
				population.setIndividual(ind, i, j);
			}
		}
		// Debug: Ausgabe der Population
		if(debug) population.output();
		
		// Initialisiere Raueber
		for (int i=0; i<predators.size(); i++) {
			int randomValueX, randomValueY;
		
			// zufaellige x- und y- Werte fuer die Raeuber-Position
			randomValueX =  Math.abs(rand.nextInt(SIZE_X));
			randomValueY =  Math.abs(rand.nextInt(SIZE_Y));

			// ArrayList mit der Raeuber-Position
			ArrayList pos = new ArrayList();
			pos.add(new Integer(randomValueX));
			pos.add(new Integer(randomValueY));
			
			// Debug-Ausgabe: Wo ist der Raeuber?
			if(debug) System.out.println("R"+i+" pos: " + pos);

			// Position des Raeubers setzen
			((Predator)predators.get(i)).setPosition(pos);
		}
	}// Ende: init
	
	

	/**
	 * Fuehrt eine Generation der Evolution durch, also einen Verfahrensschritt
	 * eines Raeuber-Beute-Systems. <br>
	 * Wird von startAlgo <b>maxGenerations</b>-mal aufgerufen. 
	 */
	private void evol() throws Exception{
		
		// Debug-Ausgabe: Nummer der Generation
		if(debug) System.out.println("\nevol(), Generation:"+ currentGeneration);
		
		// Hier findet ein Generationswechsel statt, jeder Raeuber "faengt"
		// sich ein Beutetier
		for (int i=0; i<predators.size(); i++) {
			
			// Den i-ten Raeuber aus der Liste holen
			Predator curPredator = (Predator)predators.get(i);
			
			// Kriterium des Raeubers ermitteln
			int predatorObjective = curPredator.getObjective();
			
			// leere Mengen 
			selectionSet.clear();
			dominatedSelectionSet.clear();
			recombinationSet.clear();
			
			// zufaellige Bewegung des Raeubers
			randomWalk((Predator)predators.get(i));
			if (debug)
				System.out.println("R"+i+" pos:"+ curPredator.getPosition().toString());
			
			// fuelle Selektionsmenge
			buildSelectionSet((Predator)predators.get(i));
			
			Prey prey = new Prey();
						
			// bestimme Selektionsmethode
			switch(selectionMethod) {
				case '0':
					// Selektion nach Laumanns (einkriteriell fuer jeden Raeuber)
					prey = selectionLaumanns(predatorObjective);
					break;
				case '1':
					// Selektion bzgl. Pareto-Dominanz
					prey = selectionDominance();
					break;
				default:
					throw new Exception ("wrong selection parameter");
			}
			
			// Debug-Ausgabe: Wo ist das Beute-Individuum?
			if(debug) System.out.println("Prey: "+prey.getPosition());
			
			// fuehre Rekombination durch (dadurch wird das Beutetier
			// ueberschrieben, also automatisch vom Raeuber "gefressen")
			recombination(prey);
			
			// Debug-Ausgabe: welche Objekt-Variablen hat das neue Individuum?
			if(debug) System.out.println("Prey-rec: "+prey.getObjectVariables());
			
			// fuehre Mutation durch
			mutation(prey, variance);
			
			// Debug-Ausgabe: Welche Objekt-Variablen nach der Mutation?
			if(debug) System.out.println("Prey-mut: "+prey.getObjectVariables());
			
			// Fitness des neuen Individuums setzen
			try{
				prey.setFitness((ArrayList)test.execute(Collections.synchronizedList(prey.getObjectVariables())));
			} catch (Exception e){
				e.printStackTrace();
				System.exit(1);	// Direkter Abbruch des Programms
			}
			
			// Generationsnummer des neuen Individuums setzen
			prey.setGeneration(currentGeneration);
			
			// Das neue, rekombinierte und mutierte Individuum neu in den Graphen setzen
			population.setIndividual(prey, prey.getPosition());	
			
			//Evaluatinosbasierter LiveOutput
			// Archiv fr live-output fllen
			getBestIndividuals();
			EvaluationBasedLiveOutput(archive);
			
			// Debug-Ausgabe: Population ausgeben
			if(debug) population.output();
		}
	}// evol
	

	/**
	 * Zufaellige gleichverteilte Bewegung eines Raeubers ueber den Graphen innerhalb der vorgegebenen Nachbarschaft.
	 * Aufgerufen von evol() fuer alle Raeuber.
	 * 
	 * @param pred (Predator): der Raeuber, der bewegt wird 
	 */
	private void randomWalk(Predator pred){
		// zufaelliger Lauf der Raeubers ueber den Graphen

		//  Nachbarschaft bestimmen
		ArrayList neighbours = getDirectNeighbourhood((ArrayList)pred.getPosition());
		
		// errechnet eine ganzzahlige gleichverteilte Zahl
		// zwischen 0 (incl.) und 4 (excl.)
		int randomValue =  Math.abs(rand.nextInt(4));
		
		// neue Position setzen, indem die Position eines der Nachbarn
		// (zufaellig) gewaehlt wird
		pred.setPosition((ArrayList)neighbours.get(randomValue));	
	}// Ende: randomWalk
	
	

	/**
	 * Rekombination.
	 * Aufgerufen von evol().
	 * 
	 * Rekombination von Individuuen, hier diskrete Rekombination, d.h.
	 * jede Objekt-Variable wird einzeln per Zufall von der entsprechenden
	 * Objektvariable der Eltern uebernommen
	 * 
	 * @param prey (Prey): das Beute-Individuum, welches aus seiner Nachbarschaft neu rekombiniert wird.
	 */
	private void recombination(Prey prey){
		
		// Zufallswert
		int randomValue;
		
		// Die Objekt-Variablen des Individuums
		ArrayList variables = new ArrayList();
		variables = prey.getObjectVariables();
		
		// Position des Beute-Individuums im Graphen
		ArrayList pos = new ArrayList();
		
		// ArrayList mit den Koordinaten erstellen
		pos = prey.getPosition();

		// Die Rekombinationsmenge
		ArrayList recombSet = new ArrayList();
		
		// Rekombinations-Nachbarschaft erstellen
		recombSet = this.buildRecombinationSet(pos);	
		
		// Rekombination fuer alle Variablen
		for (int i = 0; i<variables.size(); i++){
		
			// errechnet eine ganzzahlige gleichverteilte Zahl
			// zwischen 0 (incl.) und 4 (excl.)
			randomValue =  Math.abs(rand.nextInt(4));

			variables.set(i, ( ((Prey)recombSet.get(randomValue)).getObjectVariables().get(i)));
		} // Ende: for (i...
		
		// Dem Individuum die neuen mutierten Gene zuweisen				
		prey.setObjectVariables(variables);
	}// recombination
	
	
	/**
	 * Mutation.
	 * Aufgerufen von evol(). 
	 * 
	 * @param ind das Individuum, das mutiert werden soll
	 * @param varianz die Mutationsschrittweite
	 */
	private void mutation(Prey ind, double varianz){
		Double var;
		// Individuum wird mutiert
		ArrayList variables = ind.getObjectVariables();
		
		// Fuer alle Objektvariablen...
		for (int i = 0; i<variables.size(); i++){
			var = (Double)variables.get(i);	// Aktuellen Variablenwert der Variable i bestimmen
			// Mutieren, indem normalverteilte Zuf.-Var. um Null mit Varianz <varianz> addiert wird
			// Normalverteilte Zufallsvariable mit der Varianz <varianz>
			var = new Double(var.doubleValue()+(rand.nextGaussian())*varianz);	
			variables.set(i, var);			// neuen Wert speichern
		}
	}// mutation

	/**
	 * Selektion zur Ersetzung.
	 * Variante nach Laummans, jeder Raeuber selektiert nach dem hinsichtl. seines Kriteriums schlechtesten Individuums
	 * 
	 * Aufgerufen von evol().
	 * 
	 * @param predatorObjective das Raeuber-Kriterium
	 * @return prey: das Beute-Individuum, das selektiert (gefressen) wird
	 */
	private Prey selectionLaumanns(int predatorObjective) {
		// Nach Laumanns selektiert der Raeuber das Individuum aus der 
		// Selektionsnachbarschaft,  das bezueglich des Raeuber-Kriteriums
		// am "unfittesten" ist.
		// Hier wird nun selektiert anhand des Kriteriums des aktuellen Raeubers,
		// d.h. das Individuum des SelectionSets mit dem hoechsten Fitnesswert
		// bezgl. des Raeuber-Kriteriums wird ausgewaehlt und 
		// spaeter ueberschrieben.
		
		double curFitness;		// aktuelle Fitness 
		double worstFitness;	// bisher schlechteste Fitness
		Prey prey = new Prey();	// zu selektierendes Individuum
			
		prey = (Prey)selectionSet.get(0); // erstes Individuum der Selektionsmenge
		
		// die Fitness des aktuellen (bisher schlechtesten preys) 
		worstFitness =  ((Double)prey.getFitness().get(predatorObjective)).doubleValue();
		if (debugSel) System.out.println("sel: "+worstFitness);

		// Fuer alle Individuen in der Selektionsmenge...
		for (int k=1; k<selectionSet.size(); k++){
		
			// die Fitness des naechsten prey im SelectionSet
			curFitness = ((Double)((Prey)selectionSet.get(k)).getFitness().get(predatorObjective)).doubleValue();
				
			if (debugSel) System.out.print("sel: "+ curFitness);
				
			// Individuum schlechter als das bisher schlechteste Individuum?
			if ( worstFitness < curFitness ) {
				worstFitness = curFitness;
				prey = (Prey)selectionSet.get(k);
				
				if (debugSel) System.out.print("  schlecht");
			
			} // End: if
			if (debugSel) System.out.println();
		} // End. for
		
		return prey;
		
	}// Ende: selectionLaumanns
	
	
	/**
	 * Selektion zur Ersetzung.
	 * Methode selektiert bzgl Dominanz.
	 * 
	 * Aufgerufen von evol().
	 * 
	 * @return prey: das Beute-Individuum, das selektiert (gefressen) werden soll
	 */
	private Prey selectionDominance() {
		// waehle dominierte Individuen aus selectionSet
		buildDominatedSelectionSet();
		
		// Debug-Ausgabe
		if (debugSel) {
			liveOutput(dominatedSelectionSet, currentGeneration);
		}
		
		// waehle gleichverteilt ein Individuum aus dominatedSelectionSet
		// zwischen 0 (incl.) und dominatedSelectionSet.size() (excl.)
		int randomNumber = rand.nextInt(dominatedSelectionSet.size());
			
		// ausgewaehltes Individuum wird spaeter ueberschrieben
		Prey prey = ((Prey)dominatedSelectionSet.get(randomNumber));
		
		return prey;
		
	}// Ende: selectionDominance

	
	/**
	 * Erstelle Selektionmenge aus der Nachbarschaft.
	 * 
	 * Aufgerufen von evol().
	 * 
	 * @param pred der Raeuber, zu dem eine Selektionsmenge erstellt werden soll.
	 */
	private void buildSelectionSet(Predator pred){
		// Selektionsmenge erstellen zum Raeuber pred
		// Waehle die 4 direkten Nachbarn des Knotens auf dem sich der Raeber
		// befindet und den Knoten selbst.
		
		// position des Raeubers im Graphen ermitteln
		ArrayList predatorPos = pred.getPosition();
		int predatorPosX= ((Integer)predatorPos.get(0)).intValue();
		int predatorPosY= ((Integer)predatorPos.get(1)).intValue();
		//	selection enthaelt Objekte vom Typ Prey
		
		// Beute direkt auf dem Knoten, auf dem sich der Raeuber befindet 
		Prey prey0 = (Prey) population.getIndividual(predatorPosX, predatorPosY);
		prey0.setPosition(predatorPos);
		
		// Zur Selektionsmenge hinzufuegen
		selectionSet.add(prey0);
		
		// Die Nachbarn...
		ArrayList neighbourPosition = getDirectNeighbourhood(predatorPos);
		
		// die 4 Nachbarn: links, rechts, oben, unten
		for (int i=0; i<4; i++) {
			Prey prey = (Prey)population.getIndividual((ArrayList)neighbourPosition.get(i));
			prey.setPosition((ArrayList)neighbourPosition.get(i));
			selectionSet.add(prey);
		}
	}// Ende: buildSelectionSet
	
	

	/**
	 * Erstellt die dominierte Selektionsmenge aus den dominerten Individuen
	 * der "normalen" Selektionsmenge.
	 *
	 */
	private void buildDominatedSelectionSet(){
		// Selektionsmenge der dominierten Individuen (aus der urspruenglichen Selektionsmenge) erstellen
		
		dominatedSelectionSet = ParetoArchive.refreshDominatedArchive(selectionSet);
		
		// Kein Individuum der Menge wird dominiert?-> dann nimm einfach alle!
		if (dominatedSelectionSet.isEmpty())
			dominatedSelectionSet = selectionSet; 		
	}// Ende: buildDominatedSelectionSet
	
	
		
	/**
	 *	Erstellt die Rekombinationsmenge (die Nachbarschaft) 
	 * @param pos Zentrum der Nachbarschaft = Position des zu ersetztenden Prey-Objekts
	 * @return recombinationSet: Menge, die zur Rekombination zur Auswahl steht
	 */
	private ArrayList buildRecombinationSet(ArrayList pos){
		// Rekombinationsmenge erstellen
		ArrayList neighbourhood = new ArrayList();
		
		// Die Nachbarschaft ermitteln
		neighbourhood = getDirectNeighbourhood(pos);
		for(int i=0; i< neighbourhood.size(); i++){
			recombinationSet.add(i, population.getIndividual((ArrayList)neighbourhood.get(i)));
		}
		
		return(recombinationSet);
	}// Ende: buildRecombinationSet
	
	
	
	/**
	 * Bestimme die direkte Nachbarschaft, speichere diese Individuenin der Nachbarschaftsmenge
	 * 
	 * @param pos Zentrum der Nachbarschaft
	 * @return neighbourhood Menge der direkten Nachbarn der Position pos
	 */
	private ArrayList getDirectNeighbourhood(ArrayList pos) {
		// Gibt ein ArrayListe zurueck, die die Positionen der direkten
		// Nachbarn des Knoten pos im Graphen enthaelt
		// Die Positionen der Knoten sind jeweils wieder als
		// ArrayListe mit zwei Koordinaten gesspeichert
		
		ArrayList neighbourhood = new ArrayList();
		
		// center ist das Zentrum der Nachbarschaft, also der Knoten auf den
		// sich die Nachbarschaft bezieht (sprich: das Individuum, um das es geht)

		int centerX= ((Integer)pos.get(0)).intValue();
		int centerY= ((Integer)pos.get(1)).intValue();
						
		// Knoten links vom aktuellen Knoten
		ArrayList left = new ArrayList(2);
		left.add(new Integer(getLeft(centerX)));
		left.add(new Integer(centerY));
		neighbourhood.add(left);
		
		//	Knoten rechts vom aktuellen Knoten
		ArrayList right = new ArrayList(2);
		right.add(new Integer(getRight(centerX)));
		right.add(new Integer(centerY));
		neighbourhood.add(right);
		
		//	Knoten unter dem aktuellen Knoten
		ArrayList down = new ArrayList(2);
		down.add(new Integer(centerX));
		down.add(new Integer(getDown(centerY)));
		neighbourhood.add(down);
		
		//	Knoten rechts vom aktuellen Knoten
		ArrayList up = new ArrayList(2);
		up.add(new Integer(centerX));
		up.add(new Integer(getUp(centerY)));
		neighbourhood.add(up);
		
		return neighbourhood;
	}// Ende: getDirectNeighbourhood

	
	
	/**
	 * Bestimme den rechten Nachbarn. 
	 * Wird das rechte Ende des Graphen erreicht, wird am linken Ende fortgefahren (Torus-Eigenschaft). 
	 * 
	 * @param posX x-Koordinate
	 * @return pos x-Koordinate der rechts-benachbarten Position von posX 
	 */
	private int getRight(int posX) {
		// Ursprung des Populationsgraphen ist oben links (wie eine Matrix)
		// zu grosse oder negative Eingaben werden nicht als Fehler behandelt
		// z.B. getRight(SIZE_X +5) = 0
		
		// x-Koordinate rechts von der Position posX
		// Da der Populationsgraph ein Torus ist, ist der Knoten rechts vom
		// letzten der erste
		
		
		int pos = posX;
		
		pos++;
		if (pos > SIZE_X - 1) {
			pos = 0;
		}
		return pos;
	}// getRight
	
	/**
	 * Bestimme den linken Nachbarn. 
	 * Wird das linke Ende des Graphen erreicht, wird am rechten Ende fortgefahren (Torus-Eigenschaft).
	 * 
	 * @param posX x-Koordinate
	 * @return pos (int)  x-Koordinate der links-benachbarten Position von posX 
	 */
	private int getLeft(int posX) {
		// x-Koordinate links von der Position posX
		// Da der Populationsgraph ein Torus ist, ist der Knoten links vom
		// ersten der letzte
		int pos = posX;
	
		pos--;
		if (pos < 0) {
			pos = SIZE_X - 1;
		}
		return pos;
	}// getLeft
	
	/**
	 * 
	 * Bestimme den oberen Nachbarn. 
	 * Wird das obere Ende des Graphen erreicht, wird am unteren Ende fortgefahren (Torus-Eigenschaft).
	 *
	 * @param posY y-Koordinate
	 * @return pos (int)  y-Koordinate der oben-benachbarten Position von posY 
	 */
	private int getUp(int posY) {
		// y-Koordinate oberhalb von der Position posY
		// Da der Populationsgraph ein Torus ist, ist der Knoten ueber
		// dem obersten der unterste
		int pos = posY;
	
		pos--;
		if (pos < 0) {
			pos = SIZE_Y - 1;
		}
		return pos;
	}// getUp
	
	/**
	 * Bestimme den unteren Nachbarn. 
	 * Wird das untere Ende des Graphen erreicht, wird am oberen Ende fortgefahren (Torus-Eigenschaft). 
	 * 
	 * @param posY y-Koordinate
	 * @return pos (int)  y-Koordinate der unten-benachbarten Position von posY 
	 */
	private int getDown(int posY) {
		// y-Koordinate unterhalb von der Position posY
		// Da der Populationsgraph ein Torus ist, ist der Knoten unter
		// dem untersten der oberste
		int pos = posY;
		
		pos++;
		if (pos > SIZE_Y - 1) {
			pos = 0;
		}
		return pos;
	}// getDown
	
	/**
	 * 
	 * Bestimme die besten Individuuen,
	 * und zwar soviele, wie die globale Variable archiveSize angibt. Speichert die Individuen
	 * in archive.
	 *
	 */
	private void getBestIndividuals() {
		// filtere dominierte Individuen
		for (int i=0; i<SIZE_X; i++){
			for(int j=0; j<SIZE_Y; j++) {
				archive = ParetoArchive.refreshArchive(archiveSize, archive, population.getIndividual(i,j));
			}
		}
	}// Ende: getBestIndividuals
	
	
	
	/**
	 * Schreibe die besten Individuen in die Ausgabedatei.
	 *
	 */
	private void writeArchiveFitness2File() {
		// Gibt die Fitnesswerte (Paretofront) des Archivs aus	
		try {
    		 /* Ausgabedatei-Handle erstellen */	
			 FileWriter fw = new FileWriter(outputFileNamePrefix);
			 
			 /* Durch alle Archiv-Elemente iterieren */
			 for (Iterator iter = archive.iterator(); iter.hasNext();) {
				Prey outputIndi = (Prey)iter.next();  // Individuum als Liste (ihrer Koordinaten) casten
				StringBuffer row=new StringBuffer();  // Neuen Stringbuffer fuer Ausgabe erstellen
				 /* Durch alle Fitnesswerte eines Individuums iterieren */	
				 for (Iterator iter2 = outputIndi.getFitness().iterator(); iter2.hasNext();) {
						   Double value = (Double) iter2.next();
						   row.append(value.toString()+" ");
				}
				/* Ausgabe */	
				fw.write(row.toString()+"\n");
				//	System.out.println(row.toString());
			 }
			 /* Ausgabedatei schliessen */	
			 fw.close();
		 } 
		 catch (IOException e) {
				   e.printStackTrace();
		 }
    }// Ende: writeArchiveFitness2File

	/**
	 * 
	 * toString-Methode, wird hier benutzt, um eine Algorithmen-Beschreibung zu uebergeben.
	 * 
	 * @see tools.io.Help
	 */
	public String toString() {
		String description;
		description = "Klasse PredatorPrey, Paket algo \n\n";
		description += "Beschreibung :\n\n";
		description += "Klassischer Raeuber-Beute-Algorithmus\n";
		description += "Operiert auf einem Torus-Graphen von sizeX mal sizeY Knoten (mit jew.\n";
		description += "4 Nachbarn), in denen die Beuteindividuem leben. Ueber diesen Graphen\n";
		description += "wandern so viele Raeuber, wie es Kriterien gibt in zufaelliger Richtung.\n";
		description += "Sie fressen dabei das jeweils schwaechste Individuum bezueglich ihres\n";
		description += "Kriteriums. Gefressene Individuen werden durch Rekombination aus ihren \n";
		description += "Nachbarn neu gebildet und anschliessend mutiert.\n\n";
		description += "Kommandozeilenparameter koennen auch in der mit -c angegebenen Konfigurations-\n";
		description += "Datei zeilenweise angegeben werden\n";
		description += "-t: Klasse der Testfunktion (z.B. FonsecaF1)\n";
		description += "-g: (int): Anzahl der Generationen/Iterationen\n";
		description += "-r: (int): optional, Random-Seed (Initialisierung der Zufallsklasse)\n";
		description += "-o: optional, Praefix fuer die Dateinamen der Ausgabe (An diese wird noch" +
	    				" das Kuerzel fuer die Art der Ausgabedatei angehaengt, z.B. .best.res);" +
						" Wird der Wert nicht gesetzt, so wird ein Praefix aus Datum und Uhrzeit gewaehlt\n";
		description += "-d: optional, wenn gesetzt, werden Debug-Ausgaben angezeigt\n";
		description += "-live -> Abstand, wieviele Evaluationen zwischen 2 liveOutputs liegen soll.\n";
		description += "-archive (int): optional, Maximalgroesse des globalen Archives, welches am Ende ausgegeben wird; 0=unbegrenzt (Standardwert)\n";

		description += "-c: Dateiname der Parameter-Eingebedatei\n";
		description += "-variance (double): Mutationsschrittweite\n";
		description += "-selection (int): Art der Selektion (0, 1, ...) \n";
		description += "\n";

		return description;
	} // Ende: toString




/* Innere Klassen */
	
	/**
	 * Populationsgraph,
	 * auf diesem leben die Beuteindividuen und bewegen sich die Raeuber.
	 * 
	 * @author Ralf
	 *	
	 *	
	 */
	
	private class PopulationGraph{
		/**
		 * Der Graph an sich als 2-dim Array von Prey-Objekten.
		 */
		private Prey [] [] torus;

		/**
		 * @param x die Groesse in X-Richtung
		 * @param y die Groesse in Y-Richtung
		 */
		private PopulationGraph(int x, int y){
			torus = new Prey [x] [y];
		}	

		/**
		 * Setzt ein Beute-Individuum an die angegebene Stelle im Populationsgraphen,
		 * uebergeben werden die Koordinaten als int-Werte.
		 * 
		 * @param i das zu setzende Individuum
		 * @param x die X-Koordinate
		 * @param y die Y-Koordinate
		 */
		private void setIndividual(Prey i, int x, int y){
			torus[x][y] = i;
		}

		
		/**
		 * Setzt ein Beute-Individuum an die angegebene Stelle im Populationsgraphen,
		 * uebergeben wird eine ArrayList mit den Koordinaten.
		 * 
		 * @param ind das zu setzende Individuum
		 * @param pos die Position im Graphen
		 */
		private void setIndividual(Prey ind, ArrayList pos){
			int posX = ((Integer)pos.get(0)).intValue();
			int posY = ((Integer)pos.get(1)).intValue();
			torus[posX][posY] = ind;
		}

		/**
		 * Individuum aus dem Populationsgraphen holen,
		 * uebergeben werden die Koordinaten als int-Werte.
		 * 
		 * @param x die X-Koordinate im Graphen
		 * @param y die Y-Koordinate im Graphen
		 * @return Prey
		 */
		private Prey getIndividual(int x, int y){
			return(torus[x][y]);
		}

		/**
		 * Individuum aus dem Populationsgraphen holen,
		 * uebergeben werden die Koordinaten als ArrayList.
		 * 
		 * @param pos die Position im Graphen in einer ArrayList
		 * @return Prey: das Individuum, das gewuenscht wurde
		 */
		private Prey getIndividual(ArrayList pos){
			int posX = ((Integer)pos.get(0)).intValue();
			int posY = ((Integer)pos.get(1)).intValue();
			return(torus[posX][posY]);
		}
		
		/**
		 * Ausgabe zum Debuggen.
		 * Formatiert fuer einen 3x3 Torus, sonst erkennt man nichts.
		 * Ansicht fuer grosse Tabs.
		 *
		 */
		private void outputBigTab() {
			NumberFormat nf = NumberFormat.getNumberInstance();
			nf.setMinimumFractionDigits(4);
			nf.setMaximumFractionDigits(4);
			
			StringBuffer rowPosition = new StringBuffer();
			rowPosition.delete(0,rowPosition.length());
			StringBuffer rowObjVar = new StringBuffer();
			rowObjVar.delete(0,rowObjVar.length());
			StringBuffer rowFitness = new StringBuffer();
			rowFitness.delete(0,rowFitness.length());
			StringBuffer rowGeneration = new StringBuffer();
			rowGeneration.delete(0,rowGeneration.length());

			System.out.println("#### Population ####");
			//System.out.println("*** Generation: "+ currentGeneration);
			
			for(int i=0; i<SIZE_Y; i++) {
				rowGeneration.delete(0,rowGeneration.length());
				rowGeneration.append("Gen: ");
				rowPosition.delete(0,rowPosition.length());
				rowPosition.append("Pos: ");
				rowObjVar.delete(0,rowObjVar.length());
				rowObjVar.append("Obj: ");
				rowFitness.delete(0,rowFitness.length());
				rowFitness.append("Fit: ");
				
				for (int j=0; j<SIZE_X; j++) {
					Prey ind = population.getIndividual(i,j);
					
					// Geburts-Generation des Individuum
					rowGeneration.append("|" + ind.getGeneration() + "|\t\t\t\t\t");
				
					// Position im Graphen
					List pos = ind.getPosition();
					rowPosition.append("|");
					for (Iterator posIter = pos.iterator(); posIter.hasNext(); ) {
						 Integer s = (Integer)(posIter.next());
						 rowPosition.append(s+" ");
					}
					rowPosition.append("|\t\t\t\t");
					
					// Objektvariablen
					List obj = ind.getObjectVariables();
					rowObjVar.append("|");
					for (Iterator objIter = obj.iterator(); objIter.hasNext(); ) {
						String s = nf.format(objIter.next());
						rowObjVar.append(s+" ");
					}
					rowObjVar.append("|\t\t\t");

					// Fitness
					List fit = ind.getFitness();
					rowFitness.append("|");
					for (Iterator fitIter = fit.iterator(); fitIter.hasNext(); ) {
						String s = nf.format(fitIter.next());
						rowFitness.append(s+" ");
					}
					rowFitness.append("|\t");
					
				}// end: for(SIZE_X)
				System.out.println(rowGeneration.toString());
				System.out.println(rowPosition.toString());
				System.out.println(rowObjVar.toString());
				System.out.println(rowFitness.toString());
				System.out.println("\t ---------------- \t ---------------- \t ---------------");
			
			}// end: for(SIZE_Y)
			
		}// PopulationGraph:output

		/**
		* Ausgabe zum Debuggen.
		* Formatiert fuer einen 3x3 Torus, sonst erkennt man nichts.
		* Ansicht fuer kleine Tabs.
		*/
		private void output() {
			NumberFormat nf = NumberFormat.getNumberInstance();
			nf.setMinimumFractionDigits(4);
			nf.setMaximumFractionDigits(4);

			StringBuffer rowPosition = new StringBuffer();
			rowPosition.delete(0,rowPosition.length());
			StringBuffer rowObjVar = new StringBuffer();
			rowObjVar.delete(0,rowObjVar.length());
			StringBuffer rowFitness = new StringBuffer();
			rowFitness.delete(0,rowFitness.length());
			StringBuffer rowGeneration = new StringBuffer();
			rowGeneration.delete(0,rowGeneration.length());

			System.out.println("#### Population ####");
			//System.out.println("*** Generation: "+ currentGeneration);
	
			for(int i=0; i<SIZE_Y; i++) {
				rowGeneration.delete(0,rowGeneration.length());
				rowGeneration.append("Gen:\t");
				rowPosition.delete(0,rowPosition.length());
				rowPosition.append("Pos:\t");
				rowObjVar.delete(0,rowObjVar.length());
				rowObjVar.append("Obj:\t");
				rowFitness.delete(0,rowFitness.length());
				rowFitness.append("Fit:\t");
		
				for (int j=0; j<SIZE_X; j++) {
					Prey ind = population.getIndividual(i,j);
			
					// Geburts-Generation des Individuum
					rowGeneration.append("|" + ind.getGeneration() + "|\t\t\t\t\t\t\t\t");
		
					// Position im Graphen
					List pos = ind.getPosition();
					rowPosition.append("|");
					for (Iterator posIter = pos.iterator(); posIter.hasNext(); ) {
						 Integer s = (Integer)(posIter.next());
						 rowPosition.append(s+" ");
					}
					rowPosition.append("|\t\t\t\t\t\t\t");
			
					// Objektvariablen
					List obj = ind.getObjectVariables();
					rowObjVar.append("|");
					for (Iterator objIter = obj.iterator(); objIter.hasNext(); ) {
						String s = nf.format(objIter.next());
						rowObjVar.append(s+" ");
					}
					rowObjVar.append("|\t\t\t");

					// Fitness
					List fit = ind.getFitness();
					rowFitness.append("|");
					for (Iterator fitIter = fit.iterator(); fitIter.hasNext(); ) {
						String s = nf.format(fitIter.next());
						rowFitness.append(s+" ");
					}
					rowFitness.append("|\t\t\t");
				
				}// end: for(SIZE_X)
				System.out.println(rowGeneration.toString());
				System.out.println(rowPosition.toString());
				System.out.println(rowObjVar.toString());
				System.out.println(rowFitness.toString());
				System.out.println("\t\t-----------------------\t\t-----------------------\t\t----------------------");
			
			}// end: for(SIZE_Y)
			
		}// PopulationGraph:output

	}//End Class PopulationGraph

	
	/**
	 * Predator (Raeuber)-Klasse
	 * 
	 * @author Ralf
	 *
	 */
	private class Predator{
		/**
		 * die Raeuber-Position im Populations-Graphen
		 */
		private ArrayList position;		
		
		/**
		 * Nummer des zu optimierenden Kriteriums
		 */
		private int objective;			
		
		/**
		 * Konstruktor
		 * @param obj Nummer des Kriteriums, nach dem dieser Raeuber selektieren soll
		 * @param pos die Position des Raeubers im Graphen
		 */
		private Predator(int obj, ArrayList pos){
			objective = obj;
			position = pos;

		}
		
		/**
		 * Position des Raeubers im Graphen setzen.
		 * @param pos die Position im Graphen
		 */
		private void setPosition(ArrayList pos){
			position = pos;
		}
		
		/**
		 * Gibt die Position des Raeubers im Graphen zurueck.
		 * @return ArrayList
		 */
		private ArrayList getPosition(){
			return(position);
		}
		
		/**
		 * Gibt die Nummer des Kriteriums zurueck, nach dem dieser Raeuber selektiert.
		 * @return int
		 */
		private int getObjective(){
			return(objective);
		}
		
	}//End class Predator
	
	
	/**
	 * Beute-Individuum.
	 * @see Individual 
	 * @author Ralf
	 *
	 */
	private class Prey extends Individual{
		
		/**
		 * Die Position des Beute-Individuums im Graphen.
		 */
		private ArrayList position;
		
		/**
		 * Leerer Konstruktor.
		 *
		 */
		private Prey(){
			
		}
		
		/**
		 * Konstruktor von Prey Die Position im Graphen wird als Parameter
		 * verwendet und gesetzt.
		 * @param pos ArrayList der Position im Populationsgraphen
		 */
		private Prey(ArrayList pos) {
			position = pos;
		}
		
		/**
		 * Gibt die Position des Beute-Individuums im Graphen zurueck.
		 * @return position ArrayList der Position im Graphen
		 */
		private ArrayList getPosition() {
			return position;
		}

		/**
		 * Setzt die Position im Populationsgraphen.
		 * @param pos ArrayList der Position im Graphen
		 */
		private void setPosition(ArrayList pos) {
			position = pos;
		}

	}// End class Prey
	

} // End class PredatorPrey