/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on Jun 24, 2004
 * 
 *
 */
package org.nobeltje.algo;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.nobeltje.tools.Individual;
import org.nobeltje.tools.ParetoArchive;
import org.nobeltje.tools.io.AlgoOutput;
import org.nobeltje.tools.io.ArgsInputBeautifier;

/**<p>
 *Algorithmus fuer das Rucksackproblem 
 *</p>
 *<p>
 *Implementiert von Hendrik Noot und Miriam Padberg. <br />
 *Dieser Algorithmus berechnet mit Hilfer von Tabu Search eine Rucksackbepackung.<br />
 *Als Startwert fuer eine Bepackung wird der Startwert der zugehoerigen Testfunktion genommen.
 *Der Algorithmus berechnet deterministisch eine Bepackung aus der Startbepackung.<br />
 *Momentan ist die Listenlaenge und die Tabudauer variable als Parameter zu uebergeben.<br />
 *Dieser Algorithmus arbeitet nicht nur auf einkriterielle, sondern auch auf mehrkriterielle
 *Rucksackproblemeingabe, dadurch, dass er die Nutzenwerte addiert und die Summe dieser
 *minimiert. Hierbei ist es natuerlich wichtig, dass die Nutzenwerte negativ sind, da ansonsten 
 *die ideale Bepackung die leere Bepackung ist.<br />
 *Die Nachbarschaft fuer dieses Problem ist klar definiert. Eine Bepackung a ist Nachbar von Bepackung b,
 *genau dann, wenn man nur ein Element von b anfasst (also hinzufuegt oder wieder aus dem 
 *Rucksack entfernt) und dann a erhaelt.<br />
 *  Der Algorithmus berechnet nur eine Lsung, nicht alle moeglichen gleichwertigen Loesungen.
 * </p>
  * @author $Author: noot $
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 */
public class TabuSearchKnapsack extends AbstractAlgo {
	  //dynamische TabuList
	  List tabulist;
	
	  // Tabudauer eines Elementes in der TabuListe
	  int tabuiterations;
	  //maximale Anzahl von Eintraegen in der TabuListe
	  int maxtabulistlength;
	  //maximale groee des Paretoarchives
	  private int archive;
	  
	  private String classnameTestfkt;
	  private int dimension;
	  private Individual bestindividual;
	  private double bestfitness;
	  private ArrayList paretoarchive = null;
	  private int livetime;
		
	  /**
	   * Klasse fuer die Elemente in der TabuListe mit set und get Methoden
	   */
	  class TabuListElement{
		int livetime;
		int index;
		/**
		 * gibt den Index des Elements in der TabuListe
		 */
		public int getIndex() {
			return index;
		}
		/**
		 * gibt die Lebensdauer eines Elements in der TabuListe zurueck
		 */
		public int getLivetime() {
			return livetime;
		}
		/**
		 * setzt den Index eines Elementes in der TabuListe
		 */
		public void setIndex(int i) {
			index = i;
		}
		/**
		 * setzt die Lebensdauer von Element i in der TabuListe
		 */
		public void setLivetime(int i) {
			livetime = i;
		}
	  }//Ende innere Klasse
	
	/**
	 * 
	 */
	public TabuSearchKnapsack() {
		super();
	}
	/**<p>
	 * Konstruktor
	 * </p>
	 * <p>
	 * Die uebergebenen Argumente werden durch die abstrakte Algoklasse umgesetzt.<br />
	 * Die Parameter, die speziell fuer diesen Algorithmus wichtig sind, werden auch abgefragt und "gespeichert".
	 * Zu diesen Paramtern gehoert die Anzahl der Iterationen (tabuiterations), die maximale Laenge der 
	 * Tabuliste (tabulistlength) sowie die Angabe der Verweildauer eines Elementes in der Tabuliste (livetime).<br />
	 * Falls eine Eingabe nicht alle notwendigen Paramter hat, so werden default Werte von 3 fuer die Anzahl der 
	 * Iterationen benutzt, eine Anzahl von 15 fuer die maximale Laenge der Tabuliste, sowie ein Verweildauer in der
	 * Tabuliste von 3 Iterationen. <br />
	 * Falls dennoch Werte in der Eingabe fehlen oder falsche Werttypen im Aufruf stehen, wird eine dementsprechende 
	 * Exception geworfen, die darauf hinweist,
	 * dass die Eingabe nicht korrekt bzw nicht komplett war. </p>  
	 * @param args
	 */
	public TabuSearchKnapsack(String[] args) {
		super(args);
		
		try {
			tabuiterations = ArgsInputBeautifier.getIntProperty(specificParameters, "tabuiterations", 3, false);
			maxtabulistlength = ArgsInputBeautifier.getIntProperty(specificParameters, "maxtabulistlength", 15, false);
			livetime = ArgsInputBeautifier.getIntProperty(specificParameters, "livetime", 3, false);
			archive = ArgsInputBeautifier.getIntProperty(specificParameters, "archive", 0, false);
		} catch (Exception e1) {
			System.err.println("Eingabe ist nicht komplett oder fehlerhaft");
			e1.printStackTrace();
		}
		
		dimension=test.getDimension();
		
		paretoarchive = new ArrayList();
		tabulist = new ArrayList();
		
		if (debug) System.out.println("TabuSearchKnapsack Algo Objekt erzeugt: \n" + test);
		if (debug) System.out.println("maxGenerations:" + maxGenerations);
		
		if (debug) System.out.println("tabuiterations:" + tabuiterations);
		if (debug) System.out.println("maxtabulistlength:" + maxtabulistlength);
		if (debug) System.out.println("livetime:" + livetime);
	}	
/**<p>
 * start Algorithmus
 * </p>
 * <p>
 * Diese Methode umfasst eigentlich die Hauptroutine. Hier werden zuerst einmal die internen wichtigen 
 * Variablen initialisiert. Anschliessend wird eine Startbepackung des Rucksacks von der Testfuntkion erfragt.<br />
 * Anschliessend wird (fuer insgesamt tabuiterations viele) Iterationen immer das Element aus der Nachbarschaft
 * gesucht, das die beste Summe der Nutzenwerte bringt und in den Rucksack gepackt bzw. das Element aus dem 
 * Rucksack gepackt, was zu der niedrigsten Verschlechterung fuehrt und zudem nicht in der Tabuliste steht.<br />
 *  Anschliessend wird der Index des Elementes fuer die Verweildauer von lifetime in die Tabuliste eingefuegt
 * Falls es kein Element in der Nachbarschaft mit diesen Eigenschaften gibt, so wird die Tabuliste gelscht und eine neue 
 * Iteration begonnen.</p>
 */
	public boolean startAlgo() throws Exception {
		Individual individual = null;
		Individual bestneighbour = null;
		TabuListElement tabu;
		
		//Startbepackung von der Testfunktion sich geben lassen. Dieses wird als Momentan beste Bepackung 
		//angesehen und hab fuer die erste Iteration die beste erreichte Fitness
		//Um sich das Startelement auch im Durchlauf zu sehen, wird auf der Konsole dieses einmal ausgegeben
		try {
			if (debug) System.out.println("Startelement: " + test.getStartPoint());
			individual = new Individual((ArrayList)test.getStartPoint(), (ArrayList)test.execute(test.getStartPoint()));
			bestindividual = individual;
			bestfitness = calculateFitness(individual);
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(-1);
		}
		
		for (int iterationcounter = 0; iterationcounter < maxGenerations; iterationcounter++) {
			if (debug) System.out.println("Iteration: " + iterationcounter);
				
			//besten Nachbarn ermitteln
			bestneighbour = findLeadingNeighbour(individual);
		
			//falls kein geeigneter Nachbar gefunden wird, die Tabuliste loeschen. Falls eine debug Ausgabe gewuenscht wird, 
			// wird auch dieses in der Konsole ausgegeben
			//ansonsten wird das Paretoarchiv ueberprueft
			// und ggf. um diesen nachbarn erweitert. Dazu halt das Paretoarchiv durchlaufen, mit allen 
			// Elementen vergleichen und falls es "besser" ist, hinzugepackt und die Elemente, die von der neuen Bepackung 
			// dominiert wird, wird gelscht
			if(bestneighbour == null){
				tabulist.clear();
				if (debug)System.out.println("Tabuliste geloescht");
			}else{
				paretoarchive = ParetoArchive.refreshArchive(archive, paretoarchive, bestneighbour);
				System.out.println("Paretoarchive ueberprueft");
				//debug-ausgabe: Falls es erwuenscht wird, wird nach jeder Iteration die momentane Bepackung (individual) sowie die 
				//Nachbarsbepackung (die beste) ausgegeben
				if(debug)System.out.println("Individuum " + booleanArrayListToInt(individual.getObjectVariables()) + individual.getObjectVariables());
				if(debug)System.out.println("bestneighbour " + booleanArrayListToInt(bestneighbour.getObjectVariables()) + bestneighbour.getObjectVariables());
				
				//Zug zur Tabulist hinzufuegen: dazu wird zuerst geprueft, in welchem Index sich die beiden Bepackungen unterscheiden
				//(dies entspricht dem Index des Elements, das eingepackt bzw aus dem Rucksack ausgepackt wurde).
				if (tabuIndex(bestneighbour, individual)!=-1){		//falls individual und bestnneighbour gleich sind. Dies sollte nicht passieren, deswegen 
					// wird ein Fehler auf die Konsole geschrieben, falls dieser Fall doch eintritt... denn dann funktioniert die Nachbarschaftsberechnung nicht
					//korrekt
					//Andernfalls wird das Element in die Tabuliste eingefuegt.
					addMoveToTabulist(individual, bestneighbour);
				}else {
					//Fehler, es drfte kein gleiches Element geliefert werden
					System.out.println("Fehler, Nachbarschaftsberechnung luft fehlerhaft!");
				}
				//nun ist die Iteration beendet, die Nachfolgebepackung wird zur aktuellen Bepackung fuer die kommende Iteration
			individual = bestneighbour;
			}
		}
			
			/* Tabuliste anpassen
			 * Die Tabudauer eines jeden Elementes in der Tabuliste wird um 1 reduzieren und ggf. aus der liste entfernen,
			 *falls die Verweildauer dadurch auf 0 gesetzt wird.
			 */
			for (int i = 0; i < tabulist.size(); i++) {
				tabu = (TabuListElement) tabulist.get(i);
				tabu.setLivetime(tabu.getLivetime()-1);
				if(tabu.getLivetime()==0){
					tabulist.remove(i);
					i--;
				}
			}
		
		//Ausgabe
		//die gehrt zwar hier eigentlich nicht hin, aber in main auch nicht
		AlgoOutput.writeFitness(paretoarchive, outputFileNamePrefix, false);
		AlgoOutput.writeObjectVariables(paretoarchive, outputFileNamePrefix, false);
		test.finalize();
		return true;
	}
	
	/**<p>
	 * Methode, die den Index des Elements zurueckgibt, das angefasst wurde
	 * </p>
	 * <p>
	 * Vergleicht die booleschen Indizes des 0-1 Strings, der bei uns die Bepackung kodiert und 
	 * gibt die Stelle der letzten Abweichung an und -1 wenn die beiden dargestellten Bepackungen 
	 * identisch sind.<br />
	 * Da mit dieser Methode nur 2 an sich benachbarter Konfigurationen / Bepackungen getestet werden,
	 * sollte immer nur ein Index unterschiedlich sein.
	 *  </p>
	 * @param indi1 : Individual (Bepackung)
	 * @param indi2 : Individual (Bepackung) 
	 * @return Index int, der das Element, welches im letzten Zug angefasst wurde
	 */
	private int tabuIndex (Individual indi1, Individual indi2) {
		int help=-1; //muss auf einen Indexwert gesetzt werden, den es auf keinen Fall gibt
		for (int i = 0; i < indi1.getObjectVariables().size();i++) {
			if (indi1.getObjectVariables().get(i)!=indi2.getObjectVariables().get(i)) 		
			help = i;
		}
		return help;
	}
	/**<p>
	 * Fitnessberechnung bei Eingabe einer Bepackung. </p>
	 * <p>Methode, die bei Eingabe eines Individuums den Fitneswert berechnet. Hierfuer ist der
	 * Fitnesswert einfach die Summe der einzelnen Nutzenwerte unseres Individuums (ein Element
	 * des Rucksacks). </p>
	 * @param individual : Bepackung als boolscher String
	 * @return Fitnesswert als Summe der einzelnen Nutzenwerte, geteilt durch die Anzahl der Nutzenwerte
	 */
	private double calculateFitness(Individual individual) {
		double sum = 0;
		//initialisierung der Summe mit 0, anschliessendes Durchgehen der einzenen Nutzenwerte aller Elemente, die 
		//im Rucksack sind, also im String mit 1 stehen.
		for (Iterator iter = individual.getFitness().iterator(); iter.hasNext(); ) {
			sum += ((Double)iter.next()).doubleValue();
		}
		return sum/dimension;
	}
	
	/**<p>
	 * Methode zur Nachbarschaftsberechnung</p>
	 * <p>
	 * Ermittelt die benachbarten Individuen (Bepackungen), indem erst einmal jedes Element einzeln 
	 * angefasst wird und in den Rucksack gepackt wird (falls es noch nicht drin ist) bzw aus dem Rucksack
	 * gepackt wird. Anschliessend wird von der neuen Bepackung der Fitnesswert bestimmt, nachdem getestet
	 * wird, ob dieser nicht in der Tabuliste steht. Es wird der beste Nachbar zurueckgegeben, dabei ist der beste Nachbar
	 * definiert als die Bepackung, die zu der besten Fitness fuehrt bzw zur kleinsten Verschlechterung.<br />
	 * Bei einer Bepackung, wobei das gekippte Bit (also das angefasste Element) in der Tabuliste steht, 
	 * wird ebenfalls ueberprueft, ob es eine bessere Fitness liefert als die bisher best erreichte (Aspirations-
	 * kriterium). Falls dieses zutrifft, wird diese Bepackung gewhlt.</p>
	 * @param individual : Bepackung
	 * @return beste Nachbarbepackung
	 */
	private Individual findLeadingNeighbour(Individual individual) {
		ArrayList neighbours = new ArrayList();
		int index=0;
		//das beste Individuum
		Individual leader = null;
		double leaderfitness = 0;
		for (int i = 0; i < dimension; i++) {
			//Variablen der Testfunktion von denen wir ausgehen
			ArrayList neighbourvariables = (ArrayList) individual.getObjectVariables().clone();
			//Bit i kippen!!
			if (((Boolean) neighbourvariables.get(i)).booleanValue()) {
				neighbourvariables.set(i,new Boolean(false));
			} else {
				neighbourvariables.set(i,new Boolean(true));
			}
			//nicht definierte Loesungen fangen
			try {
				Individual neighbour = new Individual(neighbourvariables,(ArrayList)test.execute(neighbourvariables));
				double fitness = calculateFitness(neighbour);			
				if (!isTabu(i)) {
						neighbours.add(neighbour);
					if (fitness <  bestfitness) {
						bestfitness = fitness;
	//					if (debug) System.out.println("Betrachtet wird Fitness von Individuum mit Index " + i);
						index=i;
					} 
					if (leader == null || fitness < leaderfitness) {
						leader = neighbour;
						leaderfitness = fitness;
						index=i;
					}
				} else {
					if (debug)System.out.println("Aspirationskriterium ueberpruefen");
					//aspirationskriterium: besser ist das!
					if (fitness <  bestfitness) {
						if (debug)System.out.println("Unterschied fitness vgl bestfitness: " + fitness + " : " + bestfitness);
						neighbours.add(neighbour);
						bestfitness = fitness;
						if (leader == null || fitness < leaderfitness) {
							leader = neighbour;
							leaderfitness = fitness;
						}
					}
					leader =null;
				}
			} catch (Exception e) {
/*				if (debug) {
					System.out.println("Kanditat war: " + neighbourvariables);
					System.out.println("Dies sollte der booleschen Darstellung von " + i +"entsprechen");
					System.out.println("\nDiese Loesung ist nicht definiert: Es gibt keine geeignete Nachbarschaft");
					} 
*/			}
			
		} 
		//debug Ausgabe der besten gefunden Nachbarns
		if (debug) System.out.println("bester Nachbar ist Element mit Index " + (index));
//		if(debug)System.out.println("leader " + leader.getObjectVariables());
		return leader;
		
	}
	
	/**
	 * <p>
	 * Zug in die Tabuliste einfuegen
	 * </p>
	 * <p>
	 * Methode, die den Index des Elements, was gerade geandert worden ist in der Bepackung 
	 * in die Tabuliste packt.<br />
	 * Dazu wird zuerst einmal geprueft, welches Bit sich denn in den beiden Eingabebepackungen 
	 * unterscheidet, anschliessend geprueft, ob das Element bereits in der Tabuliste ist (durch 
	 * Aspirationskriterium kann dieses sein). Falls dies der Fall ist, wird einfach die Tabudauer auf livetime 
	 * gesetzt. Falls das Element nicht in der Tabuliste steht, so wird der Index des Bits, was den Unterschied
	 * der Bepackungen klassifiziert in die Tabuliste mit der Verweildauer "livetime" eingefuegt. </p>
	 * @param from  Bepackung
	 * @param to Bepackung
	 */
	private void addMoveToTabulist(Individual from, Individual to) {
		boolean isin = false;
		int tabuindexvalue = tabuIndex(from, to);
		for(int i = 0; i < tabulist.size();i++) {
			if (((TabuListElement)tabulist.get(i)).getIndex() == tabuindexvalue) {
				isin = true;
				((TabuListElement)tabulist.get(i)).setLivetime(tabuiterations);
				if (debug)System.out.println("Verweildauer des Elements mit Index " + tabuindexvalue + " in der TabuListe zurueckgesetzt");
			}
		}
		if(!isin) {
			TabuListElement tabuelement= new TabuListElement();
			tabuelement.setLivetime(tabuiterations);
			tabuelement.setIndex(tabuIndex(from, to));
			tabulist.add(tabuelement);

			if (debug)System.out.println("Element mit Index " + tabuindexvalue + " in die TabuListe eingefuegt");
		}
		
	}
	
	
	/**
	 * <p>ueberprueft, ob ein Individuum in der Tabuliste steht</p>
	 * 
	 * <p>Dazu wird die Tabuliste von vorne nach hinten nach einem Element mit dem uebergebenen Index
	 * durchsucht. Falls dieses Element in der Tabuliste bereits vorhanden ist, wird true zurueckgegeben, 
	 * falls nicht, so wird false als Ruckgabewert zurueckgegeben.</p>
	 * @param index 
	 * @return boolean, ob Element in der TabuListe steht
	 */
	public boolean isTabu(int index) {
		for(int iter = 0; iter <tabulist.size(); iter++){
			if (((TabuListElement)tabulist.get(iter)).getIndex()==index){ 
			
				if (debug) System.out.println("Tabuliste enthaelt Index " + index);
			return true;
			}	
		}
		return false;
	}
	
	/**
	 * <p>Generiert eine Liste mit Boolean Variablen die eine Integer Zahl repraesentieren.</p>
	 * @param n die Integerzahl
	 * @return eine ArrayList mit Boolean Objekten die n Binaer darstellen
	 */
	private ArrayList intToBooleanArrayList(int n){
		ArrayList binaryNumber = new ArrayList();
		//zahl in einen binaerstring wandeln
		String binaryNumberString = Integer.toBinaryString(n);
		//fuehrende nullen (nicht im string) in die liste einf?gen
		for (int j = 0; j < (test.getNumberOfObjectives() - binaryNumberString.length()); j++) {
			binaryNumber.add(new Boolean(false));
		}
		//Boolean nach binaerdarstellung anhaengen
		for (int j = 0; j < binaryNumberString.length(); j++) {
			if (binaryNumberString.charAt(j)=='1') {
				binaryNumber.add(new Boolean(true));
			} else {
				binaryNumber.add(new Boolean(false));
			}
		}
		return binaryNumber;
	}
	
	/**
	 * Gibt eine Kurzbeschreibung des Algos zurueck.
	 * 
	 * @return Kurzbeschreibung des Algos	 
	 */
	public String toString() {
		String description = new String();
		description += "Tabu Search Algorithmus\n";
		description += "Beschreibung:\n";
		description += "Diese Klasse stellt eine Tabu Search Implementierung bereit. Sie kann kombinatorische Probleme\n";
		description += "die dem mehrkriteriellen 0/1 Rucksackproblem entsprechen lsen.\n";
		description += "Dazu wird in jeder Iteration die Nachbarschaft, dies sind alle Kombination, welche durch entfernen\n";
		description += "oder hinzufuegen eines Gegenstandes erzeugt werden knnen, berprft. Der Algorithmus whlt dann\n";
		description += "den besten Gegenstand dieser Nachbarschaft, falls der Zug dorthin nicht als Tabu deklariert ist,\n";
		description += "aus. Die Bewertung der Gegenstnde erfolgt durch Addition aller Nutzwerte, so dass ein\n";
		description += "einkriterieller Fitnesswert existiert. Das hinzunehmen oder entfernen eines Gegenstandes wird als\n";
		description += "Zug bezeichnet. Wenn ein Zug durchgefrt wir, darf dieser eine bestimmte Anzahl von Runden, der\n";
		description += "Tabudauer, nicht rckgngig gemacht werden. Dies nenen wir Tabu.\n";
		description += "Die Tabuliste enthlt alle Tabuisierten Zge. Es werden jedoch auch tabuisierte Zge durchgefhrt,\n" +
						"wenn dadurch die bisher beste Gesammtfitness erreicht werden kann (Aspiration).\n\n";
		description += "Kommandozeilenparameter:\n";
		description += "-t: Klasse der Testfunktion (nur Kobinatorische, z.B. KnapsackProblem)\n";
		description += "-g: (int): Anzahl der Generationen/Iterationen\n";
		description += "-o: optional, Praefix fuer die Dateinamen der Ausgabe (An diese wird noch" +
	    				" das Kuerzel fuer die Art der Ausgabedatei angehaengt, z.B. .best.res);" +
						" Wird der Wert nicht gesetzt, so wird ein Praefix aus Datum und Uhrzeit gewaehlt\n";
		description += "-d: optional, wenn gesetzt, werden Debug-Ausgaben angezeigt\n";
		description += "-archive (int): optional, Maximalgroesse des globalen Archives, welches am Ende ausgegeben wird; 0=unbegrenzt (Standardwert)\n";description += "-\n";
		description += "\n";
		description += "-maxtabulistlength: Die maximale Anzahl Zge in der Tabuliste\n";
		description += "-tabuiterations: Die maximale Verweildauer eines Zuges in der Tabuliste\n";
		description += "-\n";
		return description;
	}
	
	/**
	 * <p>Erstellt aus einer ArrayList mit Booleans den entsprechenden int Wert.</p>
	 * @param binaryNumber
	 * @return int-Wert
	 */
	private int booleanArrayListToInt(ArrayList binaryNumber) {
		StringBuffer binaryNumberStringBuffer = new StringBuffer ();
		for (Iterator iter = binaryNumber.iterator(); iter.hasNext();) {
			if (((Boolean)iter.next()).booleanValue()) {
				binaryNumberStringBuffer.append("1");
			} else {
				binaryNumberStringBuffer.append("0");
			}
		}
		return Integer.valueOf(binaryNumberStringBuffer.toString(), 2).intValue();
	}
	
	/**
	 * <p>Aufruf des TabuSearchAlgos, Mainmethode. </p>
	 * 
	 *<p> Um diesen Algorithmus starten zu koennen, muss eine neue Instanz der Kasse erzeugt werden, 
	 * anschliessend diese mit dem Aufruf der Methode startAlgo() starten.</p>
	 * @param args Angabe des Problems, Lebensdauer eines Elements in der TabuListe, maximale Laenge der TabuListe, maximale Anzahl Iterationen
	 */
	public static void main (String[] args) {
		TabuSearchKnapsack tsk = new TabuSearchKnapsack(args);
		try {
			tsk.startAlgo();
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	}
	  
}

