/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 package org.nobeltje.analysis;

import gif.LZWEncoder;
import gif.NeuQuant;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.image.BufferedImage;
import java.awt.image.DataBufferByte;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.util.Arrays;
import java.util.Properties;
import java.util.StringTokenizer;

import javax.swing.ImageIcon;

import org.nobeltje.tools.io.ArgsInputBeautifier;
import org.nobeltje.tools.io.SortFile;
import org.nobeltje.exceptions.MalInputFormatException;

/**
 * Die Klasse AnimatedGifEncoder enkodiert ein animiertes GIF File, das aus einem
 * oder mehreren Frames besteht.<br>
 * Untersttzte Bildformate fr die Frames sind GIF, JPG und PNG.
 * 
 * @author $Author: noot $
 * @version $Revision: 727 $
 * $Date: 2005-02-08 09:14:18 +0100 (Di, 08 Feb 2005) $
 */

public class AnimatedGifEncoder
{
	/**
	 * <p>Breite der Bilder</p>
	 */
	private int width;
	/**
	 * <p>Hhe der Bilder</p>
	 */
	private int height;
	/**
	 * <p>Transparenz-Farbe falls gegeben</p>
	 */
	private Color transparent = null;
	/**
	 * <p>Index der Transparenz-Farbe in der Farbtabelle</p>
	 */
	private int transIndex;
	/**
	 * <p>Anzahl der Wiederholungen<br>
	 * Default = 1, kein Loop</p>
	 */
	private int repeat = 1;
	/**
	 * Verzgerung zwischen den Frames in Hundertsteln<br>
	 * Default = 0</p>
	 */
	private int delay = 0;
	/**
	 * <p>Bereit, die Frames auszugeben</p>
	 */
	private boolean started = false;
	/**
	 * <p>Verzeichnisname, in dem sich die Eingabedaten befinden</p>
	 */
	private File dir;
	/**
	 * Sortierkriterium fr die Eingabedateien 
	 */
	private String sort = null;

	private OutputStream out;
	/**
	 * <p>Name der Ausgabedatei</p>
	 */
	private String outputFile;
	/**
	 * <p>Aktuelles Frame</p>
	 */
	private BufferedImage image;
	/**
	 * BGR Byte Array des Frames</p>
	 */
	private byte[] pixels;
	/**
	 * <p>Byte Array des Frames mit den Pixeln konvertiert zur Index Palette
	 */
	private byte[] indexedPixels;
	/**
	 * <p>Anzahl der Bits pro Pixel</p>
	 */
	private int colorDepth;
	/**
	 * <p>RGB Palette</p>
	 */
	private byte[] colorTab;
	/**
	 * <p>Aktive Palette-Eintrge</p>
	 */
	private boolean[] usedEntry = new boolean[256];
	/**
	 * <p>Gre der Farbtabelle (bits-1)</p>
	 */
	private int palSize = 7;
	/**
	 * <p>Disposal Code (-1 = Verwende Default)</p>
	 */
	private int dispose = -1;
	/**
	 * <p>Schliee den Stream beim Beenden</p>
	 */
	private boolean closeStream = false;
	
	private boolean firstFrame = true;
	/**
	 * <p>Fall false, ermittle die Gre aus dem ersten Frame</p<
	 */
	private boolean sizeSet = false;
	/**
	 * <p>Default Sample Intervall fr den Quantisierer</p>
	 */
	private int sample = 10;

	/**
	 * Setzt die Gre der GIF Frames.<br>
	 * Wenn diese Methode nicht aufgerufen wird, wird die Gre automatisch aus dem
	 * ersten Frame ermittelt.
	 *
	 * @param width - Breite der Frames
	 * @param height - Hhe der Frames
	 */
	public void setSize(int width, int height) {
		if (started && !firstFrame) return;
		this.width = width;
		this.height = height;

		if (this.width < 1) this.width = 320;
		if (this.height < 1) this.height = 240;
		sizeSet = true;
	}

	/**
	 * Setzt die Transparenz-Farbe fr das letzte hinzugefgte und alle folgenden
	 * Frames.<br>
	 * Da alle Farben beim Quantisieren modifiziert werden, wird in der endgltigen
	 * Palette von jedem Frame die Farbe genommen, die dieser am nchsten kommt.
	 * Kann auch auf Null gesetzt werden, falls es keine Transparenz gibt.
	 *
	 * @param color - Farbe, die als Transparenz-Wert behandelt wird
	 */
	public void setTransparent(Color color) {
		transparent = color;
	}

	/**
	 * Mit dem Repeat-Wert kann man angeben, wie oft die Animation wiederholt wird.<br>
	 * Default ist 1. 0 bedeutet eine Endlosschleife.<br>
	 * Die Methode mu aufgerufen werden, bevor das erste Frame hinzugefgt wird.
	 *
	 * @param repeat - Anzahl der Wiederholungen
	 */
	public void setRepeat(int repeat) {
		if (repeat >= 0) {
			this.repeat = repeat;
		}
	}

	/**
	 * Setzt die Verzgerung zwischen den einzelnen Frames oder ndert sie fr
	 * nachfolgende Frames. Bezieht sich immer auf das zuletzt eingefgte Frame.
	 *
	 * @param delay - Verzgerung im Millisekunden
	 */
	public void setDelay(int delay) {
		this.delay = Math.round(delay / 10.0f);
	}
	
	/**
	 * Setzt den Disposal Code fr das zuletzt hinzugefgte Frame und alle
	 * folgenden.<br>
	 *  Default ist 0, wenn keine Transparenz-Farbe gesetzt wurde, ansonsten 2.
	 *
	 * @param code - Disposal Code
	 */
	public void setDispose(int code) {
		if (code >= 0) {
			dispose = code;
		}
	}

	/**
	 * Setzt die Sample Qualitt fr die Quantisierung (In der GIF Spezifizierung
	 * ist ein Maximum von 256 Farben festgelegt).<br>
	 * Kleinere Werte (Minimum = 1) produzieren bessere Farben bei deutlich
	 * langsamerer Verarbeitung.<br>
	 * 10 ist Default und ergibt ein gutes Color Mapping bei angemessener
	 * Geschwindigkeit. Werte grer als 20 bringen keine bemerkenswerten
	 * Verbesserung der Geschwindigkeit mehr.
	 *
	 * @param quality - Sample Rate ( > 0 )
	 */
	public void setQuality(int quality) {
		if (quality < 1) quality = 1;
		sample = quality;
	}

	/**
	 * Initialisiert die Erstellung eines GIF Files im angegebenen Stream.<br>
	 * Der Stream wird nicht automatisch geschlossen.
	 *
	 * @param os - OutputStream in den die GIF Bilder geschrieben werden
	 * @return false - falls die Initialisierung fehlschlug
	 */

	public boolean start(OutputStream os) {
		if (os == null) return false;
		boolean ok = true;
		closeStream = false;
		out = os;
		try {
			// Header
			writeString("GIF89a");
		} catch (IOException e) {
			ok = false;
		}
		return started = ok;
	}

	/**
	 * Initialisiert das Schreiben eines GIF Files mit dem angegebenen Namen.<br>
	 *
	 * @param file - Name der Ausgabedatei
	 * @return false - falls die Initialisierung fehlschlug
	 */
	public boolean start(String file) {
		boolean ok = true;
		try {
			out = new BufferedOutputStream(new FileOutputStream(file));
			ok = start(out);
			closeStream = true;
		} catch (IOException e) {
			ok = false;
		}
		return started = ok;
	}

	/**
	 * Fgt das nchste GIF Frame hinzu.<br>
	 *
	 * Das Frame wird nicht sofort geschrieben, sondern bis zum nchsten Frame
	 * zurckgestellt, so dass noch Zeitvorgaben eingefgt werden knnen.<br>
	 * 
	 * Der Aufruf von <code>finish()</code> leert die Warteschlange.<br>
	 *
	 * Falls <code>setSize()</code> nicht aufgerufen wurde, wird die Gre des 
	 * ersten Bildes fr alle Frames genommen.
	 *
	 * @param bufferedImage - Enthlt das zu schreibende Bild
	 * @return true - wenn erfolgreich
	 */
	public boolean addFrame(BufferedImage bufferedImage) {
		if ((bufferedImage == null) || !started)
			return false;

		try {
			if (!sizeSet)
				// Benutze die Gre des ersten Frames
				setSize(bufferedImage.getWidth(), bufferedImage.getHeight());

			image = bufferedImage;
			// Konvertiere ins korrekte Format, falls ntig
			getImagePixels();
			// Erstelle Farbtabelle und mappe die Pixel
			analyzePixels();
			if (firstFrame) {
				// Logical Screen Descriptor
				writeLSD();
				// Globale Farbtabelle
				writePalette();
				if (repeat >= 0)
					// Benutze Netscape Extension, um die Anzahl der Wiederholungen anzugeben
					writeNetscapeExt();
			}
			// Schreibe Graphic Control Extension
			writeGraphicCtrlExt();
			// Schreibe Image Descriptor
			writeImageDesc();
			if (!firstFrame)
				// Lokale Farbtabelle
				writePalette();

			// Kodiere und schreibe Pixel-Daten
			writePixels();
			firstFrame = false;
		} catch (IOException e) {
			return false;
		}

		return true;
	}
	
	/**
	 * Leert die Warteschlange der anstehenden Daten und schliet die Ausgabedatei.<br>
	 * Falls in einen OutputStream geschrieben wird, bleibt dieser geffnet.
	 * 
	 * @return true - wenn erfolgreich
	 */
	public boolean finish() {
		if (!started) return false;

		started = false;

		try {
			// Schreibe GIF Header
			out.write(0x3b);
			out.flush();
			if (closeStream)
				out.close();
		} catch (IOException e) {
			return false;
		}

		// Reset fr folgende Aufrufe
		transIndex		= 0;
		out				= null;
		image			= null;
		pixels			= null;
		indexedPixels	= null;
		colorTab		= null;
		closeStream		= false;
		firstFrame		= true;

		return true;
	}

	/**
	 * Liest ein Bild ein und speichert es in einem BufferedImage
	 * 
	 * @param file - der Name der Bilddatei
	 * @return das Bild in einem BufferedImage
	 */
	public BufferedImage createBufferedImage(String file) {
		// Hole das Bild
		ImageIcon icon = new ImageIcon(file);
		Image image = icon.getImage();

		// Erstelle ein leeres BufferedImage mit der Gre des Bildes
		BufferedImage buffImage = 
			new BufferedImage(image.getWidth(null), 
							  image.getHeight(null),
							  BufferedImage.TYPE_INT_ARGB);

		// Zeichne das Bild ins BufferedImage
		buffImage.getGraphics().drawImage(image, 0, 0, null);

		return buffImage;
	}

	/**
	 * Wandelt einzelne Bilder in ein animiertes GIF um.
	 * 
	 * @param args - enthlt die Kommandozeilenparameter
	 */
	public void runEncoder(String[] args) {
		System.out.println("Beginne Enkodierung...\n");
		
		// Initialisiere den Encoder
		try {
			Properties input = ArgsInputBeautifier.beautify(args);
			dir = new File(ArgsInputBeautifier.getStringProperty(input, "d", "", true));
			if (!dir.exists()) {
				System.err.println("Verzeichnis nicht vorhanden");
				System.exit(1);	// Direkter Abbruch des Programms
			}
			outputFile = ArgsInputBeautifier.getStringProperty(input, "o", "AnimatedGIF.gif", false);

			setDelay(ArgsInputBeautifier.getIntProperty(input, "delay", 500, false));

			setRepeat(ArgsInputBeautifier.getIntProperty(input, "repeat", 0, false));
	
			sort = ArgsInputBeautifier.getStringProperty(input, "sort", null, false);
		}
		catch (MalInputFormatException mife) {
			mife.printStackTrace();
			System.exit(1);
		}

		start(outputFile);
		
		// Hole aus dem angegebenen Verzeichnis alle Dateinamen mit der
		// Endung die im Fitness Filter festgelegt ist
		String suffixes[] = { "gif", "jpg", "png" };
		File imageFiles[] = dir.listFiles(new CustomFilter(suffixes));

		if (sort != null) {
			StringTokenizer tokenizer = new StringTokenizer(sort, ",");
			while(tokenizer.hasMoreElements())
				imageFiles = SortFile.sortFileList(imageFiles, tokenizer.nextToken());
		} else
			Arrays.sort(imageFiles);

		for (int i = 0; i < imageFiles.length; i++) {
			addFrame(createBufferedImage(imageFiles[i].getAbsolutePath()));
			System.out.println("\"" + imageFiles[i].getName() + "\" hinzugefgt");
		}

		finish();
		System.out.println("\nEnkodierung beendet");
	}

	/**
	 * Kommandozeilenparameter des AnimatedGifEncoders:
	 * <ul>
	 *   <li><b>-d</b> Name des Quellverzeichnisses fr die Grafiken</li>
	 *   <li><b>-o</b> Name der Ausgabedatei (optional; default="AnimatedGIF.gif")</li>
	 *   <li><b>-delay</b> Verzgerung zwischen den Einzelframes in Millisekunden (optional; default=500)</li>
	 *   <li><b>-repeat</b> Anzahl der Wiederholungen (optional; default=Endlosschleife</li>
	 * </ul><br>
	 * <p>Beispiel: -d PictureSrc -o MyAni.gif -delay 200 -repeat 1
	 * </p>
	 */
	public static void main (String[] args) {
		AnimatedGifEncoder encoder = new AnimatedGifEncoder();

		encoder.runEncoder(args);
	}

	/**
	 * Analysiere die Bildfarben und erstelle eine Farbtabelle
	 */
	private void analyzePixels() {
		int len = pixels.length;
		int nPix = len / 3;
		indexedPixels = new byte[nPix];
		NeuQuant nq = new NeuQuant(pixels, len, sample);

		// Initialisiere Quantisierer
		// Erstelle die reduzierte Farbpalette
		colorTab = nq.process();
		// Konvertiere von BGR nach RGB
		for (int i = 0; i < colorTab.length; i += 3) {
			byte temp = colorTab[i];
			colorTab[i] = colorTab[i + 2];
			colorTab[i + 2] = temp;
			usedEntry[i / 3] = false;
		}
		// Mappe die Bildpixel auf die neue Palette
		for (int i = 0, k = 0; i < nPix; i++) {
			int index = nq.map(pixels[k++] & 0xff,
							   pixels[k++] & 0xff,
							   pixels[k++] & 0xff);
			usedEntry[index] = true;
			indexedPixels[i] = (byte)index;
		}
		pixels = null;
		colorDepth = 8;
		palSize = 7;
		// Nehme die beste Entsprechung fr die Transparenz-Farbe, falls spezifiziert
		if (transparent != null) {
			transIndex = findClosest(transparent);
		}
	}
	
	/**
	 * Sucht die naheliegendste Farbe aus der Tabelle zu einer gegebenen Farbe.
	 * 
	 * @return Palette Index der gefundenen Farbe 
	 */
	private int findClosest(Color c) {
		if (colorTab == null) return -1;
		int r = c.getRed();
		int g = c.getGreen();
		int b = c.getBlue();
		int minpos = 0;
		int dmin = 256 * 256 * 256;
		int len = colorTab.length;
		for (int i = 0; i < len;) {
			int dr = r - (colorTab[i++] & 0xff);
			int dg = g - (colorTab[i++] & 0xff);
			int db = b - (colorTab[i] & 0xff);
			int d = dr * dr + dg * dg + db * db;
			int index = i / 3;
			if (usedEntry[index] && (d < dmin)) {
				dmin = d;
				minpos = index;
			}
			i++;
		}
		return minpos;
	}
	
	/**
	 * Schreibt die Bildpixel in das Byte Array "pixels"
	 */
	private void getImagePixels() {
		int w = image.getWidth();
		int h = image.getHeight();
		int type = image.getType();
		if ((w != width)
			|| (h != height)
			|| (type != BufferedImage.TYPE_3BYTE_BGR)) {
			// Erstelle ein neues Bild mit korrekter Gre/Format
			BufferedImage temp = new BufferedImage(width, height, BufferedImage.TYPE_3BYTE_BGR);
			Graphics2D g = temp.createGraphics();
			g.drawImage(image, 0, 0, null);
			image = temp;
		}
		pixels = ((DataBufferByte) image.getRaster().getDataBuffer()).getData();
	}
	
	/**
	 * Schreibt die Graphic Control Extension
	 */
	private void writeGraphicCtrlExt() throws IOException {
		// Extension Introducer
		out.write(0x21);
		// GCE Label
		out.write(0xf9);
		// Gre der Datenblcke
		out.write(4);
		int transp, disp;
		if (transparent == null) {
			transp = 0;
			// dispose = keine Aktion
			disp = 0;
		} else {
			transp = 1;
			// Erzwinge Clear, falls eine Transparenz-Farbe benutzt wird
			disp = 2;
		}
		if (dispose >= 0)
			// Benutze berschreiben
			disp = dispose & 7;

		disp <<= 2;

		// Gepackte Felder
		out.write(0 |			// 1:3 reserved
			   disp |			// 4:6 disposal
			      0 |			// 7   user input - 0 = none
		     transp);			// 8   transparency flag

		writeShort(delay);		// delay x 1/100 sec
		out.write(transIndex);	// transparent color index
		out.write(0);			// block terminator
	}
	
	/**
	 * Schreibt den Image Descriptor
	 * 
	 * @throws IOException
	 */
	private void writeImageDesc() throws IOException {
		out.write(0x2c);	// Bild Separator
		writeShort(0);		// Bild Position x,y = 0,0
		writeShort(0);
		writeShort(width);	// Bildgre
		writeShort(height);
		// Gepackte Felder
		if (firstFrame)
			// keine LCT  - GCT wird fr das erste (oder einzige) Frame benutzt
			out.write(0);
		else {
			// Spezifiziere normale LCT
			out.write(0x80 |	// 1 Lokale Farbtabelle	- 1 = ja
						 0 |	// 2 Interlaced			- 0 = nein
						 0 |	// 3 Sortiert			- 0 = nein
						 0 |	// 4-5 Reserviert
				   palSize);	// 6-8 Gre der Farbtabelle
		}
	}
	
	/**
	 * Schreibe den logischen Screen Descriptor
	 */
	private void writeLSD() throws IOException {
		// Logische Screen Size
		writeShort(width);
		writeShort(height);
		// Gepackte Felder
		out.write((0x80 |	// 1	: Globales Farbtabellen-Flag = 1 (Benutze GCT)
				   0x70 |	// 2-4	: Farbauflsung = 7
				   0x00 |	// 5	: GCT Sortierungs-Flag = 0
			   palSize));	// 6-8	: GCT Gre

		out.write(0);		// Hintergrund Farbindex
		out.write(0);		// Pixel Aspect Ratio - 1:1 angenommen
	}
	
	/**
	 * Schreibe Netscape Extension um die Anzahl der Wiederholungen zu definieren.
	 * 
	 * @throws IOException
	 */
	private void writeNetscapeExt() throws IOException {
		out.write(0x21);	// Extension Introducer
		out.write(0xff);	// App Extension Label
		out.write(11);		// Block Gre
		writeString("NETSCAPE" + "2.0"); // App ID + Auth Code
		out.write(3);		// Sub-Block Gre
		out.write(1);		// Loop Sub-Block ID
		writeShort(repeat);	// Loop Count (Extra Iterationen, 0 = Endlosschleife)
		out.write(0);		// Block Terminator
	}
	
	/**
	 * Schreibt Farbtabelle
	 * 
	 * @throws IOException
	 */
	private void writePalette() throws IOException {
		out.write(colorTab, 0, colorTab.length);
		int n = (3 * 256) - colorTab.length;
		for (int i = 0; i < n; i++)
			out.write(0);
	}
	
	/**
	 * Enkodiert und schreibt Pixel Daten
	 * 
	 * @throws IOException
	 */
	private void writePixels() throws IOException {
		LZWEncoder encoder = new LZWEncoder(width, height, indexedPixels, colorDepth);
		encoder.encode(out);
	}
	
	/**
	 * Schreibt 16-Bit Werte in den OutputStream, LSB zuerst
	 */
	private void writeShort(int value) throws IOException {
		out.write(value & 0xff);
		out.write((value >> 8) & 0xff);
	}
	
	/**
	 * Schreibt einen String in den OutputStream
	 */
	private void writeString(String s) throws IOException {
		for (int i = 0; i < s.length(); i++)
			out.write((byte) s.charAt(i));
	}
}
