/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on Jun 14, 2004
 *
 */
package org.nobeltje.metrics;

import java.util.ArrayList;
import java.util.List;

import org.nobeltje.tools.ParetoDomination;
import org.nobeltje.tools.io.FileParser;
import org.nobeltje.exceptions.MalInputFormatException;

/**
 * Die Klasse CMetric berechnet die schwach dominierten Punkte einer Loesungsmenge eines Algorithmus,<br>
 * im Verhltnis zu einer Loesungsmenge eines anderen Algorithmus.<br>
 * Dazu wird fuer jedes Element eines Algorithmus geprueft, ob dieses durch ein Element des anderen<br> 
 * Algorithmus schwach dominiert wird. Diese Anzahl wird durch die Anzahl der Elemente des zu <br>
 * bewertenden Algortihmus genommen. <br>
 * 
 * 
 * @author schulteb
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 * 
 */


public class CMetric implements Metric
{
	private final int parameterCount = 2;	// Gibt an, wieviele Parameter die
											// Metrik als Eingabe erwartet

	/**
	 * Methode berechnet fuer jeden Vektor einer Liste, ob dieser durch einen
	 * Vektor der anderen Liste dominiert wird, und zaehlt die Anzahl der
	 * dominierten Vektoren in der ersten Liste.<br>  
	 * 
	 * @param vectorList1 eine Liste von Vektoren eines zu bewertenden Algortihmus
	 * @param vectorList2 eine zweite Liste von Vektoren des anderen Algorithmus
	 * @return Anzahl der dominierten Punkte der Liste von Vektoren des zu bewertenden Algorithmus
	 */
	public double getNumberOfDominatedVectors(List vectorList1, List vectorList2){
		double number = 0.0;
		
		for (int i = 0; i < vectorList1.size(); i++) {
			for (int j = 0; j < vectorList2.size(); j++) {
				if (ParetoDomination.checkDomination((ArrayList)vectorList1.get(i), (ArrayList)vectorList2.get(j)) == ParetoDomination.DOMINATION_IND2)
				  number++;
			}
		}
		return number;
	}
	
	
	/**
	 * Methode berechnet die Anzahl der dominierten Punkte einer Liste durch die Anzahl der Punkte <br> 
	 * dieser Liste fuer die zu bewertenden Algortihmen.<br>
	 * Das Ergebnis der Metrik spiegelt die schwache Dominanz der Algorithmen untereinander wieder.
	 * 
	 * @param list1 eine Liste von Vektoren
	 * @param list2 eine zweite Liste
	 * @return Verhaeltnis der Domination der einzelnen Listen
	 * @throws Exception, wenn berechnete Werte gleich sind
	 */
	public double computeMetric(List list1, List list2) throws Exception {
		double a = (getNumberOfDominatedVectors(list2,list1)/(double)list2.size());//Berechne C(A,B)
		double b = (getNumberOfDominatedVectors(list1,list2)/(double)list1.size());//Berechne C(B,A)
 
		if(a == b)//Die Mengen koennen zwar die gleichen Ergebnisse der C-Metrik haben, aber eine Menge kann der ParetoFront trotdem naeher sein
			throw new Exception ("Diese Metrik kann leider nicht diese Mengen bewerten");
		else
		{
		    if(a == 0.0)
		    	System.out.println("Kein Punkt aus Algorithmus2 wird durch einen Punkt aus Algorithmus1 schwach dominiert!");
		    if(a == 1.0)
		    	System.out.println("Alle Punkte aus Algorithmus2 werden durch mindestens einen Punkt aus Algorithmus1 schwach dominiert!");
		}
		return a;
	}
	
	/**
	 * @throws Exception, das die CMetric zwei vektolisten vergleicht
	 */
	public double computeMetric(List list1) throws Exception{
		throw new Exception("Diese Methode wird von der C-Metrik nicht unterst?tzt");
	}
	
	/**
	 * Gibt eine Kurzbeschreibung des Algos zurueck.
	 * @return Beschreibung der CMetrik
	 */
	public String toString() 
	{
		String str;
		str =  "\n\nKlasse C - Metric\n";
		str += "=================\n";
		str += "\n";
		str += "Beschreibung:\n";
		str += "Implementation der C - Metrik im mehrdimensionalen Raum," +
			"nach Ideen von ??\n";
		str +="Die CMetric vergleicht zwei Mengen zweier verschiedener Algorithmen.\n";
		str +="Es wird berechnet, ob die Menge zwei die Menge eins dominiert. \n";
		str +="Dazu werden alle dominierten Punkte gezaehlt und durch die Anzahl aller Punkte der zweiten Menge geteilt \n";
		str +="Bekommt man als Ergebnis 0 heraus , wird kein Punkt aus Algorithmus2 durch einen Punkt aus Algorithmus1 schwach dominiert!\n";
		str +="Bekommt man als Ergebnis 1, werden  Punkte aus Algorithmus2 durch mindestens einen Punkt aus Algorithmus1 schwach dominiert\n";
		str +="\n";
	    str +="Kommandozeilenparameter:\n";
		str +="-liste(ArrayList1): Menge des Algorithmus1\n";
		str +="-liste(ArrayList2): Menge des Algorithmus2\n";
		return str;
	}

	/**
	 * @return Anzahl der benoetigten Vektorlisten
	 */
	public int getParameterCount() {
		return parameterCount;
	}

	/**
	 * Die Main Methode legt ein neues Objekt der Metrik an. Sie liest die Daten
	 * aus den Files, die als Parameter uebergen wurden, in ArrayLists ein.
	 * Schlielich werden die Daten durch den Aufruf von computeMetric() ausgewertet.<br>
	 * 
	 * @param args enthaelt die Dateinamen fuer die Daten
	 * @throws Exception
	 */
	public static void Main(String[] args) throws Exception
	{
		SMetric metric = new SMetric();

		if (args.length != metric.getParameterCount())
			throw new MalInputFormatException("Die Metrik erwartet " +
					metric.getParameterCount() + " Dateinamen als Parameter");

		List list1 = new ArrayList();
		List list2 = new ArrayList();
		FileParser.parseFile(args[0], list1);
		FileParser.parseFile(args[1], list2);

		double output = metric.computeMetric(list1, list2);

		System.out.println(output);
	}
}