/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on 28.05.2004
 */
package org.nobeltje.metrics;

import java.util.ArrayList;
import java.util.List;

import org.nobeltje.tools.io.FileParser;
import org.nobeltje.exceptions.MalInputFormatException;

/**
 * @author $Author: noot $
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 */

public class SMetric implements Metric
{
	private final int parameterCount = 1;	// Gibt an, wieviele Parameter die
											// Metrik als Eingabe erwartet

	/**
	 * Berechnet, ob 'vector2' von 'vector1' dominiert wird unter Beruecksichtigung
	 * der ersten 'nObjectives' Werte
	 * @param vector1
	 * @param vector2
	 * @return boolean
	 * @throws Exception
	 */
	private boolean dominates(List vector1, List vector2, int nObjectives) throws Exception
	{
		if (vector1.size() != vector2.size())
			throw new Exception("Die Dimension der Vektoren stimmt nicht ?berein!");

		boolean  betterInAnyObjective = false;
		int  i;

		for (i = 0; i < nObjectives && ((Double)vector1.get(i)).doubleValue() >= ((Double)vector2.get(i)).doubleValue(); i++)
			if (((Double)vector1.get(i)).doubleValue() > ((Double)vector2.get(i)).doubleValue())
				betterInAnyObjective = true;
		return (i >= nObjectives && betterInAnyObjective);
	}


	/**
	 * Tauscht die Vektoren 'i' und 'j'
	 * @param vectorList Menge der Vektoren
	 * @param i
	 * @param j
	 */
	private void swap(List vectorList, int i, int j)
	{
		List temp;

		temp = (List)vectorList.get(i);
		vectorList.set(i, vectorList.get(j));
		vectorList.set(j, temp);
	}


	/**
	 * Alle nichtdominierten Punkte werden betrachtet in Bezug auf die ersten
	 * 'nObjectives' Werte. 'vectorList' wird so sortiert, dass die nichtdominierten
	 * Vektoren an den Stellen [0..n-1] stehen
	 * @param vectorList Menge der Vektoren
	 * @param nPoints
	 * @param nObjectives
	 * @return n
	 * @throws Exception
	 */
	public int sortNondominatedSet(List vectorList, int nPoints, int nObjectives) throws Exception
	{
		int i = 0;
		int j;

		while (i < nPoints) {
			j = i + 1;
			while (j < nPoints) {
				if (dominates((List)vectorList.get(i), (List)vectorList.get(j), nObjectives)) {
					// Entferne Punkt 'j'
					nPoints--;
					swap(vectorList, j, nPoints);
				} else if (dominates((List)vectorList.get(j), (List)vectorList.get(i), nObjectives)) {
					// Entferne Punkt 'i'
					// Der Punkt muss im naechsten Durchlauf nochmal betrachtet
					// werden; deswegen dekrementiere 'i'
					nPoints--;
					swap(vectorList, i, nPoints);
					i--;
					break;
				} else
					j++;
			}
			i++;
		}
		return nPoints;
	}


	/**
	 * Betrachte fuer alle Vektoren den Wert an der Stelle 'objective'
	 * und berechne den kleinsten Wert
	 * @param vectorList Menge der Vektoren
	 * @param nPoints Die ersten 'nPoints' Vektoren werden betrachtet
	 * @param objective Der Wert an der Stelle 'objective' wird betrachtet
	 * @return Der kleinste, gefundene Wert
	 */
	private double surfaceUnchangedTo(List vectorList, int nPoints, int objective) throws Exception
	{
		double minValue, temp;

		if (nPoints < 1)
			throw new Exception("Keine Punkte mehr vorhanden");

		minValue = ((Double)((List)vectorList.get(0)).get(objective)).doubleValue();
		for (int i = 1; i < nPoints; i++) {
			temp = ((Double)((List)vectorList.get(i)).get(objective)).doubleValue();
			if (minValue > temp)
				minValue = temp;
		}
		return minValue;
	}


	/**
	 * Entferne alle Punkte, die an der Stelle 'objective' einen kleineren
	 * Wert als 'threshold' haben.
	 * Die verbleibenden Vektoren stehen in 'vectorList' an den Stellen [0..n-1]
	 * @param vectorList Menge der Vektoren
	 * @param nPoints Die ersten 'nPoints' Vektoren werden betrachtet
	 * @param objective Der Wert an der Stelle 'objective' wird betrachtet
	 * @param threshold
	 * @return n
	 */
	private int reduceNondominatedSet(List vectorList, int nPoints, int objective, double threshold)
	{
		for (int i = 0; i < nPoints; i++)
			if (((Double)((List)vectorList.get(i)).get(objective)).doubleValue() <= threshold) {
				nPoints--;
				swap(vectorList, i, nPoints);
			}
		return nPoints;
	}


	/**
	 * Berechnet das Hypervolumen, das durch die Vektorenfront aus 'vectorList'
	 * dominiert wird.
	 * @param vectorList Menge der Vektoren
	 * @param nPoints
	 * @param nObjectives
	 * @return Das berechnete Hypervolumen
	 * @throws Exception
	 */
	private double calculateHypervolume(List vectorList, int nPoints, int nObjectives) throws Exception
	{
		double volume = 0;
		double distance = 0;

		while (nPoints > 0) {
			int     nNondominatedPoints;
			double  tempVolume, tempDistance;

			nNondominatedPoints = sortNondominatedSet(vectorList, nPoints, nObjectives - 1);

			if (nObjectives < 3) {
				if (nNondominatedPoints < 1)
					throw new Exception("Keine Punkte mehr vorhanden");
				tempVolume = ((Double)((List)vectorList.get(0)).get(0)).doubleValue();
			}
			else
				tempVolume = calculateHypervolume(vectorList, nNondominatedPoints, nObjectives - 1);

			tempDistance = surfaceUnchangedTo(vectorList, nPoints, nObjectives - 1);
			volume += (tempVolume * (tempDistance - distance));
			distance = tempDistance;
			nPoints = reduceNondominatedSet(vectorList, nPoints, nObjectives - 1, distance);
		}
		return volume;
	}


	/**
	 * @param vectorList Menge von nicht dominierten Individuen
	 * @return Das berechnete Hypervolumen
	 * @throws Exception
	 */
	public double computeMetric(List vectorList) throws Exception
	{
		int redSizeFront;
		int nObjectives;

		if ((nObjectives = ((List)vectorList.get(0)).size()) < 2)
			throw new Exception("Dimension der Vektoren muss >= 2 sein");

		// Berechne dominiertes Hypervolumen
		redSizeFront = sortNondominatedSet(vectorList, vectorList.size(), nObjectives);
		return calculateHypervolume(vectorList, redSizeFront, nObjectives);
	}


	public double computeMetric(List list1, List list2) throws Exception
	{
		throw new Exception("Diese Methode wird von der S-Metrik nicht unterstuetzt");
	}

	//die Beschreibung der SMetrik in toString() weiterfuehren
	public String toString()
	{
		String str;
		str =  "\n\nKlasse SMetric\n";
		str += "=================\n";
		str += "\n";
		str += "Beschreibung:\n";
		str += "Implementation der S - Metrik, fuer den mehrdimensionalen Raum" +
			"nach Ideen von E. Zitzler, K. Deb, L. Thiele\n";
		return str;
	}


	/**
	 * Gibt an, wieviele Parameter (Dateien) die Metrik erwartet
	 * @return Anzahl der Dateien
	 */
	public int getParameterCount()
	{
		return parameterCount;
	}

	/**
	 * Die Main Methode legt ein neues Objekt der Metrik an. Sie liest die Daten
	 * aus den Files, die als Parameter uebergen wurden, in ArrayLists ein.
	 * Schlie?lich werden die Daten durch den Aufruf von computeMetric() ausgewertet.<br>
	 * 
	 * @param args enthaelt die Dateinamen fuer die Daten
	 * @throws Exception
	 */
	public static void Main(String[] args) throws Exception
	{
		SMetric metric = new SMetric();

		if (args.length != metric.getParameterCount())
			throw new MalInputFormatException("Die Metrik erwartet " +
					metric.getParameterCount() + " Dateinamen als Parameter");

		List list = new ArrayList();
		FileParser.parseFile(args[0], list);

		double output = metric.computeMetric(list);

		System.out.println(output);
	}
}