/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 package org.nobeltje.simulator.elevatorcon;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;
import java.util.logging.Level;
import java.util.logging.Logger;

import org.nobeltje.testfkt.AbstractTestfkt;
import org.nobeltje.tools.Individual;
import org.nobeltje.tools.Particle;
import org.nobeltje.tools.RandomPG;
import org.nobeltje.tools.io.PropertiesPG;
import org.nobeltje.algo.AbstractAlgo;
import org.nobeltje.algo.MueRhoLES;

/**
 * Diese Klasse bindet einen in C programmierten S-Ring Fahrstuhlsimulator in das Nobeltje Framework ein.
 * Es wird davon ausgegangen, dass die Verzeichnisstruktur des CVS-Repositories gegeben ist. In diesem Falle wird
 * der Simulator automatisch gefunden und ggf. kompiliert (sofern make installiert und in der Systemumgebung
 * verfuegbar ist). Ansonsten kann der Speicherort des Simulators als Kommandozeilenparameter angegeben werden.
 * 
 * Als Kriterium wird die Ankunftswahrscheinlichkeit verwendet. Unter Beruecksichtigung der Anzahl der Stockwerke
 * und der Anzahl der eingesetzten Wagen wird fuer einen Gewichtsvektor die Guete des beschriebenen Fahrstuhlsystems
 * bei der gegebenen Ankunftswahrscheinlichkeit berechnet. Wenn ein Fahrstuhlsystem fuer mehr als eine Wahrscheinlichkeit
 * gute Ergebnisse leifern soll, wird das Problem mehrkriteriell; die Anzahl der Optimierungskriterien entspricht
 * also der Anzahl der Ankunftswahrscheinlichkeiten, fuer die optimiert werden soll.
 * 
 * @author $Author: noot $
 * @version $Revision: 726 $<br>
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 *
 */
public class SRingCon extends AbstractTestfkt {

	private static final int POSTPROCESSING_REPEATS = 100;
	
	private Logger log;
	private Level logLevel=Level.WARNING;
	private int logLevelInt;
	
	// Leider noetig fuer vernuenftiges LogLevel-Abfragen per if
	private final int LOG_OFF=0;
	private final int LOG_SEVERE = 1;
	private final int LOG_WARNING = 2;
	private final int LOG_INFO = 3;
	private final int LOG_CONFIG =4;
	private final int LOG_FINE = 5;
	private final int LOG_FINER = 6;
	private final int LOG_FINEST = 7;
	

//	private float[] hallCallProb={0.3f,0.4f}; //Ankunftswahrscheinlichkeiten
	private ArrayList hallCallProb; //Ankunftswahrscheinlichkeiten
	private int floorCount=6; //Anzahl der Stockwerke
	private int cars=2; //Anzahl der Wagen im Aufzugsystem
	private int seed=13; //Random-Seed
	private int iterations=10000; //Anzahl der zu simulierenden Iterationen

	private String pathToSim = ""; //Absoluter Pfad zum Simulator
	private String[] cmd; 
	private Runtime runtime=Runtime.getRuntime();

	private String[] resultList;

	
	/**
	 * Diese Methode liefert Informationen ueber die Verwendung an den Benutzer und wird von der Help-Klasse
	 * ausgelesen. 
	 */
	public String toString()
	{
		String info;
		info="\nSRingCon.java - Java-Schnittstelle zum SRing-Fahrstuhlsimulator\n";
		info+="---------------------------------------------------------------\n";
		info+="Der Simulator liegt als C-Quelle in diesem Repository vor und wird ggf. automatisch kompiliert\n";
		info+="(sofern die Umgebung des OS dies zulaesst).\n";
		info+="Am Ende der Simulation bereinigt diese Schnittstelle die verrauschten Ergebnisse der gefundenen\n";
		info+="Individuen und gibt diese zurueck.\n";
		info+="Moegliche Optionen fuer diese Schnittstelle (alle optional):\n";
		info+="-tti: Anzahl der vom Simulator zu verwendenden Iterationen (Standard:"+iterations+")\n";
		info+="-ttn: Anzahl der vom Simulator zu verwendenden Stockwerke (Standard:"+floorCount+")\n";
		info+="-ttm: Anzahl der Fahrstuhlwagen im zu simulierenden Gebaeude (Standard: "+cars+")\n";
		info+="-ttp: Wahrscheinlichkeiten fuer Ankunft einer Person auf einem Stockwerk. " +
				"Mehrere Angaben mit \":\" getrennt (Bsp: -ttp 0.3:0.4). " +
				"Hiermit wird die Anzahl der Kriterien festgelegt (Standard: 0.3:0.4)\n";
		info+="-ttr: Random-Seed (Standard: randomisiert)\n";
		info+="-ttd: Debug-Modus\n";
		info+="    0 -> Keine Ausgaben\n";
		info+="    1 -> SEVERE (bislang noch nicht implementiert)\n";
		info+="         Schwere Fehler werden in Form von \"Exceptions\" immer gemeldet.\n";
		info+="    2 -> WARNING (Standard)\n";
		info+="    3 -> INFO\n";
		info+="    4 -> CONFIG\n";
		info+="    5-7 -> FINER\n";
		info+="-ttpath: Pfad zum Simulator (nur notwendig, wenn nicht die Verzeichnisstruktur des CVS-Repositories\n";
		info+="      gegeben ist oder ein anderer Simulator verwendet werden soll).\n";
		info+="Die Ankunftswahrscheinlichkeiten sind momentan fest vorgegeben (0.3 und 0.4); somit auch die\n";
		info+="Dimension und Kriterialitaet dieses Simulators (=Anzahl der Wahrscheinlichkeiten), da jede\n";
		info+="Wahrscheinlichkeit als ein Optimierungskriterium angesehen wird.\n\n";
		return info;
	}
	
	
	/**
	 * Initialisierung des Adapters: Variablen initialisiern und ggf. mit Werten der Kommandozeile
	 * oder einer Konfigurationsdatei belegen.
	 * ring-Aufruf initialisieren
	 * 
	 * @see testfkt.TestFunctions#init(tools.io.PropertiesPG)
	 */
	public void init(PropertiesPG params) {
		// Super-Konstruktor
		super.init(params);
		// Anzahl Stockwerke
		floorCount = specificTestfktParameters.getIntProperty("n", floorCount, false);
		// Anzahl Aufzuege
		cars = specificTestfktParameters.getIntProperty("m", cars, false);
		// Anzahl Iterationen
		iterations = specificTestfktParameters.getIntProperty("i", iterations, false);
		// Absoluter Pfad zum Simulator
		pathToSim = specificTestfktParameters.getStringProperty("path", "", false);
		// Random-Seed
		if (specificTestfktParameters.containsKey("r")) {
			seed = specificTestfktParameters.getIntProperty("r", 0, true);
		} else {
			RandomPG rand = new RandomPG();
			seed = rand.nextInt();
		}
		// Logging-Level (siehe http://java.sun.com/j2se/1.4.2/docs/api/java/util/logging/Level.html)
		logLevelInt = specificTestfktParameters.getIntProperty("d", LOG_WARNING, false);
		switch (logLevelInt)
		{
		    case LOG_OFF: logLevel=Level.OFF; break;
			case LOG_SEVERE: logLevel=Level.SEVERE; break;
			case LOG_WARNING: logLevel=Level.WARNING; break;
			case LOG_INFO: logLevel=Level.INFO; break;
			case LOG_CONFIG: logLevel=Level.CONFIG; break;
			case LOG_FINE: logLevel=Level.FINER; break;
			case LOG_FINER: logLevel=Level.FINER; break;
			case LOG_FINEST: logLevel=Level.FINER; break;
			
			default: logLevel=Level.WARNING; break;
		}
		log = Logger.getLogger(this.getClass().getName());
		log.setLevel(logLevel);
		

		/* Hall-Call-Prob kann derzeit nicht gesetzt werden, weil dies ein Array und kein einzelner Wert ist */
		hallCallProb = new ArrayList();
		if (specificTestfktParameters.containsKey("p")) {
			StringTokenizer pToken = new StringTokenizer(specificTestfktParameters.getProperty("p"),":");
			while (pToken.hasMoreElements()){
				try {
					String p = pToken.nextToken();
					hallCallProb.add(new Float(p));
					log.finer("Using hall-call-propability "+p);
				} catch (Exception e) {
					log.warning("Error casting given hall-call-propability to float. Skipping.");
				}
			}
		}
		if (hallCallProb.size() == 0) {
			log.info("No hall-call-propability specified or errors occured. Using default-values 0.3 and 0.4.");
			hallCallProb.add(new Float(0.3));
			hallCallProb.add(new Float(0.4));
		}

		//private float[] hallCallProb={0.3f,0.4f}; //Ankunftswahrscheinlichkeiten
		
		
		// Umgebung fuer Simulator definieren
		String fileSep = System.getProperty("file.separator"); //OS-Abhaengiger Pfad-Seperator (z. B. "/" oder "\")
		// Da der Sep. unter DOS gleichzeitig das String-Escape-Zeichen ist, mu verdoppelt werden!
		if (fileSep == "\\")
			fileSep=fileSep+fileSep; 
		if (pathToSim.compareTo("") == 0) {	// Falls nicht durch Parameter gesetzt
			pathToSim = "simulator"+fileSep+"FahrStuhl";
		}
		
		/* Pruefen, welches Betriebssystem laeuft, um Dateinamen mit Endung korrekt zu setzen.
		 * Dies ist notwendig fuer die Pruefung, ob der Simulator existiert.
		 */
		String simName;
		if (fileSep.equalsIgnoreCase("\\")) {
			log.finer("Windows: Accepted Simulator-Name is \"ring.exe\".");
			simName = "ring.exe"; // Name des Simulator-Binary
		} else {
			log.finer("Unix/Linux: Accepted Simulator-Name is \"ring\".");
			simName = "ring"; // Name des Simulator-Binary
		}
		String completePathWSim = pathToSim+fileSep+simName;
		
		
		try {
			//		 Was tun wir, wenn das binary nicht da ist?? Kompilieren natuerlich!
			File simBin = new File(completePathWSim);
			if (! simBin.exists())
			{
				log.warning("Simulator nicht gefunden..rufe make auf..");
				String[] makeCmd = {"make","-C",pathToSim};
				Process makeSim= runtime.exec(makeCmd);
				// Auf Ende von "make" warten
				try {
					makeSim.waitFor();
				}
				catch (InterruptedException e) {
				}
				if (makeSim.exitValue() != 0)
					throw new Exception("SimulatorCompileException: make and gcc installed? Sources in right place?");
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
				
		/* Aufruf bauen */
		cmd = new String[7+2*floorCount]; 
		// Programmname und -pfad
		cmd[0] = completePathWSim;
		// Feste Parameter fuer Simulator
		cmd[1] = Integer.toString(floorCount);
		cmd[2] = Integer.toString(cars);
		cmd[3] = Integer.toString(iterations);
		cmd[6] = "0"; // Kein Debugging
		
		/* Debug-infos */
		log.info(getConfig());
	}
	
	
	/** 
	 * Haupt-Methode:
	 * Zunaechst wird der Simulator gesucht und ggf. kompiliert, anschlieend werden die Simulationen gestartet.
	 * 
	 * @see testfkt.TestFunctions#execute(java.util.List)
	 */
	public List execute(List input) throws Exception
	{
		List results = null;
		
		/* Ueberpruefung, ob init() ausgefuert wurde, sonst Fehler ausgeben und null zurueckgeben */
		if (hallCallProb.size() == 0) {
			log.severe("Object has not been initialized. Please run init() first! \n !! Returning null-value !!!");
			return null;
		} 
		
		/* Normal fortfahren (init() wurde korrekt ausgefuehrt) */
		else {
			cmd[5] = Integer.toString(seed);
			
			String simulatorOutputLine;
			results = new ArrayList();
	
			// Init 
			for (int j=7; j<cmd.length; j++) {
				cmd[j] = "";
			}
			// Individuum als Gewichtungsvektor uebergeben
			for (int j=0; j<input.size(); j++)
			{
				cmd[7+j] = input.get(j).toString();
			}
	
			/* Programm starten */
			for (int i=0; i<hallCallProb.size(); i++)
			{
				// Hall-Call-Wahrscheinlichkeit setzen, wodurch mehrere Zielkriterien entstehen
				cmd[4] = ((Float)hallCallProb.get(i)).toString();
				// Programm aufrufen
				Process simulator= runtime.exec(cmd);
				// Auf Ende der Simulation warten
				try {
					simulator.waitFor();
				} catch (InterruptedException e) {
				}
				// Output lesen
				BufferedReader simulatorOutputBuffer = new BufferedReader(
						new InputStreamReader(simulator.getInputStream()));
				if ((simulatorOutputLine = simulatorOutputBuffer.readLine()) != null) {
					results.add(new Double(simulatorOutputLine));
					// Debug-Ausgaben
					if (logLevelInt >= LOG_FINER) System.out.println("Simulator-Ergebnis: " + simulatorOutputLine);
					//log.finer("Simulator-Ergebnis: " + simulatorOutputLine);
				}
			}
			if (logLevelInt >= LOG_FINER)
				System.out.println("");
		}
		return results;
	}
	

	/* (non-Javadoc)
	 * @see testfkt.TestFunctions#writeParetoPoints(int, java.lang.String)
	 */
	public void writeParetoPoints(int n, String filename) throws Exception {
		}

	
	/**
	 * Da es sich hierbei um ein "echtes" Problem und keine Testfunktion handelt, koennen an dieser
	 * Stelle keine optimalen Grenzen gegeben werden. Um die Kompatibilitaet zu Algos zu erhalten,
	 * die diese Angaben benoetigen, werden nicht auf Korrektheit gepruefte Testwerte zurueckgeliefert.
	 * 
	 * @see testfkt.TestFunctions#getOptimalBounds()
	 */
	public List getOptimalBounds() {
		List bounds = new ArrayList();

		// DIES SIND NUR TESTWERTE, DAMIT DIE ALGOS NICHT ABBRECHEN !
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		bounds.add(new Double(-1000));	
		bounds.add(new Double(1000));
		
		return (bounds);
	}

	
	/* (non-Javadoc)
	 * @see testfkt.TestFunctions#getStartPoint()
	 */
	public List getStartPoint() {
		List bounds = new ArrayList();

		// DIES SIND NUR TESTWERTE, DAMIT DIE ALGOS NICHT ABBRECHEN !
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		bounds.add(new Double(0));	
		
		return (bounds);
	}

	
	/**
	 * Liefert die benoetigte Dimension der einzelnen Individuen zurueck; diese ist hier abhaengig
	 * von der Anzahl der Stockwerke.
	 * 
	 * @see testfkt.TestFunctions#getDimension()
	 */
	public int getDimension() {
		//Anzahl Stockwerke*2, z.B. 6 Stockwerke=12 Dimensionen
		return floorCount*2;
	}

	
	/* (non-Javadoc)
	 * @see testfkt.TestFunctions#getNumberOfObjectives()
	 */
	public int getNumberOfObjectives() {
		/* Die Anzahl der Wahrscheinlichkeiten, fuer die die Loesungsvektoren
		 * berechnet werden, ist unsere Kriterienzahl.
		 */
		return hallCallProb.size();
	}
	
	
	/**
	 * Diese Methode mu nach dem kompletten Simualtionslauf vom Algorithmus aufgerufen werden;
	 * sie bereinigt die verrauschten Fitnesswerte der vom Algo gewonnenen Individuen und gibt
	 * diese zurueck.
	 * 
	 * @see testfkt.TestFunctions#postProcessing(java.util.List)
	 */
	public List postProcessing(List individuals) {
		
		if (logLevelInt >= LOG_INFO) {
			// Postprocessing-Ergebnisse in Datei speichern (hier nur Leerzeile)
			try {
				FileWriter fw = new FileWriter("postprocessing.txt", true);
				fw.write("\n--------------------------------------- \n");
				fw.write(getConfig()+"\n");
				fw.close();
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		/* Liste enthaelt mindestens ein Individum-Objekt */
//		if ((individuals.size()>0) && (individuals.get(0) instanceof Individual)) {
		if (individuals.size()>0) {

			// Debug-Ausgabe vor Bearbeitung */
			if (logLevelInt >= LOG_CONFIG)
			{
				System.out.println("Alte Fitnesswerte vor Rauschbereinigung:");
				for (int i=0; i<individuals.size(); i++) {
					if (individuals.get(0) instanceof Individual) {
						Individual ind = (Individual)individuals.get(i);
						System.out.println(ind.getFitness().toString());
					} 
					else if (individuals.get(0) instanceof Particle) {
						Particle ind = (Particle)individuals.get(i);
						System.out.println(ind.getFitness().toString());
					}					
				}
			}
			
			log.config("Starting PostProcessing Elevator at i="+Integer.toString(iterations)
					+ " with "+individuals.size()+" individuals..");
			// Eigentliche Bearbeitung, d.h. Neu-Simulation
			try {
				/* Alle Individuen einzeln durchgehen: 
				 * Fitness erneut berechnen, Mittelwert fuer jedes Kriterium bilden */
				RandomPG rand = new RandomPG();
				
				for (int i=0; i<individuals.size(); i++) {
					if (logLevelInt >= LOG_CONFIG) System.out.print(i + ": ");
					List oldFitness = null;
					List objVars = null;
					
					if (individuals.get(0) instanceof Individual) {
						Individual ind = (Individual)individuals.get(i);
						oldFitness = (List)ind.getFitness();
						objVars = (List)ind.getObjectVariables();
					} 
					else if (individuals.get(0) instanceof Particle) {
						Particle ind = (Particle)individuals.get(i);
						oldFitness = (List)ind.getFitness();
						objVars = (List)ind.getPosition();
					}	
					
					List newFitness = new ArrayList();
					for (int k=0; k<oldFitness.size(); k++) {	// fuer jedes Kriterium
						newFitness.add(new Double(0));
					}
					for (int j=0; j<POSTPROCESSING_REPEATS; j++) {	// POSTPROCESSING_REPEATS oft mal
						seed = rand.nextInt();
						List tempFitness = execute(objVars);
						for (int k=0; k<tempFitness.size(); k++) {	// fuer jedes Kriterium
							double sumFitness = ((Double)newFitness.get(k)).doubleValue();
							double addFitness = ((Double)tempFitness.get(k)).doubleValue();
							newFitness.set(k, new Double(sumFitness + addFitness));
						}
						if ((logLevelInt >= LOG_CONFIG) && (j % 10 == 0)) System.out.print(".");
					}
					if (logLevelInt >= LOG_CONFIG) System.out.println("complete.");
					for (int k=0; k<newFitness.size(); k++) {	// fuer jedes Kriterium
						newFitness.set(k, new Double(((Double)newFitness.get(k)).doubleValue() / POSTPROCESSING_REPEATS));
					}
					
					if (logLevelInt >= LOG_INFO) {
						try {
					 		// Postprocessing-Ergebnisse in Datei speichern (hier nur Leerzeile)
					 		FileWriter fw = new FileWriter("postprocessing.txt", true);
					 		if (individuals.get(0) instanceof Individual) {
								Individual ind = (Individual)individuals.get(i);
								fw.write(ind.getFitness().toString()+" -> "+newFitness.toString()+"\n");
							} 
							else if (individuals.get(0) instanceof Particle) {
								Particle ind = (Particle)individuals.get(i);
								fw.write(ind.getFitness().toString()+" -> "+newFitness.toString()+"\n");
							}	
							fw.close();
						} catch (IOException e) {
							e.printStackTrace();
						}
					}	
					
					if (individuals.get(0) instanceof Individual) {
						Individual ind = (Individual)individuals.get(i);
						ind.setFitness(newFitness);
					} 
					else if (individuals.get(0) instanceof Particle) {
						Particle ind = (Particle)individuals.get(i);
						ind.setFitness((ArrayList)newFitness);
					}	
				}
			} catch (Exception e) {
				e.printStackTrace();
				// nix -> Fitnesswert bleibt einfach unveraendert
			}

			log.config("PostProcessing Elevator finished.\n");
			
			// Debug-Ausgabe nach Bearbeitung */
			if (logLevelInt >= LOG_CONFIG)
			{
				System.out.println("Neue Fitnesswerte:");
				for (int i=0; i<individuals.size(); i++) {
					Individual ind = (Individual)individuals.get(i);
					System.out.println(ind.getFitness().toString());
				}
			}
		}
		return individuals;
	}
	
	
	/**
	 * Liefert die Konfiguration als String zurueck. Dies soll Debugging und Konsoleninfos erleichtern.
	 */
	public String getConfig() {
		return("ring-Config: Floors="+floorCount + 
				", Cars="+cars + 
				", Iterations="+iterations +
				", Seed="+seed);
	}
	
	
	/**
	 * main-Methode: Ruft den Simulator mit verschiedenen Iterationen und stets den gleichen
	 * Individuen auf. Dies dient dem Austesten der ring-Parameter.
	 */
	public static void main (String[] args) {
		SRingCon test = new SRingCon();
		PropertiesPG props = new PropertiesPG();
		RandomPG rand = new RandomPG(Integer.parseInt(args[0]));
		AbstractAlgo algo = new MueRhoLES(); // nur fuer Methode AbstractAlgo.liveOutput(..) verwendet!
		
	   /* Objektvariablen des Indiviuums erzeugen */	
		ArrayList generation = new ArrayList();
		try {
			for (int k=0; k<5; k++) {
				ArrayList obj = new ArrayList();
				for (int j = 0; j < test.getDimension(); j++)
				{
					obj.add(new Double(rand.nextInt() + rand.nextDouble()));						
				}
				System.out.println("Individuum " + k + ": " + obj);
				generation.add(new Individual(obj, new ArrayList()));
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
	   /* Aufruf der Testfunktion */
		try
		{
			//System.in.read();
			props.setProperty("d", "4");
			props.setProperty("r", Integer.toString(rand.nextInt()));
			//props.setProperty("r", "2");
			for (int i=10; i<=1000000; i=i*10) {
				System.out.println(i);
			   /* Objektvariablen des Indiviuums erzeugen */	
				props.setProperty("i", Integer.toString(i));
				test.init(props);
				for (int k=0; k<generation.size(); k++) {
					Individual ind = (Individual)generation.get(k);
					ind.setFitness((ArrayList)test.execute(ind.getObjectVariables()));
				}
				test.postProcessing(generation);
				//algo.liveOutput(generation, i);
			}
		}
		catch (Exception e)
		{
			 e.printStackTrace();
		}
	}
	
}
		

