/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on 01.07.2004
 *
 * 
 */
package org.nobeltje.tools;

import org.nobeltje.exceptions.BinaryStringFormatException;

/**
 * Diese Klasse implemetiert eine bin&auml;re Flie&szlig;kommazahlen-Darstellung und -Umrechnung
 * mit doppleter Genauigkeit (64 bit) nach IEEE 754.<br>
 * TODO: Momentan sind noch keine Sonderf&auml;lle (Unendlichkeit, Null etc.) implementiert bzw. Verhalten bei Sonderf&auml;llen.noch ist nicht bekannt.
 * 
 * @author $Author: noot $
 * @version $Revision: 726 $<br>
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 */
public class DoubleIEEE754

{
	/**
	 * Anzahl der Bits, die dem Exponenten zugeordnet sind.
	 */
	public static int EXPSIZE=11;
	/**
	 * Anzahl der Bits, die der Mantisse zugeordnet sind.
	 */
	public static int MANTISSIZE=52;
	/**
	 * Anzahl der Bits, die f&uuml; das Vorzeichen der mantisse verwendet werden.
	 */
	public static int ALGSIGNSIZE=1;
	/**
	 * Die gesamte Anzahl an Bits, die eine Flie&szlig;kommazahl mit doppleter Genauigkeit nach IEEE 754 enth&auml;lt. 
	 */
	public static int VARSIZE = EXPSIZE+MANTISSIZE+ALGSIGNSIZE; //64 bit nach IEEE 754

	/**
	 * Die Methode wandelt einen double-Wert in einen bin&auml;ren String (nach IEEE 754).<br>
	 * Da die verwendeten Java-Methoden keine f&uuml;hrenden Nullen ausgeben, muss der String
	 * ggf. aufgef&uuml;llt werden, um die Normgr&ouml;sse zu erhalten.
	 * 
	 * @param inputDouble double: Flie&szlig;kommawert, der umgerechnet werden soll.
	 * @return String: Bin&auml;rer String
	 * @see Long#toBinaryString(long)
	 * @see Double#doubleToLongBits(double)
	 */
	public static String getBinaryString(double inputDouble)
	{
		String output=Long.toBinaryString(Double.doubleToLongBits(inputDouble));
		while (output.length() < DoubleIEEE754.VARSIZE)
    		output="0"+output;
		return output;
	}
	
	/**
	 * Die Methode wandelt nach IEEE 754 einen String in eine Flie&szlig;kommazahl um.
	 * @param binaryString String: Ein String, bestehend aus genau VARSIZE Zeichen aus der Menge {1,0} 
	 * @return double: Den Wert des Strings als Fliesskommazahl doppelter Genauigkeit (nach IEEE 754)
	 * @throws BinaryStringFormatException falls der String-Parameter ung&uuml;ltige Zeichen enth&auml;lt oder eine falsche L&auml;nge besitzt.
	 */
	public static double getDouble(String binaryString) throws BinaryStringFormatException
	{
		int i,j;
		int exponent=0;
		double mantis=0;
		int currentIndex=binaryString.length();
		if (currentIndex != VARSIZE)
			throw new BinaryStringFormatException(VARSIZE);
		currentIndex--;
		
		// Mantisse extrahieren und in Integer wandeln 
		for (i=0; i<MANTISSIZE;i++)
		{
			if (binaryString.charAt(currentIndex)=='1')
				j=1;
			else
				if (binaryString.charAt(currentIndex)=='0')
					j=0;
				else
					throw new BinaryStringFormatException(VARSIZE);
			mantis+=j*(double)Math.pow((double)2,(double)(-1*MANTISSIZE)+i);
			currentIndex--;
		}
		mantis=mantis+1;
		
		// Exponent extrahieren und in Integer wandeln 
		for (i=0; i<=EXPSIZE-1;i++)
		{
			if (binaryString.charAt(currentIndex)=='1')
				j=1;
			else
				if (binaryString.charAt(currentIndex)=='0')
					j=0;
				else
					throw new BinaryStringFormatException(VARSIZE);
			exponent=exponent+j*(int)Math.pow((double) 2,(double)i);
			currentIndex--;
		}
		exponent=exponent-1023; //Nach IEEE wird der Exponent in Bias-Darstellung verwendet
		
		// Vorzeichen
		int mantisAlgSign=0;
		if (binaryString.charAt(currentIndex)=='1')
			mantisAlgSign=-1; //negatives Vorzeichen
		else
			if (binaryString.charAt(currentIndex)=='0')
				mantisAlgSign=1; //negatives Vorzeichen
			else
				throw new BinaryStringFormatException(VARSIZE);

		
		/* Dezimalwert zusammenstoepseln */
		return (mantisAlgSign*mantis*Math.pow((double)2,(double)exponent));
	}
}
