/*
 * Copyright (C) 2005  PG447 - University Dortmund
 * 
 * This file is part of NObELTJE.
 * 
 * NObELTJE is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as published by
 * the Free Software Foundation.
 * 
 * NObELTJE is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with NObELTJE; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

 /*
 * Created on 28.05.2004
 */
package org.nobeltje.tools;

import java.util.ArrayList;
import java.util.List;

/**
 * Individual kapselt die Objekt- und Strategievariablen mit den zugehoerigen Funktions- oder Fitnesswerten in einem Objekt.
 * <br>Da die Klasse eigentlich nur Daten enthaelt, besitzt sie auch fast nur get- und set-Methoden.
 *  
 * @author $Author: noot $
 * @version $Revision: 726 $
 * $Date: 2005-02-02 00:50:20 +0100 (Mi, 02 Feb 2005) $
 */

public class Individual implements Cloneable
{
	/**
	 * *** Variabeln ***
	 * WICHTIG: Bei Erweitern der Variablen unbedingt die Methode clone() erweitern !
	 */
	
    /**
     * Liste der Objektvariablen 
     */
    private List objectVariables;
    
    /**
     * Liste der Strategievariablen 
     */
	private List strategyVariables;
	
	/**
	 * Liste der Fitnesswerte
	 */
	private List fitness;
	
	/**
	 * Generation, in der das Individuum generiert wurde
	 */
	private int generation = 0; 
	

	/**
	 * Leerer Konstruktor, der nur alle Listen mit Objekten initialisiert.
	 */
	public Individual()
	{
		objectVariables = new ArrayList();
		strategyVariables = new ArrayList();
		fitness = new ArrayList();
		generation = 0;
	}
	
	
	/**
	 * Konstruktor, der alle Listen mit Objekten initialisiert und die aktuelle Generation setzt.
	 * 
	 * @param generation Generation, in der das Individuum generiert wurde
	 */
	public Individual(int generation)
	{
		objectVariables = new ArrayList();
		strategyVariables = new ArrayList();
		fitness = new ArrayList();
		this.generation = generation;
	}
	
	
	/**
	 * Konstruktor, der die Objektvariablen und die Fitnesswerte setzt, die Strategievariablen und die aktuelle Generation initalisiert.
	 * 
	 * @param obj Liste der Objektvariablen
	 * @param fit Liste der Fitensswerte
	 */
	public Individual(List obj, List fit)
	{
		objectVariables = obj;
		strategyVariables = new ArrayList();
		fitness = fit;
		createLists();
	}
	
	/**
	 * Konstruktor, der alle Listen und die aktuelle Generation setzt.
	 * 
	 * @param obj Liste der Objektvariablen
	 * @param strategy Liste der Strategievariablen
	 * @param fit Liste der Fitensswerte
	 * @param generation Generation, in der das Individuum generiert wurde
	 */	
	public Individual(List obj, List strategy, List fit, int generation)
	{
		objectVariables = obj;
		strategyVariables = strategy;
		fitness = fit;
		this.generation = generation;
		createLists();
	}

	
	/**
	 * Kopiert das eigene Objekt und gibt es als Object zurueck.
	 * 
	 * (non-Javadoc)
	 * @see java.lang.Object#clone()
	 */
	public Object clone()
	{
		createLists();
		return new Individual ((List)((ArrayList)this.objectVariables).clone(),
							   (List)((ArrayList)this.strategyVariables).clone(),	
							   (List)((ArrayList)this.fitness).clone(),
							   generation);
	}
	
	
	/**
	 * Prueft, ob die Klassenvariablen (Listen) auf null zeigen und generiert
	 * entsprechend Listen fuer diese. (insb. wichtig fuer clone())
	 */
	private void createLists() {
		if (objectVariables==null) {
			objectVariables = new ArrayList();
		}
		if (strategyVariables==null) {
			strategyVariables = new ArrayList();
		}
		if (fitness==null) {
			fitness = new ArrayList();
		}
		
	}
	
	
	/**
	 * Vergleicht, ob die Objektvariabeln von 2 Individuen identisch sind.
	 * <br>Vorgehensweise: alle Negativ-Tests durchfueren und ggf. mit return false abbrechen.
	 * Wenn alle Tests bestanden sind, wird am Ende true zurueckgegeben.
	 * 
	 * @param ind1 Erstes Individuum
	 * @param ind2 Zweites Individuum
	 * @return Status, ob Objektvariablen der uebergebenen Individuen gleich sind.
	 */
	public static boolean compareObjectVariables(Individual ind1, Individual ind2) {
		// Pruefen der Dimensionen der Objektvariablen
		if (ind1.getObjectVariables().size() != ind2.getObjectVariables().size())
		{
			return false;
		}
		// Vergleichen aller Objektvariablen
		for (int i=0; i < ind1.getObjectVariables().size(); i++) {
			if (!ind1.getObjectVariables().get(i).equals(ind2.getObjectVariables().get(i)))
			{
				return false;
			}
		}			
		// Wenn bis hierher durchgekommen: Alle Tests bestanden -> true zurueckgeben
		return true;
	}
	
	
	/**
	 * Liefert die Objektvariablen des Objekts zurueck. 
	 * 
	 * @return Liste der Objektvariablen
	 */
	public ArrayList getObjectVariables()
	{
		return (ArrayList)objectVariables;
	}
	
	/**
	 * Liefert die Strategievariablen des Objekts zurueck. 
	 * 
	 * @return Liste der Strategievariablen
	 */
	public ArrayList getStrategyVariables() {
		return (ArrayList)strategyVariables;
	}

	/**
	 * Liefert die Fitnesswerte des Objekts zurueck. 
	 * 
	 * @return Liste der Fitnesswerte
	 */
	public ArrayList getFitness()
	{
		return (ArrayList)fitness;
	}

	/**
	 * Liefert die Nummer der Generation zurueck, in der das Individuum generiert wurde.  
	 * 
	 * @return Nummer der Generation
	 */
	public int getGeneration()
	{
		return generation;
	}

	/**
	 * Setzt die Objektvariablen des Objekts.
	 * 
	 * @param list Liste der Objektvariablen
	 */
	public void setObjectVariables(List list)
	{
		objectVariables = list;
		createLists();
	}
	
	/**
	 * Setzt die Strategievariablen des Objekts.
	 * 
	 * @param list Liste der Strategievariablen
	 */
	public void setStrategyVariables(List list) {
		strategyVariables = list;
		createLists();
	}

	/**
	 * Setzt die Fitnesswerte des Objekts.
	 * 
	 * @param list Liste der Fitnesswerte
	 */
	public void setFitness(List list)
	{
		fitness = list;
		createLists();
	}

	/**
	 * Setzt die Generation, in der das Indiviuum generiert wurde.
	 * 
	 * @param generation Nummer der Generation
	 */
	public void setGeneration(int generation)
	{
		this.generation = generation;
	}
	
	
	/**
	 * Erstellt aus zwei Listen mit Objektvariablen und Fitnesswerten eine Liste mit Individuen
	 * 
	 * @param objList Liste mit Listen der Objektvariablen
	 * @param fitList Liste mit Listen der Fitnesswerten
	 * @return List Liste von generierten Individuen
	 */
	public static List makeIndividuals(List objList, List fitList) {
		ArrayList result = new ArrayList();
		for(int i=0; i<objList.size(); i++)
		{
			Individual ind = new Individual((ArrayList)objList.get(i), (ArrayList)fitList.get(i));
			result.add(ind);
		}
		return result;
	}
	
	

	/**
	 * Erstellt aus Liste mit Individuen eine Liste mit Listen der Objektvariablen und 
	 * eine Liste mit Listen der Fitnesswerten.<br>
	 * Da wir hier 2 Rueckgaben haben, werden die Ergebnisse per Call-by-reference zurueckgegeben.
	 * 
	 * @param indList Liste der zu verarbeitenden Individuen (input)
	 * @param objList Liste mit Listen der Objektvariablen (output)
	 * @param fitList Liste mit Listen der Fitnesswerten (output)
	 */
	public static void extractIndividuals(List indList, List objList, List fitList) {
		/* Listenobjekte fuer Ausgabe neu erstellen (also damit ggf. loeschen */
		objList = new ArrayList();
		fitList = new ArrayList();
		
		/* Individuen umwandeln */
		for(int i=0; i<indList.size(); i++)
		{
			Individual ind = (Individual)indList.get(i);
			objList.add(ind.getObjectVariables());
			fitList.add(ind.getFitness());
		}
	}

}
